//	TorusGames-Common.h
//
//	Platform-independent definitions for Torus Games's internal code.
//	The internal code doesn't know or care what platform (Win32 or MacOS)
//	it's running on.
//	However, the platform-specific user interface *does* include
//	the present file.  For that reason we take care not to use any
//	OpenGL-specific definitions (i.e. nothing from gl.h) because
//	the user interface doesn't know about OpenGL.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#pragma once

#include <stdlib.h>	//	for malloc() and free()
#include <stdbool.h>


//	For the Shape of Space talk, start with Intro instead of Pool,
//	and offset Tic-Tac-Toe by half a square.
//#define TORUS_GAMES_FOR_TALK

//	For the middle school workshop Exploring the Shape of Space
//	I need cylinder-only tic-tac-toe.  This is just a quick hack,
//	not for general distribution.
//#define CYLINDER_HACK


//	Acknowledge unused parameters.
#define UNUSED_PARAMETER(x)	((void)(x))

//	Define a byte.
typedef unsigned char	Byte;

//	Define our own UTF-16 type.
//	Our Char16 corresponds to a WCHAR on Win32 or to a UniChar on Mac OS.
//	We do not support surrogate pairs — in fact we often pass a character
//	as a single Char16 — so really this is more like UCS-2 than UTF-16.
typedef unsigned short int	Char16;

//	Internal functions pass error messages as pointers
//	to zero-terminated UTF-8 strings.  The strings are defined
//	statically, so the caller need not (must not!) free them.
typedef const char		*ErrorText;

//	Define GET_MEMORY and FREE_MEMORY macros to manage memory so
//	that in future we can easily change memory allocation schemes.
//	Use a variable gMemCount to test for memory leaks.
//	Define FREE_MEMORY_SAFELY() to test for and set NULL pointers.
#define GET_MEMORY(n)			(gMemCount++, malloc(n))
#define FREE_MEMORY(p)			(gMemCount--, free(p))
#define FREE_MEMORY_SAFELY(p)	{if (p != NULL) {FREE_MEMORY(p); p = NULL;}}

//	BUFFER_LENGTH() measures the number of items in an array,
//	not the number of bytes, and automatically adjusts to changes
//	in the number of elements or the size of each element.
#define BUFFER_LENGTH(a)	( sizeof(a) / sizeof((a)[0]) )

//	Macro to compare two 2-letter language codes.
#define SAME_LANGUAGE_CODE(A,B)	((A)[0] == (B)[0] && (A)[1] == (B)[1])

//	RUNNING_OPENGL_3 marks places where changes will be required
//	for an OpenGL 3.0 forward-compatible context.
//	When this program supports such a context,
//	the fixed macro RUNNING_OPENGL_3 will get replaced
//	by a flag saying whether or not an OpenGL 3.0 context
//	has been loaded at runtime.  It's possible that more than
//	one such flag will be required, depending on how closely
//	OpenGL ES 2.0 mirrors OpenGL 3.0.  For now, this macro
//	suffices to mark places that will require attention.
#define RUNNING_OPENGL_3	false


//	Represent colors as (αR, αG, αB, α) instead of the traditional (R,G,B,α)
//
//		1.	to allow rigorously correct blending
//			[see comments preceding the call to glBlendFunc()], and
//
//		2.	to let mipmap generation automatically average pixels
//			weighted by their alpha values [see comments preceding
//			the call to glGenerateMipmap()].
//
#define PREMULTIPLY_RGBA(r,g,b,a)	(float [4]){(a)*(r), (a)*(g), (a)*(b), (a)}


//	Assign the vertex attributes to well-defined locations in the vertex shader.
//	ATTRIBUTE_POSITION must bind to location 0, but the remaining choices are arbitrary.
#define ATTRIBUTE_POSITION				0
#define ATTRIBUTE_TEX_COORD				1
#define ATTRIBUTE_COLOR					2
#define ATTRIBUTE_TRANSFORMATION_X		3
#define ATTRIBUTE_TRANSFORMATION_Y		4
#define ATTRIBUTE_TEX_COORD_TRANSFORM	5

//	Keep an array of shader programs, each referenced by an unsigned integer 
//	that glCreateProgram() provides to refer to the given program.
//	Each program contains a vertex shader and a fragment shader.
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.
typedef enum
{
	ShaderTexture = 0,
	ShaderColor,
	ShaderColorWithAlphaMask,
	ShaderColorWithTexture,
	NumShaders
} ShaderIndex;

//	Keep an array of texture "names", each "name" being an unsigned integer 
//	that glGenTextures() generates to refer to the given texture.  
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.
//	This enum lists only the common game-independent textures.
//	Each game will manage some additional textures of its own.
typedef enum
{
	TextureCommonHandFlat = 0,
	TextureCommonHandGrab,
	TextureCommonClockBase,
	TextureCommonClockSweep,
	TextureCommonClockRim,
	TextureCommonClockArrow,
	TextureCommonLineDashed,
	TextureCommonHalo,
	NumTexturesCommon
} TextureIndexCommon;

//	Keep an array of Vertex Array Object (VAO) "names" along with
//	two parallel arrays of Vertex Buffer Object (VBO) names, one for 
//	the vertex buffer proper and the other for the so-called index buffer.
//	Each "name" is an unsigned integer that glGenVertexArrays() 
//	or glGenBuffers() generates to refer to the given VAO or VBO, respectively.
//	Here we define humanly meaningful synonyms for the array indices {0, 1, 2, … }.
typedef enum
{
	VertexBufferRectangle = 0,
	VertexBufferRectangleWithEndcaps,
	VertexBufferTriangle,
	NumVertexBuffers
} VertexBufferIndex;


//	When composing kana, e.g. 's' + 'u' = 'す',
//	let 0 represent no pending character.
#define NO_PENDING_CHARACTER	0

//	π
#define PI		3.14159265358979323846
#define TWOPI	(2.0 * PI)

//	Some games refer to recent cursor positions
//	to determine the cursor heading.
#define DRAG_HISTORY_DEPTH	17

//	Handy utilities.
//	(Caution:  Arguments may be evaluated twice!)
#ifndef MAX
	#define MAX(A,B)	((A) > (B) ? (A) : (B))
#endif
#ifndef ABS
	#define ABS(A)		((A) >= 0 ? (A) : (-(A)))
#endif
#define SIGN(A)			((A) != 0 ? ((A) > 0 ? (+1) : (-1)) : 0)


//	Enumerate available games.
typedef enum
{
	GameNone,
	GameIntro,
	GameTicTacToe,
	GameMaze,
	GameCrossword,
	GameWordSearch,
	GameJigsaw,
	GameChess,
	GamePool,
	GameApples,
	NumGameTypes
} GameType;

//	Enumerate available topologies.
typedef enum
{
	TopologyTorus,
	TopologyKlein
} TopologyType;

//	Enumerate available views.
typedef enum
{
	ViewDomain,	//	View fundamental domain.
	ViewSmall,	//	View small fundamental domain, as transition to tiling.
	ViewTiling	//	View tiling.
} ViewType;

//	Enumerate hand cursor modes.
typedef enum
{
	HandNone,	//	We don't have the cursor.
	HandFree,	//	The cursor is moving freely within our space.
	HandScroll,	//	The cursor is scrolling the space relative to the window.
	HandDrag	//	The cursor is dragging an object within the space.
} HandStatus;

//	To place an object (either a sprite or a cursor)
//	in the fundamental domain we apply a dilation,
//	a rotation, a possible side-to-side reflection
//	and a translation, in that order.
typedef struct
{
	double	itsH,		//	horizontal translation, in range [-0.5, +0.5]
			itsV;		//	 vertical  translation, in range [-0.5, +0.5]
	bool	itsFlip;	//	reflection, always side-to-side
	double	itsAngle,	//	rotation, in radians counterclockwise
			itsSizeH,	//	horizontal dilation
			itsSizeV;	//	vertical dilation
} Placement;

//	The fundamental domain itself may be translated and flipped
//	within the window, but not rotated or dilated.
typedef struct
{
	double	itsH,		//	horizontal offset, in range [-0.5, +0.5]
			itsV;		//	 vertical  offset, in range [-0.5, +0.5]
	bool	itsFlip;	//	reflection, always side-to-side
} Offset;


//	RGBA pixel
typedef struct
{
	Byte	r,	//	red,   premultiplied by alpha
			g,	//	green, premultiplied by alpha
			b,	//	blue,  premultiplied by alpha
			a;	//	alpha = opacity
} PixelRGBA;

//	RGBA image, as read from a .rgba file
typedef struct
{
	//	If the image came from an RGBA file, keep a pointer 
	//	to the raw data, including the 8 header bytes.
	//	If the image was constructed manually,
	//	set itsNumRawBytes to 0 (it will be ignored) and 
	//	set itsRawBytes = (Byte *)itsPixels to the pixel data 
	//	(with no header bytes).
	unsigned int	itsNumRawBytes;
	Byte			*itsRawBytes;
	
	//	Interpret the data as a width, a height and some pixels.
	unsigned int	itsWidth,
					itsHeight;
	PixelRGBA		*itsPixels;	//	= itsRawBytes + 8

} ImageRGBA;

//	Texture coordinates may wrap or clamp.
typedef enum
{
	TextureWrap,
	TextureClamp
} TextureBoundary;

//	Most textures will be RGBA, but a few will be alpha-only.
typedef enum
{
	TextureRGBA,
	TextureAlpha
} TextureFormat;

//	TorusGamesRectangle, TorusGamesRectangleWithEndcaps and TorusGamesTriangle
//	create vertex buffer objects with fixed texture coordinates.
//	At runtime we may wish to rescale and/or shift those coordinates.
typedef struct
{
	float	itsScaleFactor[2],
			itsOffset[2];
} TexCoordTransform;


//	The platform-independent code keeps track of fonts using
//	pointers to platform-dependent information.
//	Here's the opaque typedef.
//	The platform-dependent code will work out the details to its satisfaction.
typedef struct Font		Font, *OpaqueFont;

//	Use an enum to index the various fonts.
typedef enum
{
	FontCrosswordPuzzle = 0,	//	Start at zero to allow use as index.
	FontCrosswordClue,
	FontWordSearchPuzzle,
	FontWordSearchList,
	FontPoolStatus,
	FontApplesDigit,
	NumberOfFonts				//	Total number of fonts.
} FontIndex;

//	For some games the platform-independent code keeps a pointer 
//	to a platform-dependent side window, such as the Crossword clues window,
//	the Word Search word list or the Pool status window.  
//	The platform-independent code can't possibly know the window's true definition,
//	so it keeps an opaque pointer instead.
//	(Note:  This is slightly safer and cleaner than keeping a pointer to a void,
//	because the present scheme requires no typecasting.)
typedef struct SideWindow		SideWindow, *OpaqueWindow;

//	What's happening during idle time?
typedef enum
{
	SimulationNone,	//	typically the case, if nothing special is happening
	SimulationTicTacToeWaitToMove,
	SimulationTicTacToeWaitToDisplayWin,
	SimulationTicTacToeFlash,
	SimulationMazeFlash,
	SimulationCrosswordFlash,
	SimulationWordSearchFlash,
	SimulationChessWaitToMove,
	SimulationChessChooseComputerMove,
	SimulationChessAnimateComputerMove,
	SimulationPoolRollBalls,
	SimulationPoolPlaceCueStick,
	SimulationPoolPauseBeforeShooting
} SimulationType;

//	Use an enum to index the various sounds.

typedef enum
{
	SoundTicTacToeMove =  0,	//	Begin at 0 to permit use as array index.
	SoundTicTacToeWin,
	SoundMazeSolved,
	SoundCrosswordKeypress,		//	touch-based UI only
	SoundCrosswordComplete,
	SoundWordSearchComplete,
	SoundJigsawJoin,
	SoundJigsawComplete,
	SoundChessWrongColor,
	SoundChessCheck,
	SoundChessCheckmate,
	SoundChessStalemate,
	SoundPoolCue,
	SoundPoolClick,
	SoundPoolSink,
	SoundPoolLoss,
	SoundPoolWin,
	SoundApplesLoss,
	SoundApplesWin,
	NumSounds					//	number of defined sounds
} SoundType;


//	for Intro

typedef struct
{
	Placement		itsSpritePlacement[4];
	unsigned int	itsDragSprite;
	unsigned int	itsTextureNameBackground,
					itsTextureNameSprites[4];
} IntroData;


//	for Tic-Tac-Toe

typedef enum
{
	PlayerNone = 0,	//	Start at 0 to permit use as array index.
	PlayerX,
	PlayerO
} TicTacToePlayer;

typedef struct
{
	TicTacToePlayer	itsBoard[3][3];
	bool			itsCellFlip[3][3];
	TicTacToePlayer	itsWhoseTurn;
	Placement		itsWinLine;
	unsigned int	itsTextureNameBackground,
					itsTextureNameX,
					itsTextureNameO,
					itsTextureNameWin;
} TicTacToeData;


//	for Maze

typedef enum
{
	WallWest = 0,	//	Start at 0 to permit use as array index.
	WallSouth,
	WallEast,
	WallNorth
} MazeWallIndex;

typedef struct
{
	bool			itsWalls[4];	//	indexed by MazeWallIndex
	unsigned int	itsIndex;		//	used temporarily during maze generation
} MazeCell;

typedef struct
{
	unsigned int	itsSize;
	bool			itsMazeIsValid;
	MazeCell		**itsCells;	//	 See TorusGamesMaze.c for coordinate conventions.
	Placement		itsMousePlacement,
					itsCheesePlacement;
	unsigned int	itsTextureNameGrass,
					itsTextureNameMouse,
					itsTextureNameCheese,
					itsTextureNameMaze;
} MazeData;


//	for Crossword

#define MAX_CROSSWORD_SIZE			 6	//	Must be ≤ 15 to accommodate texture indexing scheme.
#define	MAX_CROSSWORD_CLUES			12
#define MAX_CROSSWORD_CLUE_LENGTH	80
#define MAX_CROSSWORD_PREFIX_LENGTH	32

#define CROSSWORD_NO_CLUE	0xFF

//	In the desktop Torus Games, Crossword puzzle letters look best at 256×256 resolution,
//	but you may instead use 128×128 if you want to conserve video RAM.
#define CROSSWORD_CHARACTER_TEXTURE_SIZE_DESKTOP	256

//	In the mobile Torus Games, 128×128 looks best, 
//	but crossword puzzles load noticeably faster at 64×64,
//	with only a tiny bit of blurring. 
#define CROSSWORD_CHARACTER_TEXTURE_SIZE_MOBILE		64

//	Assemble a Hangul syllable as the user types the jamo.
typedef struct
{
	//	The syllable's constituent jamo are stored in unicode
	//	not as explicit leading (0x1100-0x1112), vowel (0x1161-0x1175)
	//	and trailing (0x11A8-0x11C2) jamo, but rather 
	//	as "Hangul compatibility jamo" (0x3131-0x3163), because
	//
	//		1.	the Korean Input Method passes compatibility jamo,
	//			with no distinction for leading or trailing consonants, and
	//
	//		2.	as the Crossword code assembles syllables,
	//			it may need to remove a trailing consonant from one syllable
	//			and re-assign it as the leading consonant in the next syllable,
	//			for example  질 + ㅣ  →  지리 .
	//
	Char16	itsLeadingConsonant,
			itsVowel,
			itsTrailingConsonant;
	
	//	The combination of the current (possibily incomplete) set of jamo
	//	gets assembled into a tentative Hangul syllable, which is stored
	//	not in this structure but rather in itsPendingCharacter (see below).
			
} HangulSyllable;

//	Initialize a HangulSyllable to all zeros.
#define EMPTY_HANGUL_SYLLABLE	((HangulSyllable) {0x0000, 0x0000, 0x0000})


typedef struct
{
	bool			itsPuzzleIsValid;
	char			itsPreviousLanguageCode[3];	//	Two-letter language code, such as "en" or "ja", with terminating zero
	unsigned int	itsPuzzleSize,
					itsIndexTorus,
					itsIndexKlein;
	Char16			itsBoard[MAX_CROSSWORD_SIZE][MAX_CROSSWORD_SIZE],
					itsSolution[MAX_CROSSWORD_SIZE][MAX_CROSSWORD_SIZE];
	bool			itsBoardFlips[MAX_CROSSWORD_SIZE][MAX_CROSSWORD_SIZE],
					itsLeftwardWord[MAX_CROSSWORD_SIZE][MAX_CROSSWORD_SIZE];
	unsigned int	itsHotCellH,
					itsHotCellV;
	bool			itsHotCellFlip;
	Char16			itsHotDirection;	//	← → ↓
	bool			itsHotWord[MAX_CROSSWORD_SIZE][MAX_CROSSWORD_SIZE];
	Char16			itsPendingCharacter;	//	supports accents (e.g. ':' + 'u' = 'ü') and kana composition (e.g. 's' + 'u' = 'す')
	HangulSyllable	itsPendingHangulSyllable;
	Char16			itsCluePrefixAcross[MAX_CROSSWORD_PREFIX_LENGTH],
					itsCluePrefixDown[MAX_CROSSWORD_PREFIX_LENGTH],
					itsCluesText[MAX_CROSSWORD_CLUES][MAX_CROSSWORD_CLUE_LENGTH];
	Byte			itsClueIndexAcross[MAX_CROSSWORD_SIZE][MAX_CROSSWORD_SIZE],
					itsClueIndexDown  [MAX_CROSSWORD_SIZE][MAX_CROSSWORD_SIZE];
	unsigned int	itsTextureNameGridLine,
					itsTextureNamePendingCharacter,
					itsTextureNameCell[MAX_CROSSWORD_SIZE][MAX_CROSSWORD_SIZE];
} CrosswordData;


//	for WordSearch

#define	MAX_WORDSEARCH_SIZE			12	//	Must be ≤ 15 to accommodate texture indexing scheme.
#define MAX_WORDSEARCH_NUM_WORDS	48
#define MAX_WORDSEARCH_WORD_LENGTH	26	//	not including terminating zero

#define WORDSEARCH_CHARACTER_TEXTURE_SIZE_DESKTOP	128
#define WORDSEARCH_CHARACTER_TEXTURE_SIZE_MOBILE	 64

typedef struct
{
	char	itsStartH,
			itsStartV,
			itsDeltaH,
			itsDeltaV;
	float	itsColor[4];	//	premultiplied alpha
} WordSearchLine;

typedef struct
{
	bool			itsPuzzleIsValid;
	char			itsPreviousLanguageCode[3];	//	Two-letter language code, such as "en" or "ja", with terminating zero
	unsigned int	itsPuzzleSize,
					itsIndexTorus,
					itsIndexKlein;
	Char16			itsBoard[MAX_WORDSEARCH_SIZE][MAX_WORDSEARCH_SIZE];
	bool			itsBoardFlips[MAX_WORDSEARCH_SIZE][MAX_WORDSEARCH_SIZE];
	Char16			itsWords[MAX_WORDSEARCH_NUM_WORDS][MAX_WORDSEARCH_WORD_LENGTH + 1];
	bool			itsWordSelectionIsPending;
	unsigned int	itsStartH,
					itsStartV;
	double			itsDeltaH,
					itsDeltaV;
	double			itsDragStartH,				//	In the touch-based interface, a move may
					itsDragStartV;				//		consist of several separate drags,
	bool			itsDragWasTap;				//		concluded with a simple tap.
	float			itsColor[4];	//	premultiplied alpha
	unsigned int	itsNumLinesFound;
	WordSearchLine	itsLines[MAX_WORDSEARCH_NUM_WORDS];
	unsigned int	itsTextureNameBackground,
					itsTextureNameMark,
					itsTextureNameCell[MAX_WORDSEARCH_SIZE][MAX_WORDSEARCH_SIZE];
} WordSearchData;


//	for Jigsaw

#define MAX_JIGSAW_SIZE	5

typedef enum
{
	TabFemale	= 0,
	TabMale		= 1
} TabGender;

#define OPPOSITE_GENDER(g)	((TabGender)!(unsigned int)(g))

#define BORDER_WEST		0x08
#define BORDER_EAST		0x04
#define BORDER_SOUTH	0x02
#define BORDER_NORTH	0x01

typedef struct _JigsawPiece JigsawPiece;
struct _JigsawPiece
{
	unsigned int	itsLogicalH,			//	logical placement in puzzle
					itsLogicalV;			//		= 0, 1, ..., PuzzleSize - 1
	Placement		itsPlacement;			//	current placement on board
	JigsawPiece		*itsNeighborWest,		//	pointers to the neighbors
					*itsNeighborEast,		//		in the Klein bottle, west-east
					*itsNeighborSouth,		//		gets defined in the obvious way,
					*itsNeighborNorth;		//		relative to the standard fundamental domain
	bool			itsNeighborBondWest,	//	attached to neighboring pieces?
					itsNeighborBondEast,	//		(values must be 0 or 1 because we use them as array indices)
					itsNeighborBondSouth,
					itsNeighborBondNorth;
	TabGender		itsTabGenderWest,
					itsTabGenderEast,
					itsTabGenderSouth,
					itsTabGenderNorth;
	bool			itsClumpIsNonorientable,
					itsVisitedFlag,			//	for local use in managing a recursion
					itsReflectionFlag;		//	also for local use
};

typedef struct
{
	unsigned int	itsSize;
	JigsawPiece		itsPieces[MAX_JIGSAW_SIZE * MAX_JIGSAW_SIZE],
					*itsDrawOrder[MAX_JIGSAW_SIZE * MAX_JIGSAW_SIZE];
	ImageRGBA		*itsPuzzleImageWithBorders,
					*itsPuzzleImageWithoutBorders;
	Byte			*itsPuzzleImagePixelAssignments,	//	Indices of each pixel's containing JigsawPiece, as hhhhvvvv
					*itsPuzzleImagePixelBorderFlags;	//	Which piece-borders each pixel belongs to, as 0000wesn
	double			itsPuzzleImageFirstCenter,			//	For locating piece images in the puzzle image texture,
					itsPuzzleImageStride;				//		in [0.0, 2.0] coordinates.
	unsigned int	itsCurrentTorusPuzzle,
					itsCurrentKleinPuzzle;
	bool			itsPuzzleTextureIsUpToDate;
	JigsawPiece		*itsDragPiece;
	unsigned int	itsTextureNameBackground,
					itsTextureNamePuzzle;
} JigsawData;


//	for Chess

typedef struct
{
	unsigned int	itsStartH,
					itsStartV;
	signed int		itsDeltaH,
					itsDeltaV;
	unsigned int	itsFinalH,
					itsFinalV;
} ChessMove;

typedef struct
{
	unsigned int	itsSquares[8][8];			//	[ h (rightward 0..7) ][ v (downward 0..7) ]
	unsigned int	itsWhoseTurn;				//	WHITE or BLACK

	bool			itsUserMoveIsPending,
					itsComputerMoveIsPending;	//	Is the program dragging a piece for the computer's move?
	unsigned int	itsMoveStartH,
					itsMoveStartV;
	Placement		itsMovePlacement;
	double			itsDragStartH,				//	In the touch-based interface, a move may
					itsDragStartV;				//		consist of several separate drags,
	bool			itsDragWasTap;				//		concluded with a simple tap.

	bool			itsThreadThinkingFlag,		//	for separate thread to signal completion
					itsThreadAbortFlag;			//	for main thread to ask separate thread to terminate ASAP
	ChessMove		itsThreadBestMove;			//	for separate thread to post its output

	ChessMove		itsCheckThreat;				//	Will be displayed iff itsDeltaH != 0 || itsDeltaV != 0.
	bool			itsCheckmateFlag;			//	When game is over, flag will be true for checkmate or false for stalemate.
	unsigned int	itsTextureNameBoard,
					itsTextureNameArrow,
					itsTextureNameTarget,
					itsTextureNameCheckmate,
					itsTextureNameStalemate,
					itsTextureNamePiece[2][6];	//	indexed by i = white/black (i=0,1) and j = piece index - 1 (1 ≤ j+1 ≤ 6)
} ChessData;


//	for Pool

//	Ball  0   is the cue ball.
//	Balls 1-3 are the solid colors.
//	Ball  4   is black (in effect an "eight ball").
//	Balls 5-7 are the striped colors.
#define NUM_POOL_BALLS	8

//	Two players compete in each pool game.
//	If the computer plays, it typically plays for Player B
//	(exception:  the user may switch from human vs. human
//	to human vs. computer mode at any time, in which case
//	the computer might end up playing for Player A).
typedef enum
{
	PoolPlayerA = 0,	//	These values permit toggling with the ! operator.
	PoolPlayerB = 1
} PoolPlayer;

//	The first player to sink a ball gets that type (solids or stripes).
//	Once all balls of that types get sunk, the player shoots
//	for the black ball (the 4-ball).
typedef enum
{
	PoolUndecided,
	PoolSolids,
	PoolStripes,
	PoolBlack
} PoolGoal;

typedef enum
{
	PoolCueNone,		//	Pool cue is not visible
						//		(for example, while balls are rolling)
	PoolCueVisible,		//	Pool cue is visible, but not under direct user control
						//		(for example, while user is scrolling the board)
	PoolCueActive		//	Pool cue is under direct user control
						//		(for example, while user is lining up his/her shot)
} PoolCueStatus;

typedef struct
{
	PoolPlayer		itsWhoseTurn;		//	Is it PoolPlayerA's turn or PoolPlayerB's turn?
	PoolGoal		itsPlayerAGoal,		//	Who's shooting for what?
					itsPlayerBGoal;
	bool			itsWinFlag,			//	Has the current player won?
					itsLoseFlag;		//	Has the current player lost?
	PoolPlayer		itsComputerPlayer;	//	If itsHumanVsComputer is true, is the computer playing for PoolPlayerA or PoolPlayerB?
	bool			itsBallInPlay[NUM_POOL_BALLS];		//	Is the given ball on the table?  If so...
	Placement		itsBallPlacement[NUM_POOL_BALLS];	//	...where is it?
	PoolCueStatus	itsCueStatus;
	Placement		itsCueGrip;			//	The middle part of the cue stick, where the hand grabs it.
	double			itsDragStartH,				//	In the touch-based interface, a cue stick movement may
					itsDragStartV;				//		consist of several separate drags,
	bool			itsDragWasTap;				//		concluded with a simple tap.
	double			itsBallVelocityH[NUM_POOL_BALLS],	//	relative to ball's local coordinates, in units
					itsBallVelocityV[NUM_POOL_BALLS];	//		of distance per second
	bool			itsPassiveScratchFlag,	//	If the user sinks no balls, that's a passive scratch.
					itsActiveScratchFlag;	//	If the user sinks an illegal ball, that's an active scratch.
	unsigned int	itsTextureNameTable,
					itsTextureNamePocket,
					itsTextureNameStick,
					itsTextureNameSight,
					itsTextureNameBall[NUM_POOL_BALLS];
} PoolData;


//	for Apples

//	A single AppleInfo byte records the status of each square
//	in the Apples games.
//	The four low-order bits (bits 0 through 3) record how many
//		of the square's eight neighbors contains a worm.
//	Bit 4 records whether the apple itself has a worm.
//	Bit 5 records whether the apple has been exposed or not.
//	Bit 6 records whether user has marked the cell.
//	Bit 7 records whether the cell is drawn with a flip or not.
//
//	Define masks as unsigned integers rather than signed integers,
//	so that (itsBoard[h][v] & APPLE_NBR_COUNT_MASK) will be unsigned.
#define APPLE_NBR_COUNT_MASK	0x0Fu
#define APPLE_HAS_WORM_MASK		0x10u
#define APPLE_EXPOSED_MASK		0x20u
#define APPLE_MARKED_MASK		0x40u
#define APPLE_FLIPPED_MASK		0x80u
typedef unsigned char AppleStatus;

//	What is the biggest allowable board size?
#define MAX_APPLE_BOARD_SIZE	64

//	The numbers look best at 256×256 resolution,
//	but soak up large amounts of both video RAM and regular RAM.
//	To conserve RAM, let's render the numbers at 128×128,
//	which looks almost as good.
#define APPLES_CHARACTER_TEXTURE_SIZE	128

typedef struct
{
	unsigned int	itsBoardSize;
	AppleStatus		itsBoard[MAX_APPLE_BOARD_SIZE][MAX_APPLE_BOARD_SIZE];
	unsigned int	itsTextureNameBackground,
					itsTextureNameApple,
					itsTextureNameWorm,
					itsTextureNameFaceYucky,
					itsTextureNameFaceHappy,
					itsTextureNameNumeral[9];
} ApplesData;


//	All platform-independent data about the space and
//	how it's displayed live in the ModelData structure.
typedef struct ModelData	ModelData;
struct ModelData
{
	//	Have the various OpenGL elements been pre-initialized?
	bool			itsPreparedExtensions,
					itsPreparedShaders,
					itsPreparedTextures,
					itsPreparedVAOs;
	
	//	Which OpenGL extensions does the current context support?
	bool			itsMultisamplingAvailableFlag,
					itsAnisotropicAvailableFlag,
					itsShadersAvailableFlag;
	
	//	Which OpenGL extensions would the user like enabled?
	bool			itsMultisamplingEnabledFlag,	//	always false
					itsShadersEnabledFlag;			//	true by default, keyboard control only
	
	//	OpenGL shaders and textures.
	unsigned int	itsShaderPrograms[NumShaders],
					itsTextureNamesCommon[NumTexturesCommon],	//	game-independent textures only
						//	Each game defines and manages additional textures as it sees fit.
					itsVertexArrayNames [NumVertexBuffers],
					itsVertexBufferNames[NumVertexBuffers],
					itsIndexBufferNames [NumVertexBuffers];		//	currently unused

	//	Current OpenGL state.
	unsigned int	itsCurrentShaderIndex,
					itsCurrentTextureName,
					itsCurrentVertexBufferIndex;

	//	Synchronize to the monitor's refresh rate?
	bool			itsSyncFlag;
	
	//	When some part of the program (for example the mouse-handling code)
	//	wants to request a redraw, it sets itsRedrawRequestFlag.
	//	The idle-time routine will handle the request and clear the flag.
	//
	//	Note:  It's essential that all redraws funnel through the idle-time routine.
	//	For example, if mouse movements generated redraw events directly, this could
	//	saturate the event loop, thus blocking normal idle-time motion.
	bool			itsRedrawRequestFlag;

	//	Our basic drawing code will draw into a square
	//	fundamental domain with x and y coordinates each
	//	running from -0.5 to +0.5.  The window has the
	//	same dimensions as the fundamental domain,
	//	and initially the two coincide.  However,
	//	the user may choose to scroll the fundamental domain
	//	relative to the window.  Keep track of the offset.
	//	We're still in the same coordinate system, so the
	//	offset will also have x and y components
	//	in the range -0.5 to +0.5.
	//	Also keep track of whether the image is reflected,
	//	as can happen in the Klein bottle.
	Offset			itsOffset;

	//	Which game?
	GameType		itsGame;

	//	Which topology?
	TopologyType	itsTopology;

	//	Which view?
	ViewType		itsView;

	//	Play human vs. computer in tic-tac-toe, chess and pool?
	bool			itsHumanVsComputer;

	//	How difficult in maze, jigsaw and chess?
	//	0 = too easy, 1 = easy, 2 = medium, 3 = hard, 4 = too hard
	unsigned int	itsDifficultyLevel;

	//	Draw the glide reflection axes in Klein bottle mode?
	bool			itsShowGlideAxes;

	//	Has someone won yet?
	bool			itsGameIsOver,
					itsFlashFlag;

	//	Hand cursor support
	HandStatus		itsHandStatus;
	Placement		itsHandPlacement;		//	cursor placement in fundamental domain, [-0.5, +0.5] coordinates
	bool			itsDragIsReversed;		//	Do the dragged object's local coordinates agree or disagree
											//	with the hand's local coordinates?  Individual Games may
											//	choose to use this field or ignore it.
	double			itsDragHistory[DRAG_HISTORY_DEPTH][2];	//	Only GameIntro and GameMaze use this field.

	//	When the left mouse button goes down,
	//	keep track of whether it's a double-click or not,
	//	then respond appropriately when the mouse button comes up.
	bool			itsDoubleClickFlag;
	Placement		itsPrevClickPlacement;	//	for judging double-clicks
	
	//	Touches have momentum (iPhone OS only)
	bool			itsTouchIsCoasting;
	double			itsTouchVelocity[2];

	//	Fonts (maintained by platform-dependent code)
	OpaqueFont		itsFonts[NumberOfFonts];
	
	//	Side window (maintained by platform-dependent code)
	OpaqueWindow	itsSideWindow;
	
	//	Keep track of what simulation (typically none) that's running during idle time.
	//	If so, keep track of how long the current simulation period has been running.
	SimulationType	itsSimulationStatus;
	double			itsSimulationElapsedTime,	//	total elapsed time in seconds
					itsSimulationDeltaTime;		//	time since previous update, in seconds
	
	//	Long computation support
	bool			itsProgressMeterIsActive;
	double			itsProgressMeter;	//	runs from 0.0 (start) to 1.0 (finish)

	//	Game-dependent functions
	//
	//	When a game starts up it sets these function pointers
	//	and when it shuts down it restores them to NULL.
	//
	//	If you prefer object-oriented langauge, you may think of each game
	//	as a subclass of a generic game class.  In effect each game
	//	implements methods that the base class defines only abstractly,
	//	only here we do it in C rather than C++.
	//
	//	If a game doesn't need to implement a particular method,
	//	it may instead set the corresonding pointer to NULL.
	void	(*itsGameShutDown			)(ModelData *md);
	void	(*itsGameSetUpTextures		)(ModelData *md, ErrorText *aFirstError);
	void	(*itsGameShutDownTextures	)(ModelData *md);
	void	(*itsGameReset				)(ModelData *md);
	void	(*itsGameHandMoved			)(ModelData *md);
	bool	(*itsGameDragBegin			)(ModelData *md, bool aShiftOrControlKeyIsDown);
	void	(*itsGameDragObject			)(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
	void	(*itsGameDragEnd			)(ModelData *md);
	void	(*itsGameCharacterInput		)(ModelData *md, Char16 aCharacter);
	void	(*itsGameSpecialCommand		)(ModelData *md);
	void	(*itsGameSimulationUpdate	)(ModelData *md);
	void	(*itsGameDraw				)(ModelData *md);

	//	Game-specific data
	//
	//	A given window supports only one game at a time, 
	//	so we may store the game-specific data in a union.
	//	(I would have preferred an unnamed union, but
	//	standard C does not officially support them.)
	union
	{
		IntroData		Intro;
		TicTacToeData	TicTacToe;
		MazeData		Maze;
		CrosswordData	Crossword;
		WordSearchData	WordSearch;
		JigsawData		Jigsaw;
		ChessData		Chess;
		PoolData		Pool;
		ApplesData		Apples;
	} itsGameOf;

};


//	A dictionary will translate all user-visible phrases
//	to the currently active language.  
//	The dictionary's structure is private to TorusGamesLocalization.c.
//	For all other purposes the following opaque typedef suffices.
typedef struct LocalizedPhrase	LocalizedPhrase;


//	Platform-independent global variables

//	Two-letter language code, such as "en" or "ja".
extern char				gLanguageCode[3];	//	zero-terminated ASCII string

//	Keep a dictionary for translating all user-interface text
//	into the current language.
extern LocalizedPhrase	*gDictionary;

//	Show frame rate?
extern bool				gShowFrameRate;

//	Test for memory leaks.
extern signed int		gMemCount;

//	Enable sound effects?
extern bool				gPlaySounds;


//	Platform-dependent global functions,
//	in TorusGames-Win32-Utilities.c or TorusGames-MacOS-Utilities.c
extern bool			GetBasePath(char *aPath, unsigned int aPathSize);
extern ErrorText	GetFileContents(const char *aPathName, unsigned int *aNumRawBytes, Byte **someRawBytes);
extern void			FreeFileContents(unsigned int *aNumRawBytes, Byte **someRawBytes);
extern void			SetAlphaTextureFromCharacter(unsigned int aTextureName, Char16 aCharacter, OpaqueFont anOpaqueFont, unsigned int aSize, bool aShadersFlag, ErrorText *aFirstError);
extern void			PlayMIDI(const char *aRelativePath, SoundType aSound);
extern void			PlayWAV(const char *aRelativePath, SoundType aSound);
extern void			ClearSoundCache(void);
extern unsigned int	GetNumPuzzles(GameType aGame, TopologyType aTopology);
extern void			GetRawPuzzleData(GameType aGame, TopologyType aTopology, unsigned int aPuzzleIndex, Char16 *aBuffer, unsigned int aBufferLength);
extern void			CrosswordSetClues(ModelData *md, Char16 *aClueAcross, Char16 *aClueDown, Char16 aHotDirection);
extern void			WordSearchSetList(ModelData *md);
extern void			PoolSetStatusMessage(ModelData *md);
extern void			StartNewThread(ModelData *md, void (*aStartFuction)(ModelData *));
extern void			SleepBriefly(void);
extern void			ErrorMessage(ErrorText aMessage, ErrorText aTitle, bool aFatalFlag);

//	Platform-independent global functions

//	in TorusGamesInit.c
extern void			SetUpModelData(ModelData *md);
extern void			ShutDownModelData(ModelData *md);
extern void			ClearFunctionPointers(ModelData *md);
extern ErrorText	SetUpGraphicsAsNeeded(ModelData *md);
extern void			ShutDownGraphicsAsNeeded(ModelData *md);
extern void			SetUpOneTexture(ModelData *md, unsigned int *aTextureName, char *aTextureFileName, TextureBoundary aTextureBoundary, TextureFormat aTextureFormat, ErrorText *aFirstError);
extern ErrorText	ReadImageRGBA(char *aTextureFileName, ImageRGBA **anImageRGBA);
extern void			FreeImageRGBA(ImageRGBA **anImage);
extern void			ShutDownOneTexture(unsigned int *aTextureName);

//	in TorusGamesSimulation.c
extern void			ResetGame(ModelData *md);
extern void			SetUpGame(ModelData *md);
extern void			ShutDownGame(ModelData *md);
extern void			SpecialMenuCommand(ModelData *md);
extern void			ChangeHumanVsComputer(ModelData *md, bool aNewHumanVsComputer);
extern void			ChangeDifficultyLevel(ModelData *md, unsigned int aNewDifficultyLevel);
extern void			ChangeTopology(ModelData *md, TopologyType aNewTopology);
extern void			ChangeView(ModelData *md, ViewType aNewView);
extern void			SimulationBegin(ModelData *md, SimulationType aSimulationStatus);
extern void			SimulationEnd(ModelData *md);
extern bool			SimulationWantsUpdates(ModelData *md);
extern void			SimulationUpdate(ModelData *md, double aFramePeriod);

//	in TorusGamesMouse.c
extern void			MouseDown(ModelData *md, double aMouseH, double aMouseV, bool aShiftOrControlKeyIsDown, bool aTemporalDoubleClick);
extern void			MouseMove(ModelData *md, double aMouseDeltaH, double aMouseDeltaV);
extern void			MouseUp(ModelData *md);
extern void			MouseGone(ModelData *md);
extern void			MouseWheel(ModelData *md, double aDeltaH, double aDeltaV);
extern void			TouchBegan(ModelData *md, double aTouchH, double aTouchV, bool aFlickFlag);
extern void			TouchMoved(ModelData *md, double aTouchDeltaH, double aTouchDeltaV, double aTouchDeltaT);
extern void			TouchEnded(ModelData *md);
extern void			TouchCancelled(ModelData *md);
extern void			TouchMomentum(ModelData *md, double aTimeInterval);
extern void			CharacterInput(ModelData *md, Char16 aCharacter);

//	in TorusGamesGraphics.c
extern char			*GetRendererName(void);
extern void			SetTextureImageFancy(unsigned int aTextureName, unsigned int aWidth, unsigned int aHeight, unsigned int aDepth, Byte *aPixelArray, bool aShadersFlag);
extern bool			IsPowerOfTwo(unsigned int n);
extern void			ResizeGLView(unsigned int aNewViewWidth, unsigned int aNewViewHeight);
extern ErrorText	ReadRGBAPixels(unsigned int aWidth, unsigned int aHeight, Byte *aBuffer);
extern ErrorText	Render(ModelData *md);
extern void			DrawFundamentalDomain(ModelData *md, unsigned int aTextureName, unsigned int aNumReps);
extern void			DrawSprite(ModelData *md, unsigned int aTextureName, Placement *aPlacement);
extern void			DrawColor(ModelData *md, Placement *aPlacement, float aColor[4]);
extern void			DrawColorWithAlphaMask(ModelData *md, unsigned int aTextureName, Placement *aPlacement, float aColor[4]);
extern void			DrawSpriteTransparent(ModelData *md, unsigned int aTextureName, Placement *aPlacement);
extern void			DrawSpriteWithTextureTransform(ModelData *md, Placement *aPlacement, unsigned int aTextureName, TexCoordTransform *aTexCoordTransform);
extern void			DrawKleinAxes(ModelData *md, float aColorA[4], float aColorB[4]);
extern void			DrawSegment(ModelData *md, float aColor[4], unsigned int aLineTextureName, double aCenterH, double aCenterV, double aLength, double aThickness, double anAngle, unsigned int aNumReps);
extern void			DrawSegmentWithEnds(ModelData *md, float aColor[4], unsigned int aLineTextureName, double aCenterH, double aCenterV, double aLength, double aThickness, double anAngle);
extern void			DrawTriangle(ModelData *md, Placement *aPlacement, float aColor[4]);
extern ErrorText	GetErrorString(void);

//	in TorusGamesRectangle.c
extern void			MakeRectangleVAO(unsigned int aVertexArrayName, unsigned int aVertexBufferName, unsigned int anIndexBufferName, unsigned int anOpenGLVersion);
extern void			BindRectangleVAO(unsigned int aVertexArrayName, unsigned int aVertexBufferName, unsigned int anIndexBufferName, unsigned int anOpenGLVersion, bool aShadersFlag);
extern void			DrawRectangleVAO(void);

//	in TorusGamesRectangleWithEndcaps.c
extern void			MakeRectangleWithEndcapsVAO(unsigned int aVertexArrayName, unsigned int aVertexBufferName, unsigned int anIndexBufferName, unsigned int anOpenGLVersion);
extern void			BindRectangleWithEndcapsVAO(unsigned int aVertexArrayName, unsigned int aVertexBufferName, unsigned int anIndexBufferName, unsigned int anOpenGLVersion, bool aShadersFlag);
extern void			DrawRectangleWithEndcapsVAO(void);

//	in TorusGamesTriangle.c
extern void			MakeTriangleVAO(unsigned int aVertexArrayName, unsigned int aVertexBufferName, unsigned int anIndexBufferName, unsigned int anOpenGLVersion);
extern void			BindTriangleVAO(unsigned int aVertexArrayName, unsigned int aVertexBufferName, unsigned int anIndexBufferName, unsigned int anOpenGLVersion, bool aShadersFlag);
extern void			DrawTriangleVAO(void);

//	in TorusGamesUtilities.c
extern void			NormalizePlacement(Placement *aPlacement, TopologyType aTopology);
extern void			NormalizeOffset(Offset *anOffset, TopologyType aTopology);
extern double		ShortestDistance(double h0, double v0, double h1, double v1, TopologyType aTopology, bool *aMirrorFlag);
extern void			PlayTheSound(SoundType aSound);
extern Char16		ToLowerCase(Char16 aLetter);
extern Char16		BaseLetter(Char16 aLetter);
extern Char16		PromoteHalfWidthKatakana(Char16 aCharacter);
extern Char16		PromoteSmallKana(Char16 aCharacter);
extern Char16		ComposeHiragana(Char16 aCharacter, Char16 *aPendingCharacter);
extern Char16		ComposeHangul(Char16 aCharacter, Char16 *aPendingCharacter, HangulSyllable *aPendingHangulSyllable);
extern Char16		QWERTYtoDubeolshik(Char16 aCharacter);
extern Char16		ComposeCyrillic(Char16 anExistingCharacter, Char16 aNewCharacter);
extern Char16		ComposeLatin(Char16 aCharacter, Char16 *aPendingCharacter);
extern ErrorText	GetAbsolutePath(const char *aDirectory, const char *aFileName, char *aPath, unsigned int aPathSize);

//	in TorusGamesOptions.c
extern void			SetShadersEnabledFlag(ModelData *md, bool aShadersEnabledFlag);
extern void			SetShowGlideAxes(ModelData *md, bool aShowGlideAxesChoice);

//	in TorusGamesLocalization.c
extern void			RefreshLocalization(const char aTwoLetterLanguageCode[3]);
extern const char	*GetLocalizedText(const char *aKey);

//	in individual game files (TorusGamesIntro.c, TorusGamesTicTacToe.c, … )
extern void			IntroSetUp(ModelData *md);
extern void			TicTacToeSetUp(ModelData *md);
extern void			MazeSetUp(ModelData *md);
extern void			CrosswordSetUp(ModelData *md);
extern void			WordSearchSetUp(ModelData *md);
extern void			JigsawSetUp(ModelData *md);
extern void			ChessSetUp(ModelData *md);
extern void			PoolSetUp(ModelData *md);
extern void			ApplesSetUp(ModelData *md);
