//	TorusGamesChess.c
//
//	Code specific to the Chess game.
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


//	A single unsigned integer describes the contents
//	of each square on the chessboard via the following flags:

#define EMPTY_SQUARE	0x00000000

#define PIECE_MASK		0x00000007
#define KING			0x00000001
#define QUEEN			0x00000002
#define BISHOP			0x00000003
#define KNIGHT			0x00000004
#define ROOK			0x00000005
#define PAWN			0x00000006

#define COLOR_MASK		0x00000008
#define WHITE			0x00000000
#define BLACK			0x00000008

#define PARITY_MASK		0x00000010
#define PLAIN			0x00000000
#define REFLECTED		0x00000010

//	When animating the computer's move,
//	how fast should the piece travel?
#define CHESS_PIECE_SPEED					2.0		//	squares per second

//	In the touch-based interface, when continuing a previously started move,
//	how close must the new touch be to the selected piece to count as a hit?
#define CHESS_MOVE_CONTINUATION_HIT_RADIUS	0.1

//	In the touch-based interface, the user taps a selected piece to put it down.
//	How much accidental movement should we allow in the definition of a tap?
#define CHESS_TAP_TOLERANCE					0.01


typedef struct
{
	char			itsDirections[8][2];
	unsigned int	itsNumDirections;	//	maximum of 8 valid directions
	bool			itsUnlimitedDistanceFlag;
} MoveRule;

typedef struct
{
	unsigned int	itsMovesConsidered,
					itsMovesTotal;
	double			*itsProgressMeter;
} ChessProgress;


static const MoveRule	gMoveRules[7] =
{
	//	unused
	{
		{{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0}},
		0,
		false
	},
	//	king
	{
		{{-1, 0},{+1, 0},{ 0,-1},{ 0,+1},{-1,-1},{-1,+1},{+1,+1},{+1,-1}},
		8,
		false
	},
	//	queen
	{
		{{-1, 0},{+1, 0},{ 0,-1},{ 0,+1},{-1,-1},{-1,+1},{+1,+1},{+1,-1}},
		8,
		true
	},
	//	bishop
	{
		{{-1,-1},{-1,+1},{+1,+1},{+1,-1}},
		4,
		true
	},
	//	knight
	{
		{{-2,-1},{-2,+1},{+2,-1},{+2,+1},{-1,-2},{+1,-2},{-1,+2},{+1,+2}},
		8,
		false
	},
	//	rook
	{
		{{-1, 0},{+1, 0},{ 0,-1},{ 0,+1}},
		4,
		true
	},
	//	pawn (capturing or not capturing)
	{
		{{-1, 0},{+1, 0},{ 0,-1},{ 0,+1},{-1,-1},{-1,+1},{+1,+1},{+1,-1}},
		8,
		false
	}
};

//	I have no idea how to accurately judge the relative values
//	of the pieces in torus or Klein bottle chess, so let's just
//	use the standard values and hope for the best.
//	The value of the king should exceed the combined value of all other pieces,
//	in effect making the king infinitely valuable.
//	The NO_PLAUSIBLE_MOVE_VALUE should be considerably less than
//	the negative of the king's value, so any other result will be preferred.
//	Define gPieceValues as signed ints for easy negation.
static const signed int	gPieceValues[7] = {0, 1024, 9, 3, 3, 5, 1};
#define NO_PLAUSIBLE_MOVE_VALUE	(-2048)


//	Public functions
static void			ChessShutDown(ModelData *md);
static void			ChessSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void			ChessShutDownTextures(ModelData *md);
static void			ChessReset(ModelData *md);
static bool			ChessDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void			ChessDragObject(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
static void			ChessDragEnd(ModelData *md);
static void			ChessSpecialCommand(ModelData *md);
static void			ChessSimulationUpdate(ModelData *md);
static void			ChessDraw(ModelData *md);

//	Private functions
static void			LetComputerMove1(ModelData *md);
static void			LetComputerMove2(ModelData *md);
static void			LetComputerMove3(ModelData *md);
static void			LetComputerMove4(ModelData *md);
static void			FindHitSquare(Placement *aPlacement, TopologyType aTopology, unsigned int *h, unsigned int *v, bool *aFlipFlag);
static bool			MoveIsPlausible(unsigned int aBoard[8][8], TopologyType aTopology, unsigned int h0, unsigned int v0, unsigned int h1, unsigned int v1, ChessMove *aMove);
static bool			KingWouldBeInCheck(unsigned int aBoard[8][8], TopologyType aTopology, unsigned int h0, unsigned int v0, unsigned int h1, unsigned int v1, unsigned int aThreatenedPlayer, ChessMove *aMove);
static bool			KingIsInCheck(unsigned int aBoard[8][8], TopologyType aTopology, unsigned int aThreatenedPlayer, ChessMove *aMove);
static void			TestForCheck(ModelData *md);
static bool			HasLegalMove(unsigned int aBoard[8][8], TopologyType aTopology, unsigned int aPlayer);
static void			ComputerSelectsMove(ModelData *md);
static signed int	ValueOfSituation(unsigned int aBoard[8][8], TopologyType aTopology, unsigned int aPlayer, unsigned int aRecursionHeight, bool aCheckTestFlag, bool *anAbortFlag, ChessMove *aBestMove, ChessProgress *aProgressReport);
static void			EvaluatePieceMoves(unsigned int aBoard[8][8], TopologyType aTopology, unsigned int aPlayer, unsigned int aRecursionHeight, bool aCheckTestFlag, bool *anAbortFlag, ChessMove *aBestMove, signed int *aBestValue, unsigned int *aBestValueMultiplicity, unsigned int h0, unsigned int v0, ChessProgress *aProgressReport);
static void			TerminateThinkingThread(ModelData *md);
static void			ChessComputerGoesFirst(ModelData *md);
static bool			AnimateComputerMove(ModelData *md);
static void			DrawPiece(ModelData *md, unsigned int h, unsigned int v);


void ChessSetUp(ModelData *md)
{
	unsigned int	i,
					j;

	//	Initialize function pointers.
	md->itsGameShutDown			= &ChessShutDown;
	md->itsGameSetUpTextures	= &ChessSetUpTextures;
	md->itsGameShutDownTextures	= &ChessShutDownTextures;
	md->itsGameReset			= &ChessReset;
	md->itsGameHandMoved		= NULL;
	md->itsGameDragBegin		= &ChessDragBegin;
	md->itsGameDragObject		= &ChessDragObject;
	md->itsGameDragEnd			= &ChessDragEnd;
	md->itsGameCharacterInput	= NULL;
	md->itsGameSpecialCommand	= &ChessSpecialCommand;
	md->itsGameSimulationUpdate	= &ChessSimulationUpdate;
	md->itsGameDraw				= &ChessDraw;
	
	//	Initialize flags.
	md->itsGameOf.Chess.itsUserMoveIsPending		= false;
	md->itsGameOf.Chess.itsComputerMoveIsPending	= false;
	md->itsGameOf.Chess.itsThreadThinkingFlag		= false;
	md->itsGameOf.Chess.itsThreadAbortFlag			= false;
	
	//	Initialize game-specific texture names.
	md->itsGameOf.Chess.itsTextureNameBoard		= 0;
	md->itsGameOf.Chess.itsTextureNameArrow		= 0;
	md->itsGameOf.Chess.itsTextureNameTarget	= 0;
	md->itsGameOf.Chess.itsTextureNameCheckmate	= 0;
	md->itsGameOf.Chess.itsTextureNameStalemate	= 0;
	for (i = 0; i < 2; i++)
		for (j = 0; j < 6; j++)
			md->itsGameOf.Chess.itsTextureNamePiece[i][j] = 0;

	//	Set up the board.
	ChessReset(md);
}


static void ChessShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	ChessShutDownTextures(md);

	//	Terminate the thinking thread, if one is active.
	TerminateThinkingThread(md);
}


static void ChessSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNameBoard,		"Chess/Board.rgba",		TextureWrap,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNameArrow,		"Chess/Arrow.rgba",		TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNameTarget,		"Chess/Target.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNameCheckmate,	"Chess/Checkmate.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNameStalemate,	"Chess/Stalemate.rgba",	TextureClamp,	TextureRGBA,	aFirstError);

	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[0][0],	"Chess/King-Wh.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[0][1],	"Chess/Queen-Wh.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[0][2],	"Chess/Bishop-Wh.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[0][3],	"Chess/Knight-Wh.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[0][4],	"Chess/Rook-Wh.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[0][5],	"Chess/Pawn-Wh.rgba",	TextureClamp,	TextureRGBA,	aFirstError);

	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[1][0],	"Chess/King-Bk.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[1][1],	"Chess/Queen-Bk.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[1][2],	"Chess/Bishop-Bk.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[1][3],	"Chess/Knight-Bk.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[1][4],	"Chess/Rook-Bk.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Chess.itsTextureNamePiece[1][5],	"Chess/Pawn-Bk.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
}


static void ChessShutDownTextures(ModelData *md)
{
	unsigned int	i,
					j;

	ShutDownOneTexture(&md->itsGameOf.Chess.itsTextureNameBoard		);
	ShutDownOneTexture(&md->itsGameOf.Chess.itsTextureNameArrow		);
	ShutDownOneTexture(&md->itsGameOf.Chess.itsTextureNameTarget	);
	ShutDownOneTexture(&md->itsGameOf.Chess.itsTextureNameCheckmate	);
	ShutDownOneTexture(&md->itsGameOf.Chess.itsTextureNameStalemate	);

	for (i = 0; i < 2; i++)
		for (j = 0; j < 6; j++)
			ShutDownOneTexture(&md->itsGameOf.Chess.itsTextureNamePiece[i][j]);
}


static void ChessReset(ModelData *md)
{
	unsigned int	h,
					v;

	//	Cancel any pending user move.
	md->itsGameOf.Chess.itsUserMoveIsPending = false;

	//	Abort any pending computer move.
	TerminateThinkingThread(md);
	SimulationEnd(md);
	md->itsGameOf.Chess.itsComputerMoveIsPending = false;
	
	//	Reset the pieces.

	for (h = 0; h < 8; h++)
		for (v = 0; v < 8; v++)
			md->itsGameOf.Chess.itsSquares[h][v] = EMPTY_SQUARE;

	md->itsGameOf.Chess.itsSquares[1][1] = PLAIN | BLACK | KNIGHT;
	md->itsGameOf.Chess.itsSquares[2][1] = PLAIN | BLACK | PAWN  ;
	md->itsGameOf.Chess.itsSquares[3][1] = PLAIN | BLACK | ROOK  ;
	md->itsGameOf.Chess.itsSquares[1][2] = PLAIN | BLACK | BISHOP;
	md->itsGameOf.Chess.itsSquares[2][2] = PLAIN | BLACK | KING  ;
	md->itsGameOf.Chess.itsSquares[3][2] = PLAIN | BLACK | QUEEN ;
	md->itsGameOf.Chess.itsSquares[1][3] = PLAIN | BLACK | ROOK  ;
	md->itsGameOf.Chess.itsSquares[2][3] = PLAIN | BLACK | PAWN  ;
	md->itsGameOf.Chess.itsSquares[3][3] = PLAIN | BLACK | KNIGHT;

	md->itsGameOf.Chess.itsSquares[5][5] = PLAIN | WHITE | ROOK  ;
	md->itsGameOf.Chess.itsSquares[6][5] = PLAIN | WHITE | PAWN  ;
	md->itsGameOf.Chess.itsSquares[7][5] = PLAIN | WHITE | KNIGHT;
	md->itsGameOf.Chess.itsSquares[5][6] = PLAIN | WHITE | QUEEN ;
	md->itsGameOf.Chess.itsSquares[6][6] = PLAIN | WHITE | KING  ;
	md->itsGameOf.Chess.itsSquares[7][6] = PLAIN | WHITE | BISHOP;
	md->itsGameOf.Chess.itsSquares[5][7] = PLAIN | WHITE | KNIGHT;
	md->itsGameOf.Chess.itsSquares[6][7] = PLAIN | WHITE | PAWN  ;
	md->itsGameOf.Chess.itsSquares[7][7] = PLAIN | WHITE | ROOK  ;

	//	White goes first.
	md->itsGameOf.Chess.itsWhoseTurn = WHITE;

	//	Noone threatens check.
	md->itsGameOf.Chess.itsCheckThreat.itsStartH = 0;
	md->itsGameOf.Chess.itsCheckThreat.itsStartV = 0;
	md->itsGameOf.Chess.itsCheckThreat.itsDeltaH = 0;
	md->itsGameOf.Chess.itsCheckThreat.itsDeltaV = 0;
	md->itsGameOf.Chess.itsCheckThreat.itsFinalH = 0;
	md->itsGameOf.Chess.itsCheckThreat.itsFinalV = 0;

	//	Game is live.
	md->itsGameIsOver = false;
}


static bool ChessDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
	unsigned int	h,
					v;
	bool			theNormalizationReverses;

	UNUSED_PARAMETER(aShiftOrControlKeyIsDown);

	//	The touch-based and mouse-based versions of torus chess handle drags differently.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Touch-based interface
	
	//	If a move is already in progress...
	if (md->itsGameOf.Chess.itsUserMoveIsPending)
	{
		//	...and the user touches the selected piece,
		md->itsDragIsReversed = (	md->itsHandPlacement.itsFlip
								 != md->itsGameOf.Chess.itsMovePlacement.itsFlip);
		if (ShortestDistance(
					md->itsHandPlacement.itsH,
					md->itsHandPlacement.itsV,
					md->itsGameOf.Chess.itsMovePlacement.itsH,
					md->itsGameOf.Chess.itsMovePlacement.itsV,
					md->itsTopology,
					&md->itsDragIsReversed)
				< CHESS_MOVE_CONTINUATION_HIT_RADIUS)
		{
			//	then let the pending move continue.
			//	That is, start a new drag which will be part of the same pending move.
			md->itsGameOf.Chess.itsDragStartH	= md->itsGameOf.Chess.itsMovePlacement.itsH;
			md->itsGameOf.Chess.itsDragStartV	= md->itsGameOf.Chess.itsMovePlacement.itsV;
			md->itsGameOf.Chess.itsDragWasTap	= true;
			return true;
		}
		else	//	But if the user touches elsewhere,
		{
			//	treat the touch as a scroll.
			//	The pre-existing move remains valid but suspended,
			//	awaiting future touches for its final completion.
			return false;
		}
	}
	
	//	No move is in progress, so let the code below start a new one.
#else
	//	Mouse-based interface
	//	Moves are never left in suspension, so no extra code is needed here.
#endif
	
	//	If we return false for any reason, then no move will be pending.
	md->itsGameOf.Chess.itsUserMoveIsPending = false;

	//	If the game is already over, ignore all hits.
	if (md->itsGameIsOver)
		return false;

	//	What square got hit?
	FindHitSquare(&md->itsHandPlacement, md->itsTopology, &h, &v, &theNormalizationReverses);

	//	If the square's empty, report no hit.
	if (md->itsGameOf.Chess.itsSquares[h][v] == EMPTY_SQUARE)
		return false;

	//	Dismiss any previous threat graphics.
	md->itsGameOf.Chess.itsCheckThreat.itsDeltaH	= 0;
	md->itsGameOf.Chess.itsCheckThreat.itsDeltaV	= 0;

	//	If the player clicks on a piece of the wrong color...
	if ((md->itsGameOf.Chess.itsSquares[h][v] & COLOR_MASK)
	 != (md->itsGameOf.Chess.itsWhoseTurn     & COLOR_MASK))
	{
		//	Play a sound to make the error obvious.
		PlayTheSound(SoundChessWrongColor);
		
		//	Do not start a move.
		md->itsGameOf.Chess.itsUserMoveIsPending = false;

		//	Nevertheless accect the click as a drag so that we can highlight
		//	the correct pieces while the mouse remains down, to attract 
		//	the player's attention to them.
		return true;
	}

	md->itsGameOf.Chess.itsUserMoveIsPending = true;

	md->itsGameOf.Chess.itsMoveStartH = h;
	md->itsGameOf.Chess.itsMoveStartV = v;

	md->itsGameOf.Chess.itsMovePlacement.itsH		= -0.5 + 0.125*h;
	md->itsGameOf.Chess.itsMovePlacement.itsV		= -0.5 + 0.125*v;
	md->itsGameOf.Chess.itsMovePlacement.itsFlip	= ((md->itsGameOf.Chess.itsSquares[h][v] & PARITY_MASK) == REFLECTED);
	md->itsGameOf.Chess.itsMovePlacement.itsAngle	= 0.0;
	md->itsGameOf.Chess.itsMovePlacement.itsSizeH	= 0.125;
	md->itsGameOf.Chess.itsMovePlacement.itsSizeV	= 0.125;

	md->itsGameOf.Chess.itsDragStartH	= md->itsGameOf.Chess.itsMovePlacement.itsH;
	md->itsGameOf.Chess.itsDragStartV	= md->itsGameOf.Chess.itsMovePlacement.itsV;
	md->itsGameOf.Chess.itsDragWasTap	= false;	//	A move's first drag never counts as a terminating tap.
													//	Instead, an initial tap serves to select the piece.

	md->itsDragIsReversed = (	md->itsHandPlacement.itsFlip
							  ^ md->itsGameOf.Chess.itsMovePlacement.itsFlip
							  ^ theNormalizationReverses);

	return true;
}


static void ChessDragObject(
	ModelData	*md,
	double		aHandLocalDeltaH,
	double		aHandLocalDeltaV)
{
	double	thePieceLocalDeltaH,
			thePieceLocalDeltaV,
			theHandAmbientDeltaH,
			theHandAmbientDeltaV,
			thePieceAmbientDeltaH,
			thePieceAmbientDeltaV;

	//	Move the hand.

	theHandAmbientDeltaH	= md->itsHandPlacement.itsFlip
								? -aHandLocalDeltaH : aHandLocalDeltaH;
	theHandAmbientDeltaV	= aHandLocalDeltaV;

	md->itsHandPlacement.itsH += theHandAmbientDeltaH;
	md->itsHandPlacement.itsV += theHandAmbientDeltaV;
	NormalizePlacement(&md->itsHandPlacement, md->itsTopology);

	//	Move the piece.

	if (md->itsGameOf.Chess.itsUserMoveIsPending)
	{
		thePieceLocalDeltaH		= md->itsDragIsReversed
									? -aHandLocalDeltaH : aHandLocalDeltaH;
		thePieceLocalDeltaV		= aHandLocalDeltaV;

		thePieceAmbientDeltaH	= md->itsGameOf.Chess.itsMovePlacement.itsFlip
									? -thePieceLocalDeltaH : thePieceLocalDeltaH;
		thePieceAmbientDeltaV	= thePieceLocalDeltaV;

		md->itsGameOf.Chess.itsMovePlacement.itsH += thePieceAmbientDeltaH;
		md->itsGameOf.Chess.itsMovePlacement.itsV += thePieceAmbientDeltaV;
		NormalizePlacement(&md->itsGameOf.Chess.itsMovePlacement, md->itsTopology);

		//	If the piece has travelled a non-trivial distance
		//	from the start of the present drag, then the drag is not a tap.
		//	This is relevant only in the touch-based interface, not the mouse-based one.
		//	In the touch-based interface, a move may consist of several separate drags,
		//	terminated by a tap.
		if (ShortestDistance(
					md->itsGameOf.Chess.itsDragStartH,
					md->itsGameOf.Chess.itsDragStartV,
					md->itsGameOf.Chess.itsMovePlacement.itsH,
					md->itsGameOf.Chess.itsMovePlacement.itsV,
					md->itsTopology,
					NULL)
				> CHESS_TAP_TOLERANCE)
		{
			md->itsGameOf.Chess.itsDragWasTap = false;
		}
	}
}


static void ChessDragEnd(ModelData *md)
{
	unsigned int	h,
					v;
	bool			theNormalizationReverses;

	//	Ignore drags that began on a piece of the wrong color.
	if ( ! md->itsGameOf.Chess.itsUserMoveIsPending )
		return;

	//	The touch-based and mouse-based versions of torus chess handle drags differently.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Touch-based interface
	
	//	If the drag was not a simple tap, then even though the drag is over,
	//	let the move remain pending, awaiting subsequent drags.
	if ( ! md->itsGameOf.Chess.itsDragWasTap )
		return;
	
	//	A simple tap ends both the drag and the move that the drag is a part of.
	//	Continue with the code below.
#else
	//	Mouse-based interface
	//	Each move consists of a single drag.
#endif
	
	//	The move is over.
	md->itsGameOf.Chess.itsUserMoveIsPending = false;
	
	//	What square did we land on?
	FindHitSquare(&md->itsGameOf.Chess.itsMovePlacement, md->itsTopology, &h, &v, &theNormalizationReverses);

	//	If we're back to where we started,
	//	just update the flip and return.
	if (h == md->itsGameOf.Chess.itsMoveStartH
	 && v == md->itsGameOf.Chess.itsMoveStartV)
	{
		if (md->itsGameOf.Chess.itsMovePlacement.itsFlip
		  ^ theNormalizationReverses)
			md->itsGameOf.Chess.itsSquares[h][v] |= PARITY_MASK;
		else
			md->itsGameOf.Chess.itsSquares[h][v] &= (~PARITY_MASK);

		return;
	}

	//	A move is "plausible" if the piece can get from the old square
	//	to the new.  It's "legal" if, in addition, it does not put the
	//	player's king in check.

	//	If the move is implausible or illegal, ignore it.
	//	(Note:  No special action is required to ignore
	//	a move -- it happens automatically!)

	if ( ! MoveIsPlausible(	md->itsGameOf.Chess.itsSquares,
							md->itsTopology,
							md->itsGameOf.Chess.itsMoveStartH,
							md->itsGameOf.Chess.itsMoveStartV,
							h,
							v,
							NULL) )
		return;

	//	If KingWouldBeInCheck() reports a non-zero md->itsGameOf.Chess.itsCheckThreat,
	//	the standard drawing routine ChessDraw() will display the threat automatically.
	if (KingWouldBeInCheck(	md->itsGameOf.Chess.itsSquares,
							md->itsTopology,
							md->itsGameOf.Chess.itsMoveStartH,
							md->itsGameOf.Chess.itsMoveStartV,
							h,
							v,
							md->itsGameOf.Chess.itsWhoseTurn,
							&md->itsGameOf.Chess.itsCheckThreat))
	{
		PlayTheSound(SoundChessCheck);
		return;
	}

	//	The move's both plausible and legal, so make it.
	md->itsGameOf.Chess.itsSquares[h][v] = md->itsGameOf.Chess.itsSquares[md->itsGameOf.Chess.itsMoveStartH][md->itsGameOf.Chess.itsMoveStartV];
	if (md->itsGameOf.Chess.itsMovePlacement.itsFlip ^ theNormalizationReverses)
		md->itsGameOf.Chess.itsSquares[h][v] |= PARITY_MASK;
	else
		md->itsGameOf.Chess.itsSquares[h][v] &= (~PARITY_MASK);
	md->itsGameOf.Chess.itsSquares[md->itsGameOf.Chess.itsMoveStartH][md->itsGameOf.Chess.itsMoveStartV] = EMPTY_SQUARE;

	//	Now it's the other player's turn.
	md->itsGameOf.Chess.itsWhoseTurn ^= COLOR_MASK;

	//	Test for check(mate).
	TestForCheck(md);

	//	If we're playing against the computer,
	//	let the computer make its move.
	if (md->itsHumanVsComputer
	 && ! md->itsGameIsOver )
	{
		LetComputerMove1(md);
	}
}


static void LetComputerMove1(ModelData *md)
{
	//	If we just put the computer's king in check...
	if (md->itsGameOf.Chess.itsCheckThreat.itsDeltaH != 0
	 || md->itsGameOf.Chess.itsCheckThreat.itsDeltaV != 0)
	{
		//	...give the "check sound" a moment to play before
		//	letting the computer make its move,
		SimulationBegin(md, SimulationChessWaitToMove);
	}
	else
	{
		//	...otherwise let the computer make its move immediately.
		LetComputerMove2(md);
	}
}


static void LetComputerMove2(ModelData *md)
{
	//	Let the computer do its thinking in a separate thread.
	//	After completing the move it will set itsThreadThinkingFlag = false.
	md->itsGameOf.Chess.itsThreadThinkingFlag	= true;
	md->itsGameOf.Chess.itsThreadAbortFlag		= false;
	StartNewThread(md, &ComputerSelectsMove);
	
	//	Let the main thread keep the user interface responsive
	//	while the separate thread selects a move.
	//	If the user cancels the game (for example if s/he
	//	resets Chess or selects a different game),
	//	set itsThreadAbortFlag to true and wait until
	//	the separate thread sets itsThreadThinkingFlag to false
	//	to signal completion (at which point we are guaranteed
	//	that it's done accessing shared data).
	SimulationBegin(md, SimulationChessChooseComputerMove);
}


static void LetComputerMove3(ModelData *md)
{
	//	The separate thread has terminated normally.
	//	Nevertheless, perform an unnecessary error check.
	if (md->itsGameOf.Chess.itsThreadAbortFlag)
		return;	//	should never occur

	//	Animate the move that the computer has chosen.
	md->itsGameOf.Chess.itsComputerMoveIsPending	= true;
	md->itsGameOf.Chess.itsMoveStartH				= md->itsGameOf.Chess.itsThreadBestMove.itsStartH;
	md->itsGameOf.Chess.itsMoveStartV				= md->itsGameOf.Chess.itsThreadBestMove.itsStartV;
	SimulationBegin(md, SimulationChessAnimateComputerMove);
}


static void LetComputerMove4(ModelData *md)
{
	ChessMove	theMove;

	//	The animation is complete.
	md->itsGameOf.Chess.itsComputerMoveIsPending = false;

	//	For convenience
	theMove = md->itsGameOf.Chess.itsThreadBestMove;

	//	Finalize the move.
	md->itsGameOf.Chess.itsSquares[theMove.itsFinalH][theMove.itsFinalV]
		= md->itsGameOf.Chess.itsSquares[theMove.itsStartH][theMove.itsStartV];
	if (md->itsTopology == TopologyKlein
	 && (((theMove.itsStartV + theMove.itsDeltaV) - theMove.itsFinalV) & 0x00000008) )	//	moved an odd multiple of 8 ?
		md->itsGameOf.Chess.itsSquares[theMove.itsFinalH][theMove.itsFinalV] ^= PARITY_MASK;
	md->itsGameOf.Chess.itsSquares[theMove.itsStartH][theMove.itsStartV] = EMPTY_SQUARE;

	//	It's now the other player's turn.
	md->itsGameOf.Chess.itsWhoseTurn ^= COLOR_MASK;
	
	//	Check?
	TestForCheck(md);
}


static void FindHitSquare(
	Placement		*aPlacement,	//	input
	TopologyType	aTopology,		//	input
	unsigned int	*h,				//	output
	unsigned int	*v,				//	output
	bool			*aFlipFlag)		//	output
{
	//	Guard against "impossible" errors.
	if (aPlacement->itsH < -0.5 || aPlacement->itsH > +0.5
	 || aPlacement->itsV < -0.5 || aPlacement->itsV > +0.5)
	{
		ErrorMessage(	"Invalid input in FindHitSquare() in TorusGamesChess.c.",
						"Internal Error",
						false);
		*h			= 0;
		*v			= 0;
		*aFlipFlag	= false;
		return;
	}

	*h = (unsigned int) floor(8*(aPlacement->itsH + 0.5) + 0.5);
	*v = (unsigned int) floor(8*(aPlacement->itsV + 0.5) + 0.5);
	*aFlipFlag = false;

	if (*v == 8)
	{
		if (aTopology == TopologyKlein)
		{
			*h = 8 - *h;
			*aFlipFlag = true;
		}
		*v = 0;
	}
	if (*h == 8)
		*h = 0;
}


static bool MoveIsPlausible(
	unsigned int	aBoard[8][8],	//	like md->itsGameOf.Chess.itsSquares
	TopologyType	aTopology,
	unsigned int	h0,		//	start point
	unsigned int	v0,
	unsigned int	h1,		//	end point
	unsigned int	v1,
	ChessMove		*aMove)	//	output, may be NULL;  if non-NULL, search for shortest move, interpreting the "zero move" as "no move yet present"
{
	MoveRule		theMoveRule;
	bool			theResult;
	unsigned int	i,
					theCount;
	signed int		h,
					v,
					dh,
					dv,
					theDeltaH,
					theDeltaV;

	//	Just to be safe...
	if (aBoard[h0][v0] == EMPTY_SQUARE)
		return false;

	//	Going nowhere does not count as a plausible move.
	if (h0 == h1 && v0 == v1)
		return false;

	//	A piece can't capture a piece of the same color.
	if (aBoard[h1][v1] != EMPTY_SQUARE
	 && (aBoard[h0][v0] & COLOR_MASK) == (aBoard[h1][v1] & COLOR_MASK))
		return false;

	//	Select the appropriate MoveRule for the given piece.
	theMoveRule = gMoveRules[aBoard[h0][v0] & PIECE_MASK];

	//	Set the default return value.
	theResult = false;

	//	Test all plausible destinations.
	for (i = 0; i < theMoveRule.itsNumDirections; i++)
	{
		h = h0;
		v = v0;
		dh = theMoveRule.itsDirections[i][0];
		dv = theMoveRule.itsDirections[i][1];
		theCount = 0;

		do
		{
			//	Take one step.
			h += dh;
			v += dv;
			theCount++;

			//	Stay within the canonical fundamental domain.
			if (v < 0)
			{
				v += 8;
				if (aTopology == TopologyKlein)
				{
					h = 8 - h;
					dh = -dh;
				}
			}
			if (v > 7)
			{
				v -= 8;
				if (aTopology == TopologyKlein)
				{
					h = 8 - h;
					dh = -dh;
				}
			}
			if (h < 0)
				h += 8;
			if (h > 7)
				h -= 8;

			//	Did we reach the desired ending square?
			if ((unsigned int)h == h1
			 && (unsigned int)v == v1)
			{
				//	For a typical piece we've found our mark,
				//	but a pawn may capture iff it moves on a diagonal.
				if
				(
					(aBoard[h0][v0] & PIECE_MASK) == PAWN
				 &&
						(aBoard[h1][v1] != EMPTY_SQUARE)
					 != (dh != 0 && dv != 0)
				)
					return false;

				//	We've found our mark.
				theResult = true;

				//	Report the move if requested.
				if (aMove != NULL)
				{
					theDeltaH = theCount * theMoveRule.itsDirections[i][0];
					theDeltaV = theCount * theMoveRule.itsDirections[i][1];

					//	If...
					if
					(
						//	...no move is yet present,
						(aMove->itsDeltaH == 0 && aMove->itsDeltaV == 0)
					 ||
						//	...or the existing move is longer,
						(	ABS(aMove->itsDeltaH) + ABS(aMove->itsDeltaV)
						  >    ABS(theDeltaH)     +    ABS(theDeltaV)    )
					)
					{
						//	...then record this move.
						aMove->itsStartH = h0;
						aMove->itsStartV = v0;
						aMove->itsDeltaH = theDeltaH;
						aMove->itsDeltaV = theDeltaV;
						aMove->itsFinalH = h1;
						aMove->itsFinalV = v1;
					}
				}

				//	If we don't care about the particular move,
				//	we can return immediately.  Otherwise keep going,
				//	in case there's a shorter way to get to the same place.
				if (aMove == NULL)
					return true;
			}

		} while (theMoveRule.itsUnlimitedDistanceFlag
			  && aBoard[h][v] == EMPTY_SQUARE);
	}

	return theResult;
}


static bool KingWouldBeInCheck(
	unsigned int	aBoard[8][8],	//	like md->itsGameOf.Chess.itsSquares
	TopologyType	aTopology,
	unsigned int	h0,	//	start point
	unsigned int	v0,
	unsigned int	h1,	//	end point
	unsigned int	v1,
	unsigned int	aThreatenedPlayer,	//	WHITE or BLACK
	ChessMove		*aMove)				//	output;  may be NULL
{
	unsigned int	theRememberedPiece;
	bool			theKingWouldBeInCheck;

	//	Make the proposed move.
	theRememberedPiece	= aBoard[h1][v1];
	aBoard[h1][v1]		= aBoard[h0][v0];
	aBoard[h0][v0]		= EMPTY_SQUARE;

	//	Test whether it puts aThreatenedPlayer's king in check.
	theKingWouldBeInCheck = KingIsInCheck(aBoard, aTopology, aThreatenedPlayer, aMove);

	//	Un-make the proposed move.
	aBoard[h0][v0]		= aBoard[h1][v1];
	aBoard[h1][v1]		= theRememberedPiece;

	return theKingWouldBeInCheck;
}


static bool KingIsInCheck(
	unsigned int	aBoard[8][8],		//	like md->itsGameOf.Chess.itsSquares
	TopologyType	aTopology,
	unsigned int	aThreatenedPlayer,	//	WHITE or BLACK
	ChessMove		*aMove)				//	output;  may be NULL
{
	unsigned int	h,
					v,
					h1,
					v1,
					theOpponent;

	//	For safety, and to keep the compiler happy.
	h1 = 0;
	v1 = 0;

	//	Locate the king.
	for (h = 0; h < 8; h++)
		for (v = 0; v < 8; v++)
			if ((aBoard[h][v] & (COLOR_MASK | PIECE_MASK)) == (aThreatenedPlayer | KING))
			{
				h1 = h;
				v1 = v;
			}

	//	Identify the opponent.
	theOpponent = aThreatenedPlayer ^ COLOR_MASK;

	//	Clear aMove, if present.
	if (aMove != NULL)
	{
		aMove->itsStartH = 0;
		aMove->itsStartV = 0;
		aMove->itsDeltaH = 0;
		aMove->itsDeltaV = 0;
		aMove->itsFinalH = 0;
		aMove->itsFinalV = 0;
	}

	//	Check whether any of the opponent's pieces could capture the king.
	//
	//	Note:  This is a grossly inefficient algorithm for testing for check.
	//	A far more efficient algorithm would be to start with the king and ask
	//
	//	1.	Would a pawn-move   (starting from the king) land on an enemy pawn?
	//	2.	Would a knight-move (starting from the king) land on an enemy knight?
	//	3.	Would a rook-move   (starting from the king) land on an enemy rook or queen?
	//	4.	Would a bishop-move (starting from the king) land on an enemy bishop or queen?
	//
	//	At the moment, however, this routine never gets called
	//	within any time-critical code, so I'm leaving the following algorithm
	//	in place because the code is simpler and shorter.
	//	(But if ever the more efficient algorithm is needed,
	//	it won't be ridiculously complicated either.)

	for (h = 0; h < 8; h++)
		for (v = 0; v < 8; v++)
			if (aBoard[h][v] != EMPTY_SQUARE
			 && (aBoard[h][v] & COLOR_MASK) == theOpponent
			 && MoveIsPlausible(aBoard, aTopology, h, v, h1, v1, aMove))
			{
				//	If we're not recording the move, we're done.
				if (aMove == NULL)
					return true;

				//	Otherwise keep searching in hopes of finding a shorter move.
			}

	if (aMove == NULL)
		return false;
	else
		return aMove->itsDeltaH != 0 || aMove->itsDeltaV != 0;
}


static void TestForCheck(ModelData *md)
{
	if (KingIsInCheck(md->itsGameOf.Chess.itsSquares, md->itsTopology, md->itsGameOf.Chess.itsWhoseTurn, &md->itsGameOf.Chess.itsCheckThreat))
	{
		if (HasLegalMove(md->itsGameOf.Chess.itsSquares, md->itsTopology, md->itsGameOf.Chess.itsWhoseTurn))
		{
			PlayTheSound(SoundChessCheck);
		}
		else
		{
			md->itsGameIsOver						= true;
			md->itsGameOf.Chess.itsCheckmateFlag	= true;
			PlayTheSound(SoundChessCheckmate);
		}
	}
	else
	{
		if ( ! HasLegalMove(md->itsGameOf.Chess.itsSquares, md->itsTopology, md->itsGameOf.Chess.itsWhoseTurn) )
		{
			md->itsGameIsOver						= true;
			md->itsGameOf.Chess.itsCheckmateFlag	= false;
			PlayTheSound(SoundChessStalemate);
		}
	}
}


static bool HasLegalMove(
	unsigned int	aBoard[8][8],	//	like md->itsGameOf.Chess.itsSquares
	TopologyType	aTopology,
	unsigned int	aPlayer)		//	WHITE or BLACK
{
	//	This algorithm is rather inefficient, but that's OK given that
	//	we use it only once per move to detect checkmate/stalemate.
	//	If we ever needed to use it within a time-critical loop, 
	//	we'd need to design a more efficient algorithm.
	//	For present purposes, let's stick with simplicity.

	unsigned int	h0,
					v0,
					h1,
					v1;

	for (h0 = 0; h0 < 8; h0++)
		for (v0 = 0; v0 < 8; v0++)
			if (aBoard[h0][v0] != EMPTY_SQUARE
			 && (aBoard[h0][v0] & COLOR_MASK) == aPlayer)
				for (h1 = 0; h1 < 8; h1++)
					for (v1 = 0; v1 < 8; v1++)
						if (MoveIsPlausible(aBoard, aTopology, h0, v0, h1, v1, NULL)
						 && ! KingWouldBeInCheck(aBoard, aTopology, h0, v0, h1, v1, aPlayer, NULL) )
							return true;

	return false;
}


static void ComputerSelectsMove(ModelData *md)
{
	bool			*theThinkingFlag,
					*theAbortFlag;
	ChessMove		*theBestMove;
	bool			*theProgressMeterFlag;
	double			*theProgressMeter;
	TopologyType	theTopology;
	unsigned int	theWhoseTurn,			//	WHITE or BLACK
					theDifficultyLevel;		//	0…4
	ChessProgress	theDryRunProgressReport,
					theProgressReport;
	double			theDryRunProgressMeter;
	unsigned int	theBoard[8][8],
					h,
					v,
					theRecursionDepth;
	signed int		theValue;

	//	This function will run in its own separate thread,
	//	while the main thread keeps the UI responsive.
	//	If a torus cursor is present, the main thread will
	//	display it as a watch (i.e. a progress meter).

	//	Identify the shared variables
	//	that we'll use to communicate with the main thread.
	theThinkingFlag			= &md->itsGameOf.Chess.itsThreadThinkingFlag;
	theAbortFlag			= &md->itsGameOf.Chess.itsThreadAbortFlag;
	theBestMove				= &md->itsGameOf.Chess.itsThreadBestMove;
	theProgressMeterFlag	= &md->itsProgressMeterIsActive;
	theProgressMeter		= &md->itsProgressMeter;

	//	Make a copy of the chess board
	//	along with other read-only variables.
	theTopology			= md->itsTopology;
	theWhoseTurn		= md->itsGameOf.Chess.itsWhoseTurn;
	theDifficultyLevel	= md->itsDifficultyLevel;
	for (h = 0; h < 8; h++)
		for (v = 0; v < 8; v++)
			theBoard[h][v] = md->itsGameOf.Chess.itsSquares[h][v];

	//	Hereafter we ignore md.
	//	Our only interaction with its contents is
	//	via the shared variables defined previously.

	//	Activate the progress meter.
	*theProgressMeterFlag	= true;
	*theProgressMeter		= 0.0;

	//	Measure progress as a fraction of the total number
	//	of moves to be considered.  Make a dry run at recursion level 0
	//	to count the total number of initial moves to be considered.
	theDryRunProgressReport.itsMovesConsidered	= 0;
	theDryRunProgressReport.itsMovesTotal		= 1;	//	value is irrelevant just so we avoid division by zero
	theDryRunProgressReport.itsProgressMeter	= &theDryRunProgressMeter;
	(void) ValueOfSituation(
			theBoard,
			theTopology,
			theWhoseTurn,
			1,
			false,
			NULL,
			NULL,
			&theDryRunProgressReport);
	theProgressReport.itsMovesConsidered	= 0;
	theProgressReport.itsMovesTotal			= theDryRunProgressReport.itsMovesConsidered;
	theProgressReport.itsProgressMeter		= theProgressMeter;

	//	We're guaranteed to find some legal move -- otherwise we'd
	//	already be checkmated or stalemated and this function wouldn't
	//	have gotten called -- but all the same let's put a default
	//	in place to guard against the impossible.
	//	2007/01/19	Now that anAbortFlag is used, the algorithm may
	//	indeed terminate without finding a legal move.
	theBestMove->itsStartH = 0;
	theBestMove->itsStartV = 0;
	theBestMove->itsDeltaH = 0;
	theBestMove->itsDeltaV = 0;
	theBestMove->itsFinalH = 0;
	theBestMove->itsFinalV = 0;

	//	In almost all cases, ValueOfSituation() will find some
	//	legal move when running at the full depth of the recursion.
	//	Nevertheless, we want to be prepared for the possibility
	//	of a "hopeless" game, for which the recursion returns a value
	//	indicating the inevitable loss of the king (assuming optimal play
	//	on the human's part).  We need to treat such a return value
	//	very carefully, because if it occurs at the top level
	//	of the recursion, it indicates an illegal move (i.e. exposing
	//	the king to check), even though at lower levels in the recursion
	//	it's merely unfortunate, not illegal.
	//
	//	A good way to deal with this situation is to try the recursion
	//	at successively shallower levels.  Almost always it will succeed
	//	at the maximum level (say level 4, for sake of dicussion).
	//	But if it fails at level 4 (meaning no move exists that prevents
	//	loss of the king), then we can try the recursion again at level 3,
	//	and so on until it finally succeeds.  The recursion must succeed
	//	by the time we try level 1, because we know a priori that we are
	//	neither checkmated nor stalemated, i.e. a legal move exists that
	//	doesn't expose the king to check.  One advantage of this scheme
	//	is that it gives the human opponent a chance to go wrong.
	//	That is, if the computer makes a move that looks OK as far
	//	as level 3, the human might make a suboptimal move and miss
	//	the potential "guaranteed win" at level 4.

	//	Difficulty levels 0 through 4 correspond
	//	to recursion depths 2 through 6, respectively.
	//	However, recursion depth 6 is too slow, so suppress it.
	theRecursionDepth = theDifficultyLevel + 2;
	if (theRecursionDepth < 2)	//	Unnecessary test.
		theRecursionDepth = 2;
	if (theRecursionDepth > 5)	//	Suppress recursion depth 6.
		theRecursionDepth = 5;

	do
	{
		//	Let ValueOfSituation() decide the best move.
		//	In theory we could also report the value it returns
		//	(which might give some idea of how well the game
		//	is going from the computer's point of view),
		//	but for now let's just ignore it.
		theValue = ValueOfSituation(
			theBoard,
			theTopology,
			theWhoseTurn,
			theRecursionDepth,
			theRecursionDepth == 1,	//	request special check testing for height 1 recursions only
			theAbortFlag,
			theBestMove,
			&theProgressReport);

	} while (theValue < -gPieceValues[KING]/2
		  && --theRecursionDepth >= 1
		  && ! *theAbortFlag );

	//	De-activate the progress meter.
	*theProgressMeterFlag	= false;
	*theProgressMeter		= 0.0;

	//	We're done.  This thread is about to terminate.
	//	As our last act, let the main thread know it may proceed.
	*theThinkingFlag = false;
}


static signed int ValueOfSituation(
	unsigned int	aBoard[8][8],		//	input; like md->itsGameOf.Chess.itsSquares
	TopologyType	aTopology,			//	input;
	unsigned int	aPlayer,			//	input; BLACK or WHITE
	unsigned int	aRecursionHeight,	//	input
	bool			aCheckTestFlag,		//	input
	bool			*anAbortFlag,		//	input; may be NULL
	ChessMove		*aBestMove,			//	output; may be NULL
	ChessProgress	*aProgressReport)	//	input and output; may be NULL
{
	signed int		theBestValue;				//	the value of an optimal move
	unsigned int	theBestValueMultiplicity,	//	how many different optimal moves share theBestValue
					h,
					v;

	//	Has the user cancelled?
	if (anAbortFlag != NULL && *anAbortFlag)
		return 0;

	//	Thoughts on a fast algorithm:
	//	Rather than testing each move to see whether it leaves
	//	the player's king in check, let's simply ignore any possible
	//	exposure to check and let the recursion catch it at the next
	//	deeper level, where it will manifest itself as an opportunity
	//	for the opponent to capture the player's king.  For this approach
	//	to work, when we reach recursion height 0 we must report
	//	gPieceValues[KING] if the new current player can snag a king.
	//	Otherwise we report 0 to end the recursion.
	//
	//	Hmmm... in fact this call to KingIsInCheck() is the bottleneck.
	//	If the calling routine plans to actually make the dangerous move
	//	(which occurs only at the "too easy" level, where the original recursion
	//	has height 1) it will pass true for aCheckTestFlag and we'll
	//	know to do the test.  Otherwise we won't.
	//	Thus at deeper levels of a typical recursion, we may simply ignore
	//	the danger.  The time saved will go part of the way towards letting
	//	us running a deeper recursion to begin with.
	if (aRecursionHeight == 0)
	{
		if (aCheckTestFlag
		 && KingIsInCheck(aBoard, aTopology, aPlayer^COLOR_MASK, NULL))
			return gPieceValues[KING];
		else
			return 0;
	}

	//	Initialize theBestValue to NO_PLAUSIBLE_MOVE_VALUE, and then
	//	"pass it through" the calls to ValueOfBestPlausibleMove().
	//	If ValueOfBestPlausibleMove() can find a better value,
	//	it will pass back the new value and also update aBestStart
	//	and aBestDelta.
	theBestValue				= NO_PLAUSIBLE_MOVE_VALUE;
	theBestValueMultiplicity	= 0;

	for (h = 0; h < 8; h++)
		for (v = 0; v < 8; v++)
			if (aBoard[h][v] != EMPTY_SQUARE
			 && (aBoard[h][v] & COLOR_MASK) == aPlayer)
				EvaluatePieceMoves(
					aBoard, aTopology, aPlayer, aRecursionHeight, aCheckTestFlag, anAbortFlag,
					aBestMove, &theBestValue, &theBestValueMultiplicity, h, v, aProgressReport);

	//	If no legal move was found, return 0.
	//	I realize that in some situations a stalemate may represent
	//	a great success and in other a great failure, but this
	//	is good enough for a first approximation.
	if (theBestValueMultiplicity == 0)
		return 0;
	else
		return theBestValue;
}


static void EvaluatePieceMoves(
	unsigned int	aBoard[8][8],				//	input; like md->itsGameOf.Chess.itsSquares
	TopologyType	aTopology,					//	input;
	unsigned int	aPlayer,					//	input; BLACK or WHITE
	unsigned int	aRecursionHeight,			//	input; must be ≥ 1
	bool			aCheckTestFlag,				//	input
	bool			*anAbortFlag,				//	input; may be NULL
	ChessMove		*aBestMove,					//	output; may be NULL
	signed int		*aBestValue,				//	input and output
	unsigned int	*aBestValueMultiplicity,	//	input and output
	unsigned int	h0,							//	input
	unsigned int	v0,							//	input
	ChessProgress	*aProgressReport)			//	input and output; may be NULL
{
	MoveRule		theMoveRule;
	unsigned int	i,
					h,
					v,
					theCount,
					theRememberedPiece;
	signed int		hh,
					vv,
					dh,
					dv,
					theCaptureValue,
					theSituationValue,
					theTotalValue;

	//	Just to be safe...
	if (aBoard[h0][v0] == EMPTY_SQUARE
	 || (aBoard[h0][v0] & COLOR_MASK) != aPlayer)
	{		
		return;
	}
	if (aRecursionHeight == 0)
	{
		ErrorMessage(	"Recursion height zero in EvaluatePieceMoves().",
						"Internal Error",
						false);
		return;
	}

	//	Select the appropriate MoveRule for the given piece.
	theMoveRule = gMoveRules[aBoard[h0][v0] & PIECE_MASK];

	//	Test all plausible destinations.
	for (i = 0; i < theMoveRule.itsNumDirections; i++)
	{
		hh = h0;
		vv = v0;
		dh = theMoveRule.itsDirections[i][0];
		dv = theMoveRule.itsDirections[i][1];
		theCount = 0;

		do
		{
			//	Take one step.
			hh += dh;
			vv += dv;
			theCount++;

			//	Stay within the canonical fundamental domain.
			if (vv < 0)
			{
				vv += 8;
				if (aTopology == TopologyKlein)
				{
					hh = 8 - hh;
					dh = -dh;
				}
			}
			if (vv > 7)
			{
				vv -= 8;
				if (aTopology == TopologyKlein)
				{
					hh = 8 - hh;
					dh = -dh;
				}
			}
			if (hh < 0)
				hh += 8;
			if (hh > 7)
				hh -= 8;

			//	Copy the (nonnegative) signed (hh,vv) to the unsigned (h,v).
			h = hh;
			v = vv;

			//	Ignore illegal pawn moves.
			if ((aBoard[h0][v0] & PIECE_MASK) == PAWN
			 && (aBoard[h][v] != EMPTY_SQUARE) != (dh != 0 && dv != 0))
				continue;

			//	Ignore illegal captures.
			if (aBoard[h][v] != EMPTY_SQUARE
			 && (aBoard[h][v] & COLOR_MASK) == aPlayer)
				continue;

			//	Evaluate this move.
			//	Conceptually, the moves value is the sum of the value
			//	of the captured piece (if any) plus the value
			//	of the resulting board situation.

			//		Evaluate the capture value.
			theCaptureValue = gPieceValues[aBoard[h][v] & PIECE_MASK];

			//		Evaluate the situation value.
			if ((aBoard[h][v] & PIECE_MASK) == KING)
				theSituationValue = 0;	//	game over!
			else
			{
				//	Temporarily make the proposed move.
				theRememberedPiece	= aBoard[h ][v ];
				aBoard[h ][v ]		= aBoard[h0][v0];
				aBoard[h0][v0]		= EMPTY_SQUARE;

				//	The situation's value to aPlayer is
				//	the negative of its value to the opponent.
				theSituationValue = - ValueOfSituation( aBoard, aTopology,
										aPlayer^COLOR_MASK,
										aRecursionHeight - 1, aCheckTestFlag, anAbortFlag,
										NULL, NULL);

				//	Un-make the proposed move.
				aBoard[h0][v0]		= aBoard[h ][v ];
				aBoard[h ][v ]		= theRememberedPiece;
			}

			//		Evaluate the overall move.
			theTotalValue = theCaptureValue + theSituationValue;
			if (theTotalValue > *aBestValue)
			{
				*aBestValue				= theTotalValue;
				*aBestValueMultiplicity	= 0;	//	will be incremented momentarily
			}
			if (theTotalValue == *aBestValue)
			{
				(*aBestValueMultiplicity)++;

				//	Choose randomly among the equally good moves.
				if (rand()%(*aBestValueMultiplicity) == 0)
				{
					if (aBestMove != NULL)
					{
						aBestMove->itsStartH = h0;
						aBestMove->itsStartV = v0;
						aBestMove->itsDeltaH = theCount * theMoveRule.itsDirections[i][0];
						aBestMove->itsDeltaV = theCount * theMoveRule.itsDirections[i][1];
						aBestMove->itsFinalH = h;
						aBestMove->itsFinalV = v;
					}
				}
			}

			//	Let the UI know about our progress.
			if (aProgressReport != NULL
			 && aProgressReport->itsMovesTotal != 0)	//	unnecessary but safe
			{
				aProgressReport->itsMovesConsidered++;
				*(aProgressReport->itsProgressMeter)
					= (double) aProgressReport->itsMovesConsidered
					/ (double) aProgressReport->itsMovesTotal;
			}

		} while (theMoveRule.itsUnlimitedDistanceFlag
			  && aBoard[h][v] == EMPTY_SQUARE);
	}
}


static void TerminateThinkingThread(ModelData *md)
{
	//	If a thinking thread exists...
	if (md->itsGameOf.Chess.itsThreadThinkingFlag)
	{
		//	...politely ask it to stop as quickly as possible and then...
		md->itsGameOf.Chess.itsThreadAbortFlag = true;
		
		//	...wait for it to do so.  The reason we wait here
		//	for the thread to stop is that we want to rigorously guarantee
		//	that the main thread can't start a new thinking thread
		//	while the old thinking thread is still using 
		//	the shared communication variables.
		while (md->itsGameOf.Chess.itsThreadThinkingFlag)
			SleepBriefly();
	}
}


static void ChessSpecialCommand(ModelData *md)
{
	ChessComputerGoesFirst(md);
}


static void ChessComputerGoesFirst(ModelData *md)
{
	if (md->itsHumanVsComputer)
	{
		ChessReset(md);
		LetComputerMove1(md);
	}
}


static void ChessSimulationUpdate(ModelData *md)
{
	switch (md->itsSimulationStatus)
	{
		case SimulationChessWaitToMove:
			//	Give the "check sound" a second to play before
			//	letting the computer make its move.
			if (md->itsSimulationElapsedTime >= 1.0)
			{
				SimulationEnd(md);
				LetComputerMove2(md);
			}
			break;
		
		case SimulationChessChooseComputerMove:
		
			//	While the computer chooses its move in a separate thread,
			//	the main thread is keeping the user interface responsive.
			//	This is good but we don't want to hog too many cycles.
			//	So let's let the main thread take a little nap.
			SleepBriefly();

			//	When the computer has finished thinking...
			if ( ! md->itsGameOf.Chess.itsThreadThinkingFlag )
			{
				//	...let it make the chosen move.
				SimulationEnd(md);
				LetComputerMove3(md);
			}
			break;
		
		case SimulationChessAnimateComputerMove:
			//	Draw the moving piece.
			//	When the motion is complete...
			if (AnimateComputerMove(md))
			{
				//	...finalize the move.
				SimulationEnd(md);
				LetComputerMove4(md);
			}
			break;

		default:
			break;
	}
}


static bool AnimateComputerMove(ModelData *md)
{
	ChessMove	theMove;
	double		theTotalDistance,
				theTotalTime,
				dh,
				dv;
	
	//	For convenience
	theMove = md->itsGameOf.Chess.itsThreadBestMove;

	//	How far will the piece move?
	theTotalDistance = sqrt(theMove.itsDeltaH * theMove.itsDeltaH
						  + theMove.itsDeltaV * theMove.itsDeltaV);
	if (theTotalDistance == 0.0)
	{
		ErrorMessage(	"theTotalDistance is zero in AnimateComputerMove().",
						"Internal Error",
						false);
		return true;	//	Give up on the animation.
	};
	
	//	How long would we like it to take?
	theTotalTime = theTotalDistance / CHESS_PIECE_SPEED;
	
	//	Has the full time elapsed?
	if (md->itsSimulationElapsedTime >= theTotalTime)
		return true;	//	The animation is complete.

	//	Compute (dh, dv)...
	
	//	...as a unit vector
	dh = (theMove.itsDeltaH / theTotalDistance);
	dv = (theMove.itsDeltaV / theTotalDistance);
	
	//	...as a velocity vector
	dh *= CHESS_PIECE_SPEED;
	dv *= CHESS_PIECE_SPEED;
	
	//	...as a distance vector
	dh *= md->itsSimulationElapsedTime;
	dv *= md->itsSimulationElapsedTime;

	//	Set up the placement.
	md->itsGameOf.Chess.itsMovePlacement.itsH		= -0.5 + 0.125*(theMove.itsStartH + dh);
	md->itsGameOf.Chess.itsMovePlacement.itsV		= -0.5 + 0.125*(theMove.itsStartV + dv);
	md->itsGameOf.Chess.itsMovePlacement.itsFlip	= ((md->itsGameOf.Chess.itsSquares[theMove.itsStartH][theMove.itsStartV] & PARITY_MASK) == REFLECTED);
	md->itsGameOf.Chess.itsMovePlacement.itsAngle	= 0.0;
	md->itsGameOf.Chess.itsMovePlacement.itsSizeH	= 0.125;
	md->itsGameOf.Chess.itsMovePlacement.itsSizeV	= 0.125;

	//	Normalize to the fundamental domain.
	NormalizePlacement(&md->itsGameOf.Chess.itsMovePlacement, md->itsTopology);
	
	//	Because md->itsSimulationStatus != SimulationNone,
	//	the board will get redrawn.
	//	Because md->itsGameOf.Chess.itsComputerMoveIsPending == true,
	//	the drag will get displayed properly.
	//	No further action is required here!

	return false;	//	Keep going with the animation.
}


static void ChessDraw(ModelData *md)
{
	unsigned int	h,
					v;
	Placement		thePlacement;
	unsigned int	theDragPiece;

	//	Draw the background.
	DrawFundamentalDomain(md, md->itsGameOf.Chess.itsTextureNameBoard, 4);

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(1.0, 0.0, 0.0, 1.0),
					PREMULTIPLY_RGBA(0.0, 0.0, 1.0, 1.0) );

	//	If the user has attempted to drag a piece of the wrong color,
	//	draw halos under the pieces of the right color.
	if (md->itsHandStatus == HandDrag && ! md->itsGameOf.Chess.itsUserMoveIsPending)
		for (h = 0; h < 8; h++)
			for (v = 0; v < 8; v++)
				if (md->itsGameOf.Chess.itsSquares[h][v] != EMPTY_SQUARE
				 && (md->itsGameOf.Chess.itsSquares[h][v] & COLOR_MASK) == md->itsGameOf.Chess.itsWhoseTurn)
				{
					thePlacement.itsH		= -0.5 + h*0.125;
					thePlacement.itsV		= -0.5 + v*0.125;
					thePlacement.itsFlip	= false;
					thePlacement.itsAngle	= 0.0;
					thePlacement.itsSizeH	= 0.125;
					thePlacement.itsSizeV	= 0.125;

					DrawColorWithAlphaMask(	md,
											md->itsTextureNamesCommon[TextureCommonHalo],
											&thePlacement,
											PREMULTIPLY_RGBA(0.0, 1.0, 0.5, 1.0));
				}

	//	Draw the pieces.
	for (h = 0; h < 8; h++)
		for (v = 0; v < 8; v++)
			DrawPiece(md, h, v);

	//	Draw the check, if any.
	//	Note that we draw it after drawing the pieces,
	//	so that it will sit on top of (not underneath) neighboring pieces.
	if (md->itsGameOf.Chess.itsCheckThreat.itsDeltaH != 0
	 || md->itsGameOf.Chess.itsCheckThreat.itsDeltaV != 0)
	{
		//	If the king has been checkmated or stalemated,
		//	don't draw the "target" graphic underneath,
		//	but instead just wait for the checkmate or stalemate
		//	graphic to go on top.
		if ( ! md->itsGameIsOver )
		{
			//	Draw the "target" graphic.
			thePlacement.itsH		= -0.5 + md->itsGameOf.Chess.itsCheckThreat.itsFinalH*0.125;
			thePlacement.itsV		= -0.5 + md->itsGameOf.Chess.itsCheckThreat.itsFinalV*0.125;
			thePlacement.itsFlip	= false;
			thePlacement.itsAngle	= 0.0;
			thePlacement.itsSizeH	= 1.5 * 0.125;
			thePlacement.itsSizeV	= 1.5 * 0.125;
			DrawSprite(md, md->itsGameOf.Chess.itsTextureNameTarget, &thePlacement);

			//	Redraw the threatened piece, if any, on top of the "target".
			DrawPiece(md, md->itsGameOf.Chess.itsCheckThreat.itsFinalH, md->itsGameOf.Chess.itsCheckThreat.itsFinalV);
		}

		//	Draw the attacking piece's arrow no matter what.
		thePlacement.itsH		= -0.5 + md->itsGameOf.Chess.itsCheckThreat.itsStartH*0.125;
		thePlacement.itsV		= -0.5 + md->itsGameOf.Chess.itsCheckThreat.itsStartV*0.125;
		thePlacement.itsFlip	= false;
		thePlacement.itsAngle	= atan2(md->itsGameOf.Chess.itsCheckThreat.itsDeltaV, md->itsGameOf.Chess.itsCheckThreat.itsDeltaH);
		thePlacement.itsSizeH	= 1.5 * 0.125;
		thePlacement.itsSizeV	= 1.5 * 0.125;
		DrawSprite(md, md->itsGameOf.Chess.itsTextureNameArrow, &thePlacement);

		//	Redraw the attacking piece on top of the arrow.
		DrawPiece(md, md->itsGameOf.Chess.itsCheckThreat.itsStartH, md->itsGameOf.Chess.itsCheckThreat.itsStartV);
	}

	//	Draw the checkmate or stalemate graphic if the game is over.
	if (md->itsGameIsOver)
	{
		//	In the case of a checkmate the king is known to be
		//	at (itsGameOf.Chess.itsCheckThreat.itsFinalH, itsGameOf.Chess.itsCheckThreat.itsFinalV)
		//	but in the case of a stalemate we don't know where it is,
		//	so we must find it manually.
		thePlacement.itsH = 0.0;	//	initialize for safety only
		thePlacement.itsV = 0.0;
		for (h = 0; h < 8; h++)
			for (v = 0; v < 8; v++)
				if ((md->itsGameOf.Chess.itsSquares[h][v] & (COLOR_MASK | PIECE_MASK)) == (md->itsGameOf.Chess.itsWhoseTurn | KING))
				{
					thePlacement.itsH = -0.5 + h*0.125;
					thePlacement.itsV = -0.5 + v*0.125;
				}
		thePlacement.itsFlip	= false;
		thePlacement.itsAngle	= 0.0;
		thePlacement.itsSizeH	= 1.5 * 0.125;
		thePlacement.itsSizeV	= 1.5 * 0.125;

		DrawSprite(	md,
					md->itsGameOf.Chess.itsCheckmateFlag ?
						md->itsGameOf.Chess.itsTextureNameCheckmate :
						md->itsGameOf.Chess.itsTextureNameStalemate,
					&thePlacement);
	}

	//	Draw the drag piece, if any.
	if (md->itsGameOf.Chess.itsUserMoveIsPending
	 || md->itsGameOf.Chess.itsComputerMoveIsPending)
	{
		theDragPiece = md->itsGameOf.Chess.itsSquares[md->itsGameOf.Chess.itsMoveStartH][md->itsGameOf.Chess.itsMoveStartV];

#ifdef TORUS_GAMES_FOR_OPENGL_ES
		if ( ! md->itsHumanVsComputer || md->itsGameOf.Chess.itsWhoseTurn == WHITE)
			DrawColorWithAlphaMask(	md,
									md->itsTextureNamesCommon[TextureCommonHalo],
									&md->itsGameOf.Chess.itsMovePlacement,
									PREMULTIPLY_RGBA(0.0, 0.5, 1.0, 1.0));
#endif

		DrawSprite(	md,
					md->itsGameOf.Chess.itsTextureNamePiece
						[ (theDragPiece & COLOR_MASK) >> 3 ]
						[ (theDragPiece & PIECE_MASK) - 1  ],
					&md->itsGameOf.Chess.itsMovePlacement);
	}
}


static void DrawPiece(
	ModelData		*md,
	unsigned int	h,
	unsigned int	v)
{
	Placement	thePlacement;
	void		(*theDrawSpriteFunction)(ModelData *, unsigned int, Placement *);

	if (md->itsGameOf.Chess.itsSquares[h][v] != EMPTY_SQUARE)
	{
		thePlacement.itsH		= -0.5 + h*0.125;
		thePlacement.itsV		= -0.5 + v*0.125;
		thePlacement.itsFlip	= (md->itsGameOf.Chess.itsSquares[h][v] & PARITY_MASK) == REFLECTED;
		thePlacement.itsAngle	= 0.0;
		thePlacement.itsSizeH	= 0.125;
		thePlacement.itsSizeV	= 0.125;

		if
		(
			(
				md->itsGameOf.Chess.itsUserMoveIsPending
			 || md->itsGameOf.Chess.itsComputerMoveIsPending
			)
		 && md->itsGameOf.Chess.itsMoveStartH == h
		 && md->itsGameOf.Chess.itsMoveStartV == v
		)
			theDrawSpriteFunction = DrawSpriteTransparent;
		else
			theDrawSpriteFunction = DrawSprite;

		theDrawSpriteFunction(	md,
								md->itsGameOf.Chess.itsTextureNamePiece
									[ (md->itsGameOf.Chess.itsSquares[h][v] & COLOR_MASK) >> 3 ]
									[ (md->itsGameOf.Chess.itsSquares[h][v] & PIECE_MASK) - 1  ],
								&thePlacement);
	}
}
