//	TorusGamesCrossword.c
//
//	Code specific to the Crossword game.
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	The crossword puzzle's cells are indexed by coordinates (h,v), with
//	h running left-to-right from 0 to n-1 and
//	v running bottom-to-top from 0 to n-1,
//	where n is of course the size of the puzzle.
//	I chose to have v run from bottom to top for consistency
//	with OpenGL coordinates.
//	Cell centers align with the natural n×n grid, so for example
//	cells straddle the edges of the fundamental domain
//	and a single cell sits centered at the fundamental domain's corners.
//	In other words, cell (h,v) sits with its center at
//	( -1/2 + h/n,  -1/2 + v/n ).
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt


#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


#ifdef TORUS_GAMES_FOR_OPENGL_ES
#define CROSSWORD_CHARACTER_TEXTURE_SIZE	CROSSWORD_CHARACTER_TEXTURE_SIZE_MOBILE
#else
#define CROSSWORD_CHARACTER_TEXTURE_SIZE	CROSSWORD_CHARACTER_TEXTURE_SIZE_DESKTOP
#endif

//	In the mobile version of Torus Games the grid lines are very thin
//	and require antialiasing to look good (and in particular to have equal widths).
//	The iPhone OS doesn't support multisampling, and in any case the Torus Games 
//	don't use multisampling but instead rely on texture filtering for all antialiasing.
//	So, to draw a grid line, we'll draw a swath of twice the desired width,
//	but with only the middle half drawn opaque.
//
//	CROSSWORD_GRID_LINE_WIDTH is the width of the final visible grid line,
//	expressed as a fraction of the cell width.
//
//	CROSSWORD_GRID_LINE_WIDTH_X2 is the width of the swath that gets rendered,
//	using a texture whose middle half is opaque and whose two end quarters are transparent.
//	This texturing is important in the mobile version of the torus games,
#define CROSSWORD_GRID_LINE_WIDTH			(1.0/32.0)
#define CROSSWORD_GRID_LINE_WIDTH_X2		(2.0 * CROSSWORD_GRID_LINE_WIDTH)
#define CROSSWORD_CHARACTER_INSET_FACTOR	(1.0 - CROSSWORD_GRID_LINE_WIDTH)


typedef enum
{
	WordHorizontal,		//	→ or ←
	WordVertical		//	always ↓
} WordOrientation;


//	Public functions
static void	CrosswordShutDown(ModelData *md);
static void	CrosswordSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void	CrosswordShutDownTextures(ModelData *md);
static void	CrosswordReset(ModelData *md);
static void	CrosswordHandMoved(ModelData *md);
static bool	CrosswordDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void	CrosswordCharacterInput(ModelData *md, Char16 aCharacter);
static void	CrosswordSimulationUpdate(ModelData *md);
static void	CrosswordDraw(ModelData *md);

//	Private functions
static void	SetHotWordFromCursor(ModelData *md);
static void	SetHotWord(ModelData *md, unsigned int h, unsigned int v, bool aFlip, WordOrientation aWordOrientation);
static bool	MoveToPrecedingNonblackCell(ModelData *md, unsigned int *h, unsigned int *v, bool *aFlip, Char16 aDirection, bool aCyclicFlag);
static bool	MoveToNextNonblackCell(ModelData *md, unsigned int *h, unsigned int *v, bool *aFlip, Char16 aDirection, bool aCyclicFlag);
static void	SetClue(ModelData *md);
static bool	RespondToArrow(ModelData *md, Char16 aCharacter);
static bool	CrosswordIsSolved(ModelData *md);
static void	FixAccentsAsNecessary(ModelData *md);


void CrosswordSetUp(ModelData *md)
{
	unsigned int	h,
					v;

	//	Initialize function pointers.
	md->itsGameShutDown			= &CrosswordShutDown;
	md->itsGameSetUpTextures	= &CrosswordSetUpTextures;
	md->itsGameShutDownTextures	= &CrosswordShutDownTextures;
	md->itsGameReset			= &CrosswordReset;
	md->itsGameHandMoved		= &CrosswordHandMoved;
	md->itsGameDragBegin		= &CrosswordDragBegin;
	md->itsGameDragObject		= NULL;
	md->itsGameDragEnd			= NULL;
	md->itsGameCharacterInput	= &CrosswordCharacterInput;
	md->itsGameSpecialCommand	= NULL;
	md->itsGameSimulationUpdate	= &CrosswordSimulationUpdate;
	md->itsGameDraw				= &CrosswordDraw;

	//	Initialize variables for robust error handling.
	md->itsGameOf.Crossword.itsPuzzleIsValid	= false;
	md->itsGameOf.Crossword.itsPuzzleSize		= 0;
	
	//	Note the initial language.
	md->itsGameOf.Crossword.itsPreviousLanguageCode[0] = gLanguageCode[0];
	md->itsGameOf.Crossword.itsPreviousLanguageCode[1] = gLanguageCode[1];
	md->itsGameOf.Crossword.itsPreviousLanguageCode[2] = 0;

	//	Start at the beginning of each puzzle list.
	md->itsGameOf.Crossword.itsIndexTorus		= 0;
	md->itsGameOf.Crossword.itsIndexKlein		= 0;

	//	Initialize the hot cell, just for good form.
	//	CrosswordReset() will reset it to avoid the black cells.
	md->itsGameOf.Crossword.itsHotCellH		= 0;
	md->itsGameOf.Crossword.itsHotCellV		= 0;
	md->itsGameOf.Crossword.itsHotDirection	= L'→';

	//	Set up an empty board to ensure safe texture initialization.
	for (h = 0; h < MAX_CROSSWORD_SIZE; h++)
		for (v = 0; v < MAX_CROSSWORD_SIZE; v++)
			md->itsGameOf.Crossword.itsBoard[h][v] = ' ';
	
	//	Initialize game-specific texture names.
	md->itsGameOf.Crossword.itsTextureNameGridLine			= 0;
	md->itsGameOf.Crossword.itsTextureNamePendingCharacter	= 0;
	for (h = 0; h < MAX_CROSSWORD_SIZE; h++)
		for (v = 0; v < MAX_CROSSWORD_SIZE; v++)
			md->itsGameOf.Crossword.itsTextureNameCell[h][v] = 0;

	//	Reset the crossword puzzle.
	CrosswordReset(md);
}


static void CrosswordShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	CrosswordShutDownTextures(md);
}


static void CrosswordSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	unsigned int	h,
					v;

	//	Set the fixed texture.
	SetUpOneTexture(md, &md->itsGameOf.Crossword.itsTextureNameGridLine,	"Crossword/GridLine.rgba",	TextureWrap,	TextureAlpha,	aFirstError);

	//	Set the pending-character texture.
	SetUpOneTexture(md, &md->itsGameOf.Crossword.itsTextureNamePendingCharacter,	NULL,				TextureClamp,	TextureAlpha,	aFirstError);
	if (md->itsGameOf.Crossword.itsPendingCharacter != NO_PENDING_CHARACTER)
	{
		SetAlphaTextureFromCharacter(
			md->itsGameOf.Crossword.itsTextureNamePendingCharacter,
			md->itsGameOf.Crossword.itsPendingCharacter,
			md->itsFonts[FontCrosswordPuzzle],
			CROSSWORD_CHARACTER_TEXTURE_SIZE,
			md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
			aFirstError);
	}

	//	Set the textures for the characters in the cells.
	for (h = 0; h < MAX_CROSSWORD_SIZE; h++)
		for (v = 0; v < MAX_CROSSWORD_SIZE; v++)
		{
			SetUpOneTexture(md, &md->itsGameOf.Crossword.itsTextureNameCell[h][v],	NULL,				TextureClamp,	TextureAlpha,	aFirstError);
			SetAlphaTextureFromCharacter(
				md->itsGameOf.Crossword.itsTextureNameCell[h][v],
				md->itsGameOf.Crossword.itsBoard[h][v],
				md->itsFonts[FontCrosswordPuzzle],
				CROSSWORD_CHARACTER_TEXTURE_SIZE,
				md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
				aFirstError);
		}
}


static void CrosswordShutDownTextures(ModelData *md)
{
	unsigned int	h,
					v;

	ShutDownOneTexture(&md->itsGameOf.Crossword.itsTextureNameGridLine			);
	ShutDownOneTexture(&md->itsGameOf.Crossword.itsTextureNamePendingCharacter	);

	for (h = 0; h < MAX_CROSSWORD_SIZE; h++)
		for (v = 0; v < MAX_CROSSWORD_SIZE; v++)
			ShutDownOneTexture(&md->itsGameOf.Crossword.itsTextureNameCell[h][v]);
}


static void CrosswordReset(ModelData *md)
{
	ErrorText		theErrorMessage		= NULL;
	unsigned int	theNumPuzzles,
					*theIndex;
	Char16			theRawPuzzleData[2048],
					*r,				//	the read position
					theDirection;	//	the arrow direction:  ← → ↓
	unsigned int	h,
					v,
					theCharacterCount;
	Byte			theClueCount;

	//	Initialize itsGameOf.Crossword.itsPuzzleIsValid to false
	//	so that if a subroutine encounters an error and displays
	//	a message box, any WM_PAINT messages that get sent
	//	to the main window won't try to draw an invalid puzzle.
	md->itsGameOf.Crossword.itsPuzzleIsValid = false;

	//	Initialize the clues to empty values.
	for (theClueCount = 0; theClueCount < MAX_CROSSWORD_CLUES; theClueCount++)
		for (theCharacterCount = 0; theCharacterCount < MAX_CROSSWORD_CLUE_LENGTH; theCharacterCount++)
			md->itsGameOf.Crossword.itsCluesText[theClueCount][theCharacterCount] = 0;
	for (h = 0; h < MAX_CROSSWORD_SIZE; h++)
		for (v = 0; v < MAX_CROSSWORD_SIZE; v++)
		{
			md->itsGameOf.Crossword.itsClueIndexAcross[h][v] = CROSSWORD_NO_CLUE;
			md->itsGameOf.Crossword.itsClueIndexDown  [h][v] = CROSSWORD_NO_CLUE;
			md->itsGameOf.Crossword.itsLeftwardWord   [h][v] = false;
		}
	
	//	If the user changed languages, reset the puzzle indices to 0 
	//	so the user will get the easiest puzzle in the new language.
	if ( ! SAME_LANGUAGE_CODE(gLanguageCode, md->itsGameOf.Crossword.itsPreviousLanguageCode) )
	{
		md->itsGameOf.Crossword.itsIndexTorus = 0;
		md->itsGameOf.Crossword.itsIndexKlein = 0;

		md->itsGameOf.Crossword.itsPreviousLanguageCode[0] = gLanguageCode[0];
		md->itsGameOf.Crossword.itsPreviousLanguageCode[1] = gLanguageCode[1];
		md->itsGameOf.Crossword.itsPreviousLanguageCode[2] = 0;
	}

	//	Read the raw puzzle data as a string resource.
	theNumPuzzles	= GetNumPuzzles(GameCrossword, md->itsTopology);
	theIndex		= md->itsTopology == TopologyTorus ?
						&md->itsGameOf.Crossword.itsIndexTorus :
						&md->itsGameOf.Crossword.itsIndexKlein;
	*theIndex		%= theNumPuzzles;	//	unnecessary but safe
	GetRawPuzzleData(	GameCrossword,
						md->itsTopology,
						*theIndex,
						theRawPuzzleData,
						BUFFER_LENGTH(theRawPuzzleData));

	//	Increment the index for next time.
	(*theIndex)++;
	*theIndex %= theNumPuzzles;

	//	Did the raw data arrive OK?
	if (theRawPuzzleData[0] == 0)
	{
		theErrorMessage = "Couldn't load raw crossword puzzle data.";
		goto CleanUpCrosswordReset;
	}

	//	Initialize a "read pointer".
	r = theRawPuzzleData;

	//	Confirm the leading 'T' or 'K'.
	if (*r++ != (md->itsTopology == TopologyTorus ? 'T' : 'K'))
	{
		theErrorMessage = "Puzzle data lacks leading 'T' or 'K'.";
		goto CleanUpCrosswordReset;
	}

	//	Parse the puzzle size.
	if (r[0] >= '1' && r[0] <= ('0' + MAX_CROSSWORD_SIZE)
	 && r[1] == '/' && r[2] == '/')
	{
		md->itsGameOf.Crossword.itsPuzzleSize = *r - '0';
		r += 3;	//	Skip past the two separators "//" as well as puzzle size itself.
	}
	else
	{
		theErrorMessage = "Puzzle data must begin with <T|K><puzzle size>//, for example T5// or K4// .  Make sure MAX_CROSSWORD_SIZE is at least as large as your puzzle size.";
		goto CleanUpCrosswordReset;
	}

	//	Read the puzzle solution.
	//
	//	Note #1:  The internal v-coordinate runs bottom-to-top,
	//	while raw puzzle data runs top-to-bottom.
	//
	//	Note #2:  The internal coordinates are ordered (h,v),
	//	so the outer loop iterates over v, even though v is
	//	the second coordinate.
	for (v = md->itsGameOf.Crossword.itsPuzzleSize; v-- > 0; )
	{
		for (h = 0; h < md->itsGameOf.Crossword.itsPuzzleSize; h++)
		{
			if (*r != 0		//	essential for robust string parsing
			 && *r != '/')	//	not essential, but convenient,
							//	  because '/' most likely serves only
							//		as a marker not as puzzle data
			{
				md->itsGameOf.Crossword.itsSolution[h][v] = *r++;
			}
			else
			{
				theErrorMessage = "Incomplete puzzle solution data.";
				goto CleanUpCrosswordReset;
			}
		}

		if (*r++ != '/')
		{
			theErrorMessage = "Missing '/' between puzzle solution lines (or some line is too long).";
			goto CleanUpCrosswordReset;
		}
	}
	if (*r++ != '/')
	{
		theErrorMessage = "Missing '//' at end of puzzle solution.";
		goto CleanUpCrosswordReset;
	}

	//	Read the clues.
	theClueCount = 0;
	while (*r != '/')
	{
		if (theClueCount >= MAX_CROSSWORD_CLUES)
		{
			theErrorMessage = "Too many clues.";
			goto CleanUpCrosswordReset;
		}

		if (*r == 0)
		{
			theErrorMessage = "Missing // at end of last clue.";
			goto CleanUpCrosswordReset;
		}

		//	Read the clue's position and direction.

		h = *r++ - '0';
		if (h >= md->itsGameOf.Crossword.itsPuzzleSize)	//	h is unsigned so h >= 0 is guaranteed
		{
			theErrorMessage = "Clue has bad h coordinate.";
			goto CleanUpCrosswordReset;
		}

		v = *r++ - '0';
		if (v >= md->itsGameOf.Crossword.itsPuzzleSize)	//	v is unsigned so v >= 0 is guaranteed
		{
			theErrorMessage = "Clue has bad v coordinate.";
			goto CleanUpCrosswordReset;
		}

		theDirection = *r++;
		if (theDirection != L'←'
		 && theDirection != L'→'
		 && theDirection != L'↓')
		{
			theErrorMessage = "Clue has no direction arrow.";
			goto CleanUpCrosswordReset;
		}

		//	Read the clue's text.
		theCharacterCount = 0;
		while (*r != '/')
		{
			if (*r == 0)
			{
				theErrorMessage = "Incomplete clue.";
				goto CleanUpCrosswordReset;
			}

			if (theCharacterCount >= MAX_CROSSWORD_CLUE_LENGTH - 1)
			{
				theErrorMessage = "Clue is too long.";
				goto CleanUpCrosswordReset;
			}

			md->itsGameOf.Crossword.itsCluesText[theClueCount][theCharacterCount] = *r++;

			theCharacterCount++;
		}
		md->itsGameOf.Crossword.itsCluesText[theClueCount][theCharacterCount] = 0;
		r++;

		//	Assign the clue to the appropriate puzzle squares.
		switch (theDirection)
		{
			case L'←':
				while (md->itsGameOf.Crossword.itsSolution[h][v] != '*')
				{
					if (md->itsGameOf.Crossword.itsClueIndexAcross[h][v] != CROSSWORD_NO_CLUE)
					{
						theErrorMessage = "Puzzle contains either an unbounded (cyclic) horizontal word or overlapping horizontal clues.";
						goto CleanUpCrosswordReset;
					}

					md->itsGameOf.Crossword.itsClueIndexAcross[h][v] = theClueCount;
					md->itsGameOf.Crossword.itsLeftwardWord   [h][v] = true;
					if (h > 0)
						h--;
					else
						h = md->itsGameOf.Crossword.itsPuzzleSize - 1;
				}
				break;

			case L'→':
				while (md->itsGameOf.Crossword.itsSolution[h][v] != '*')
				{
					if (md->itsGameOf.Crossword.itsClueIndexAcross[h][v] != CROSSWORD_NO_CLUE)
					{
						theErrorMessage = "Puzzle contains either an unbounded (cyclic) horizontal word or overlapping horizontal clues.";
						goto CleanUpCrosswordReset;
					}

					md->itsGameOf.Crossword.itsClueIndexAcross[h][v] = theClueCount;
					if (h < md->itsGameOf.Crossword.itsPuzzleSize - 1)
						h++;
					else
						h = 0;
				}
				break;

			case L'↓':
				while (md->itsGameOf.Crossword.itsSolution[h][v] != '*')
				{
					if (md->itsGameOf.Crossword.itsClueIndexDown[h][v] != CROSSWORD_NO_CLUE)
					{
						theErrorMessage = "Puzzle contains either an unbounded (cyclic) vertical word or overlapping vertical clues.";
						goto CleanUpCrosswordReset;
					}

					md->itsGameOf.Crossword.itsClueIndexDown[h][v] = theClueCount;
					if (v > 0)
						v--;
					else
					{
						v = md->itsGameOf.Crossword.itsPuzzleSize - 1;
						if (md->itsTopology == TopologyKlein)
							h = (h > 0) ? md->itsGameOf.Crossword.itsPuzzleSize - h : 0;
					}
				}
				break;
		}

		//	Increment the count.
		theClueCount++;
	}

	//	Initialize the current board to all ' '.
	for (h = 0; h < md->itsGameOf.Crossword.itsPuzzleSize; h++)
		for (v = 0; v < md->itsGameOf.Crossword.itsPuzzleSize; v++)
		{
			md->itsGameOf.Crossword.itsBoard[h][v]		= ' ';
			md->itsGameOf.Crossword.itsBoardFlips[h][v]	= false;
		}

	//	Initialize the hot word to the first horizontal word in the top row.
	//	(Hmmm... a singleton word is a theoretical possibility.
	//	Fortunately I don't think this occurs in any of the puzzles,
	//	and even if it did it shouldn't be a disaster.)
	h = 0;
	v = md->itsGameOf.Crossword.itsPuzzleSize - 1;
	while (md->itsGameOf.Crossword.itsSolution[h][v] == '*')
	{
		h++;
		if (h == md->itsGameOf.Crossword.itsPuzzleSize)
		{
			theErrorMessage = "Crossword puzzle's top row contains all black squares.";
			goto CleanUpCrosswordReset;
		}
	}

	//	Move the hot cell to the beginning of the hot word.
	//	In a Klein bottle puzzle, the word may begin from either the left or the right.
	while (MoveToPrecedingNonblackCell(	md,
										&h,
										&v,
										NULL,
										md->itsGameOf.Crossword.itsLeftwardWord[h][v] ? L'←' : L'→',
										false))
		;

	//	Color the hot word.  (SetHotWord() will call SetClue() only if the hot cell has changed.)
	SetHotWord(md, h, v, false, WordHorizontal);

	//	Set the clue.  (May or may not be redundant -- see comment immediately above.)
	SetClue(md);
	
	//	Clear itsPendingCharacter and itsPendingHangulSyllable.
	md->itsGameOf.Crossword.itsPendingCharacter			= NO_PENDING_CHARACTER;
	md->itsGameOf.Crossword.itsPendingHangulSyllable	= EMPTY_HANGUL_SYLLABLE;

	//	We'll need new textures.
	md->itsPreparedTextures = false;

	//	Abort any pending simulation.
	SimulationEnd(md);

	//	The game is not over.
	md->itsGameIsOver = false;

CleanUpCrosswordReset:
	if (theErrorMessage != NULL)
	{
		md->itsGameOf.Crossword.itsPuzzleIsValid = false;
		ErrorMessage(theErrorMessage, "CrosswordReset() Error", false);
	}
	else
		md->itsGameOf.Crossword.itsPuzzleIsValid = true;
}


static void CrosswordHandMoved(ModelData *md)
{
	SetHotWordFromCursor(md);
}


static void SetHotWordFromCursor(ModelData *md)
{
	double			theRawH,
					theRawV;
	unsigned int	h,
					v;
	double			theOffsetH,
					theOffsetV;
	bool			theFlip;
	WordOrientation	theWordOrientation;

	//	The hand always has coordinates in the range [-0.5, +0.5],
	//	corresponding to cell coordinates h and v in the range [0, n].

	theRawH = md->itsGameOf.Crossword.itsPuzzleSize*(md->itsHandPlacement.itsH + 0.5);
	theRawV = md->itsGameOf.Crossword.itsPuzzleSize*(md->itsHandPlacement.itsV + 0.5);

	h = (unsigned int) floor(theRawH + 0.5);
	v = (unsigned int) floor(theRawV + 0.5);
	
	theOffsetH = theRawH - h;
	theOffsetV = theRawV - v;
	
	//	Note the tentative hand chirality,
	//	which we'll correct below in the special case of v == n in a Klein bottle.
	theFlip = md->itsHandPlacement.itsFlip;

	//	Wrap the boundary cases  h == n  and/or  v == n .
	//	This code also guards against an "impossible" hand placement,
	//	giving legal (h,v) in all cases.
	if (v >= md->itsGameOf.Crossword.itsPuzzleSize)
	{
		v = 0;
		if (md->itsTopology == TopologyKlein)
		{
			h		= md->itsGameOf.Crossword.itsPuzzleSize - h;
			theFlip	= ! theFlip;
		}
	}
	if (h >= md->itsGameOf.Crossword.itsPuzzleSize)
		h = 0;
	
	//	Ignore hand cursor motion over black squares.
	if (md->itsGameOf.Crossword.itsSolution[h][v] == '*')
		return;
	
	//	Select a horizontal word or a vertical one?
	theWordOrientation = (
		(md->itsGameOf.Crossword.itsLeftwardWord[h][v] ?
			(theOffsetV - theOffsetH >= 0.0) : (theOffsetV + theOffsetH >= 0.0))
		? WordHorizontal : WordVertical);
	
	//	Let the hand cursor pass its chirality to the cell it lies on.
	md->itsGameOf.Crossword.itsBoardFlips[h][v] = theFlip;
	
	//	Set the word the passes through the selected cell in the selected direction.
	SetHotWord(md, h, v, theFlip, theWordOrientation);
}

static void SetHotWord(
	ModelData		*md,
	unsigned int	h,
	unsigned int	v,
	bool			aFlip,
	WordOrientation	aWordOrientation)
{
	Char16			theDirection;	//	← → ↓
	unsigned int	hh,
					vv;

	//	Which direction does the word run?
	if (aWordOrientation == WordHorizontal)
		theDirection = (md->itsGameOf.Crossword.itsLeftwardWord[h][v] ? L'←' : L'→');
	else
		theDirection = L'↓';

	//	If the hot cell hasn't changed, we're done.
	if (md->itsGameOf.Crossword.itsHotCellH		== h
	 && md->itsGameOf.Crossword.itsHotCellV		== v
	 && md->itsGameOf.Crossword.itsHotCellFlip	== aFlip
	 && md->itsGameOf.Crossword.itsHotDirection	== theDirection)
	{
		return;
	}

	//	Set the new hot cell.
	md->itsGameOf.Crossword.itsHotCellH		= h;
	md->itsGameOf.Crossword.itsHotCellV		= v;
	md->itsGameOf.Crossword.itsHotCellFlip	= aFlip;
	md->itsGameOf.Crossword.itsHotDirection	= theDirection;

	//	Clear any previous hot word.
	for (hh = md->itsGameOf.Crossword.itsPuzzleSize; hh-- > 0; )
		for (vv = md->itsGameOf.Crossword.itsPuzzleSize; vv-- > 0; )
			md->itsGameOf.Crossword.itsHotWord[hh][vv] = false;

	//	Mark the hot word's letters.
	hh = h;
	vv = v;
	do
	{
		md->itsGameOf.Crossword.itsHotWord[hh][vv] = true;
		MoveToNextNonblackCell(md, &hh, &vv, NULL, theDirection, true);
	} while (hh != h || vv != v);

	//	Set the clue.
	SetClue(md);
}

static bool MoveToPrecedingNonblackCell(
	ModelData		*md,
	unsigned int	*h,
	unsigned int	*v,
	bool			*aFlip,			//	may be NULL
	Char16			aDirection,		//	forward direction of word at (*h, *v)
	bool			aCyclicFlag)	//	wrap around cyclically?
{
	switch (aDirection)
	{
		case L'←':	return MoveToNextNonblackCell(md, h, v, aFlip, L'→', aCyclicFlag);
		case L'→':	return MoveToNextNonblackCell(md, h, v, aFlip, L'←', aCyclicFlag);
		case L'↓':	return MoveToNextNonblackCell(md, h, v, aFlip, L'↑', aCyclicFlag);
		case L'↑':	return MoveToNextNonblackCell(md, h, v, aFlip, L'↓', aCyclicFlag);
		default:	return false;	//	should never occur
	}
}

static bool MoveToNextNonblackCell(
	ModelData		*md,
	unsigned int	*h,
	unsigned int	*v,
	bool			*aFlip,			//	may be NULL
	Char16			aDirection,
	bool			aCyclicFlag)	//	wrap around cyclically?
{
	unsigned int	n,
					theNextH,
					theNextV;
	bool			theNextFlip;
	
	if (aCyclicFlag)
	{
		//	Handle the cyclic case in terms of the non-cyclic case.

		if ( ! MoveToNextNonblackCell(md, h, v, aFlip, aDirection, false))
		{
			while (MoveToPrecedingNonblackCell(md, h, v, aFlip, aDirection, false))
				;
		}
		
		return true;	//	cyclic case is never blocked
	}
	else
	{
		//	Handle the non-cyclic case.

		n = md->itsGameOf.Crossword.itsPuzzleSize;

		theNextFlip = (aFlip != NULL ? *aFlip : false);

		switch (aDirection)
		{
			case L'←':
				if (*h > 0)
					theNextH = *h - 1;
				else
					theNextH =  n - 1;
				theNextV = *v;
				break;

			case L'→':
				if (*h < n - 1)
					theNextH = *h + 1;
				else
					theNextH = 0;
				theNextV = *v;
				break;

			case L'↓':
				theNextH = *h;
				if (*v > 0)
					theNextV = *v - 1;
				else
				{
					theNextV =  n - 1;
					if (md->itsTopology == TopologyKlein)
					{
						theNextH = (theNextH > 0) ? n - theNextH : 0;
						theNextFlip = ! theNextFlip;
					}
				}
				break;

			case L'↑':
				theNextH = *h;
				if (*v < n - 1)
					theNextV = *v + 1;
				else
				{
					theNextV =  0;
					if (md->itsTopology == TopologyKlein)
					{
						theNextH = (theNextH > 0) ? n - theNextH : 0;
						theNextFlip = ! theNextFlip;
					}
				}
				break;
				
			default:
				return false;	//	should never occur
		}
		
		if (md->itsGameOf.Crossword.itsSolution[theNextH][theNextV] != '*')
		{
			*h = theNextH;
			*v = theNextV;
			if (aFlip != NULL)
				*aFlip = theNextFlip;
			
			return true;
		}
		else
			return false;
	}
}

static void SetClue(
	ModelData	*md)
{
	unsigned int	theClueIndexAcross,
					theClueIndexDown;
	Char16			*theClueAcross,
					*theClueDown;
	
	static Char16	theEmptyClue[1] = {0};

	theClueIndexAcross	= md->itsGameOf.Crossword.itsClueIndexAcross[md->itsGameOf.Crossword.itsHotCellH][md->itsGameOf.Crossword.itsHotCellV];
	if (theClueIndexAcross < MAX_CROSSWORD_CLUES)
		theClueAcross	= md->itsGameOf.Crossword.itsCluesText[theClueIndexAcross];
	else
		theClueAcross	= theEmptyClue;

	theClueIndexDown	= md->itsGameOf.Crossword.itsClueIndexDown  [md->itsGameOf.Crossword.itsHotCellH][md->itsGameOf.Crossword.itsHotCellV];
	if (theClueIndexDown < MAX_CROSSWORD_CLUES)
		theClueDown		= md->itsGameOf.Crossword.itsCluesText[theClueIndexDown];
	else
		theClueDown		= theEmptyClue;

	CrosswordSetClues(md, theClueAcross, theClueDown, md->itsGameOf.Crossword.itsHotDirection);
}


static bool CrosswordDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
	UNUSED_PARAMETER(aShiftOrControlKeyIsDown);

	//	Set the hot cell...
	SetHotWordFromCursor(md);

	//	...and then treat *all* mouse clicks as scrolls.
	return false;
}


static void CrosswordCharacterInput(
	ModelData	*md,
	Char16		aCharacter)
{
	unsigned int	h,
					v;
	bool			theFlip;
	unsigned int	hh,
					vv;
	bool			ff;

	//	Just to be safe...
	if ( ! md->itsGameOf.Crossword.itsPuzzleIsValid
	 || md->itsGameOf.Crossword.itsPuzzleSize < 1
	 || md->itsGameOf.Crossword.itsPuzzleSize > MAX_CROSSWORD_SIZE
	 || md->itsGameOf.Crossword.itsHotCellH >= md->itsGameOf.Crossword.itsPuzzleSize
	 || md->itsGameOf.Crossword.itsHotCellV >= md->itsGameOf.Crossword.itsPuzzleSize)
		return;
	
	//	Let arrows ← ↑ → ↓ move the hot cell.
	if (RespondToArrow(md, aCharacter))
		return;

	//	Ignore character input once the game is solved.
	if (md->itsGameIsOver)
		return;
	
	//	Let a tab '\t' move the hot cell forwards.
	if (aCharacter == L'\t')
	{
		MoveToNextNonblackCell(	md,
								&md->itsGameOf.Crossword.itsHotCellH,
								&md->itsGameOf.Crossword.itsHotCellV,
								&md->itsGameOf.Crossword.itsHotCellFlip,
								md->itsGameOf.Crossword.itsHotDirection,
								true);	//	wrap cyclically
		return;
	}
	
	//	Let a backspace '\b' move the hot cell backwards.
	//	In a moment we'll delete its contents.
	if (aCharacter == L'\b')
	{
		MoveToPrecedingNonblackCell(md,
									&md->itsGameOf.Crossword.itsHotCellH,
									&md->itsGameOf.Crossword.itsHotCellV,
									&md->itsGameOf.Crossword.itsHotCellFlip,
									md->itsGameOf.Crossword.itsHotDirection,
									true);	//	wrap cyclically
	}

	//	For easy legibility
	h		= md->itsGameOf.Crossword.itsHotCellH;
	v		= md->itsGameOf.Crossword.itsHotCellV;
	theFlip	= md->itsGameOf.Crossword.itsHotCellFlip;

	//	If the user typed a backspace '\b', we've already moved the hot cell backwards.
	//	Now delete the hot cell's contents and return.
	if (aCharacter == L'\b')
	{
		aCharacter = L' ';

		md->itsGameOf.Crossword.itsBoard[h][v]		= aCharacter;
		md->itsGameOf.Crossword.itsBoardFlips[h][v]	= theFlip;

		SetAlphaTextureFromCharacter(
			md->itsGameOf.Crossword.itsTextureNameCell[h][v],
			aCharacter,
			md->itsFonts[FontCrosswordPuzzle],
			CROSSWORD_CHARACTER_TEXTURE_SIZE,
			md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
			NULL);

		md->itsGameOf.Crossword.itsPendingCharacter			= NO_PENDING_CHARACTER;
		md->itsGameOf.Crossword.itsPendingHangulSyllable	= EMPTY_HANGUL_SYLLABLE;

		return;
	}

	//	In a Korean puzzle, re-interpret Latin characters by mapping 
	//	the QWERTY keyboard layout to the Dubeolshik (두벌식) layout.
	//
	//		Minor advantage:  Mac users can solve the Korean puzzles
	//			without needing to switch to (or even enable) 
	//			a Korean keyboard layout.
	//
	//		Major advantage:  Windows users, by choosing the Latin keyboard layout,
	//			can assemble jamo into syllables directly within
	//			the crossword puzzle itself (just like on the Mac)
	//			rather than in a separate Hangul composition window.
	//			Either way, the user types as if s/he were using
	//			the Dubeolshik (두벌식) layout.
	//
	//	This operation is case sensitve (for example, 'p' → 'ㅔ' while 'P' → 'ㅖ')
	//	so call QWERTYtoDubeolshik() before calling ToLowerCase().
	if (SAME_LANGUAGE_CODE(gLanguageCode, "ko"))
		aCharacter = QWERTYtoDubeolshik(aCharacter);

	//	Convert all input to lowercase (in languages which make
	//	an uppercase/lowercase distinction -- for all other languages
	//	this call should have no effect).
	aCharacter = ToLowerCase(aCharacter);

	//	In the unlikely (but nevertheless possible) event
	//	that a Japanese user types a half-width katakana,
	//	promote it to a full-width katakana, for example ﾐ → ミ.
	aCharacter = PromoteHalfWidthKatakana(aCharacter);
	
	//	Promote small kana to large kana, for example っ → つ.
	aCharacter = PromoteSmallKana(aCharacter);
	
	//	In a Dutch puzzle, merge a newly arrive 'j' with an existing 'i'
	//	to get the combined character 'ĳ'.
	if (SAME_LANGUAGE_CODE(gLanguageCode, "nl"))
	{
		if (md->itsGameOf.Crossword.itsBoard[h][v] == L'i'
		 && aCharacter == L'j')
		{
			aCharacter = L'ĳ';
		}
	}
	
	//	In a Japanese puzzle, manually compose Latin letters into kana,
	//	as an alternative to the Japanese IME.
	//	This will let people with no Japanese IME use the puzzles,
	//	and may also be more convenient even for Japanese users,
	//	because they won't have to keep hitting return after each letter.
	//
	//	In a Korean puzzle, manually compose jamo into syllables.
	//	This will require the user to use a Hangul keyboard layout.
	//
	//	In a Russian puzzle, transliterate Latin letters
	//	to Cyrillic ('b'→'б', 'zh'→'ж', etc.) as an alternative
	//	to the Russian IME.  This will let people with no Russian IME
	//	use the puzzles.  Most Russians, however, will prefer
	//	direct Cyrillic input.
	//
	//	For Latin-based puzzles, provide some minimal accent support,
	//	but don't pretend to provide full support;  Vietnamese users
	//	may either enter unaccented letters (as is customary in Vietnamese
	//	crosswords, as well as French and Italian ones) and wait 
	//	for the accents to appear at the end, or they may use a Vietnamese IME
	//	in the Windows version if one is available.
	//
	//	In all cases, if aCharacter contributes to itsPendingCharacter
	//	but does not generate a character for immediate use,
	//	the Compose<Script>() function returns a space ' '.
	//	The space ' ' then gets written to the current hot cell,
	//	and serves as a neutral background for displaying itsPendingCharacter.
	//
	//	The Win32 Korean Input Method Editor (IME) passes along
	//	the return or newline used to terminate the final syllable.
	//	To avoid overwriting the next crossword cell, 
	//	ComposeHangul() returns 0 whenever a return or newline
	//	arrives without an internally pending character for it to resolve.
	if (SAME_LANGUAGE_CODE(gLanguageCode, "ja"))
		aCharacter = ComposeHiragana(aCharacter, &md->itsGameOf.Crossword.itsPendingCharacter);
	else
	if (SAME_LANGUAGE_CODE(gLanguageCode, "ko"))
		aCharacter = ComposeHangul(aCharacter, &md->itsGameOf.Crossword.itsPendingCharacter, &md->itsGameOf.Crossword.itsPendingHangulSyllable);
	else
	if (SAME_LANGUAGE_CODE(gLanguageCode, "ru"))
		aCharacter = ComposeCyrillic(md->itsGameOf.Crossword.itsBoard[h][v], aCharacter);
	else
		aCharacter = ComposeLatin(aCharacter, &md->itsGameOf.Crossword.itsPendingCharacter);
	
	//	ComposeHangul() returns 0 when a spurious return or newline arrives.
	if (aCharacter == 0)
		return;
		
	//	Make sure the pending character texture is up-to-date.
	if (md->itsGameOf.Crossword.itsPendingCharacter != NO_PENDING_CHARACTER)
		SetAlphaTextureFromCharacter(
			md->itsGameOf.Crossword.itsTextureNamePendingCharacter,
			md->itsGameOf.Crossword.itsPendingCharacter,
			md->itsFonts[FontCrosswordPuzzle],
			CROSSWORD_CHARACTER_TEXTURE_SIZE,
			md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
			NULL);
		
	//	Auto-advance feature
	//
	//	Users often make the following mistake.  Say the user has already
	//	written the horizontal word "dad", and then clicks on the first 'd'
	//	to find the clue "Man's best friend" for the vertical word.
	//	Seeing "d__", the user naturally types 'o' and 'g', and is
	//	surprised and disappointed to see the result "og_".
	//	The problem is that the 'o' overwrote the 'd' that was already present,
	//	then the 'g' followed where the 'o' should have gone.
	//	The technically correct solution would be either to overwrite the initial 'd',
	//	typing 'd', 'o', 'g', or to click into the middle square and type 'o', 'g'
	//	from there.  Nevertheless, given how common this error is, it seems worth
	//	making an exception to the usual rules to automatically correct it.
	//	More precisely, we make the following auto-advance rule:
	//		
	//		If the current square already contains the correct letter
	//		and the user types a different one (other than a space ' '),
	//		then the hot cell advances to that first empty square and 
	//		the letter gets inserted there.
	//
	//	In general I prefer to avoid ad hoc rules, but in the present case
	//	it seems justified.
	if
	(
		md->itsGameOf.Crossword.itsBoard[h][v] == md->itsGameOf.Crossword.itsSolution[h][v]
	 &&
		aCharacter != md->itsGameOf.Crossword.itsSolution[h][v]
	 &&
		(	
			//	A "true" space is a space with no pending character.
			//	A "true" space should never trigger auto-advance,
			//	but should always erase the cell, even if the existing
			//	character there is correct.
			aCharacter != L' '
		 || md->itsGameOf.Crossword.itsPendingCharacter != NO_PENDING_CHARACTER
		)
	)
	{
		hh = h;
		vv = v;
		ff = theFlip;
		
		while (md->itsGameOf.Crossword.itsBoard[hh][vv] != ' '
			&& MoveToNextNonblackCell(	md,
										&hh,
										&vv,
										&ff,
										md->itsGameOf.Crossword.itsHotDirection,
										false))
			;
		
		if (md->itsGameOf.Crossword.itsBoard[hh][vv] == ' ')
		{
			md->itsGameOf.Crossword.itsHotCellH		= hh;
			md->itsGameOf.Crossword.itsHotCellV		= vv;
			md->itsGameOf.Crossword.itsHotCellFlip	= ff;

			h		= hh;
			v		= vv;
			theFlip	= ff;
		}
	}

	//	Record aCharacter in the current hot cell.
	md->itsGameOf.Crossword.itsBoard[h][v]		= aCharacter;
	md->itsGameOf.Crossword.itsBoardFlips[h][v]	= theFlip;
	SetAlphaTextureFromCharacter(
		md->itsGameOf.Crossword.itsTextureNameCell[h][v],
		aCharacter,
		md->itsFonts[FontCrosswordPuzzle],
		CROSSWORD_CHARACTER_TEXTURE_SIZE,
		md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
		NULL);
	
	//	If a complete character arrived (not just a contribution to the pending character),
	//	then advance the hot cell.
	//	Exception:  In a Dutch puzzle, if the user typed an 'i' in a cell 
	//	whose solution character is 'ĳ', leave the hot cell there
	//	so the user may easily type a 'j' on top of the 'i'.
	if (aCharacter != ' '
	 && (	md->itsGameOf.Crossword.itsSolution[h][v] != L'ĳ'
		 || aCharacter != L'i'))
	{
		MoveToNextNonblackCell(	md,
								&md->itsGameOf.Crossword.itsHotCellH,
								&md->itsGameOf.Crossword.itsHotCellV,
								&md->itsGameOf.Crossword.itsHotCellFlip,
								md->itsGameOf.Crossword.itsHotDirection,
								true);
	}

	//	Is that a win?
	if (CrosswordIsSolved(md))
	{
		md->itsGameIsOver = true;
		PlayTheSound(SoundCrosswordComplete);
		FixAccentsAsNecessary(md);
		SimulationBegin(md, SimulationCrosswordFlash);
	}
}


static bool RespondToArrow(
	ModelData	*md,
	Char16		aCharacter)
{
	switch (aCharacter)
	{
		case L'←':
		case L'↑':
			//	Move to the preceding cell.
			MoveToPrecedingNonblackCell(md,
										&md->itsGameOf.Crossword.itsHotCellH,
										&md->itsGameOf.Crossword.itsHotCellV,
										&md->itsGameOf.Crossword.itsHotCellFlip,
										md->itsGameOf.Crossword.itsHotDirection,
										true);	//	wrap cyclically
			return true;

		case L'→':
		case L'↓':
			//	Move to the next cell.
			MoveToNextNonblackCell(	md,
									&md->itsGameOf.Crossword.itsHotCellH,
									&md->itsGameOf.Crossword.itsHotCellV,
									&md->itsGameOf.Crossword.itsHotCellFlip,
									md->itsGameOf.Crossword.itsHotDirection,
									true);	//	wrap cyclically
			return true;
		
		default:
			return false;
	}
}


static bool CrosswordIsSolved(ModelData *md)
{
	unsigned int	h,
					v;

	for (h = 0; h < md->itsGameOf.Crossword.itsPuzzleSize; h++)
		for (v = 0; v < md->itsGameOf.Crossword.itsPuzzleSize; v++)
			if (md->itsGameOf.Crossword.itsSolution[h][v] != '*'
			 && BaseLetter(md->itsGameOf.Crossword.itsBoard[h][v]) != BaseLetter(md->itsGameOf.Crossword.itsSolution[h][v]))
				return false;

	return true;
}


static void FixAccentsAsNecessary(ModelData *md)
{
	unsigned int	h,
					v;

	for (h = 0; h < md->itsGameOf.Crossword.itsPuzzleSize; h++)
		for (v = 0; v < md->itsGameOf.Crossword.itsPuzzleSize; v++)
			if (md->itsGameOf.Crossword.itsSolution[h][v] != '*'
			 && (	md->itsGameOf.Crossword.itsSolution[h][v] <= L'ぁ'	//	don't replace board's katakana
				 || md->itsGameOf.Crossword.itsSolution[h][v] >= L'ヾ')	//		with solution's hiragana
			 && md->itsGameOf.Crossword.itsBoard[h][v] != md->itsGameOf.Crossword.itsSolution[h][v]
			 && BaseLetter(md->itsGameOf.Crossword.itsBoard[h][v]) == BaseLetter(md->itsGameOf.Crossword.itsSolution[h][v]))
			{
				md->itsGameOf.Crossword.itsBoard[h][v] = md->itsGameOf.Crossword.itsSolution[h][v];
				SetAlphaTextureFromCharacter(
					md->itsGameOf.Crossword.itsTextureNameCell[h][v],
					md->itsGameOf.Crossword.itsBoard[h][v],
					md->itsFonts[FontCrosswordPuzzle],
					CROSSWORD_CHARACTER_TEXTURE_SIZE,
					md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
					NULL);
			}
}


static void CrosswordSimulationUpdate(ModelData *md)
{
	switch (md->itsSimulationStatus)
	{
		case SimulationCrosswordFlash:

			//	Color the background green for three seconds.
			if (md->itsSimulationElapsedTime < 3.0)
			{
				md->itsFlashFlag = true;
			}
			else
			{
				SimulationEnd(md);
				md->itsFlashFlag = false;
			}

			break;

		default:
			break;
	}
}


static void CrosswordDraw(ModelData *md)
{
	double			theCellSize;
	Placement		theCellPlacement;
	unsigned int	h,
					v;
	Placement		theArrowPlacement;

	static float	theHotWordColor[4]			= PREMULTIPLY_RGBA(0.50, 0.75, 1.00, 1.00),
					theHotCellColor[4]			= PREMULTIPLY_RGBA(0.00, 0.50, 1.00, 1.00),
					theArrowColor[4]			= PREMULTIPLY_RGBA(1.00, 1.00, 0.00, 1.00),
					theWhiteColor[4]			= PREMULTIPLY_RGBA(1.00, 1.00, 1.00, 1.00),
					theBlackColor[4]			= PREMULTIPLY_RGBA(0.00, 0.00, 0.00, 1.00),
					thePendingCharColor[4]		= PREMULTIPLY_RGBA(0.00, 0.00, 0.00, 0.50),
					theVictoryColor[4]			= PREMULTIPLY_RGBA(0.25, 1.00, 0.50, 1.00);

	//	The background is a solid color.  Let glClear() fill it efficiently.
	if (md->itsFlashFlag)
		glClearColor(theVictoryColor[0], theVictoryColor[1], theVictoryColor[2], theVictoryColor[3]);
	else
		glClearColor(theWhiteColor[0],   theWhiteColor[1],   theWhiteColor[2],   theWhiteColor[3]  );
	glClear(GL_COLOR_BUFFER_BIT);
	
	//	Just to be safe...
	if ( ! md->itsGameOf.Crossword.itsPuzzleIsValid )
		return;
	
	//	Precompute the cell size.
	theCellSize = 1.0 / md->itsGameOf.Crossword.itsPuzzleSize;
	
	//	Color the hot word cells.
	theCellPlacement.itsFlip	= false;
	theCellPlacement.itsAngle	= 0.0;
	theCellPlacement.itsSizeH	= theCellSize;
	theCellPlacement.itsSizeV	= theCellSize;
	for (h = 0; h < md->itsGameOf.Crossword.itsPuzzleSize; h++)
		for (v = 0; v < md->itsGameOf.Crossword.itsPuzzleSize; v++)
			if ( ! md->itsGameIsOver
			 && md->itsGameOf.Crossword.itsHotWord[h][v])
			{
				theCellPlacement.itsH = -0.5 + h*theCellSize;
				theCellPlacement.itsV = -0.5 + v*theCellSize;

				DrawColor(	md,
							&theCellPlacement,
							(	h == md->itsGameOf.Crossword.itsHotCellH
							 && v == md->itsGameOf.Crossword.itsHotCellV ) ?
									theHotCellColor : theHotWordColor);
			}
	
	//	Draw the grid.
	for (h = 0; h < md->itsGameOf.Crossword.itsPuzzleSize; h++)
	{
		DrawSegment(md,
					theBlackColor,
					md->itsGameOf.Crossword.itsTextureNameGridLine,
					-0.5 + (0.5 + h)*theCellSize,
					0.0,
					1.0,
					CROSSWORD_GRID_LINE_WIDTH_X2*theCellSize,
					0.5*PI,
					0);
	}
	for (v = 0; v < md->itsGameOf.Crossword.itsPuzzleSize; v++)
	{
		DrawSegment(md,
					theBlackColor,
					md->itsGameOf.Crossword.itsTextureNameGridLine,
					0.0,
					-0.5 + (0.5 + v)*theCellSize,
					1.0,
					CROSSWORD_GRID_LINE_WIDTH_X2*theCellSize,
					0.0,
					0);
	}

	//	Draw the black cells.
	theCellPlacement.itsFlip	= false;
	theCellPlacement.itsAngle	= 0.0;
	theCellPlacement.itsSizeH	= theCellSize;
	theCellPlacement.itsSizeV	= theCellSize;
	for (h = 0; h < md->itsGameOf.Crossword.itsPuzzleSize; h++)
		for (v = 0; v < md->itsGameOf.Crossword.itsPuzzleSize; v++)
			if (md->itsGameOf.Crossword.itsSolution[h][v] == '*')
			{
				theCellPlacement.itsH = -0.5 + h*theCellSize;
				theCellPlacement.itsV = -0.5 + v*theCellSize;

				DrawColor(md, &theCellPlacement, theBlackColor);
			}

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(1.0, 0.0, 0.0, 1.0),
					PREMULTIPLY_RGBA(0.0, 0.0, 1.0, 1.0) );

	//	Draw the cell contents.
	theCellPlacement.itsAngle	= 0.0;
	theCellPlacement.itsSizeH	= CROSSWORD_CHARACTER_INSET_FACTOR * theCellSize;
	theCellPlacement.itsSizeV	= CROSSWORD_CHARACTER_INSET_FACTOR * theCellSize;
	for (h = 0; h < md->itsGameOf.Crossword.itsPuzzleSize; h++)
	{
		theCellPlacement.itsH = -0.5 + h*theCellSize;

		for (v = 0; v < md->itsGameOf.Crossword.itsPuzzleSize; v++)
		{
			theCellPlacement.itsV		= -0.5 + v*theCellSize;
			theCellPlacement.itsFlip	= md->itsGameOf.Crossword.itsBoardFlips[h][v];

			//	Draw the character if it's present.
			if (md->itsGameOf.Crossword.itsSolution[h][v] != '*'
			 && md->itsGameOf.Crossword.itsBoard   [h][v] != ' ')
			{
				DrawColorWithAlphaMask(md, md->itsGameOf.Crossword.itsTextureNameCell[h][v], &theCellPlacement, theBlackColor);
			}
			
			//	If this is the hot cell, draw the word direction selection arrows.
			if ( ! md->itsGameIsOver
			 && h == md->itsGameOf.Crossword.itsHotCellH
			 && v == md->itsGameOf.Crossword.itsHotCellV)
			{
				if (md->itsGameOf.Crossword.itsHotDirection == L'↓')
				{
					//	vertical hot word
					theArrowPlacement.itsH		= theCellPlacement.itsH;
					theArrowPlacement.itsV		= theCellPlacement.itsV  -  0.25 * theCellSize;
					theArrowPlacement.itsFlip	= false;
					theArrowPlacement.itsAngle	= -0.5 * PI;
					theArrowPlacement.itsSizeH	= 0.25 * theCellSize;
					theArrowPlacement.itsSizeV	= 0.25 * theCellSize;
				}
				else
				{
					//	horizontal hot word (leftward or rightward)
					theArrowPlacement.itsH		= theCellPlacement.itsH
												+ (md->itsGameOf.Crossword.itsLeftwardWord[h][v] ? -0.25 : 0.25)
												* theCellSize;
					theArrowPlacement.itsV		= theCellPlacement.itsV;
					theArrowPlacement.itsFlip	= md->itsGameOf.Crossword.itsLeftwardWord[h][v];
					theArrowPlacement.itsAngle	= 0.0;
					theArrowPlacement.itsSizeH	= 0.25 * theCellSize;
					theArrowPlacement.itsSizeV	= 0.25 * theCellSize;
				}
				DrawTriangle(md, &theArrowPlacement, theArrowColor);
			}
		}
	}
	
	//	Draw the pending character, if any.
	if (md->itsGameOf.Crossword.itsPendingCharacter != NO_PENDING_CHARACTER
	 && md->itsGameOf.Crossword.itsSolution[md->itsGameOf.Crossword.itsHotCellH][md->itsGameOf.Crossword.itsHotCellV] != '*')
	{
		theCellPlacement.itsH		= -0.5 + md->itsGameOf.Crossword.itsHotCellH * theCellSize;
		theCellPlacement.itsV		= -0.5 + md->itsGameOf.Crossword.itsHotCellV * theCellSize;
		theCellPlacement.itsFlip	= md->itsGameOf.Crossword.itsBoardFlips[md->itsGameOf.Crossword.itsHotCellH][md->itsGameOf.Crossword.itsHotCellV];
		theCellPlacement.itsAngle	= 0.0;
		theCellPlacement.itsSizeH	= CROSSWORD_CHARACTER_INSET_FACTOR * theCellSize;
		theCellPlacement.itsSizeV	= CROSSWORD_CHARACTER_INSET_FACTOR * theCellSize;

		DrawColorWithAlphaMask(md, md->itsGameOf.Crossword.itsTextureNamePendingCharacter, &theCellPlacement, thePendingCharColor);
	}
}
