//	TorusGamesJigsaw.c
//
//	Code specific to the Jigsaw game.
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


//	How close must two pieces be to snap together?
#define JIGSAW_EPSILON	0.02

//	To keep the file size down, the mobile version of Torus Games
//	provides only one jigsaw puzzle of each topology (the best one!).
#ifdef TORUS_GAMES_FOR_OPENGL_ES
#define NUM_TORUS_PUZZLE_IMAGES		2
#define NUM_KLEIN_PUZZLE_IMAGES		1
#else
#define NUM_TORUS_PUZZLE_IMAGES		5
#define NUM_KLEIN_PUZZLE_IMAGES		2
#endif

//	When drawing a piece, expand the bounding rectangle as necessary
//	to accommodate either a black outline (on each TabFemale side)
//	or a tab (on each TabMale side).
#define	MARGIN_FOR_BORDER	1.05
#define MARGIN_FOR_TAB		1.53

//	In order to get the black borders correct at corners
//	(and do so efficiently) we need to know the approximate
//	half width of the black border curve, in piece coordinates.
#define INNER_BORDER	0.488
#define OUTER_BORDER	0.510


//	Send a given clump to the front or the back of the draw order?
typedef enum
{
	SendClumpToFront,
	SendClumpToBack
} SendClumpDirection;


//	Public functions
static void			JigsawShutDown(ModelData *md);
static void			JigsawSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void			JigsawShutDownTextures(ModelData *md);
static void			JigsawReset(ModelData *md);
static bool			JigsawDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void			JigsawDragObject(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
static void			JigsawDragEnd(ModelData *md);
static void			JigsawSpecialCommand(ModelData *md);
static void			JigsawDraw(ModelData *md);

//	Private functions
static ErrorText	RefreshPuzzleTexture(ModelData *md);
static ImageRGBA	*ResampleImage2x(ImageRGBA *aSrcImage, TopologyType aTopology);
static ErrorText	MakeJigsawImage(ImageRGBA *anImageRGBA, unsigned int aPuzzleSize, TopologyType aPuzzleTopology,
						JigsawPiece *aPieceList, ImageRGBA **aPuzzleImageWithBorders, ImageRGBA **aPuzzleImageWithoutBorders,
						Byte **somePuzzleImagePixelAssignments, Byte **somePuzzleImagePixelBorderFlags,
						double *aPuzzleImageFirstCenter, double *aPuzzleImageStride);
static void			SendClumpToFrontOrBack(ModelData *md, unsigned int aPieceIndex, SendClumpDirection aDirection);
static void			MarkClump(JigsawPiece *aPiece);
static void			MoveClump(ModelData *md, JigsawPiece *aPiece, double aDeltaH, double aDeltaV);
static void			SetNeighborPlacements(JigsawPiece *aPiece, TopologyType aTopology);
static bool			BindNearbyNeighbors(ModelData *md);
static void			AlignAllClumps(ModelData *md);
static bool			SetPieceOrientation(JigsawPiece *aPiece, bool aDesiredReflection);
static void			SetNonorientableClump(JigsawPiece *aPiece);
static void			AlignToSymmetryAxes(JigsawPiece *aPiece);
static void			GatherClumps(ModelData *md);
static void			TransferClumpToScratchList(JigsawPiece *aPiece, JigsawPiece **aScratchList, unsigned int *aScratchListIndex);
static bool			PuzzleIsComplete(ModelData *md);
static void			UpdatePuzzleTexture(ModelData *md);
static void			DrawPuzzlePieces(ModelData *md);


void JigsawSetUp(ModelData *md)
{
	//	Initialize function pointers.
	md->itsGameShutDown			= &JigsawShutDown;
	md->itsGameSetUpTextures	= &JigsawSetUpTextures;
	md->itsGameShutDownTextures	= &JigsawShutDownTextures;
	md->itsGameReset			= &JigsawReset;
	md->itsGameHandMoved		= NULL;
	md->itsGameDragBegin		= &JigsawDragBegin;
	md->itsGameDragObject		= &JigsawDragObject;
	md->itsGameDragEnd			= &JigsawDragEnd;
	md->itsGameCharacterInput	= NULL;
	md->itsGameSpecialCommand	= &JigsawSpecialCommand;
	md->itsGameSimulationUpdate	= NULL;
	md->itsGameDraw				= &JigsawDraw;
	
	//	MakeJigsawImage() will cache the processed image with and without black borders.
	md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders		= NULL;
	md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders	= NULL;
	md->itsGameOf.Jigsaw.itsPuzzleImagePixelAssignments	= NULL;
	md->itsGameOf.Jigsaw.itsPuzzleImagePixelBorderFlags	= NULL;
	md->itsGameOf.Jigsaw.itsPuzzleImageFirstCenter		= 0.0;
	md->itsGameOf.Jigsaw.itsPuzzleImageStride			= 0.0;

	//	Start with the last puzzle for each topology,
	//	so the index will wrap around to the first puzzle.
	md->itsGameOf.Jigsaw.itsCurrentTorusPuzzle	= NUM_TORUS_PUZZLE_IMAGES - 1;
	md->itsGameOf.Jigsaw.itsCurrentKleinPuzzle	= NUM_KLEIN_PUZZLE_IMAGES - 1;
	
	//	Initialize game-specific texture names.
	md->itsGameOf.Jigsaw.itsTextureNameBackground	= 0;
	md->itsGameOf.Jigsaw.itsTextureNamePuzzle		= 0;

	//	Reset the puzzle.
	JigsawReset(md);
}


static void JigsawShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	JigsawShutDownTextures(md);

	//	Free the cached puzzle images.
	FreeImageRGBA(&md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders);
	FreeImageRGBA(&md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders);
	FREE_MEMORY_SAFELY(md->itsGameOf.Jigsaw.itsPuzzleImagePixelAssignments);
	FREE_MEMORY_SAFELY(md->itsGameOf.Jigsaw.itsPuzzleImagePixelBorderFlags);
}


static void JigsawSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	SetUpOneTexture(md, &md->itsGameOf.Jigsaw.itsTextureNameBackground, "Jigsaw/Grey.rgba",	TextureWrap,	TextureRGBA,	aFirstError);
	
	//	Initialize an empty texture for the puzzle image, then clear 
	//	the itsPuzzleTextureIsUpToDate flag so the current image data
	//	will get loaded at render time.
	SetUpOneTexture(md, &md->itsGameOf.Jigsaw.itsTextureNamePuzzle,		NULL,				TextureClamp,	TextureRGBA,	aFirstError);
	md->itsGameOf.Jigsaw.itsPuzzleTextureIsUpToDate = false;
}


static void JigsawShutDownTextures(ModelData *md)
{
	ShutDownOneTexture(&md->itsGameOf.Jigsaw.itsTextureNameBackground	);
	ShutDownOneTexture(&md->itsGameOf.Jigsaw.itsTextureNamePuzzle		);
}


static void JigsawReset(ModelData *md)
{
	unsigned int	h,
					v,
					n,
					i,
					hh,
					vv;
	JigsawPiece		*thePiece;
	ErrorText		theErrorMessage;

	//	Advance to the next puzzle image.
	if (md->itsTopology == TopologyTorus)
		md->itsGameOf.Jigsaw.itsCurrentTorusPuzzle = (md->itsGameOf.Jigsaw.itsCurrentTorusPuzzle + 1) % NUM_TORUS_PUZZLE_IMAGES;
	else
		md->itsGameOf.Jigsaw.itsCurrentKleinPuzzle = (md->itsGameOf.Jigsaw.itsCurrentKleinPuzzle + 1) % NUM_KLEIN_PUZZLE_IMAGES;

	//	We'll need to load the new puzzle image and/or process it at the new difficulty level.
	md->itsPreparedTextures = false;

	//	Choose a puzzle size.
	switch (md->itsDifficultyLevel)
	{
		case 0:  md->itsGameOf.Jigsaw.itsSize = 1;  break;
		case 1:  md->itsGameOf.Jigsaw.itsSize = 2;  break;
		case 2:  md->itsGameOf.Jigsaw.itsSize = 3;  break;
		case 3:  md->itsGameOf.Jigsaw.itsSize = 4;  break;
		case 4:  md->itsGameOf.Jigsaw.itsSize = 5;  break;
		default: md->itsGameOf.Jigsaw.itsSize = 2;  break;	//	should never occur
	}
	if (md->itsGameOf.Jigsaw.itsSize > MAX_JIGSAW_SIZE)
		md->itsGameOf.Jigsaw.itsSize = MAX_JIGSAW_SIZE;

	//	For easier legibility...
	n = md->itsGameOf.Jigsaw.itsSize;

	//	Initialize the pieces.
	for (h = 0; h < n; h++)
		for (v = 0; v < n; v++)
		{
			i = h*n + v;

			md->itsGameOf.Jigsaw.itsDrawOrder[i] = &md->itsGameOf.Jigsaw.itsPieces[i];

			md->itsGameOf.Jigsaw.itsPieces[i].itsLogicalH	= h;
			md->itsGameOf.Jigsaw.itsPieces[i].itsLogicalV	= v;

			md->itsGameOf.Jigsaw.itsPieces[i].itsPlacement.itsH		= -0.5 + (double)rand() / (double)RAND_MAX;
			md->itsGameOf.Jigsaw.itsPieces[i].itsPlacement.itsV		= -0.5 + (double)rand() / (double)RAND_MAX;
			md->itsGameOf.Jigsaw.itsPieces[i].itsPlacement.itsFlip	= (md->itsTopology == TopologyKlein ? rand()%2 : false);
			md->itsGameOf.Jigsaw.itsPieces[i].itsPlacement.itsAngle	= 0.0;
			md->itsGameOf.Jigsaw.itsPieces[i].itsPlacement.itsSizeH	= 1.0 / n;
			md->itsGameOf.Jigsaw.itsPieces[i].itsPlacement.itsSizeV	= 1.0 / n;

			if (h != 0)
				hh = h - 1;
			else
				hh = n - 1;
			vv = v;
			md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborWest	= &md->itsGameOf.Jigsaw.itsPieces[hh*n + vv];

			if (h != n - 1)
				hh = h + 1;
			else
				hh = 0;
			vv = v;
			md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborEast	= &md->itsGameOf.Jigsaw.itsPieces[hh*n + vv];

			hh = h;
			if (v != 0)
				vv = v - 1;
			else
			{
				vv = n - 1;
				if (md->itsTopology == TopologyKlein)
					hh = (n - 1) - hh;
			}
			md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborSouth	= &md->itsGameOf.Jigsaw.itsPieces[hh*n + vv];

			hh = h;
			if (v != n - 1)
				vv = v + 1;
			else
			{
				vv = 0;
				if (md->itsTopology == TopologyKlein)
					hh = (n - 1) - hh;
			}
			md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborNorth	= &md->itsGameOf.Jigsaw.itsPieces[hh*n + vv];

			md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborBondWest	= false;
			md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborBondEast	= false;
			md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborBondSouth	= false;
			md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborBondNorth	= false;

			md->itsGameOf.Jigsaw.itsPieces[i].itsClumpIsNonorientable = false;
		}

	//	Set the tab genders randomly.
	for (i = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
	{
		thePiece = &md->itsGameOf.Jigsaw.itsPieces[i];

		thePiece->itsTabGenderWest  = ((rand() & 0x1) ? TabFemale : TabMale);
		thePiece->itsNeighborWest->itsTabGenderEast   = OPPOSITE_GENDER(thePiece->itsTabGenderWest);

		thePiece->itsTabGenderSouth = ((rand() & 0x1) ? TabFemale : TabMale);
		thePiece->itsNeighborSouth->itsTabGenderNorth = OPPOSITE_GENDER(thePiece->itsTabGenderSouth);
	}
	
	//	Prepare the image for the new puzzle texture.
	//	(Note that we don't yet pass the data to OpenGL -- we'll pass it
	//	at render time when the OpenGL context is active.)
	if ((theErrorMessage = RefreshPuzzleTexture(md)) != NULL)
		ErrorMessage(theErrorMessage, "Jigsaw puzzle set-up error", false);

	//	Abort any pending simulation.
	SimulationEnd(md);	//	unused in Jigsaw

	//	Ready to go.
	md->itsGameIsOver = false;
}

static ErrorText RefreshPuzzleTexture(ModelData *md)
{
	ErrorText		theErrorMessage		= NULL;
	char			*theFileName		= NULL;
	ImageRGBA		*theImageRGBA		= NULL,
					*theResampledImage	= NULL,
					*theFullSizeImage	= NULL;
	
	static char	*theTorusPuzzleFileNames[] =
				{
					"Jigsaw/Vine.rgba",
					"Jigsaw/Vegetables.rgba",
					"Jigsaw/Stones.rgba",
					"Jigsaw/Pansies.rgba",
					"Jigsaw/Gumdrops.rgba"
				},
				*theKleinPuzzleFileNames[] =
				{
					"Jigsaw/Home.rgba",
					"Jigsaw/Raccoon.rgba"
				};

#ifdef TORUS_GAMES_FOR_OPENGL_ES
	static const unsigned int	theDesiredImageSize = 256;
#else
	static const unsigned int	theDesiredImageSize = 512;
#endif

	//	Release any previously cached puzzle images.
	FreeImageRGBA(&md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders);
	FreeImageRGBA(&md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders);
	FREE_MEMORY_SAFELY(md->itsGameOf.Jigsaw.itsPuzzleImagePixelAssignments);
	FREE_MEMORY_SAFELY(md->itsGameOf.Jigsaw.itsPuzzleImagePixelBorderFlags);
	
	//	The puzzle image depends on the current topology and puzzle number.
	switch (md->itsTopology)
	{
		case TopologyTorus:
			if (md->itsGameOf.Jigsaw.itsCurrentTorusPuzzle < BUFFER_LENGTH(theTorusPuzzleFileNames))
				theFileName = theTorusPuzzleFileNames[md->itsGameOf.Jigsaw.itsCurrentTorusPuzzle];
			else
			{
				theErrorMessage = "No image filename available for current torus puzzle.";
				goto CleanUpRefreshPuzzleTexture;
			}
			break;

		case TopologyKlein:
			if (md->itsGameOf.Jigsaw.itsCurrentKleinPuzzle < BUFFER_LENGTH(theKleinPuzzleFileNames))
				theFileName = theKleinPuzzleFileNames[md->itsGameOf.Jigsaw.itsCurrentKleinPuzzle];
			else
			{
				theErrorMessage = "No image filename available for current Klein bottle puzzle.";
				goto CleanUpRefreshPuzzleTexture;
			}
			break;
		
		default:
			theErrorMessage = "Invalid topology in RefreshPuzzleTexture().";
			goto CleanUpRefreshPuzzleTexture;
	}

	//	Read the texture file.
	if ((theErrorMessage = ReadImageRGBA(theFileName, &theImageRGBA)) != NULL)
		goto CleanUpRefreshPuzzleTexture;
	
	//	Puzzle image must be square.
	if (theImageRGBA->itsWidth != theImageRGBA->itsHeight)
	{
		theErrorMessage = "Encountered non-square jigsaw puzzle image";
		goto CleanUpRefreshPuzzleTexture;
	}

	//	We expect a square puzzle image of size 512 × 512 (desktop) or 256 × 256 (mobile).
	//	If we've received a half-size image -- 256 × 256 (desktop) or 128 × 128 (mobile) --
	//	then resample it to the desired full size.  The image itself doesn't get any better,
	//	of course, but the border template and the black border outline will be smoother,
	//	because they won't get downsampled to match the low-resolution base image.
	if (theImageRGBA->itsWidth == theDesiredImageSize)
	{
		theFullSizeImage = theImageRGBA;
	}
	else
	if (theImageRGBA->itsWidth == theDesiredImageSize/2)
	{
		theResampledImage = ResampleImage2x(theImageRGBA, md->itsTopology);
		if (theResampledImage != NULL)
		{
			FreeImageRGBA(&theImageRGBA);
			theFullSizeImage = theResampledImage;
		}
		else
			theFullSizeImage = theImageRGBA;
	}
	else
	{
		theErrorMessage = "Encountered jigsaw puzzle image of unexpected size";
		goto CleanUpRefreshPuzzleTexture;
	}

	//	Create a single 2n × 2n image containing each puzzle piece in its own separate block.
	if ((theErrorMessage = MakeJigsawImage(	theFullSizeImage,
											md->itsGameOf.Jigsaw.itsSize,
											md->itsTopology,
											md->itsGameOf.Jigsaw.itsPieces,
											&md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders,
											&md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders,
											&md->itsGameOf.Jigsaw.itsPuzzleImagePixelAssignments,
											&md->itsGameOf.Jigsaw.itsPuzzleImagePixelBorderFlags,
											&md->itsGameOf.Jigsaw.itsPuzzleImageFirstCenter,
											&md->itsGameOf.Jigsaw.itsPuzzleImageStride)) != NULL)
		goto CleanUpRefreshPuzzleTexture;

	//	Clear the itsPuzzleTextureIsUpToDate flag so the new image data
	//	will get loaded at render time.
	md->itsGameOf.Jigsaw.itsPuzzleTextureIsUpToDate = false;

CleanUpRefreshPuzzleTexture:

	//	Free the images as necessary.
	FreeImageRGBA(&theImageRGBA);
	FreeImageRGBA(&theResampledImage);
	
	return theErrorMessage;
}

static ImageRGBA *ResampleImage2x(
	ImageRGBA		*aSrcImage,
	TopologyType	aTopology)
{
	ImageRGBA		*theDstImage;
	unsigned int	theSrcWidth,
					theSrcHeight,
					theDstWidth,
					theDstHeight,
					theDstRow,
					theDstCol,
					theSrcRow,
					theSrcCol;
	PixelRGBA		theSrcPixelA,
					theSrcPixelB,
					theSrcPixelC,
					theSrcPixelD;
	bool			theRowIsFlipped;
	PixelRGBA		*theDstPixel;

	//	Just to be safe...
	if (aSrcImage == NULL)
		return NULL;
	
	//	Allocate space for the 2x resampled image.
	theDstImage = GET_MEMORY(sizeof(ImageRGBA));
	if (theDstImage == NULL)
		return NULL;
	theDstImage->itsWidth		= 2 * aSrcImage->itsWidth;
	theDstImage->itsHeight		= 2 * aSrcImage->itsHeight;
	theDstImage->itsPixels		= GET_MEMORY( theDstImage->itsWidth * theDstImage->itsHeight * sizeof(PixelRGBA) );
	theDstImage->itsNumRawBytes	= 0;	//	unused
	theDstImage->itsRawBytes	= (Byte *) theDstImage->itsPixels;
	if (theDstImage->itsPixels == NULL)
	{
		FREE_MEMORY_SAFELY(theDstImage);
		return NULL;
	}
	
	//	Resample.
	//
	//	Each sample represents the color at the center of a pixel.
	//	If you take a pixel grid
	//
	//			 |       |       |       |
	//			-+-------+-------+-------+-
	//			 |       |       |       |
	//			 |       |       |       |
	//			 |       |       |       |
	//			-+-------+-------+-------+-
	//			 |       |       |       |
	//			 |       |       |       |
	//			 |       |       |       |
	//			-+-------+-------+-------+-
	//			 |       |       |       |
	//
	//	and subdivide each pixel into four quadrants
	//
	//			 |   :   |   :   |   :   |
	//			-+-------+-------+-------+-
	//			 |   :   |   :   |   :   |
	//			 | - + - | - + - | - + - |
	//			 |   :   |   :   |   :   |
	//			-+-------+-------+-------+-
	//			 |   :   |   :   |   :   |
	//			 | - + - | - + - | - + - |
	//			 |   :   |   :   |   :   |
	//			-+-------+-------+-------+-
	//			 |   :   |   :   |   :   |
	//
	//	you'll see that none of the new pixel centers (N) coincide
	//	with the old pixel centers (O)
	//
	//			 | N : N | N : N | N : N |
	//			-+-------+-------+-------+-
	//			 | N : N | N : N | N : N |
	//			 | - O - | - O - | - O - |
	//			 | N : N | N : N | N : N |
	//			-+-------+-------+-------+-
	//			 | N : N | N : N | N : N |
	//			 | - O - | - O - | - O - |
	//			 | N : N | N : N | N : N |
	//			-+-------+-------+-------+-
	//			 | N : N | N : N | N : N |
	//
	//	Exercise:  Convince yourself that each new pixel center
	//	is the weighted average of four of the old pixel centers,
	//	with weighting factors {9/16, 3/16, 3/16, 1/16}.
	//
	//	Technical note:  For a torus image, we could shift all (old) pixel centers
	//	1/4 pixel to the left and 1/4 pixel down, so each old pixel centers would
	//	coincide with one of the new pixel centers.  Such a shift would let us
	//	write simpler and faster resampling code.
	//	For a Klein bottle image, however, such a shift would move the image relative
	//	to its symmetry axes, and in particular would introduce a 1/2 pixel error
	//	along the "seam".
	
	theSrcWidth		= aSrcImage->itsWidth;
	theSrcHeight	= aSrcImage->itsHeight;
	theDstWidth		= theDstImage->itsWidth;
	theDstHeight	= theDstImage->itsHeight;
	
	for (theDstRow = theDstHeight; theDstRow-- > 0; )
	{
		for (theDstCol = theDstWidth; theDstCol-- > 0; )
		{
			//	The color at destination pixel (theDstRow, theDstCol) will be the weighted average
			//
			//		(9/16)PixelA + (3/16)PixelB + (3/16)PixelC + (1/16)PixelD.
			//
			//	If any or all of the source pixels were partially transparent, then we'd want
			//	to further weight the average according to the source pixels' opacity.  
			//	With traditional (R,G,B,α) the formula is a bit messy:
			//
			//							 9*α0*(R0,G0,B0) + 3*α1*(R1,G1,B1) + 3*α2*(R2,G2,B2) + 1*α3*(R3,G3,B3)
			//		(Ravg, Gavg, Bavg) = ---------------------------------------------------------------------
			//							                       9*α0 + 3*α1 + 3*α2 + 1*α3
			//
			//							 9*α0 + 3*α1 + 3*α2 + 1*α3
			//		              αavg = -------------------------
			//							            16
			//
			//	Fortunately with premultiplied (αR,αG,αB,α) the formula 
			//	reduces back to a simple average:
			//
			//		(αavg*Ravg, αavg*Gavg, αavg*Bavg, αavg)
			//			= (9/16) * (α0*R0, α0*G0, α0*B0, α0)
			//			+ (3/16) * (α1*R1, α1*G1, α1*B1, α1)
			//			+ (3/16) * (α2*R2, α2*G2, α2*B2, α2)
			//			+ (1/16) * (α3*R3, α3*G3, α3*B3, α3)
			//
			//	For the jigsaw puzzle images, all the alpha values are 0xFF,
			//	so these considerations are moot at the moment, but I figured
			//	it would be a good idea to document that fact for future reference,
			//	in case I ever need to resample partially transparent images.

			theSrcRow = theDstRow >> 1;	//	divide by 2, discard remainder (same as /2, maybe faster)
			theSrcCol = theDstCol >> 1;	//	divide by 2, discard remainder (same as /2, maybe faster)
			theSrcPixelA = aSrcImage->itsPixels[theSrcRow*theSrcWidth + theSrcCol];
			
			if (theDstCol & 0x00000001)	//	theDstCol is odd
			{
				theSrcCol++;					//	move to the right
				if (theSrcCol == theSrcWidth)	//	wrap if necessary
					theSrcCol = 0;
			}
			else						//	theDstCol is even
			{
				if (theSrcCol == 0)				//	wrap if necessary
					theSrcCol = theSrcWidth;
				theSrcCol--;					//	move to the left
			}
			theSrcPixelB = aSrcImage->itsPixels[theSrcRow*theSrcWidth + theSrcCol];
			
			theSrcCol = theDstCol >> 1;	//	restore theSrcCol to its original location 
			
			theRowIsFlipped = false;
			if (theDstRow & 0x00000001)	//	theDstRow is odd
			{
				theSrcRow++;					//	move upward
				if (theSrcRow == theSrcHeight)	//	wrap if necessary
				{
					theSrcRow = 0;
					if (aTopology == TopologyKlein)
					{
						theSrcCol = (theSrcWidth - 1) - theSrcCol;
						theRowIsFlipped = true;
					}
				}
			}
			else						//	theDstRow is even
			{
				if (theSrcRow == 0)				//	wrap if necessary
				{
					theSrcRow = theSrcHeight;
					if (aTopology == TopologyKlein)
					{
						theSrcCol = (theSrcWidth - 1) - theSrcCol;
						theRowIsFlipped = true;
					}
				}
				theSrcRow--;					//	move downward
			}
			theSrcPixelC = aSrcImage->itsPixels[theSrcRow*theSrcWidth + theSrcCol];

			if ((theDstCol & 0x00000001) ^ theRowIsFlipped)	//	theDstCol is odd XOR the row is flipped
			{
				theSrcCol++;					//	move to the right
				if (theSrcCol == theSrcWidth)	//	wrap if necessary
					theSrcCol = 0;
			}
			else											//	theDstCol is even XOR the row is flipped
			{
				if (theSrcCol == 0)				//	wrap if necessary
					theSrcCol = theSrcWidth;
				theSrcCol--;					//	move to the left
			}
			theSrcPixelD = aSrcImage->itsPixels[theSrcRow*theSrcWidth + theSrcCol];
			
			//	Average the four pixels.
			//
			//	Note #1.  With premultiplied (αR,αG,αB,α), the weights 
			//	don't depend on the alpha values.  See explanation above.
			//
			//	Note #2.  To divide by 16, use a bitshift  >> 4  ,
			//	which might be faster than the explicit division / 16.
			theDstPixel = &theDstImage->itsPixels[theDstRow*theDstWidth + theDstCol];
			theDstPixel->r = ( 9 * theSrcPixelA.r  +  3 * theSrcPixelB.r  +  3 * theSrcPixelC.r  +  1 * theSrcPixelD.r ) >> 4;
			theDstPixel->g = ( 9 * theSrcPixelA.g  +  3 * theSrcPixelB.g  +  3 * theSrcPixelC.g  +  1 * theSrcPixelD.g ) >> 4;
			theDstPixel->b = ( 9 * theSrcPixelA.b  +  3 * theSrcPixelB.b  +  3 * theSrcPixelC.b  +  1 * theSrcPixelD.b ) >> 4;
			theDstPixel->a = ( 9 * theSrcPixelA.a  +  3 * theSrcPixelB.a  +  3 * theSrcPixelC.a  +  1 * theSrcPixelD.a ) >> 4;
		}
	}
	
	return theDstImage;
}

static ErrorText MakeJigsawImage(
	ImageRGBA		*anImageRGBA,						//	input
	unsigned int	aPuzzleSize,						//	input
	TopologyType	aPuzzleTopology,					//	input
	JigsawPiece		*aPieceList,						//	input
	ImageRGBA		**aPuzzleImageWithBorders,			//	output
	ImageRGBA		**aPuzzleImageWithoutBorders,		//	output
	Byte			**somePuzzleImagePixelAssignments,	//	output, indices of containing JigsawPiece as hhhhvvvv
	Byte			**somePuzzleImagePixelBorderFlags,	//	output, which piece-borders this pixel belongs to, as 0000wesn
	double			*aPuzzleImageFirstCenter,			//	output, in [0.0, 2.0] coordinates
	double			*aPuzzleImageStride)				//	output, in [0.0, 2.0] coordinates
{
	ErrorText		theErrorMessage		= NULL;
	unsigned int	theOldImageSize,
					theNewImageSize;
	double			thePieceSize;
	unsigned int	theSubsequentOffset,
					theInitialOffset;
	double			theBlockSize;
	PixelRGBA		*theDstPixelWithBorders,
					*theDstPixelWithoutBorders;
	Byte			*theDstPixelAssignment,
					*theDstPixelBorderFlags;
	unsigned int	theDstRow,
					theDstCol,
					theRowBlock,
					theColBlock;
	signed int		theSrcRow,
					theSrcCol;
	bool			theRowFlip;
	ImageRGBA		*thePieceTemplate	= NULL;
	double			thePixelSize,
					thePixelCenterH,
					thePixelCenterV;
	JigsawPiece		*thePiece;
	TabGender		theTabGender;
	double			theAbsoluteH,
					theAbsoluteV,
					theOutwardCoord,
					theLateralCoord;
	double			u,
					v;
	signed int		u0,
					v0,
					u1,
					v1;
	double			theWeightU0,
					theWeightU1,
					theWeightV0,
					theWeightV1;
	PixelRGBA		theTexel00,
					theTexel01,
					theTexel10,
					theTexel11;
	double			theInterpolatedAlpha,
					theInterpolatedLightness,
					theAlphaFactor;

	if ( anImageRGBA == NULL
	 ||  aPieceList == NULL
	 ||  aPuzzleImageWithBorders == NULL
	 || *aPuzzleImageWithBorders != NULL
	 ||  aPuzzleImageWithoutBorders == NULL
	 || *aPuzzleImageWithoutBorders != NULL
	 ||  somePuzzleImagePixelAssignments == NULL
	 || *somePuzzleImagePixelAssignments != NULL
	 ||  somePuzzleImagePixelBorderFlags == NULL
	 || *somePuzzleImagePixelBorderFlags != NULL
	 ||  aPuzzleImageFirstCenter == NULL
	 ||  aPuzzleImageStride == NULL)
	{
		return "Bad input parameter in MakeJigsawImage().";
	}
	
	if (aPuzzleSize == 0 || aPuzzleSize > MAX_JIGSAW_SIZE)
	{
		theErrorMessage = "Bad puzzle size in MakeJigsawImage().";
		goto CleanUpMakeJigsawImage;
	}
	
	if (anImageRGBA->itsWidth != anImageRGBA->itsHeight)
	{
		theErrorMessage = "Non-square image passed to MakeJigsawImage().";
		goto CleanUpMakeJigsawImage;
	}

	theOldImageSize = anImageRGBA->itsWidth;	//	We already checked that itsWidth == itsHeight.
	theNewImageSize = 2 * theOldImageSize;
	
	*aPuzzleImageWithBorders = GET_MEMORY(sizeof(ImageRGBA));
	if (*aPuzzleImageWithBorders == NULL)
	{
		theErrorMessage = "Couldn't allocate aPuzzleImageWithBorders in MakeJigsawImage().";
		goto CleanUpMakeJigsawImage;
	}
	(*aPuzzleImageWithBorders)->itsWidth		= theNewImageSize;
	(*aPuzzleImageWithBorders)->itsHeight		= theNewImageSize;
	(*aPuzzleImageWithBorders)->itsPixels		= NULL;
	(*aPuzzleImageWithBorders)->itsNumRawBytes	= 0;	//	unused
	(*aPuzzleImageWithBorders)->itsRawBytes		= NULL;

	(*aPuzzleImageWithBorders)->itsPixels		= GET_MEMORY( theNewImageSize * theNewImageSize * sizeof(PixelRGBA) );
	if ((*aPuzzleImageWithBorders)->itsPixels == NULL)
	{
		theErrorMessage = "Couldn't allocate pixels for aPuzzleImageWithBorders in MakeJigsawImage().";
		goto CleanUpMakeJigsawImage;
	}
	(*aPuzzleImageWithBorders)->itsRawBytes		= (Byte *) (*aPuzzleImageWithBorders)->itsPixels;
	
	*aPuzzleImageWithoutBorders = GET_MEMORY(sizeof(ImageRGBA));
	if (*aPuzzleImageWithoutBorders == NULL)
	{
		theErrorMessage = "Couldn't allocate aPuzzleImageWithoutBorders in MakeJigsawImage().";
		goto CleanUpMakeJigsawImage;
	}
	(*aPuzzleImageWithoutBorders)->itsWidth			= theNewImageSize;
	(*aPuzzleImageWithoutBorders)->itsHeight		= theNewImageSize;
	(*aPuzzleImageWithoutBorders)->itsPixels		= NULL;
	(*aPuzzleImageWithoutBorders)->itsNumRawBytes	= 0;	//	unused
	(*aPuzzleImageWithoutBorders)->itsRawBytes		= NULL;

	(*aPuzzleImageWithoutBorders)->itsPixels		= GET_MEMORY( theNewImageSize * theNewImageSize * sizeof(PixelRGBA) );
	if ((*aPuzzleImageWithoutBorders)->itsPixels == NULL)
	{
		theErrorMessage = "Couldn't allocate pixels for aPuzzleImageWithoutBorders in MakeJigsawImage().";
		goto CleanUpMakeJigsawImage;
	}
	(*aPuzzleImageWithoutBorders)->itsRawBytes		= (Byte *) (*aPuzzleImageWithoutBorders)->itsPixels;

	*somePuzzleImagePixelAssignments = GET_MEMORY( theNewImageSize * theNewImageSize * sizeof(Byte) );
	if (*somePuzzleImagePixelAssignments == NULL)
	{
		theErrorMessage = "Couldn't allocate somePuzzleImagePixelAssignments in MakeJigsawImage().";
		goto CleanUpMakeJigsawImage;
	}

	*somePuzzleImagePixelBorderFlags = GET_MEMORY( theNewImageSize * theNewImageSize * sizeof(Byte) );
	if (*somePuzzleImagePixelBorderFlags == NULL)
	{
		theErrorMessage = "Couldn't allocate somePuzzleImagePixelBorderFlags in MakeJigsawImage().";
		goto CleanUpMakeJigsawImage;
	}
	
	//	For brevity, let n = aPuzzleSize (the number of puzzle pieces in each direction),
	//	and let s be theOldImageSize (width and height of original image, in pixels).
	//	In very rough terms, anImageRGBA consists of an n × n array of puzzle pieces,
	//	interconnected by tabs and slots.  Each puzzle piece is s/n pixels wide and s/n pixels high,
	//	excluding its tabs, but keep in mind that s/n is typically not an integer.
	//
	//	Our plan, still speaking in rough terms, is to create a 2s × 2s image
	//	that provides a 2s/n × 2s/n square for each puzzle piece, with the piece itself
	//	centered in the 2s/n × 2s/n square and with its tabs extending into the "extra" space.
	//	By allowing a 2s/n × 2s/n block for each s/n × s/n piece, we can draw all necessary tabs
	//	without interfering with neighboring pieces or their tabs.
	//
	//	The tricky part is that in a typical case, say a 256 × 256 pixel image divided
	//	into a 3 × 3 array of puzzle pieces, the individual pieces do not align
	//	with the integer pixel grid.  In other words, the pieces may contain fractional pixels
	//	along their boundaries.  This is OK, just so we make sure that a given piece
	//	bears the same relation to the pixel grid in the new 2s × 2s image that it did
	//	in the original s × s image.  This requirement will be realized if we ensure
	//	that the offset from the right edge of one piece (excluding a possible tab)
	//	to the left edge of the next piece (again excluding a possible tab) is an integer
	//	number of pixels, even though the edges themselves may have fractional pixel coordinates.
	//
	//	Technical Note:  The 2s × 2s image would be unnecessary if we were willing
	//	to draw the jigsaw puzzle pieces using multitexturing (one texture for the puzzle image
	//	and a second texture for the piece outline) and were also willing to make
	//	a whole second pass to draw the black borders, but this approach was unacceptably slow
	//	on the iPhone and iPod Touch (roughly 4-5 fps), so I decided to implement
	//	this single-pass, single texture algorithm instead.
	
	//	Because the following computations round down to integer values, we typically
	//	waste a few pixels along the right and bottom edges of the final texture,
	//	but we never risk an overrun.  The final 2s/n × 2s/n blocks come out a pixel or two
	//	smaller due to the rounding down, but that's OK because a puzzle piece's tabs
	//	are significantly less wide than the half-a-piece-body-width that we tenatively
	//	allowed for them, so the rounding down is harmless.

	//	thePieceSize is the width and height of the piece body, excluding tabs and borders.
	//	Even though it's expressed in pixel coordinates, it's typically not an integer.
	thePieceSize = (double)theOldImageSize / (double)aPuzzleSize;
	
	//	theSubsequentOffset is the distance from the right side of one piece body
	//	(excluding tabs and borders) to the left side of the next piece body.
	//	It's expressed in pixel coordinates and is always an integer, so the piece bodies
	//	will align with fractional pixels in the new image exactly as they did in the old image.
	theSubsequentOffset = theOldImageSize / aPuzzleSize;	//	fractional part intentionally discarded

	//	theInitialOffset is the distance of the first piece's left side
	//	from the left edge of the new image.  It's expressed in pixel coordinates
	//	and is always an integer.
	theInitialOffset = theSubsequentOffset / 2;				//	fractional part intentionally discarded
	
	//	The block size (in pixels) is not an integer.
	//	This creates some ambiguity about what to do with pixels
	//	straddling the boundary between one block and the next,
	//	but ultimately it doesn't matter because those pixels never get drawn.
	theBlockSize = thePieceSize + (double)theSubsequentOffset;
	
	//	At render time the drawing code will need to know the offset and stride in [0.0, 2.0] coordinates.
	*aPuzzleImageFirstCenter	= (theInitialOffset + 0.5 * thePieceSize) / theOldImageSize;
	*aPuzzleImageStride			= theBlockSize / theOldImageSize;
	
	//	For each pixel in aPuzzleWith/WithoutBorders, 
	//	work backwards to find the pixel in anImageRGBA that maps to it.

	theDstPixelWithBorders		= (*aPuzzleImageWithBorders   )->itsPixels;
	theDstPixelWithoutBorders	= (*aPuzzleImageWithoutBorders)->itsPixels;

	for (theDstRow = 0; theDstRow < theNewImageSize; theDstRow++)
	{
		theRowBlock	= (unsigned int) floor(theDstRow / theBlockSize);
		theSrcRow	= theDstRow - (theInitialOffset + theRowBlock*theSubsequentOffset);
		theRowFlip	= false;
		while (theSrcRow < 0)
		{
			theSrcRow += theOldImageSize;
			if (aPuzzleTopology == TopologyKlein)
				theRowFlip = ! theRowFlip;
		}
		while (theSrcRow >= (signed int)theOldImageSize)
		{
			theSrcRow -= theOldImageSize;
			if (aPuzzleTopology == TopologyKlein)
				theRowFlip = ! theRowFlip;
		}

		for (theDstCol = 0; theDstCol < theNewImageSize; theDstCol++)
		{
			theColBlock	= (unsigned int) floor(theDstCol / theBlockSize);
			theSrcCol	= theDstCol - (theInitialOffset + theColBlock*theSubsequentOffset);
			if (theRowFlip)
				theSrcCol = (theOldImageSize - 1) - theSrcCol;
			while (theSrcCol < 0)
				theSrcCol += theOldImageSize;
			while (theSrcCol >= (signed int)theOldImageSize)
				theSrcCol -= theOldImageSize;

			if (theRowBlock < aPuzzleSize
			 && theColBlock < aPuzzleSize)
			{
				//	Copy the source pixel to the destination pixel.
				*theDstPixelWithBorders++ = *theDstPixelWithoutBorders++ = anImageRGBA->itsPixels[theSrcRow*theOldImageSize + theSrcCol];
			}
			else
			{
				//	Fill unused pixels with transparent color.
				*theDstPixelWithBorders++ = *theDstPixelWithoutBorders++ = (PixelRGBA){0x00, 0x00, 0x00, 0x00};	//	premultiplied alpha
			}
		}
	}
	
	//	For each pixel in aPuzzleWithoutBorders, assign an opacity value
	//	by interpolating appropriate values from the piece template.
	
	theErrorMessage = ReadImageRGBA("Jigsaw/PieceTemplate.rgba", &thePieceTemplate);
	if (theErrorMessage != NULL)
		goto CleanUpMakeJigsawImage;

	thePixelSize = 1.0 / theOldImageSize;	//	in [0.0, 2.0] coordinates

	theDstPixelWithBorders		= (*aPuzzleImageWithBorders   )->itsPixels;
	theDstPixelWithoutBorders	= (*aPuzzleImageWithoutBorders)->itsPixels;
	theDstPixelAssignment		= *somePuzzleImagePixelAssignments;
	theDstPixelBorderFlags		= *somePuzzleImagePixelBorderFlags;

	for (theDstRow = 0; theDstRow < theNewImageSize; theDstRow++)
	{
		theRowBlock = (unsigned int) floor(theDstRow / theBlockSize);
		
		//	Compute the vertical coordinate of the horizontal line passing
		//	through the *center* of each pixel in the current row.
		//	Compute this vertical coordinate in pixels, to get started.
		thePixelCenterV = theDstRow + 0.5;
		
		//	Convert from pixel coordinates to [0.0, 2.0] coordinates.
		thePixelCenterV *= thePixelSize;
		
		//	Convert from [0.0, 2.0] coordinates to piece coordinates.
		//	Piece coordinates are defined to be the coordinates
		//	in which the main piece body (excluding tabs and borders)
		//	has corners at (±0.5, ±0.5).
		thePixelCenterV -= (*aPuzzleImageFirstCenter)  +  theRowBlock * (*aPuzzleImageStride);
		thePixelCenterV *= (double)aPuzzleSize;

		for (theDstCol = 0; theDstCol < theNewImageSize; theDstCol++)
		{
			theColBlock = (unsigned int) floor(theDstCol / theBlockSize);

			if (theRowBlock >= aPuzzleSize
			 || theColBlock >= aPuzzleSize)
			{
				//	Set unused pixels to full transparency and no border membership.
				*(theDstPixelWithBorders++)		= (PixelRGBA){0x00, 0x00, 0x00, 0x00};
				*(theDstPixelWithoutBorders++)	= (PixelRGBA){0x00, 0x00, 0x00, 0x00};
				*theDstPixelAssignment++		= 0x00;
				*theDstPixelBorderFlags++		= 0x00;
				continue;
			}
			
			//	Technical note:  the JigsawPieces are stored in (h,v) order,
			//	which corresonds to (col, row).  That is, the JigsawPieces
			//	are stored in column-major order.
			thePiece = &aPieceList[ theColBlock * aPuzzleSize  +  theRowBlock ];
			
			//	Record this pixel's (h,v) coordinates so we can update the image
			//	quickly when the user joins two pieces together and their borders disappear.
#if MAX_JIGSAW_SIZE > 16
			The format for theDstPixelAssignment allows only four bits for h and four bits for v,
			so aPuzzleSize must not exceed 16 (which would be too big a puzzle in any case).
#endif
			*theDstPixelAssignment = ((theColBlock << 4) | theRowBlock);
			
			//	Compute thePixelCenterH the same way we computed thePixelCenterV (see above).
			thePixelCenterH = theDstCol + 0.5;
			thePixelCenterH *= thePixelSize;
			thePixelCenterH -= (*aPuzzleImageFirstCenter)  +  theColBlock * (*aPuzzleImageStride);
			thePixelCenterH *= (double)aPuzzleSize;

			//	Which side of thePiece does this pixel lie nearest?
			if (thePixelCenterH - thePixelCenterV > 0.0)
			{
				if (thePixelCenterH + thePixelCenterV > 0.0)
				{
					theTabGender = thePiece->itsTabGenderEast;

					*theDstPixelBorderFlags = BORDER_EAST;	//	tentative value -- may be set to zero later on

					if (thePixelCenterV > +INNER_BORDER && thePixelCenterV < +OUTER_BORDER)
						*theDstPixelBorderFlags |= BORDER_NORTH;

					if (thePixelCenterV < -INNER_BORDER && thePixelCenterV > -OUTER_BORDER)
						*theDstPixelBorderFlags |= BORDER_SOUTH;
				}
				else
				{
					theTabGender = thePiece->itsTabGenderSouth;

					*theDstPixelBorderFlags = BORDER_SOUTH;	//	tentative value -- may be set to zero later on
					
					if (thePixelCenterH > +INNER_BORDER && thePixelCenterH < +OUTER_BORDER)
						*theDstPixelBorderFlags |= BORDER_EAST;
					
					if (thePixelCenterH < -INNER_BORDER && thePixelCenterH > -OUTER_BORDER)
						*theDstPixelBorderFlags |= BORDER_WEST;
				}
			}
			else
			{
				if (thePixelCenterH + thePixelCenterV > 0.0)
				{
					theTabGender = thePiece->itsTabGenderNorth;

					*theDstPixelBorderFlags = BORDER_NORTH;	//	tentative value -- may be set to zero later on
					
					if (thePixelCenterH > +INNER_BORDER && thePixelCenterH < +OUTER_BORDER)
						*theDstPixelBorderFlags |= BORDER_EAST;
					
					if (thePixelCenterH < -INNER_BORDER && thePixelCenterH > -OUTER_BORDER)
						*theDstPixelBorderFlags |= BORDER_WEST;
				}
				else
				{
					theTabGender = thePiece->itsTabGenderWest;

					*theDstPixelBorderFlags = BORDER_WEST;	//	tentative value -- may be set to zero later on
					
					if (thePixelCenterV > +INNER_BORDER && thePixelCenterV < +OUTER_BORDER)
						*theDstPixelBorderFlags |= BORDER_NORTH;
					
					if (thePixelCenterV < -INNER_BORDER && thePixelCenterV > -OUTER_BORDER)
						*theDstPixelBorderFlags |= BORDER_SOUTH;
				}
			}
			
			//	At this point all we care about is the pixel's distance outward
			//	from the piece center and its lateral displacement.
			theAbsoluteH = ABS(thePixelCenterH);
			theAbsoluteV = ABS(thePixelCenterV);
			if (theAbsoluteH > theAbsoluteV)
			{
				theOutwardCoord = theAbsoluteH;
				theLateralCoord = theAbsoluteV;
			}
			else
			{
				theOutwardCoord = theAbsoluteV;
				theLateralCoord = theAbsoluteH;
			}
			
			//	Set the current pixel's opacity and lightness.

			//	(theOutwardCoord, theLateralCoord) maps to some point (u,v) in the piece texture.
			u = theOutwardCoord * thePieceTemplate->itsWidth;
			v = ((theTabGender == TabFemale ? 0.25 : 0.75) + 0.5*theLateralCoord) * thePieceTemplate->itsHeight;

			//	Adjust (u,v) by (-0.5, -0.5) to account for the fact that 
			//	each texel value gives the rgba values sampled at the texel's center
			//	(not at its lower left corner).
			u -= 0.5;
			v -= 0.5;

			//	Find the nearest texel center below and to the left of the original (u,v).
			u0 = (signed int) floor(u);
			v0 = (signed int) floor(v);
			
			//	Compute the weighting factors expressing (u,v)
			//	as an average of the four nearest texels.
			theWeightU1 = u - (double)u0;
			theWeightU0 = 1.0 - theWeightU1;
			theWeightV1 = v - (double)v0;
			theWeightV0 = 1.0 - theWeightV1;

			//	Locate neighboring texels, for interpolation.

			//	Texture coordinates clamp in the u-direction...
			if (u0 < 0)
			{
				u0 = 0;
				u1 = 0;
			}
			else
			if (u0 >= (signed int) thePieceTemplate->itsWidth - 1)
			{
				u0 = thePieceTemplate->itsWidth - 1;
				u1 = thePieceTemplate->itsWidth - 1;
			}
			else
			{
				u1 = u0 + 1;
			}

			//	...but wrap in the v-direction.
			while (v0 < 0)
				v0 += thePieceTemplate->itsHeight;
			while (v0 >= (signed int) thePieceTemplate->itsHeight)
				v0 -= thePieceTemplate->itsHeight;
			v1 = v0 + 1;
			if (v1 >= (signed int) thePieceTemplate->itsHeight)
				v1 = 0;
			
			//	Locate the four neighboring texels.
			theTexel00 = thePieceTemplate->itsPixels[v0*thePieceTemplate->itsWidth + u0];
			theTexel01 = thePieceTemplate->itsPixels[v1*thePieceTemplate->itsWidth + u0];
			theTexel10 = thePieceTemplate->itsPixels[v0*thePieceTemplate->itsWidth + u1];
			theTexel11 = thePieceTemplate->itsPixels[v1*thePieceTemplate->itsWidth + u1];
			
			//	Interpolate the four alpha values.
			theInterpolatedAlpha = theWeightU0 * theWeightV0 * theTexel00.a
								 + theWeightU0 * theWeightV1 * theTexel01.a
								 + theWeightU1 * theWeightV0 * theTexel10.a
								 + theWeightU1 * theWeightV1 * theTexel11.a;
			
			//	Round theInterpolatedAlpha explicitly,
			//	so that when we cast to a Byte we don't risk
			//	off-by-one errors like 37.9999999 → 37 .
			theInterpolatedAlpha = floor(theInterpolatedAlpha + 0.5);

			//	Set theDstPixels' opacity.
			theDstPixelWithBorders->a = theDstPixelWithoutBorders->a = (Byte) theInterpolatedAlpha;
			
			//	Interpolate the four lightness values,
			//	which are stored in thePieceTemplate's red channel.
			theInterpolatedLightness = (theWeightU0 * theWeightV0 * theTexel00.r
									  + theWeightU0 * theWeightV1 * theTexel01.r
									  + theWeightU1 * theWeightV0 * theTexel10.r
									  + theWeightU1 * theWeightV1 * theTexel11.r)
									 * (1.0 / (double)0xFF);

			//	Is this pixel a darkened border pixel?
			if (theInterpolatedLightness < 0.999)	//	theInterpolatedLightness < 1, but allow for rounding error
			{
				//	Multiply theDstPixelWithBorders color components by theInterpolatedLightness.
				//	Interior pixels will remain unchanged (theInterpolatedLightness == 1.0)
				//	but boundary pixels will be black (theInterpolatedLightness == 0.0).
				//	Don't touch theDstPixelWithoutBorders.
				theDstPixelWithBorders->r = (Byte) floor(theDstPixelWithBorders->r * theInterpolatedLightness  +  0.5);
				theDstPixelWithBorders->g = (Byte) floor(theDstPixelWithBorders->g * theInterpolatedLightness  +  0.5);
				theDstPixelWithBorders->b = (Byte) floor(theDstPixelWithBorders->b * theInterpolatedLightness  +  0.5);
			}
			else
			{
				//	No need to ever overwrite this pixel when the user joins puzzle pieces together.
				*theDstPixelBorderFlags = 0x00;
			}
			
			//	Premultiply the red, green and blue components by alpha,
			//	in keeping with the Torus Games' premultiplied (αR,αG,αB,α) convention.
			//	(Here we rely on the fact that the old alpha value was 0xFF.)
			if (theInterpolatedAlpha < 255.0)
			{
				theAlphaFactor = theInterpolatedAlpha * (1.0 / 255.0);
				theDstPixelWithBorders->r		*= theAlphaFactor;
				theDstPixelWithBorders->g		*= theAlphaFactor;
				theDstPixelWithBorders->b		*= theAlphaFactor;
				theDstPixelWithoutBorders->r	*= theAlphaFactor;
				theDstPixelWithoutBorders->g	*= theAlphaFactor;
				theDstPixelWithoutBorders->b	*= theAlphaFactor;
			}

			//	All done with this pixel.  Move on to the next one.
			theDstPixelWithBorders++;
			theDstPixelWithoutBorders++;
			theDstPixelAssignment++;
			theDstPixelBorderFlags++;
		}
	}

CleanUpMakeJigsawImage:

	FreeImageRGBA(&thePieceTemplate);

	if (theErrorMessage != NULL)
	{
		FreeImageRGBA(aPuzzleImageWithBorders);
		FreeImageRGBA(aPuzzleImageWithoutBorders);
		FREE_MEMORY_SAFELY(*somePuzzleImagePixelAssignments);
		FREE_MEMORY_SAFELY(*somePuzzleImagePixelBorderFlags);
	}

	return theErrorMessage;
}


static bool JigsawDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
	ImageRGBA		*thePuzzleImage;
	unsigned int	i;
	JigsawPiece		*thePiece;
	Placement		thePieceGlobal,
					theHandGlobal,
					theHandLocal;
	unsigned int	h,
					v;

	UNUSED_PARAMETER(aShiftOrControlKeyIsDown);

	//	If the game is already over, ignore all hits.
	if (md->itsGameIsOver)
		return false;

	//	If no puzzle image is available, ignore all hits.
	thePuzzleImage = md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders;
	if (thePuzzleImage == NULL)
		return false;	//	Should occur only if we can't read the puzzle data or the piece template.

	//	Search the array md->itsGameOf.Jigsaw.itsDrawOrder[]
	//	in forwards order for top-to-bottom hit testing.
	for (i = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
	{
		thePiece		= md->itsGameOf.Jigsaw.itsDrawOrder[i];
		thePieceGlobal	= thePiece->itsPlacement;
		theHandGlobal	= md->itsHandPlacement;

		//	Locate the hand image nearest to the piece.
		if (theHandGlobal.itsV - thePieceGlobal.itsV < -0.5)
		{
			theHandGlobal.itsV += 1.0;
			if (md->itsTopology == TopologyKlein)
			{
				theHandGlobal.itsH		= -theHandGlobal.itsH;
				theHandGlobal.itsFlip	= ! theHandGlobal.itsFlip;
			}
		}
		if (theHandGlobal.itsV - thePieceGlobal.itsV > +0.5)
		{
			theHandGlobal.itsV -= 1.0;
			if (md->itsTopology == TopologyKlein)
			{
				theHandGlobal.itsH		= -theHandGlobal.itsH;
				theHandGlobal.itsFlip	= ! theHandGlobal.itsFlip;
			}
		}
		if (theHandGlobal.itsH - thePieceGlobal.itsH < -0.5)
			theHandGlobal.itsH += 1.0;
		if (theHandGlobal.itsH - thePieceGlobal.itsH > +0.5)
			theHandGlobal.itsH -= 1.0;

		//	Get the hand location relative to the piece center.
		theHandLocal.itsH	= theHandGlobal.itsH - thePieceGlobal.itsH;
		theHandLocal.itsV	= theHandGlobal.itsV - thePieceGlobal.itsV;
		if ( ! thePieceGlobal.itsFlip )
		{
			theHandLocal.itsFlip	= theHandGlobal.itsFlip;
		}
		else
		{
			theHandLocal.itsFlip	= ! theHandGlobal.itsFlip;
			theHandLocal.itsH		= - theHandLocal.itsH;
		}
		theHandLocal.itsAngle	= 0.0;	//	unused
		theHandLocal.itsSizeH	= 0.0;	//	unused
		theHandLocal.itsSizeV	= 0.0;	//	unused
		
		//	If the hand is near the piece, test for a hit.
		//
		//	Note:  As well as speeding up the code, this test
		//	guarantees that the pixel coordinates computed below
		//	will lie safely within the array bounds.
		if (fabs(theHandLocal.itsH) * md->itsGameOf.Jigsaw.itsSize < 0.5 * MARGIN_FOR_TAB
		 && fabs(theHandLocal.itsV) * md->itsGameOf.Jigsaw.itsSize < 0.5 * MARGIN_FOR_TAB)
		{
			//	Express theHandLocal relative to the piece's center in itsPuzzleImageWithBorders.
			//	Think in terms of the [0.0, 2.0] coordinates on itsPuzzleImageWithBorders.

			theHandLocal.itsH += md->itsGameOf.Jigsaw.itsPuzzleImageFirstCenter
								+  thePiece->itsLogicalH * md->itsGameOf.Jigsaw.itsPuzzleImageStride;
			theHandLocal.itsV += md->itsGameOf.Jigsaw.itsPuzzleImageFirstCenter
								+  thePiece->itsLogicalV * md->itsGameOf.Jigsaw.itsPuzzleImageStride;
			
			//	Convert to pixel coordinates in itsPuzzleImageWithBorders.
			h = (unsigned int) (0.5 * theHandLocal.itsH * thePuzzleImage->itsWidth);
			v = (unsigned int) (0.5 * theHandLocal.itsV * thePuzzleImage->itsHeight);
			
			//	This error check should be unnecessary,
			//	but it seems like a good idea to test the indices
			//	before accessing the texel array.
			if (h >= thePuzzleImage->itsWidth)
				h = 0;
			if (v >= thePuzzleImage->itsWidth)
				v = 0;
			
			//	Did we hit an opaque texel?
			if (thePuzzleImage->itsPixels[ v * thePuzzleImage->itsWidth  +  h ].a != 0)
			{
				md->itsDragIsReversed				= theHandLocal.itsFlip;
				md->itsGameOf.Jigsaw.itsDragPiece	= thePiece;
				SendClumpToFrontOrBack(md, i, SendClumpToFront);

				return true;
			}
		}
	}

	return false;
}


static void SendClumpToFrontOrBack(
	ModelData			*md,
	unsigned int		aPieceIndex,		//	an arbitrary piece within the clump
	SendClumpDirection	aDirection)
{
	JigsawPiece		*theTempList[MAX_JIGSAW_SIZE * MAX_JIGSAW_SIZE];
	unsigned int	i,
					ii;
	bool			theFrontFlag,
					theBackFlag;

	//	"Unnecessary" error check.
	if (aPieceIndex >= md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize)
	{
		ErrorMessage(	"Bad piece index",
						"Error in SendClumpToFrontOrBack()",
						false);
		return;
	}

	//	Set itsVisitedFlag to false for all pieces.
	for (i = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
		md->itsGameOf.Jigsaw.itsDrawOrder[i]->itsVisitedFlag = false;

	//	Visit and mark all pieces in the given clump.
	MarkClump(md->itsGameOf.Jigsaw.itsDrawOrder[aPieceIndex]);

	//	Initialize a "write index" on theTempList.
	ii = 0;

	//	Which pieces (marked or unmarked) should move to the front of the list?
	theFrontFlag	= (aDirection == SendClumpToFront);
	theBackFlag		= ! theFrontFlag;

	//	Copy pointers to front-bound pieces onto theTempList.
	for (i = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
		if (md->itsGameOf.Jigsaw.itsDrawOrder[i]->itsVisitedFlag == theFrontFlag)
			theTempList[ii++] = md->itsGameOf.Jigsaw.itsDrawOrder[i];

	//	Copy pointers to back-bound pieces onto theTempList,
	//	following the front-bound pieces.
	for (i = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
		if (md->itsGameOf.Jigsaw.itsDrawOrder[i]->itsVisitedFlag == theBackFlag)
			theTempList[ii++] = md->itsGameOf.Jigsaw.itsDrawOrder[i];

	//	"Unnecessary" error check.
	if (ii != md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize)
	{
		ErrorMessage(	"Internal error while copying pieces.",
						"Error in SendClumpToFrontOrBack()",
						false);
		return;
	}

	//	Copy theTempList back to itsGameOf.Jigsaw.itsDrawOrder.
	for (i = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
		md->itsGameOf.Jigsaw.itsDrawOrder[i] = theTempList[i];
}


static void MarkClump(JigsawPiece *aPiece)
{
	if ( ! aPiece->itsVisitedFlag )
	{
		aPiece->itsVisitedFlag = true;

		if (aPiece->itsNeighborBondWest)	MarkClump(aPiece->itsNeighborWest);
		if (aPiece->itsNeighborBondEast)	MarkClump(aPiece->itsNeighborEast);
		if (aPiece->itsNeighborBondSouth)	MarkClump(aPiece->itsNeighborSouth);
		if (aPiece->itsNeighborBondNorth)	MarkClump(aPiece->itsNeighborNorth);
	}
}


static void JigsawDragObject(
	ModelData	*md,
	double		aHandLocalDeltaH,
	double		aHandLocalDeltaV)
{
	double	thePieceLocalDeltaH,
			thePieceLocalDeltaV,
			theHandAmbientDeltaH,
			theHandAmbientDeltaV,
			thePieceAmbientDeltaH,
			thePieceAmbientDeltaV;

	//	Drag the clump containing itsDragPiece,
	//	which should be the topmost clump.

	thePieceLocalDeltaH		= md->itsDragIsReversed
								? -aHandLocalDeltaH : aHandLocalDeltaH;
	thePieceLocalDeltaV		= aHandLocalDeltaV;

	theHandAmbientDeltaH	= md->itsHandPlacement.itsFlip
								? -aHandLocalDeltaH : aHandLocalDeltaH;
	theHandAmbientDeltaV	= aHandLocalDeltaV;

	thePieceAmbientDeltaH	= md->itsGameOf.Jigsaw.itsDragPiece->itsPlacement.itsFlip
								? -thePieceLocalDeltaH : thePieceLocalDeltaH;
	thePieceAmbientDeltaV	= thePieceLocalDeltaV;

	//	Move the hand.
	md->itsHandPlacement.itsH += theHandAmbientDeltaH;
	md->itsHandPlacement.itsV += theHandAmbientDeltaV;
	NormalizePlacement(&md->itsHandPlacement, md->itsTopology);

	//	In a Klein bottle a clump of pieces may reach all the way
	//	around and connect up with itself in such a way that the clump
	//	cannot move left or right.
	if (md->itsGameOf.Jigsaw.itsDragPiece->itsClumpIsNonorientable)
		thePieceAmbientDeltaH = 0.0;

	//	Move the clump of pieces containing itsDragPiece.
	MoveClump(	md,
				md->itsGameOf.Jigsaw.itsDragPiece,
				thePieceAmbientDeltaH,
				thePieceAmbientDeltaV);
}


static void MoveClump(
	ModelData	*md,
	JigsawPiece	*aPiece,
	double		aDeltaH,
	double		aDeltaV)
{
	unsigned int	i;

	//	Move aPiece.
	aPiece->itsPlacement.itsH += aDeltaH;
	aPiece->itsPlacement.itsV += aDeltaV;

	//	Move all pieces in the same clump.

	//	The recursive function SetNeighborPlacements() gets the placements
	//	right in the universal cover (using itsVisitedFlag to avoid visiting
	//	the same piece twice), then after it's done NormalizePlacement()
	//	projects each piece individually to the fundamental domain.

	for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )
		md->itsGameOf.Jigsaw.itsPieces[i].itsVisitedFlag = false;

	SetNeighborPlacements(aPiece, md->itsTopology);

	for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )	//	just do 'em all -- it's fast
		NormalizePlacement(&md->itsGameOf.Jigsaw.itsPieces[i].itsPlacement, md->itsTopology);
}


static void SetNeighborPlacements(
	JigsawPiece		*aPiece,
	TopologyType	aTopology)
{
	//	Assume aPiece->itsPlacement has already been set
	//	and transfer the placement to all connected neighbors,
	//	and so on recursively.
	//	Use itsVisitedFlag to avoid setting the same piece twice.

	aPiece->itsVisitedFlag = true;

	if (aPiece->itsNeighborBondWest
	 && ! aPiece->itsNeighborWest->itsVisitedFlag)
	{
		aPiece->itsNeighborWest->itsPlacement = aPiece->itsPlacement;

		if (aPiece->itsPlacement.itsFlip)
			aPiece->itsNeighborWest->itsPlacement.itsH += aPiece->itsPlacement.itsSizeH;
		else
			aPiece->itsNeighborWest->itsPlacement.itsH -= aPiece->itsPlacement.itsSizeH;

		SetNeighborPlacements(aPiece->itsNeighborWest, aTopology);
	}

	if (aPiece->itsNeighborBondEast
	 && ! aPiece->itsNeighborEast->itsVisitedFlag)
	{
		aPiece->itsNeighborEast->itsPlacement = aPiece->itsPlacement;

		if (aPiece->itsPlacement.itsFlip)
			aPiece->itsNeighborEast->itsPlacement.itsH -= aPiece->itsPlacement.itsSizeH;
		else
			aPiece->itsNeighborEast->itsPlacement.itsH += aPiece->itsPlacement.itsSizeH;

		SetNeighborPlacements(aPiece->itsNeighborEast, aTopology);
	}

	if (aPiece->itsNeighborBondSouth
	 && ! aPiece->itsNeighborSouth->itsVisitedFlag)
	{
		aPiece->itsNeighborSouth->itsPlacement = aPiece->itsPlacement;

		aPiece->itsNeighborSouth->itsPlacement.itsV -= aPiece->itsPlacement.itsSizeV;

		aPiece->itsNeighborSouth->itsPlacement.itsFlip
			^= ( aTopology == TopologyKlein  &&  aPiece->itsLogicalV == 0 );

		SetNeighborPlacements(aPiece->itsNeighborSouth, aTopology);
	}

	if (aPiece->itsNeighborBondNorth
	 && ! aPiece->itsNeighborNorth->itsVisitedFlag)
	{
		aPiece->itsNeighborNorth->itsPlacement = aPiece->itsPlacement;

		aPiece->itsNeighborNorth->itsPlacement.itsV += aPiece->itsPlacement.itsSizeV;

		aPiece->itsNeighborNorth->itsPlacement.itsFlip
			^= ( aTopology == TopologyKlein  &&  aPiece->itsNeighborNorth->itsLogicalV == 0 );

		SetNeighborPlacements(aPiece->itsNeighborNorth, aTopology);
	}
}


static void JigsawDragEnd(ModelData *md)
{
	//	Nudging a piece B into position next to a nearby piece C
	//	might in turn align piece B with some previously tested piece A.
	//	To allow for this "ripple effect" let's keep repeating the alignment
	//	process until no further progress occurs.  Typically, of course,
	//	one pass will suffice.

	bool	theSomePiecesGotJoinedFlag;

	theSomePiecesGotJoinedFlag = false;

	while (BindNearbyNeighbors(md))
	{
		//	Align the pieces within each clump.
		//	Detect and label nonorientable clumps,
		//	and nudge them to align with the symmetry axes.
		AlignAllClumps(md);

		//	Note that something happened.
		theSomePiecesGotJoinedFlag = true;
	}

	if (theSomePiecesGotJoinedFlag)
	{
		GatherClumps(md);
		PlayTheSound(SoundJigsawJoin);
		md->itsGameOf.Jigsaw.itsPuzzleTextureIsUpToDate = false;
	}

	if (PuzzleIsComplete(md))
	{
		md->itsGameIsOver = true;
		PlayTheSound(SoundJigsawComplete);
	}
}


static bool BindNearbyNeighbors(ModelData *md)
{
	bool			theProgressFlag,
					theFlipsAreConsistent;
	unsigned int	i;
	JigsawPiece		*thePiece;

	theProgressFlag = false;

	for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )
	{
		thePiece = &md->itsGameOf.Jigsaw.itsPieces[i];

		theFlipsAreConsistent = (thePiece->itsPlacement.itsFlip == thePiece->itsNeighborWest->itsPlacement.itsFlip);
		if
		(	! thePiece->itsNeighborBondWest
		 && ShortestDistance(
				thePiece->itsPlacement.itsFlip ?
					thePiece->itsPlacement.itsH + thePiece->itsPlacement.itsSizeH :
					thePiece->itsPlacement.itsH - thePiece->itsPlacement.itsSizeV,
				thePiece->itsPlacement.itsV,
				thePiece->itsNeighborWest->itsPlacement.itsH,
				thePiece->itsNeighborWest->itsPlacement.itsV,
				md->itsTopology,
				&theFlipsAreConsistent) < JIGSAW_EPSILON
		 && theFlipsAreConsistent )
		{
			//	Strictly speaking it's redundant to "set the bond"
			//	in both directions (because the loop will eventually
			//	visit both thePiece and thePiece->itsNeighborWest)
			//	but setting both directions at once guarantees
			//	"bond symmetry" without depending on consistent
			//	floating point results.
			thePiece->itsNeighborBondWest						= true;
			thePiece->itsNeighborWest->itsNeighborBondEast		= true;

			theProgressFlag = true;
		}

		theFlipsAreConsistent = (thePiece->itsPlacement.itsFlip == thePiece->itsNeighborEast->itsPlacement.itsFlip);
		if
		(	! thePiece->itsNeighborBondEast
		 && ShortestDistance(
				thePiece->itsPlacement.itsFlip ?
					thePiece->itsPlacement.itsH - thePiece->itsPlacement.itsSizeH :
					thePiece->itsPlacement.itsH + thePiece->itsPlacement.itsSizeH,
				thePiece->itsPlacement.itsV,
				thePiece->itsNeighborEast->itsPlacement.itsH,
				thePiece->itsNeighborEast->itsPlacement.itsV,
				md->itsTopology,
				&theFlipsAreConsistent) < JIGSAW_EPSILON
		 && theFlipsAreConsistent )
		{
			thePiece->itsNeighborBondEast						= true;
			thePiece->itsNeighborEast->itsNeighborBondWest		= true;

			theProgressFlag = true;
		}

		theFlipsAreConsistent = (
			(thePiece->itsPlacement.itsFlip == thePiece->itsNeighborSouth->itsPlacement.itsFlip)
		  ^ (md->itsTopology == TopologyKlein && thePiece->itsLogicalV == 0) );
		if
		(	! thePiece->itsNeighborBondSouth
		 && ShortestDistance(
				thePiece->itsPlacement.itsH,
				thePiece->itsPlacement.itsV - thePiece->itsPlacement.itsSizeV,
				thePiece->itsNeighborSouth->itsPlacement.itsH,
				thePiece->itsNeighborSouth->itsPlacement.itsV,
				md->itsTopology,
				&theFlipsAreConsistent) < JIGSAW_EPSILON
		 && theFlipsAreConsistent )
		{
			thePiece->itsNeighborBondSouth						= true;
			thePiece->itsNeighborSouth->itsNeighborBondNorth	= true;

			theProgressFlag = true;
		}

		theFlipsAreConsistent = (
			(thePiece->itsPlacement.itsFlip == thePiece->itsNeighborNorth->itsPlacement.itsFlip)
		  ^ (md->itsTopology == TopologyKlein && thePiece->itsNeighborNorth->itsLogicalV == 0) );
		if
		(
			! thePiece->itsNeighborBondNorth
		 && ShortestDistance(
				thePiece->itsPlacement.itsH,
				thePiece->itsPlacement.itsV + thePiece->itsPlacement.itsSizeV,
				thePiece->itsNeighborNorth->itsPlacement.itsH,
				thePiece->itsNeighborNorth->itsPlacement.itsV,
				md->itsTopology,
				&theFlipsAreConsistent) < JIGSAW_EPSILON
		 && theFlipsAreConsistent )
		{
			thePiece->itsNeighborBondNorth						= true;
			thePiece->itsNeighborNorth->itsNeighborBondSouth	= true;

			theProgressFlag = true;
		}
	}

	return theProgressFlag;
}


static void AlignAllClumps(ModelData *md)
{
	unsigned int	i;

	//	Detect nonorientable clumps.

	if (md->itsTopology == TopologyKlein)
	{
		//	Initialize itsVisitedFlag to be false and
		//	(tentatively!) assume all clumps to be orientable.
		for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )
		{
			md->itsGameOf.Jigsaw.itsPieces[i].itsVisitedFlag			= false;
			md->itsGameOf.Jigsaw.itsPieces[i].itsClumpIsNonorientable	= false;
		}

		//	For each...
		for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )
		{
			//	...unvisited piece...
			if ( ! md->itsGameOf.Jigsaw.itsPieces[i].itsVisitedFlag )
			{
				//	...imagine that piece's clump sitting in the universal cover.
				//	Set itsReflectionFlag = false (resp. true) for each piece
				//	in the clump whose orientation agrees (resp. disagrees)
				//	with the initial piece's orientation, and set itsVisitedFlag = true
				//	as well.  If the clump connects up with itself
				//	in an orientation-reversing way...
				if ( ! SetPieceOrientation(&md->itsGameOf.Jigsaw.itsPieces[i], false) )
				{
					//	...then align each piece in the clump to the symmetry axes
					//	and set itsClumpIsNonorientable to true.
					SetNonorientableClump(&md->itsGameOf.Jigsaw.itsPieces[i]);
				}
			}
		}
	}

	//	Align pieces within each clump.

	//		Initialize itsVisitedFlag to false;
	for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )
		md->itsGameOf.Jigsaw.itsPieces[i].itsVisitedFlag = false;

	//		Align the pieces in the universal cover,
	//		setting itsVisitedFlag as we go along.
	for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )
		if ( ! md->itsGameOf.Jigsaw.itsPieces[i].itsVisitedFlag )
			SetNeighborPlacements(&md->itsGameOf.Jigsaw.itsPieces[i], md->itsTopology);

	//		Project each piece to the fundamental domain.
	for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )
		NormalizePlacement(&md->itsGameOf.Jigsaw.itsPieces[i].itsPlacement, md->itsTopology);
}


static bool SetPieceOrientation(
	JigsawPiece	*aPiece,
	bool		aDesiredReflection)
{
	if ( ! aPiece->itsVisitedFlag )
	{
		//	aPiece has not yet been visited.

		//	Assign aDesiredReflection to this piece.
		aPiece->itsVisitedFlag		= true;
		aPiece->itsReflectionFlag	= aDesiredReflection;

		//	Visit its neighbors.
		//	Return false if inconsistencies arise.
		if (aPiece->itsNeighborBondWest)
		{
			if ( ! SetPieceOrientation(aPiece->itsNeighborWest, aDesiredReflection) )
				return false;
		}
		if (aPiece->itsNeighborBondEast)
		{
			if ( ! SetPieceOrientation(aPiece->itsNeighborEast, aDesiredReflection) )
				return false;
		}
		if (aPiece->itsNeighborBondSouth)
		{
			if ( ! SetPieceOrientation(
					aPiece->itsNeighborSouth,
					aPiece->itsLogicalV == 0 ? !aDesiredReflection : aDesiredReflection) )
				return false;
		}
		if (aPiece->itsNeighborBondNorth)
		{
			if ( ! SetPieceOrientation(
					aPiece->itsNeighborNorth,
					aPiece->itsNeighborNorth->itsLogicalV == 0 ? !aDesiredReflection : aDesiredReflection) )
				return false;
		}

		//	No inconsistencies arose, so return true.
		return true;
	}
	else
	{
		//	aPiece has already been visited.
		return (aPiece->itsReflectionFlag == aDesiredReflection);
	}

}


static void SetNonorientableClump(JigsawPiece *aPiece)
{
	if ( ! aPiece->itsClumpIsNonorientable )
	{
		aPiece->itsClumpIsNonorientable = true;

		AlignToSymmetryAxes(aPiece);

		if (aPiece->itsNeighborBondWest)
			SetNonorientableClump(aPiece->itsNeighborWest);
		if (aPiece->itsNeighborBondEast)
			SetNonorientableClump(aPiece->itsNeighborEast);
		if (aPiece->itsNeighborBondSouth)
			SetNonorientableClump(aPiece->itsNeighborSouth);
		if (aPiece->itsNeighborBondNorth)
			SetNonorientableClump(aPiece->itsNeighborNorth);
	}
}


static void AlignToSymmetryAxes(JigsawPiece *aPiece)
{
	//	If a piece sits in a nonorientable clump, then its logical
	//	coordinates determine its distance from the symmetry axes.
	//	The Klein bottle has two symmetry axes (at x = 0 and x = -0.5 = +0.5)
	//	so a piece's position is well-defined only up to multiples of 0.5.
	//	Moreover, when a piece is flipped it must lie on the
	//	"opposite side" of the given symmetry axis.

	double	thePredictedH,
			theError;

	//	Compute the canonical horizontal coordinate of aPiece's center.
	thePredictedH = -0.5 + (0.5 + aPiece->itsLogicalH)*(aPiece->itsPlacement.itsSizeH);

	//	If aPiece is flipped, it lies on the opposite side of the symmetry axis.
	//	(This is true for both symmetry axes simultaneously!)
	if (aPiece->itsPlacement.itsFlip)
		thePredictedH = - thePredictedH;

	//	There are two ways to assemble the jigsaw puzzle:
	//	one way preserves the symmetry axes individually while
	//	the other way interchanges them.  A horizontal translation
	//	of 1/2 (mod 1) takes one way to the other.
	theError = fabs(aPiece->itsPlacement.itsH - thePredictedH);
	if (theError > 0.5)		//	account for wraparound
		theError = 1.0 - theError;
	if (theError > 0.25)	//	switch axes if necessary
	{
		if (thePredictedH < 0.0)
			thePredictedH += 0.5;
		else
			thePredictedH -= 0.5;
	}
	aPiece->itsPlacement.itsH = thePredictedH;
}


static void GatherClumps(ModelData *md)
{
	//	Technical note:  It would almost always work fine
	//	to gather only the clump containing itsDragPiece,
	//	but for robustness let's gather all clumps (so for example
	//	if two background pieces happen to sit within epsilon
	//	of their correct relative position just by accident, and
	//	get attached with no user initiative, this algorithm
	//	will bring them together on the list).

	unsigned int	i,	//	index on main draw order list
					ii;	//	index on scratch list
	JigsawPiece		*theScratchList[MAX_JIGSAW_SIZE * MAX_JIGSAW_SIZE];

	//	Initialize itsVisitedFlag to false;
	for (i = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
		md->itsGameOf.Jigsaw.itsPieces[i].itsVisitedFlag = false;

	//	Go down the permanent list and copy pieces to theScratchList
	//	on a clump by clump basis.
	//	Use a recursion to "sniff out" each clump,
	//	marking pieces as visited as we copy them.
	for (i = 0, ii = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
		TransferClumpToScratchList(md->itsGameOf.Jigsaw.itsDrawOrder[i], theScratchList, &ii);

	//	An "unnecessary" error check.
	if (ii != md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize)
	{
		ErrorMessage(	"Wrong number of pieces transferred to theScratchList",
						"Error in GatherClumps()",
						false);
		return;
	}

	//	Copy theScratchList back to the permanent list.
	for (i = 0; i < md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i++)
		md->itsGameOf.Jigsaw.itsDrawOrder[i] = theScratchList[i];
}


static void TransferClumpToScratchList(
	JigsawPiece		*aPiece,
	JigsawPiece		**aScratchList,
	unsigned int	*aScratchListIndex)
{
	if ( ! aPiece->itsVisitedFlag )
	{
		aPiece->itsVisitedFlag = true;

		aScratchList[(*aScratchListIndex)++] = aPiece;

		if (aPiece->itsNeighborBondWest)
			TransferClumpToScratchList(aPiece->itsNeighborWest,  aScratchList, aScratchListIndex);
		if (aPiece->itsNeighborBondEast)
			TransferClumpToScratchList(aPiece->itsNeighborEast,  aScratchList, aScratchListIndex);
		if (aPiece->itsNeighborBondSouth)
			TransferClumpToScratchList(aPiece->itsNeighborSouth, aScratchList, aScratchListIndex);
		if (aPiece->itsNeighborBondNorth)
			TransferClumpToScratchList(aPiece->itsNeighborNorth, aScratchList, aScratchListIndex);
	}
}


static bool PuzzleIsComplete(ModelData *md)
{
	unsigned int	i;

	for (i = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; i-- > 0; )
	{
		if (! md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborBondWest
		 || ! md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborBondEast
		 || ! md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborBondSouth
		 || ! md->itsGameOf.Jigsaw.itsPieces[i].itsNeighborBondNorth )
		{
			return false;
		}
	}

	return true;
}


static void JigsawSpecialCommand(ModelData *md)
{
	//	Send the frontmost clump (the one containing the piece at index 0
	//	on itsDrawOrder) to the back of itsDrawOrder.
	SendClumpToFrontOrBack(md, 0, SendClumpToBack);
}


static void JigsawDraw(ModelData *md)
{
	//	If some neighboring pieces have recently been joined to each other,
	//	revise the puzzle texture to draw those pieces without tabs or borders
	//	on the affected sides.
	if ( ! md->itsGameOf.Jigsaw.itsPuzzleTextureIsUpToDate )
	{
		if (md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders    != NULL
		 && md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders != NULL)
		{
			UpdatePuzzleTexture(md);

			SetTextureImageFancy(	md->itsGameOf.Jigsaw.itsTextureNamePuzzle,
									md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders->itsWidth,
									md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders->itsHeight,
									4,
									(Byte *) md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders->itsPixels,
									md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
		}
		else	//	Puzzle image failed to load.
			SetTextureImageFancy(	md->itsGameOf.Jigsaw.itsTextureNamePuzzle,
									1,
									1,
									4,
									(Byte [4]){0x00, 0x00, 0xFF, 0xFF},	//	premultiplied alpha
									md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);

		md->itsGameOf.Jigsaw.itsPuzzleTextureIsUpToDate = true;
	}

	//	Draw the background.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	DrawFundamentalDomain(md, md->itsGameOf.Jigsaw.itsTextureNameBackground,  4);
#else
	DrawFundamentalDomain(md, md->itsGameOf.Jigsaw.itsTextureNameBackground,  8);
#endif

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(1.0, 0.0, 0.0, 1.0),
					PREMULTIPLY_RGBA(0.0, 0.0, 1.0, 1.0) );

	//	Draw the pieces.
	DrawPuzzlePieces(md);
}


static void UpdatePuzzleTexture(ModelData *md)
{
	unsigned int	thePuzzleSize,
					theImageSize;
	PixelRGBA		*theSrcPixel,
					*theDstPixel;
	Byte			*theDstPixelAssignment,
					*theDstPixelBorderFlags;
	unsigned int	theCount,
					h,
					v;
	JigsawPiece		*thePiece;

	if (md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders	== NULL
	 || md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders		== NULL
	 || md->itsGameOf.Jigsaw.itsPuzzleImagePixelAssignments	== NULL
	 || md->itsGameOf.Jigsaw.itsPuzzleImagePixelBorderFlags	== NULL
	 || md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders->itsWidth    != md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders->itsHeight
	 || md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders->itsWidth != md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders->itsHeight
	 || md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders->itsWidth != md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders->itsWidth)
	{
		return;
	}
	
	thePuzzleSize	= md->itsGameOf.Jigsaw.itsSize;
	theImageSize	= md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders->itsWidth;

	//	Wherever a piece joins to its neighbor,
	//	overwrite bordered piece quadrants with unbordered ones.

	theSrcPixel				= md->itsGameOf.Jigsaw.itsPuzzleImageWithoutBorders->itsPixels;
	theDstPixel				= md->itsGameOf.Jigsaw.itsPuzzleImageWithBorders->itsPixels;
	theDstPixelAssignment	= md->itsGameOf.Jigsaw.itsPuzzleImagePixelAssignments;
	theDstPixelBorderFlags	= md->itsGameOf.Jigsaw.itsPuzzleImagePixelBorderFlags;
	
	for (theCount = theImageSize * theImageSize; theCount-- > 0; )
	{
		//	Typically no border flags are set and no work needs to be done.
		if ( *theDstPixelBorderFlags )
		{
			//	We're in the atypical case of a pixel that sits on one or more boundary curves.
			//	If all its boundaries have been joined to neighboring pieces,
			//	then overwrite this pixel with an unblackened one.
			
			h = *theDstPixelAssignment >> 4;
			v = *theDstPixelAssignment & 0x0F;
			
			thePiece = &md->itsGameOf.Jigsaw.itsPieces[ h * thePuzzleSize  +  v ];
			
			if ((!(*theDstPixelBorderFlags & BORDER_WEST ) || thePiece->itsNeighborBondWest )
			 && (!(*theDstPixelBorderFlags & BORDER_EAST ) || thePiece->itsNeighborBondEast )
			 && (!(*theDstPixelBorderFlags & BORDER_SOUTH) || thePiece->itsNeighborBondSouth)
			 && (!(*theDstPixelBorderFlags & BORDER_NORTH) || thePiece->itsNeighborBondNorth))
			{
				*theDstPixel = *theSrcPixel;
			}
		}
		
		//	Move on to the next pixel.
		theSrcPixel++;
		theDstPixel++;
		theDstPixelAssignment++;
		theDstPixelBorderFlags++;
	}

}


static void DrawPuzzlePieces(ModelData *md)
{
	double				theNominalPieceSize;
	unsigned int		thePieceIndex;
	JigsawPiece			*thePiece;
	double				theSideWest,
						theSideEast,
						theSideSouth,
						theSideNorth,
						theOffsetH,
						theOffsetV;
	Placement			theOptimizedPiecePlacement;
	TexCoordTransform	theOptimizedTexCoordTransform;
	double				theCenterH,
						theCenterV,
						theSizeH,
						theSizeV;

	//	In an n×n puzzle each piece has nominal size 1/n (ignoring slots and tabs)
	//	both in the physical space and in the texture image.
	theNominalPieceSize = 1.0 / md->itsGameOf.Jigsaw.itsSize;

	//	Draw the pieces in reverse order for bottom-to-top layering.
	for (thePieceIndex = md->itsGameOf.Jigsaw.itsSize * md->itsGameOf.Jigsaw.itsSize; thePieceIndex-- > 0; )
	{
		//	Note the current piece.
		thePiece = md->itsGameOf.Jigsaw.itsDrawOrder[thePieceIndex];
		
		//	On the iPhone and iPod Touch, the jigsaw puzzle's pixel fill rate is a bottleneck.
		//	So choose an efficient bounding rectangle that encloses thePiece,
		//	taking into account that some sides may have a tab, some may have only a border,
		//	and some may have no decorations at all.

		theSideWest =
			thePiece->itsTabGenderWest == TabFemale ?
				-0.5 * MARGIN_FOR_BORDER :	//	narrow margin allowing for border outline
				-0.5 * MARGIN_FOR_TAB;		//	wide margin allowing for tab
		theSideEast =
			thePiece->itsTabGenderEast == TabFemale ?
				+0.5 * MARGIN_FOR_BORDER :	//	narrow margin allowing for border outline
				+0.5 * MARGIN_FOR_TAB;		//	wide margin allowing for tab
		theSideSouth =
			thePiece->itsTabGenderSouth == TabFemale ?
				-0.5 * MARGIN_FOR_BORDER :	//	narrow margin allowing for border outline
				-0.5 * MARGIN_FOR_TAB;		//	wide margin allowing for tab
		theSideNorth =
			thePiece->itsTabGenderNorth == TabFemale ?
				+0.5 * MARGIN_FOR_BORDER :	//	narrow margin allowing for border outline
				+0.5 * MARGIN_FOR_TAB;		//	wide margin allowing for tab
		
		theOffsetH = 0.5 * (theSideEast  + theSideWest ) * theNominalPieceSize;
		theOffsetV = 0.5 * (theSideNorth + theSideSouth) * theNominalPieceSize;
		theOptimizedPiecePlacement.itsH		= thePiece->itsPlacement.itsH
											+ (thePiece->itsPlacement.itsFlip ? -theOffsetH : theOffsetH);
		theOptimizedPiecePlacement.itsV		= thePiece->itsPlacement.itsV + theOffsetV;
		theOptimizedPiecePlacement.itsFlip	= thePiece->itsPlacement.itsFlip;
		theOptimizedPiecePlacement.itsAngle	= 0.0;
		theOptimizedPiecePlacement.itsSizeH	= theNominalPieceSize * (theSideEast  - theSideWest );
		theOptimizedPiecePlacement.itsSizeV	= theNominalPieceSize * (theSideNorth - theSideSouth);

		//	Visualize thePiece's texture coordinates in [0.0, 2.0] coordinates,
		//	but then multiply by 0.5 at the last minute to convert to [0.0, 1.0] coordinates.
		theCenterH = md->itsGameOf.Jigsaw.itsPuzzleImageFirstCenter
				   + thePiece->itsLogicalH * md->itsGameOf.Jigsaw.itsPuzzleImageStride;
		theCenterV = md->itsGameOf.Jigsaw.itsPuzzleImageFirstCenter
				   + thePiece->itsLogicalV * md->itsGameOf.Jigsaw.itsPuzzleImageStride;
		theSizeH = theNominalPieceSize * (theSideEast  - theSideWest );
		theSizeV = theNominalPieceSize * (theSideNorth - theSideSouth);
		theOptimizedTexCoordTransform.itsScaleFactor[0]	= 0.5 * theSizeH;
		theOptimizedTexCoordTransform.itsScaleFactor[1]	= 0.5 * theSizeV;
		theOptimizedTexCoordTransform.itsOffset[0]		= 0.5 * (theCenterH + theOffsetH - 0.5*theSizeH);
		theOptimizedTexCoordTransform.itsOffset[1]		= 0.5 * (theCenterV + theOffsetV - 0.5*theSizeV);

		//	Draw translated images of the piece.
		DrawSpriteWithTextureTransform(	md,
										&theOptimizedPiecePlacement,
										md->itsGameOf.Jigsaw.itsTextureNamePuzzle,
										&theOptimizedTexCoordTransform);
	}
}
