//	TorusGamesPool.c
//
//	Code specific to the Pool game.
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


#ifdef TORUS_GAMES_FOR_OPENGL_ES
#define HOLE_RADIUS			0.08
#define BALL_RADIUS			0.05
#else
#define HOLE_RADIUS			0.05
#define BALL_RADIUS			0.03
#endif
#define CUE_DEAD_ZONE		(2.0 * BALL_RADIUS)
#define STICK_FACTOR		1.50	//	how long to draw the stick [for rigorously correct drawing, 0.5*STICK_FACTOR - CUE_DEAD_ZONE shouldn't exceed 1.0]
#define GRAB_CUE_EPSILON	0.2		//	be generous with this, so most of the shaft is a valid grab target
#define SHOT_STRENGTH		2.0		//	(units/sec)/(cue length)
#define BALL_DECELERATION	0.05	//	units/sec²
#define FINAL_SPEED			0.1		//	final object ball speed when computer takes a shot
									//		0.001 works great, but a larger value makes the shot seem more human.
#define POCKET_FACTOR		17		//	Within the pocket, balls accelerate towards the center with this dimensionless multiple of the BALL_DECELERATION.
#define ENDPOINT_EPSILON	0.0001
#define STICK_PLACE_TIME	1.0		//	time interval to place cue stick for computer's shot	
#define STICK_PAUSE_TIME	1.0		//	time interval to pause cue stick before computer's shot	
#define TIME_EPSILON		0.00001	//	will depend on how accurately we can compute collision times
#define VERY_LONG_TIME		1.0e8	//	impossibly long time interval
#define ROOT3				1.73205080756887729353
#define STRETCH_FACTOR		1.001
#define GRIP_SIZE			0.1

//	When drawing a pool ball, it's best to leave a 1-pixel transparent margin
//	around the ball's image within the square texture, to avoid stray pixelization
//	artifacts where the ball's circumference is tangent to the square.
//	Because this makes the square texture 2 pixels wider and 2 pixels higher
//	than the ball itself, we need to draw the sprite that much bigger than
//	the ball's true diameter.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
//	iPhone Torus Games uses 32×32 ball textures
#define BALL_MARGIN_FACTOR	(32.0/30.0)
#else
//	desktop Torus Games uses 64×64 ball textures
#define BALL_MARGIN_FACTOR	(64.0/62.0)
#endif

//	In the touch-based interface, when continuing a previously started cue stick movement,
//	how close (in board units) must the new touch be to the previous cue stick grab point 
//	to count as a hit?
#define POOL_AIM_CONTINUATION_HIT_RADIUS	0.1

//	In the touch-based interface, the user taps to confirm the cue stick position
//	and initiate the shot.  How close must the touch point stay to the cue stick's
//	grip point in order to count as a tap?  POOL_TAP_TOLERANCE is currently set to zero,
//	to let the user make arbitrarily small adjustments without prematurely triggering the shot.
#define POOL_TAP_TOLERANCE					0.0

//	Provide a cue ball (0), three solid balls (1,2,3),
//	a black ball (4) and three striped balls (5,6,7).
#if NUM_POOL_BALLS != 8
	This pool implementation assumes NUM_POOL_BALLS == 8.
#endif


//	BallPair holds position and velocity data
//	for a pair of potentially colliding balls.
typedef struct
{
	double	xa,	//	position of Ball A
			ya,
			xb,	//	position of Ball B
			yb,
			ua,	//	velocity of Ball A
			va,
			ub,	//	velocity of Ball B
			vb;
	bool	fa,	//	Is Ball A flipped?
			fb;	//	Is Ball B flipped?
} BallPair;

//	BallPath describes a potential path that might prove useful
//	when the computer takes its shot (in human vs. computer mode).
typedef struct
{
	double	itsStartH,
			itsStartV,
			itsDirectionH,
			itsDirectionV,
			itsLength,
			itsNormalH,
			itsNormalV,
			itsStartSpeed;
} BallPath;

//	Complex multiplication is convenient for adding
//	a Gaussian error to the computer's shots.
typedef struct
{
	double	re,
			im;
} Complex;


//	Public functions
static void		PoolShutDown(ModelData *md);
static void		PoolSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void		PoolShutDownTextures(ModelData *md);
static void		PoolReset(ModelData *md);
static bool		PoolDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void		PoolDragObject(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
static void		PoolDragEnd(ModelData *md);
static void		PoolSimulationUpdate(ModelData *md);
static void		PoolDraw(ModelData *md);

//	Private functions
static void		CueVector(ModelData *md, double *aCueVectorH, double *aCueVectorV, double *aCueVectorLength);
static void		SetCueGripForVector(ModelData *md, double aCueVectorH, double aCueVectorV);
static void		Shoot1(ModelData *md);
static void		Shoot2(ModelData *md, double aTimeInterval);
static void		Shoot3(ModelData *md);
static bool		BallsAreRolling(ModelData *md);
static void		SinkBalls(ModelData *md);
static void		SinkOneBall(ModelData *md, unsigned int aBall);
static void		SimulateBallMotion(ModelData *md, double aTimeInterval);
static bool		GetFirstCollision(ModelData *md, double aTimeInterval, double *aCollisionTime, unsigned int *aBallA, unsigned int *aBallB);
static double	GetCollisionTime(ModelData *md, double aTimeInterval, unsigned int aBallA, unsigned int aBallB);
static void		SetBallPair(ModelData *md, BallPair *p, unsigned int aBallA, unsigned int aBallB);
static void		RollAllBalls(ModelData *md, double aTimeInterval);
static void		TransferMomentum(ModelData *md, unsigned int aBallA, unsigned int aBallB);
static void		PullIntoPocket(ModelData *md, double aTimeInterval);
static void		SlowTheBalls(ModelData *md, double aTimeInterval);
static void		RestoreBall(ModelData *md, unsigned int aBall);
static bool		BallPositionIsLegal(ModelData *md, unsigned int aBall);
static void		ComputerChoosesShot(ModelData *md);
static bool		ObjectBallIsValid(ModelData *md, unsigned int aBall);
static void		SetPath(BallPath *aPath, double h0, double v0, double h1, double v1, double s1);
static bool		PathIsClear(ModelData *md, BallPath *aPath, unsigned int aBall, bool aPocketTest);
static bool		PathAvoidsObstacle(BallPath *aPath, double anObstacleH, double anObstacleV, double aRadius);
static void		AddGaussianError(ModelData *md);
static double	GetRandomGaussian(void);
static Complex	ComplexMultiply(Complex z, Complex w);


void PoolSetUp(ModelData *md)
{
	unsigned int	i;

	//	Initialize function pointers.
	md->itsGameShutDown			= &PoolShutDown;
	md->itsGameSetUpTextures	= &PoolSetUpTextures;
	md->itsGameShutDownTextures	= &PoolShutDownTextures;
	md->itsGameReset			= &PoolReset;
	md->itsGameHandMoved		= NULL;
	md->itsGameDragBegin		= &PoolDragBegin;
	md->itsGameDragObject		= &PoolDragObject;
	md->itsGameDragEnd			= &PoolDragEnd;
	md->itsGameCharacterInput	= NULL;
	md->itsGameSpecialCommand	= NULL;
	md->itsGameSimulationUpdate	= &PoolSimulationUpdate;
	md->itsGameDraw				= &PoolDraw;

	//	Check the compile time parameters.
	if (0.5*STICK_FACTOR - CUE_DEAD_ZONE > 1.0)
		ErrorMessage(	"The torus games framework draws objects correctly iff they extend no further than 0.5 units beyond the boundary of the fundamental domain in which they are defined.",
						"Bad Pool Parameters",
						false);

	//	When playing human vs. computer,
	//	let the computer play for PoolPlayerB by default.
	md->itsGameOf.Pool.itsComputerPlayer = PoolPlayerB;
	
	//	Initialize game-specific texture names.
	md->itsGameOf.Pool.itsTextureNameTable	= 0;
	md->itsGameOf.Pool.itsTextureNamePocket	= 0;
	md->itsGameOf.Pool.itsTextureNameStick	= 0;
	md->itsGameOf.Pool.itsTextureNameSight	= 0;
	for (i = 0; i < NUM_POOL_BALLS; i++)
		md->itsGameOf.Pool.itsTextureNameBall[i] = 0;

	//	Initialize the game.
	PoolReset(md);
}


static void PoolShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	PoolShutDownTextures(md);
}


static void PoolSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameTable,	"Pool/Felt.rgba",			TextureWrap,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNamePocket,	"Pool/Pocket.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameStick,	"Pool/Stick.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameSight,	"Pool/LineOfSight.rgba",	TextureClamp,	TextureRGBA,	aFirstError);

	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameBall[0],	"Pool/Ball0.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameBall[1],	"Pool/Ball1.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameBall[2],	"Pool/Ball2.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameBall[3],	"Pool/Ball3.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameBall[4],	"Pool/Ball4.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameBall[5],	"Pool/Ball5.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameBall[6],	"Pool/Ball6.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Pool.itsTextureNameBall[7],	"Pool/Ball7.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
}


static void PoolShutDownTextures(ModelData *md)
{
	unsigned int	i;

	ShutDownOneTexture(&md->itsGameOf.Pool.itsTextureNameTable	);
	ShutDownOneTexture(&md->itsGameOf.Pool.itsTextureNamePocket	);
	ShutDownOneTexture(&md->itsGameOf.Pool.itsTextureNameStick	);
	ShutDownOneTexture(&md->itsGameOf.Pool.itsTextureNameSight	);

	for (i = 0; i < NUM_POOL_BALLS; i++)
		ShutDownOneTexture(&md->itsGameOf.Pool.itsTextureNameBall[i]);
}


static void PoolReset(ModelData *md)
{
	unsigned int	i;

	//	Note:  With the balls packed tightly together,
	//	the program can hang on the break (very rarely, but it happens).
	//	To avoid this, introduce a STRETCH_FACTOR to space the balls
	//	very slightly apart.  Confession:  I haven't rigorously tested
	//	this fix, because the bug is not readily reproducible.

	static const double	theIntialBallPlacement[NUM_POOL_BALLS][2] =
	{
		//	cue ball
		{	0.0,										 0.25											},
		
		//	solid balls
		{	0.0  -      BALL_RADIUS * STRETCH_FACTOR,	-0.25  -  ROOT3 * BALL_RADIUS * STRETCH_FACTOR	},
		{	0.0  +  2 * BALL_RADIUS * STRETCH_FACTOR,	-0.25											},
		{	0.0  -      BALL_RADIUS * STRETCH_FACTOR,	-0.25  +  ROOT3 * BALL_RADIUS * STRETCH_FACTOR	},

		//	4-ball
		{	0.0,										-0.25											},
		
		//	striped balls
		{	0.0  +      BALL_RADIUS * STRETCH_FACTOR,	-0.25  +  ROOT3 * BALL_RADIUS * STRETCH_FACTOR	},
		{	0.0  -  2 * BALL_RADIUS * STRETCH_FACTOR,	-0.25											},
		{	0.0  +      BALL_RADIUS * STRETCH_FACTOR,	-0.25  -  ROOT3 * BALL_RADIUS * STRETCH_FACTOR	}
	};

	
	//	All balls in play.
	for (i = 0; i < NUM_POOL_BALLS; i++)
		md->itsGameOf.Pool.itsBallInPlay[i] = true;

	//	Place all balls.
	for (i = 0; i < NUM_POOL_BALLS; i++)
	{
		md->itsGameOf.Pool.itsBallPlacement[i].itsH		= theIntialBallPlacement[i][0];
		md->itsGameOf.Pool.itsBallPlacement[i].itsV		= theIntialBallPlacement[i][1];
		md->itsGameOf.Pool.itsBallPlacement[i].itsFlip	= false;
		md->itsGameOf.Pool.itsBallPlacement[i].itsAngle	= 0.0;
		md->itsGameOf.Pool.itsBallPlacement[i].itsSizeH	= 2 * BALL_RADIUS * BALL_MARGIN_FACTOR;	//	sprite width  > ball diameter
		md->itsGameOf.Pool.itsBallPlacement[i].itsSizeV	= 2 * BALL_RADIUS * BALL_MARGIN_FACTOR;	//	sprite height > ball diameter
	}
	
	//	Set all ball velocities to zero.
	for (i = 0; i < NUM_POOL_BALLS; i++)
	{
		md->itsGameOf.Pool.itsBallVelocityH[i] = 0.0;
		md->itsGameOf.Pool.itsBallVelocityV[i] = 0.0;
	}
	
	//	Initialize the cue stick.
	md->itsGameOf.Pool.itsCueStatus = PoolCueVisible;
	SetCueGripForVector(md, 0.125, 0.250);	//	relative to cue ball position
	
	//	PlayerA goes first.
	md->itsGameOf.Pool.itsWhoseTurn = PoolPlayerA;
	
	//	We don't know yet who will shoot for solids
	//	and who will shoot for stripes.
	md->itsGameOf.Pool.itsPlayerAGoal = PoolUndecided;
	md->itsGameOf.Pool.itsPlayerBGoal = PoolUndecided;

	//	Nobody has yet won or lost.
	md->itsGameOf.Pool.itsWinFlag	= false;
	md->itsGameOf.Pool.itsLoseFlag	= false;

	//	Abort any pending simulation.
	SimulationEnd(md);
	
	//	Display the initial status message.
	PoolSetStatusMessage(md);
	
	//	Go.
	md->itsGameIsOver = false;
}


static bool PoolDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
	UNUSED_PARAMETER(aShiftOrControlKeyIsDown);
	
	if (md->itsGameIsOver)
		return false;

	if (md->itsGameOf.Pool.itsCueStatus == PoolCueNone)
		return false;

	//	The touch-based and mouse-based versions of torus pool handle drags differently.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Touch-based interface

	//	If a move is already in progress...
	if (md->itsGameOf.Pool.itsCueStatus == PoolCueActive)
	{
		//	...and the user has touched the cue grip...
		if (ShortestDistance(
					md->itsHandPlacement.itsH,
					md->itsHandPlacement.itsV,
					md->itsGameOf.Pool.itsCueGrip.itsH,
					md->itsGameOf.Pool.itsCueGrip.itsV,
					md->itsTopology,
					NULL)
				< POOL_AIM_CONTINUATION_HIT_RADIUS)
		{
			//	...then let the pending cue stick movement continue.

			//	That is, start a new drag which will be part of the same pending cue stick movement.
			//	Note that we do not force itsCueGrip to move to itsHandPlacement:
			//	a simple tap, even if slightly off-center, will serve to confirm 
			//	and initiate the intended shot without disturbing its direction.

			md->itsGameOf.Pool.itsDragStartH	= md->itsHandPlacement.itsH;
			md->itsGameOf.Pool.itsDragStartV	= md->itsHandPlacement.itsV;
			md->itsGameOf.Pool.itsDragWasTap	= true;

			return true;
		}
		else	//	But if the user touches elsewhere,
		{
			//	treat the touch as a scroll.
			//	The pre-existing cue stick movement remains valid but suspended,
			//	awaiting future touches for its final completion.
			return false;
		}
	}

	//	No move is in progress, so let the code below start a new one.
#else
	//	Mouse-based interface
	
	//	Moves are never left in suspension.
	if (md->itsGameOf.Pool.itsCueStatus == PoolCueActive)
		return false;	//	should never occur
#endif

	if (ShortestDistance(	md->itsHandPlacement.itsH,
							md->itsHandPlacement.itsV,
							md->itsGameOf.Pool.itsCueGrip.itsH,
							md->itsGameOf.Pool.itsCueGrip.itsV,
							md->itsTopology,
							NULL)
		< GRAB_CUE_EPSILON)
	{
		md->itsGameOf.Pool.itsCueStatus	= PoolCueActive;

#ifdef TORUS_GAMES_FOR_OPENGL_ES
		//	The touch-based interface doesn't move itsCueGrip until it knows 
		//	the drag is not a tap, so as not to disturb the user's carefully 
		//	aligned shot when s/he taps to execute it.
		md->itsGameOf.Pool.itsDragStartH	= md->itsHandPlacement.itsH;
		md->itsGameOf.Pool.itsDragStartV	= md->itsHandPlacement.itsV;
		md->itsGameOf.Pool.itsDragWasTap	= false;	//	A cue stick movement's first drag never counts as a terminating tap.
#else
		//	Let itsCueGrip track itsHandPlacement immediately.
		md->itsGameOf.Pool.itsCueGrip	= md->itsHandPlacement;
#endif

		return true;
	}
	else
	{
		return false;
	}
}


static void PoolDragObject(
	ModelData	*md,
	double		aHandLocalDeltaH,
	double		aHandLocalDeltaV)
{
	//	Move the hand.
	md->itsHandPlacement.itsH += (md->itsHandPlacement.itsFlip ? -aHandLocalDeltaH : aHandLocalDeltaH);
	md->itsHandPlacement.itsV += aHandLocalDeltaV;
	NormalizePlacement(&md->itsHandPlacement, md->itsTopology);

#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	If the cue stick grip point has travelled a non-trivial distance
	//	from the start of the present drag, then the drag is not a tap.
	//	A cue stick motion may consist of several separate drags, terminated by a tap.
	if (md->itsGameOf.Pool.itsDragWasTap)
	{
		if (ShortestDistance(
					md->itsGameOf.Pool.itsDragStartH,
					md->itsGameOf.Pool.itsDragStartV,
					md->itsHandPlacement.itsH,
					md->itsHandPlacement.itsV,
					md->itsTopology,
					NULL)
				> POOL_TAP_TOLERANCE)
		{
			md->itsGameOf.Pool.itsDragWasTap = false;
		}
	}

	//	The touch-based interface doesn't move itsCueGrip until it knows 
	//	the drag is not a tap, so as not to disturb the user's carefully 
	//	aligned shot when s/he taps to execute it.
	if ( ! md->itsGameOf.Pool.itsDragWasTap )
		md->itsGameOf.Pool.itsCueGrip = md->itsHandPlacement;
#else
	//	Let the cue stick grip track the hand at all times.
	md->itsGameOf.Pool.itsCueGrip = md->itsHandPlacement;
#endif
}


static void PoolDragEnd(ModelData *md)
{
	//	The touch-based and mouse-based versions of torus pool handle drags differently.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Touch-based interface
	
	//	If the drag was not a simple tap, then even though the drag is over,
	//	let the cue stick movement remain pending, awaiting subsequent drags.
	if ( ! md->itsGameOf.Pool.itsDragWasTap )
		return;
	
	//	A simple tap ends both the drag and the cue stick movement 
	//	that the drag is a part of.  Continue with the code below.
#else
	//	Mouse-based interface
	//	Each move consists of a single drag.
#endif

	//	Animate the human's shot if the pool cue's grip lies outside the dead zone.
	Shoot1(md);
}


static void CueVector(
	ModelData	*md,
	double		*aCueVectorH,
	double		*aCueVectorV,
	double		*aCueVectorLength)
{
	double	theBallH,
			theBallV,
			theGripH,
			theGripV,
			theFactor;

	theBallH = md->itsGameOf.Pool.itsBallPlacement[0].itsH;
	theBallV = md->itsGameOf.Pool.itsBallPlacement[0].itsV;
	
	theGripH = md->itsGameOf.Pool.itsCueGrip.itsH;
	theGripV = md->itsGameOf.Pool.itsCueGrip.itsV;
	
	if (theGripV - theBallV > +0.5)
	{
		theGripV -= 1.0;
		if (md->itsTopology == TopologyKlein)
			theGripH = - theGripH;
	}
	if (theGripV - theBallV < -0.5)
	{
		theGripV += 1.0;
		if (md->itsTopology == TopologyKlein)
			theGripH = - theGripH;
	}
	
	if (theGripH - theBallH > +0.5)
		theGripH -= 1.0;
	if (theGripH - theBallH < -0.5)
		theGripH += 1.0;

	*aCueVectorH = theBallH - theGripH;
	*aCueVectorV = theBallV - theGripV;
	
	*aCueVectorLength = sqrt(	(*aCueVectorH) * (*aCueVectorH)
							  + (*aCueVectorV) * (*aCueVectorV) );

	if (*aCueVectorLength > CUE_DEAD_ZONE)
	{
		theFactor = (*aCueVectorLength - CUE_DEAD_ZONE) / (*aCueVectorLength);
		
		*aCueVectorH		*= theFactor;
		*aCueVectorV		*= theFactor;
		*aCueVectorLength	*= theFactor;
	}
	else
	{
		*aCueVectorH		= 0.0;
		*aCueVectorV		= 0.0;
		*aCueVectorLength	= 0.0;
	}
}

static void SetCueGripForVector(
	ModelData	*md,
	double		aCueVectorH,
	double		aCueVectorV)
{
	double		theCueVectorLength;

	md->itsGameOf.Pool.itsCueGrip.itsH		= md->itsGameOf.Pool.itsBallPlacement[0].itsH;
	md->itsGameOf.Pool.itsCueGrip.itsV		= md->itsGameOf.Pool.itsBallPlacement[0].itsV;
	md->itsGameOf.Pool.itsCueGrip.itsFlip	= false;	//	ignored
	md->itsGameOf.Pool.itsCueGrip.itsAngle	= 0.0;		//	ignored
	md->itsGameOf.Pool.itsCueGrip.itsSizeH	= 1.0;		//	ignored
	md->itsGameOf.Pool.itsCueGrip.itsSizeV	= 1.0;		//	ignored

	theCueVectorLength = sqrt(aCueVectorH * aCueVectorH
							+ aCueVectorV * aCueVectorV);
	
	if (theCueVectorLength > 0.0)
	{
		aCueVectorH *= (theCueVectorLength + CUE_DEAD_ZONE) / theCueVectorLength;
		aCueVectorV *= (theCueVectorLength + CUE_DEAD_ZONE) / theCueVectorLength;
		
		md->itsGameOf.Pool.itsCueGrip.itsH -= aCueVectorH;
		md->itsGameOf.Pool.itsCueGrip.itsV -= aCueVectorV;
		
		NormalizePlacement(&md->itsGameOf.Pool.itsCueGrip, md->itsTopology);
	}
}


static void Shoot1(ModelData *md)
{
	double	theCueVectorH;
	double	theCueVectorV;
	double	theCueVectorLength;

	//	CueVector() returns a cue vector with the dead zone already subtracted off,
	//	so the shot will be valid iff theCueVectorLength > 0.
	CueVector(md, &theCueVectorH, &theCueVectorV, &theCueVectorLength);
	if (theCueVectorLength == 0.0)
	{
		//	Ignore null shots.
		md->itsGameOf.Pool.itsCueStatus = PoolCueVisible;
		SetCueGripForVector(md, 0.125, 0.250);	//	relative to cue ball position
		return;
	}

	//	The current player (human or computer) has chosen a shot.
	//	Set the balls rolling!

	//	Play the cue-stick-strikes-cue-ball sound.
	PlayTheSound(SoundPoolCue);
	
	//	Hide the cue stick while the balls are rolling.
	md->itsGameOf.Pool.itsCueStatus = PoolCueNone;
	
	//	Keep track of what balls get sunk.
	//	Assume a passive scratch until a valid ball drops.
	md->itsGameOf.Pool.itsPassiveScratchFlag	= true;
	md->itsGameOf.Pool.itsActiveScratchFlag		= false;
	
	//	CueVector

	//	Start the cue ball rolling.
	//	Store the velocity relative to the ball's local coordinates
	//	(which may be flipped in a Klein bottle).
	md->itsGameOf.Pool.itsBallVelocityH[0] = SHOT_STRENGTH * theCueVectorH;
	md->itsGameOf.Pool.itsBallVelocityV[0] = SHOT_STRENGTH * theCueVectorV;
	if (md->itsGameOf.Pool.itsBallPlacement[0].itsFlip)
		md->itsGameOf.Pool.itsBallVelocityH[0] = - md->itsGameOf.Pool.itsBallVelocityH[0];
	
	//	Animate the balls (by calling Shoot2()) until all have stopped rolling,
	//	then finalize the shot (by calling Shoot3()).
	SimulationBegin(md, SimulationPoolRollBalls);
}


static void Shoot2(
	ModelData	*md,
	double		aTimeInterval)	//	in seconds
{
	//	The balls are in motion.
	//	Advance the simulation by one frame.

	//	Is any ball completely over the pocket?
	//	(Note that we sink a ball just after letting the UI draw it,
	//	rather than just before, so that the user sees the ball
	//	fall all the way into the hole before it disappears.)
	SinkBalls(md);

	//	Simulate the balls' physics through aTimeInterval.
	SimulateBallMotion(md, aTimeInterval);
	
	//	The pocket attracts balls within it.
	PullIntoPocket(md, aTimeInterval);
	
	//	Rolling resistance slows the balls.
	SlowTheBalls(md, aTimeInterval);
}


static void Shoot3(ModelData *md)
{
	//	The balls have come to rest.
	//	Evaluate the result.

	//	What balls dropped?
	if (md->itsGameOf.Pool.itsWinFlag)
	{
		md->itsGameIsOver = true;
		PlayTheSound(SoundPoolWin);
	}
	else if (md->itsGameOf.Pool.itsLoseFlag)
	{
		md->itsGameIsOver = true;
		PlayTheSound(SoundPoolLoss);
	}
	else
	{
		if (md->itsGameOf.Pool.itsPassiveScratchFlag || md->itsGameOf.Pool.itsActiveScratchFlag)
			md->itsGameOf.Pool.itsWhoseTurn = ! md->itsGameOf.Pool.itsWhoseTurn;
	}

	//	Update the status message.
	PoolSetStatusMessage(md);
	
	//	Restore the cue ball if necessary.
	if ( ! md->itsGameOf.Pool.itsBallInPlay[0] )
		RestoreBall(md, 0);

	//	Show the cue stick.
	md->itsGameOf.Pool.itsCueStatus = PoolCueVisible;

	//	If it's the computer's turn to shoot, let it
	//	(regardless of whether it or the human took the preceding shot).
	if ( ! md->itsGameIsOver
	 && md->itsHumanVsComputer
	 && md->itsGameOf.Pool.itsWhoseTurn == md->itsGameOf.Pool.itsComputerPlayer)
	{
		//	Choose a shot and execute it.
		ComputerChoosesShot(md);
		AddGaussianError(md);
		SimulationBegin(md, SimulationPoolPlaceCueStick);
	}
	else
	{
		//	Set a default cue stick position in preparation for the human's turn.
		SetCueGripForVector(md, 0.125, 0.250);	//	relative to cue ball position
	}
}


static bool BallsAreRolling(ModelData *md)
{
	unsigned int	i;

	for (i = 0; i < NUM_POOL_BALLS; i++)
	{
		if (md->itsGameOf.Pool.itsBallInPlay[i]
		 && (	md->itsGameOf.Pool.itsBallVelocityH[i] != 0.0
			 || md->itsGameOf.Pool.itsBallVelocityV[i] != 0.0))
		{
			return true;
		}
	}

	return false;
}


static void SinkBalls(ModelData *md)
{
	unsigned int	i,
					theNumBallsOverPocket;
	double			theDistanceFromHole[NUM_POOL_BALLS],
					theSinkRadius;
	
	//	On rare occasions two balls can jam against each other
	//	in the pocket, with neither ball wholly contained therein.
	//	In and of itself that's not a problem, but in practice
	//	the pocket's attractive force makes the two balls bounce
	//	against each other, and thus the simulation never
	//	terminates because the balls never stop moving.
	//	To avoid this problem, make a rule that if the centers
	//	of two balls are simultaneously in the pocket,
	//	both balls get sunk.  Yes, I know this is a real hack,
	//	but it seems to be the cleanest solution possible.

	//	We'll count how many balls have their center-of-mass over the pocket.
	theNumBallsOverPocket = 0;

	//	Precompute each ball's distance from the hole.
	for (i = 0; i < NUM_POOL_BALLS; i++)
	{
		if (md->itsGameOf.Pool.itsBallInPlay[i])
		{
			theDistanceFromHole[i] = sqrt(	md->itsGameOf.Pool.itsBallPlacement[i].itsH * md->itsGameOf.Pool.itsBallPlacement[i].itsH
										  + md->itsGameOf.Pool.itsBallPlacement[i].itsV * md->itsGameOf.Pool.itsBallPlacement[i].itsV);

			if (theDistanceFromHole[i] < HOLE_RADIUS)
				theNumBallsOverPocket++;
		}
		else
			theDistanceFromHole[i] = 0.0;	//	unused
	}
	
	//	If two balls sit over the pocket both get sunk immediately (cf. above).
	//	If only one ball sits over the pocket, sink it only if it's fallen
	//	all the way in.
	if (theNumBallsOverPocket >= 2)
		theSinkRadius = HOLE_RADIUS;
	else
		theSinkRadius = HOLE_RADIUS - BALL_RADIUS;
	
	//	Sink a ball if it's within theSinkRadius.
	for (i = 0; i < NUM_POOL_BALLS; i++)
	{
		if (md->itsGameOf.Pool.itsBallInPlay[i]
		 && theDistanceFromHole[i] < theSinkRadius)
		{
			SinkOneBall(md, i);
		}
	}
}


static void SinkOneBall(
	ModelData		*md,
	unsigned int	aBall)
{
	PoolGoal	theGoal;

	//	Play the sinking sound.
	PlayTheSound(SoundPoolSink);

	//	Which balls is the current player shooting for?
	theGoal = (md->itsGameOf.Pool.itsWhoseTurn == PoolPlayerA ?
				md->itsGameOf.Pool.itsPlayerAGoal : md->itsGameOf.Pool.itsPlayerBGoal);

	//	Before sinking this ball,
	//	figure out what it will mean for the game.
	if (theGoal == PoolUndecided)
	{
		switch (aBall)
		{
			case 0:		//	cue ball
				md->itsGameOf.Pool.itsActiveScratchFlag = true;
				break;
			
			case 1:		//	solid ball
			case 2:
			case 3:
				if (md->itsGameOf.Pool.itsBallInPlay[0] && md->itsGameOf.Pool.itsBallInPlay[4])
				{
					if (md->itsGameOf.Pool.itsWhoseTurn == PoolPlayerA)
					{
						md->itsGameOf.Pool.itsPlayerAGoal = PoolSolids;
						md->itsGameOf.Pool.itsPlayerBGoal = PoolStripes;
					}
					else
					{
						md->itsGameOf.Pool.itsPlayerAGoal = PoolStripes;
						md->itsGameOf.Pool.itsPlayerBGoal = PoolSolids;
					}
					md->itsGameOf.Pool.itsPassiveScratchFlag = false;
				}
				break;
			
			case 4:		//	4-ball
				md->itsGameOf.Pool.itsLoseFlag = true;
				break;

			case 5:		//	striped ball
			case 6:
			case 7:
				if (md->itsGameOf.Pool.itsBallInPlay[0] && md->itsGameOf.Pool.itsBallInPlay[4])
				{
					if (md->itsGameOf.Pool.itsWhoseTurn == PoolPlayerA)
					{
						md->itsGameOf.Pool.itsPlayerAGoal = PoolStripes;
						md->itsGameOf.Pool.itsPlayerBGoal = PoolSolids;
					}
					else
					{
						md->itsGameOf.Pool.itsPlayerAGoal = PoolSolids;
						md->itsGameOf.Pool.itsPlayerBGoal = PoolStripes;
					}
					md->itsGameOf.Pool.itsPassiveScratchFlag = false;
				}
				break;
		}
	}
	else	//	itsGameOf.Pool.itsPlayerGoals (solids or stripes) have already been decided
	{
		if ( ! md->itsGameOf.Pool.itsBallInPlay[4])		//	4-ball already sunk
		{
			md->itsGameOf.Pool.itsWinFlag	= false;
			md->itsGameOf.Pool.itsLoseFlag	= true;
		}
		else if ( ! md->itsGameOf.Pool.itsBallInPlay[0])	//	cue ball already sunk
		{
			if (aBall == 4)
				md->itsGameOf.Pool.itsLoseFlag	= true;
		}
		else	//	neither 4-ball nor cue ball previously sunk
		{
			switch (aBall)
			{
				case 0:		//	cue ball
					md->itsGameOf.Pool.itsActiveScratchFlag = true;
					break;
				
				case 1:		//	solid ball
				case 2:
				case 3:
					if (theGoal == PoolSolids)
						md->itsGameOf.Pool.itsPassiveScratchFlag = false;
					else
						md->itsGameOf.Pool.itsActiveScratchFlag = true;
					break;
				
				case 4:		//	4-ball
					if (theGoal == PoolBlack)
						md->itsGameOf.Pool.itsWinFlag	= true;
					else
						md->itsGameOf.Pool.itsLoseFlag	= true;
					break;

				case 5:		//	striped ball
				case 6:
				case 7:
					if (theGoal == PoolStripes)
						md->itsGameOf.Pool.itsPassiveScratchFlag = false;
					else
						md->itsGameOf.Pool.itsActiveScratchFlag = true;
					break;
			}
		}
	}
	
	//	Sink the ball.
	md->itsGameOf.Pool.itsBallInPlay[aBall] = false;
	
	//	Has either player qualified to shoot for the black ball?
	if ( ! (md->itsGameOf.Pool.itsBallInPlay[1] ||
			md->itsGameOf.Pool.itsBallInPlay[2] ||
			md->itsGameOf.Pool.itsBallInPlay[3]	))
	{
		if (md->itsGameOf.Pool.itsPlayerAGoal == PoolSolids)
			md->itsGameOf.Pool.itsPlayerAGoal = PoolBlack;
		if (md->itsGameOf.Pool.itsPlayerBGoal == PoolSolids)
			md->itsGameOf.Pool.itsPlayerBGoal = PoolBlack;
	}
	if ( ! (md->itsGameOf.Pool.itsBallInPlay[5] ||
			md->itsGameOf.Pool.itsBallInPlay[6] ||
			md->itsGameOf.Pool.itsBallInPlay[7]	))
	{
		if (md->itsGameOf.Pool.itsPlayerAGoal == PoolStripes)
			md->itsGameOf.Pool.itsPlayerAGoal = PoolBlack;
		if (md->itsGameOf.Pool.itsPlayerBGoal == PoolStripes)
			md->itsGameOf.Pool.itsPlayerBGoal = PoolBlack;
	}
}


static void SimulateBallMotion(
	ModelData	*md,
	double		aTimeInterval)	//	in seconds
{
	double			theCollisionTime;
	unsigned int	theBallA	= 0,
					theBallB	= 0;

	while (GetFirstCollision(md, aTimeInterval, &theCollisionTime, &theBallA, &theBallB))
	{
		RollAllBalls(md, theCollisionTime);
		TransferMomentum(md, theBallA, theBallB);
		aTimeInterval -= theCollisionTime;
	}

	RollAllBalls(md, aTimeInterval);
}


static bool GetFirstCollision(
	ModelData		*md,
	double			aTimeInterval,		//	input, in seconds
	double			*aCollisionTime,	//	output (valid iff function returns true)
	unsigned int	*aBallA,			//	output (valid iff function returns true)
	unsigned int	*aBallB)			//	output (valid iff function returns true)
{
	//	In principle we want to test for collisions at times in the range
	//	[0, aTimeInterval].  In practice we extend the range to
	//	[-TIME_EPSILON, aTimeInterval + TIME_EPSILON] to make sure
	//	no collisions slip through the cracks.  We won't have to worry
	//	about seeing the same collision twice if we consider only collisions
	//	in which the balls are moving towards each other.

	double			theCollisionTime;
	unsigned int	theBallA,
					theBallB;
	
	//	Initialize *aCollisionTime to an impossibly large value.
	*aCollisionTime = VERY_LONG_TIME;
	
	//	Consider all pairs of balls.
	for (theBallA = 0; theBallA < NUM_POOL_BALLS; theBallA++)
		if (md->itsGameOf.Pool.itsBallInPlay[theBallA])
			for (theBallB = theBallA + 1; theBallB < NUM_POOL_BALLS; theBallB++)
				if (md->itsGameOf.Pool.itsBallInPlay[theBallB])
				{
					theCollisionTime = GetCollisionTime(md, aTimeInterval, theBallA, theBallB);
					if (theCollisionTime < *aCollisionTime)
					{
						*aCollisionTime	= theCollisionTime;
						*aBallA			= theBallA;
						*aBallB			= theBallB;
					}
				}
	
	//	Report the collision or lack thereof.
	if (*aCollisionTime < aTimeInterval + TIME_EPSILON)
		return true;
	else
	{
		*aCollisionTime	= 0;
		*aBallA			= 0;
		*aBallB			= 0;
		return false;
	}
}


static double GetCollisionTime(
	ModelData		*md,
	double			aTimeInterval,	//	in seconds
	unsigned int	aBallA,
	unsigned int	aBallB)
{
	//	If Ball A and Ball B will collide within the time interval
	//	[-TIME_EPSILON, aTimeInterval + TIME_EPSILON], return the time
	//	of the inbound collision.  Otherwise return a VERY_LONG_TIME.
	
	BallPair	p;
	double		a,
				b,
				c,
				theDiscriminant,
				t;
	
	//	Locate images of BallA and BallB as close to each other as possible.
	SetBallPair(md, &p, aBallA, aBallB);
	
	//	The two balls will overlap when the difference vector
	//
	//		[(xa,ya) + t*(ua,va)] - [(xb,yb) + t*(ub,vb)]
	//
	//	has length less than 2*BALL_RADIUS.  Expanding that expression gives
	//
	//		[(xa - xb) + t*(ua - ub), (ya - yb) + t*(va - vb)]
	//
	//	whose squared length is
	//
	//		t^2 * [(ua - ub)^2 + (va - vb)^2]
	//	  +  t  * [2*(xa - xb)*(ua - ub) + 2*(ya - yb)*(va - vb)]
	//	  +  1  * [(xa - xb)^2 + (ya - yb)^2].
	//
	//	Set that last expression equal to (2*BALL_RADIUS)^2
	//	and write down the coefficients of the quadratic equation
	//
	//		a t^2 + b t + c = 0.
	a = (p.ua - p.ub)*(p.ua - p.ub) + (p.va - p.vb)*(p.va - p.vb);
	b = 2*(p.xa - p.xb)*(p.ua - p.ub) + 2*(p.ya - p.yb)*(p.va - p.vb);
	c = (p.xa - p.xb)*(p.xa - p.xb) + (p.ya - p.yb)*(p.ya - p.yb) - 4*BALL_RADIUS*BALL_RADIUS;
	
	//	In the typical case, the discriminant will be negative and
	//	the balls will not collide.
	theDiscriminant = b*b - 4*a*c;
	if (theDiscriminant <= 0.0)
		return VERY_LONG_TIME;
	
	//	Ignore the special case that the balls are travelling
	//	parallel to each other.  This case will occur in the original
	//	set-up (with seven motionless balls tangent to each other)
	//	but will otherwise be rare.
	if (a == 0.0)	//	could test against an epsilon, but maybe it isn't necessary
		return VERY_LONG_TIME;
	
	//	At this point we know the discriminant is positive,
	//	so we'll have an "inbound" collision (which we want)
	//	followed by an "outbound" collision (which we ignore).
	//	We know a > 0, so we can take the - part of the +/-
	//	in the quadratic formula.
	t = (-b - sqrt(theDiscriminant)) / (2*a);
	
	//	Does the collision occur within the time interval
	//	[-TIME_EPSILON, aTimeInterval + TIME_EPSILON] ?
	if (t > -TIME_EPSILON  &&  t < aTimeInterval + TIME_EPSILON)
		return t;
	else
		return VERY_LONG_TIME;
}


static void SetBallPair(
	ModelData		*md,
	BallPair		*p,
	unsigned int	aBallA,
	unsigned int	aBallB)
{
	//	Where are Ball A and Ball B now?
	p->xa = md->itsGameOf.Pool.itsBallPlacement[aBallA].itsH;
	p->ya = md->itsGameOf.Pool.itsBallPlacement[aBallA].itsV;
	p->xb = md->itsGameOf.Pool.itsBallPlacement[aBallB].itsH;
	p->yb = md->itsGameOf.Pool.itsBallPlacement[aBallB].itsV;
	
	//	Note the balls' velocities in their own local ball coordinates...
	p->ua = md->itsGameOf.Pool.itsBallVelocityH[aBallA];
	p->va = md->itsGameOf.Pool.itsBallVelocityV[aBallA];
	p->ub = md->itsGameOf.Pool.itsBallVelocityH[aBallB];
	p->vb = md->itsGameOf.Pool.itsBallVelocityV[aBallB];
	
	//	...then convert to global fundamental domain coordinates.
	p->fa = md->itsGameOf.Pool.itsBallPlacement[aBallA].itsFlip;
	if (p->fa)
		p->ua = - p->ua;
	p->fb = md->itsGameOf.Pool.itsBallPlacement[aBallB].itsFlip;
	if (p->fb)
		p->ub = - p->ub;
	
	//	Leaving Ball A where it is, find the closest translate of Ball B.
	//	The two balls must be quite close in order to interact within aTimeInterval
	//	(so we can safely be loose with the definition of "nearest")
	//	but we do need to allow for the possibility that the two balls
	//	straddle a boundary of the fundamental domain.
	if (p->yb - p->ya > +0.5)
	{
		p->yb -= 1.0;
		if (md->itsTopology == TopologyKlein)
		{
			p->xb = - p->xb;
			p->ub = - p->ub;
			p->fb = ! p->fb;
		}
	}
	if (p->yb - p->ya < -0.5)
	{
		p->yb += 1.0;
		if (md->itsTopology == TopologyKlein)
		{
			p->xb = - p->xb;
			p->ub = - p->ub;
			p->fb = ! p->fb;
		}
	}
	if (p->xb - p->xa > +0.5)
		p->xb -= 1.0;
	if (p->xb - p->xa < -0.5)
		p->xb += 1.0;
}


static void RollAllBalls(
	ModelData	*md,
	double		aTimeInterval)	//	in seconds
{
	unsigned int	i;
	double			theVelocityH,
					theVelocityV;

	for (i = 0; i < NUM_POOL_BALLS; i++)
		if (md->itsGameOf.Pool.itsBallInPlay[i])
		{
			//	Convert the ball's velocity from the ball's local coordinates
			//	to global fundamental domain coordinates.
			theVelocityH = md->itsGameOf.Pool.itsBallVelocityH[i];
			theVelocityV = md->itsGameOf.Pool.itsBallVelocityV[i];
			if (md->itsGameOf.Pool.itsBallPlacement[i].itsFlip)
				theVelocityH = - theVelocityH;
		
			//	Move the ball.
			md->itsGameOf.Pool.itsBallPlacement[i].itsH += theVelocityH * aTimeInterval;
			md->itsGameOf.Pool.itsBallPlacement[i].itsV += theVelocityV * aTimeInterval;
			
			//	Normalize the ball's position.
			NormalizePlacement(&md->itsGameOf.Pool.itsBallPlacement[i], md->itsTopology);
		}
}


static void TransferMomentum(
	ModelData		*md,
	unsigned int	aBallA,
	unsigned int	aBallB)
{
	//	Assume Ball A and Ball B are just now colliding.
	//	Transfer momentum from one to the other in such a way that
	//
	//		1.	The momentum transfer is parallel to the vector running
	//			from the center of Ball A to the center of Ball B.
	//
	//		2.	Energy is conserved.
	
	BallPair	p;
	double		dx,
				dy,
				theLength,
				dp;

	//	Play the ball clicking sound.
	PlayTheSound(SoundPoolClick);

	//	Locate images of BallA and BallB as close to each other as possible.
	SetBallPair(md, &p, aBallA, aBallB);
	
	//	Compute a unit vector running from the center of Ball A
	//	to the center of Ball B.
	dx = p.xb - p.xa;
	dy = p.yb - p.ya;
	theLength = sqrt(dx*dx + dy*dy);
	dx /= theLength;
	dy /= theLength;
	
	//	We want to transfer a (possibly negative) quantity dp of momentum
	//	from Ball A to Ball B along the unit vector (dx,dy).
	//	That is, Ball B will gain momentum dp*(dx,dy)
	//	while Ball A looses that same amount.
	//	The total kinetic energy before the collision (assuming all balls have
	//	unit mass) is
	//
	//		(1/2)*[ua^2 + va^2 + ub^2 + vb^2].
	//
	//	The total kinetic energy after the collision will be
	//
	//		(1/2)*[(ua - dp*dx)^2 + (va - dp*dy)^2 + (ub + dp*dx)^2 + (vb + dp*dy)^2].
	//
	//	Expanding those two expressions and setting them equal gives
	//
	//		dp^2 + dp*[dx*(ub - ua) + dy*(vb - va)] = 0.
	//
	//	Ignoring the trivial solution dp == 0 gives the desired
	//
	//		dp = dx*(ua - ub) + dy*(va - vb).

	dp = dx*(p.ua - p.ub) + dy*(p.va - p.vb);
	
	p.ua -= dp*dx;
	p.va -= dp*dy;
	
	p.ub += dp*dx;
	p.vb += dp*dy;

	//	Copy the final velocities from the BallPair's global coordinates
	//	back to each ball's local coordinates.
	md->itsGameOf.Pool.itsBallVelocityH[aBallA] = (p.fa ? - p.ua : p.ua);
	md->itsGameOf.Pool.itsBallVelocityV[aBallA] = p.va;
	md->itsGameOf.Pool.itsBallVelocityH[aBallB] = (p.fb ? - p.ub : p.ub);
	md->itsGameOf.Pool.itsBallVelocityV[aBallB] = p.vb;
}


static void PullIntoPocket(
	ModelData	*md,
	double		aTimeInterval)	//	in seconds
{
	//	When a ball rolls over the pocket,
	//	bend its trajectory towards the pocket's center.

	unsigned int	i;
	double			theRadius,
					theFactor,
					theDirectionH,
					theDirectionV;

	for (i = 0; i < NUM_POOL_BALLS; i++)
		if (md->itsGameOf.Pool.itsBallInPlay[i])
		{
			//	The pocket sits at coordinates (0,0), which simplifies the computations.

			//	How far is the ball from the hole (center-to-center) ?
			theRadius = sqrt(	md->itsGameOf.Pool.itsBallPlacement[i].itsH * md->itsGameOf.Pool.itsBallPlacement[i].itsH
							  + md->itsGameOf.Pool.itsBallPlacement[i].itsV * md->itsGameOf.Pool.itsBallPlacement[i].itsV);

			//	If the ball lies over the hole...
			if (theRadius < HOLE_RADIUS)
			{
				if (theRadius > 0.01 * HOLE_RADIUS)
					theFactor = -1.0 / theRadius;
				else
					theFactor = 0.0;	//	special case to avoid division by zero or other very small numbers

				theDirectionH = theFactor * md->itsGameOf.Pool.itsBallPlacement[i].itsH;
				theDirectionV = theFactor * md->itsGameOf.Pool.itsBallPlacement[i].itsV;
				
				//	Convert the direction to the ball's local coordinate system.
				if (md->itsGameOf.Pool.itsBallPlacement[i].itsFlip)
					theDirectionH = - theDirectionH;
				
				//	Note:  Consistent use of POCKET_FACTOR in SlowTheBalls()
				//	as well as here in PullIntoPocket() guarantees that ball
				//	speeds will never increase.
				md->itsGameOf.Pool.itsBallVelocityH[i] += (POCKET_FACTOR * BALL_DECELERATION * aTimeInterval) * theDirectionH;
				md->itsGameOf.Pool.itsBallVelocityV[i] += (POCKET_FACTOR * BALL_DECELERATION * aTimeInterval) * theDirectionV;
			}
		}
}


static void SlowTheBalls(
	ModelData	*md,
	double		aTimeInterval)	//	in seconds
{
	unsigned int	i;
	double			theSpeed,
					theDecrement,
					theFactor;

	for (i = 0; i < NUM_POOL_BALLS; i++)
		if (md->itsGameOf.Pool.itsBallInPlay[i])
		{
			theSpeed = sqrt(md->itsGameOf.Pool.itsBallVelocityH[i]*md->itsGameOf.Pool.itsBallVelocityH[i] + md->itsGameOf.Pool.itsBallVelocityV[i]*md->itsGameOf.Pool.itsBallVelocityV[i]);

			theDecrement = BALL_DECELERATION * aTimeInterval;

			//	The ball slows down faster when over the pocket.
			//
			//	Note:  Consistent use of POCKET_FACTOR in PullIntoPocket()
			//	as well as here in SlowTheBalls() guarantees that ball
			//	speeds will never increase.
			if (md->itsGameOf.Pool.itsBallPlacement[i].itsH * md->itsGameOf.Pool.itsBallPlacement[i].itsH
			  + md->itsGameOf.Pool.itsBallPlacement[i].itsV * md->itsGameOf.Pool.itsBallPlacement[i].itsV
			  < HOLE_RADIUS * HOLE_RADIUS)
			{
				theDecrement *= POCKET_FACTOR;
			}

			if (theSpeed <= theDecrement)
			{
				md->itsGameOf.Pool.itsBallVelocityH[i] = 0.0;
				md->itsGameOf.Pool.itsBallVelocityV[i] = 0.0;
			}
			else
			{
				theFactor = (theSpeed - theDecrement) / theSpeed;
				md->itsGameOf.Pool.itsBallVelocityH[i] *= theFactor;
				md->itsGameOf.Pool.itsBallVelocityV[i] *= theFactor;
			}
		}
}


static void RestoreBall(
	ModelData		*md,
	unsigned int	aBall)
{
	//	Keep trying random locations until the ball
	//	overlaps neither the pocket nor another ball.
	do
	{
		md->itsGameOf.Pool.itsBallPlacement[aBall].itsH		= ((double)rand() / (double)RAND_MAX) - 0.5;
		md->itsGameOf.Pool.itsBallPlacement[aBall].itsV		= ((double)rand() / (double)RAND_MAX) - 0.5;
		md->itsGameOf.Pool.itsBallPlacement[aBall].itsFlip	= false;

	} while ( ! BallPositionIsLegal(md, aBall) );

	//	Set the ball's velocity to zero.
	md->itsGameOf.Pool.itsBallVelocityH[aBall] = 0.0;
	md->itsGameOf.Pool.itsBallVelocityV[aBall] = 0.0;
	
	//	Put the ball into play.
	md->itsGameOf.Pool.itsBallInPlay[aBall] = true;
}


static bool BallPositionIsLegal(
	ModelData		*md,
	unsigned int	aBall)
{
	unsigned int	i;
	BallPair		p;
	double			dx,
					dy;

	//	Don't overlap the pocket.
	if (md->itsGameOf.Pool.itsBallPlacement[aBall].itsH * md->itsGameOf.Pool.itsBallPlacement[aBall].itsH
	  + md->itsGameOf.Pool.itsBallPlacement[aBall].itsV * md->itsGameOf.Pool.itsBallPlacement[aBall].itsV
	  < (HOLE_RADIUS + BALL_RADIUS)*(HOLE_RADIUS + BALL_RADIUS))
	{
		return false;
	}

	//	Don't overlap another ball.
	for (i = 0; i < NUM_POOL_BALLS; i++)
	{
		if (i != aBall
		 && md->itsGameOf.Pool.itsBallInPlay[i])
		{
			SetBallPair(md, &p, aBall, i);
			
			dx = p.xb - p.xa;
			dy = p.yb - p.ya;

			if (dx*dx + dy*dy < (2*BALL_RADIUS)*(2*BALL_RADIUS))
				return false;
		}
	}
	
	//	The ball's position is OK.
	return true;
}


static void ComputerChoosesShot(ModelData *md)
{
	//	The computer selects its shot (when playing human vs. computer)
	//	and sets itsCueVectorH and itsCueVectorV accordingly.
	//	It doesn't actually make the shot -- the usual Shoot1() function 
	//	will take care of that.
	
	double			theBestStartSpeed;
	unsigned int	theBall;
	signed int		thePocketH,
					thePocketV,
					dh,
					dv;
	double			theFirstLegEndH,
					theFirstLegEndV,
					theCueBallH,
					theCueBallV,
					theDotProduct,
					theFinalCueBallSpeed;
	BallPath		theFirstLeg,
					theSecondLeg;
	
	const double	theMaxCueVectorLength	= 0.5 - CUE_DEAD_ZONE,
					theMaxStartSpeed		= SHOT_STRENGTH * theMaxCueVectorLength;
	
	//	Evaluate the various possible shots, choosing the one that
	//	minimizes the cue ball's initial speed.  If no shot can be
	//	realized within the allowable MaxStartSpeed, we'll invoke
	//	a fallback plan below.
	theBestStartSpeed = theMaxStartSpeed;

	//	Consider each possible object ball.
	for (theBall = 0; theBall < NUM_POOL_BALLS; theBall++)
	{
		//	Is this object ball a valid target?
		if ( ! ObjectBallIsValid(md, theBall) )
			continue;
		
		//	Use the coordinate system of the tiling view ("universal cover").
		//	The object ball sits in the central copy of the fundamental domain,
		//	but we may want to make use of images of the cue ball and/or
		//	the pocket sitting in nearby copies of the fundamental domain.
		
		//	Consider 9 possible images of the pocket.
		for (thePocketH = -1; thePocketH <= +1; thePocketH++)
			for (thePocketV = -1; thePocketV <= +1; thePocketV++)
			{
				//	Let theSecondLeg be the path from the object ball
				//	to the given image of the pocket.
				SetPath(&theSecondLeg,
						md->itsGameOf.Pool.itsBallPlacement[theBall].itsH, md->itsGameOf.Pool.itsBallPlacement[theBall].itsV,
						thePocketH, thePocketV,
						FINAL_SPEED);

				//	If anything obstructs theSecondLeg,
				//	give up on it and move on to the next candidate.
				if ( ! PathIsClear(md, &theSecondLeg, theBall, false) )
					continue;

				//	The end of the first leg should be colinear with the second leg,
				//	but one ball diameter before the second leg's beginning.
				//	(Draw a picture and all will be clear!)
				
				theFirstLegEndH = theSecondLeg.itsStartH - (2*BALL_RADIUS)*theSecondLeg.itsDirectionH;
				theFirstLegEndV = theSecondLeg.itsStartV - (2*BALL_RADIUS)*theSecondLeg.itsDirectionV;
				
				//	Consider 9 possible images of the cue ball.
				for (dh = -1; dh <= +1; dh++)
					for (dv = -1; dv <= +1; dv++)
					{
						theCueBallH = md->itsGameOf.Pool.itsBallPlacement[0].itsH + dh;
						theCueBallV = md->itsGameOf.Pool.itsBallPlacement[0].itsV + dv;
						if (md->itsTopology == TopologyKlein
						 && dv != 0)
							theCueBallH = - theCueBallH;

						//	Set the cue ball's path,
						//	but with a temporary final speed of zero.
						SetPath(&theFirstLeg,
								theCueBallH, theCueBallV,
								theFirstLegEndH, theFirstLegEndV,
								0.0);

						//	A shot is possible iff the path is clear...
						if ( ! PathIsClear(md, &theFirstLeg, 0, true) )
							continue;
						
						//	..and the dot product of the direction vectors
						//	for the two legs is positive.  (A negative dot product
						//	represents an impossible shot, because the cue ball
						//	would have to pass through the object ball on the first leg.)
						theDotProduct = theFirstLeg.itsDirectionH * theSecondLeg.itsDirectionH
									  + theFirstLeg.itsDirectionV * theSecondLeg.itsDirectionV;
						if (theDotProduct <= 0.0)
							continue;
						
						//	When the (moving) cue ball strikes the (still) object ball,
						//	energy and momentum will both be conserved if the object
						//	ball carries away the momentum parallel to the line
						//	connecting the balls' centers, while the cue ball carries
						//	away the momentum perpendicular to that line.
						//	Similar triangles give the required final speed
						//	for the cue ball to achieve the desired initial speed
						//	for the object ball.  (Make a sketch and all will be clear!)
						theFinalCueBallSpeed = theSecondLeg.itsStartSpeed / theDotProduct;

						//	Recompute the first leg using the correct final speed.
						SetPath(&theFirstLeg,
								theCueBallH, theCueBallV,
								theFirstLegEndH, theFirstLegEndV,
								theFinalCueBallSpeed);

						//	Is this our best shot so far?
						//	If so, set itsGameOf.Pool.itsCueGrip.
						if (theFirstLeg.itsStartSpeed < theBestStartSpeed)
						{
							SetCueGripForVector(
								md,
								(md->itsTopology == TopologyKlein && dv != 0) ?
									- theFirstLeg.itsStartSpeed * theFirstLeg.itsDirectionH / SHOT_STRENGTH :
									+ theFirstLeg.itsStartSpeed * theFirstLeg.itsDirectionH / SHOT_STRENGTH,
								theFirstLeg.itsStartSpeed * theFirstLeg.itsDirectionV / SHOT_STRENGTH);

							theBestStartSpeed = theFirstLeg.itsStartSpeed;
						}
					}
			}
	}
	
	//	Most likely we've found a good shot by now,
	//	but if we haven't, go for Plan B.
	if (theBestStartSpeed == theMaxStartSpeed)
	{
		//	First set up a plausible Plan C,
		//	in case Plan B fails.
		SetCueGripForVector(md, 0.0, theMaxCueVectorLength);
		
		//	Plan B is to find some legal ball and blast it.
		//	Don't worry about colliding with other balls,
		//	but try not to accidently sink the cue ball.
		for (theBall = 0; theBall < NUM_POOL_BALLS; theBall++)
		{
			if (ObjectBallIsValid(md, theBall))
			{
				//	Set the cue ball's path,
				//	but with a temporary final speed of zero.
				SetPath(&theFirstLeg,
						md->itsGameOf.Pool.itsBallPlacement[   0   ].itsH,
						md->itsGameOf.Pool.itsBallPlacement[   0   ].itsV,
						md->itsGameOf.Pool.itsBallPlacement[theBall].itsH,
						md->itsGameOf.Pool.itsBallPlacement[theBall].itsV,
						0.0);
				
				if (PathAvoidsObstacle(&theFirstLeg, 0, 0, HOLE_RADIUS))
				{
					//	Set itsGameOf.Pool.itsCueGrip in the desired direction,
					//	with maximum speed.
					SetCueGripForVector(
						md,
						theMaxCueVectorLength * theFirstLeg.itsDirectionH,
						theMaxCueVectorLength * theFirstLeg.itsDirectionV);

					//	No need to look at other balls.
					break;
				}
			}
		}
	}

}


static bool ObjectBallIsValid(
	ModelData		*md,
	unsigned int	aBall)
{
	PoolGoal	theGoal;

	//	If aBall isn't in play, for sure it's not valid.
	if ( ! md->itsGameOf.Pool.itsBallInPlay[aBall] )
		return false;

	//	Which balls are we shooting for?
	theGoal = (md->itsGameOf.Pool.itsWhoseTurn == PoolPlayerA ?
				md->itsGameOf.Pool.itsPlayerAGoal : md->itsGameOf.Pool.itsPlayerBGoal);

	//	Is the given ball a valid object ball?
	switch (aBall)
	{
		case 0:			//	cue ball
			return false;
		
		case 4:			//	black ball
			if (theGoal == PoolBlack)
				return true;
			else
				return false;
		
		case 1:			//	solid ball
		case 2:
		case 3:
			if (theGoal == PoolUndecided || theGoal == PoolSolids)
				return true;
			else
				return false;
		
		case 5:			//	striped ball
		case 6:
		case 7:
			if (theGoal == PoolUndecided || theGoal == PoolStripes)
				return true;
			else
				return false;
		
		default:	//	should never occur
			return false;
	}
}


static void SetPath(
	BallPath	*aPath,	//	output
	double		h0,		//	initial ball position, in tiling view coordinates,
	double		v0,		//		assumed to lie in 9 central copies of fundamental domain
	double		h1,		//	final ball position, in tiling view coordinates
	double		v1,		//		assumed to lie in 9 central copies of fundamental domain
	double		s1)		//	final ball speed
{
	double	t0;

	aPath->itsStartH		= h0;
	aPath->itsStartV		= v0;

	aPath->itsDirectionH	= h1 - h0;
	aPath->itsDirectionV	= v1 - v0;

	aPath->itsLength		= sqrt(	aPath->itsDirectionH * aPath->itsDirectionH
								  + aPath->itsDirectionV * aPath->itsDirectionV);
	if (aPath->itsLength == 0.0)	//	should never occur
	{
		aPath->itsNormalH	= 0.0;
		aPath->itsNormalV	= 0.0;
		return;
	}

	aPath->itsDirectionH	/= aPath->itsLength;
	aPath->itsDirectionV	/= aPath->itsLength;
	
	aPath->itsNormalH		= - aPath->itsDirectionV;
	aPath->itsNormalV		=   aPath->itsDirectionH;
	
	//	By construction the ball decelerates with a constant BALL_DECELERATION,
	//	so its velocity function is
	//
	//		v(t) = v(0) - BALL_DECELERATION*t
	//
	//	If we take t = 0 to be the time the ball reaches the end
	//	of the path (which admittedly is an odd definition of t = 0,
	//	but makes sense in the present case because that's the time
	//	at which the speed s1 is known), then the velocity function becomes
	//
	//		v(t) = s1 - BALL_DECELERATION*t
	//
	//	Integrating gives the position function
	//
	//		x(t) = x(0) + s1*t - (BALL_DECELERATION/2)*t^2
	//
	//	If we measure distances relative to the end of the ball's path,
	//	the position function becomes
	//
	//		x(t) = s1*t - (BALL_DECELERATION/2)*t^2
	//
	//	Our plan is to figure out at what (negative) time t the ball
	//	began its trip, and to compute its velocity at that point.
	//
	//		-itsLength = s1*t0 - (BALL_DECELERATION/2)*t0^2
	//
	//		(BALL_DECELERATION/2)*t0^2 - s1*t0 - itsLength = 0
	//
	//		t0 = ( s1 +/- sqrt( s1^2 + 2*BALL_DECELERATION*itsLength ) )
	//			/ BALL_DECELERATION
	t0 = (s1 - sqrt(s1*s1 + 2*BALL_DECELERATION*aPath->itsLength)) / BALL_DECELERATION;
	aPath->itsStartSpeed = s1 - BALL_DECELERATION*t0;
}


static bool PathIsClear(
	ModelData		*md,
	BallPath		*aPath,			//	assumed to lie within 9 central copies of fundamental domain
	unsigned int	aBall,			//	a ball to ignore when testing for obstructions
	bool			aPocketTest)	//	test the pocket as a potential obstacle? (true on first leg, false on second leg)
{
	//	Would a ball travelling along aPath
	//	avoid other balls and also avoid the pocket?
	
	unsigned int	theBall;
	signed int		dh,
					dv;
	double			theObstacleBallH,
					theObstacleBallV;

	//	Does the path miss all other balls?
	for (theBall = 0; theBall < NUM_POOL_BALLS; theBall++)
	{
		//	Exclude the active ball itself.
		if (theBall == aBall)
			continue;
		
		//	Exclude balls no longer in play.
		if ( ! md->itsGameOf.Pool.itsBallInPlay[theBall] )
			continue;
			
		//	Consider the images of theBall within the 9 central copies
		//	of the fundamental domain.
		for (dh = -1; dh <= +1; dh++)
			for (dv = -1; dv <= +1; dv++)
			{
				theObstacleBallH = md->itsGameOf.Pool.itsBallPlacement[theBall].itsH + dh;
				theObstacleBallV = md->itsGameOf.Pool.itsBallPlacement[theBall].itsV + dv;
				if (md->itsTopology == TopologyKlein
				 && dv != 0)
				{
					theObstacleBallH = - theObstacleBallH;
				}
				
				if ( ! PathAvoidsObstacle(aPath, theObstacleBallH, theObstacleBallV, 2*BALL_RADIUS) )
					return false;
			}
	}

	//	Does the path miss the pocket?
	if (aPocketTest)
	{
		//	Consider the images of the pocket within the 9 central copies
		//	of the fundamental domain.
		for (dh = -1; dh <= +1; dh++)
			for (dv = -1; dv <= +1; dv++)
				if ( ! PathAvoidsObstacle(aPath, dh, dv, HOLE_RADIUS) )
					return false;
	}

	//	The path is clear.
	return true;
}


static bool PathAvoidsObstacle(
	BallPath	*aPath,
	double		anObstacleH,
	double		anObstacleV,
	double		aRadius)
{
	double	theSidewaysDisplacement,
			theLongwaysDisplacement,
			theFinalH,
			theFinalV,
			dh,
			dv,
			theFinalDistance;

	//	If anObstacle lies too far to either side,
	//	then aPath avoids it.  Compute a dot product to check.
	theSidewaysDisplacement = aPath->itsNormalH * (anObstacleH - aPath->itsStartH)
							+ aPath->itsNormalV * (anObstacleV - aPath->itsStartV);
	if (fabs(theSidewaysDisplacement) >= aRadius)
		return true;

	//	If anObstacle lies behind aPath's start or beyond aPath's end,
	//	then aPath avoids it.  Compute a dot product to check.
	//	Be generous on the far end, keeping in mind that the whole purpose
	//	of this path might be to hit another ball.
	theLongwaysDisplacement = aPath->itsDirectionH * (anObstacleH - aPath->itsStartH)
							+ aPath->itsDirectionV * (anObstacleV - aPath->itsStartV);
	if (theLongwaysDisplacement < 0.0)
	{
		//	On the near end we can be pretty loose, because the ball's initial position
		//	doesn't conflict with any obstacles.
		return true;
	}
	if (theLongwaysDisplacement > aPath->itsLength)
	{
		//	On the far end we need to be a little more careful,
		//	because conflicts at the final position are possible.
		//	The tricky part is that for the "first leg" of a shot,
		//	the final position will, by design, be tangent to another ball.
		theFinalH = aPath->itsStartH  +  aPath->itsLength * aPath->itsDirectionH;
		theFinalV = aPath->itsStartV  +  aPath->itsLength * aPath->itsDirectionV;
		dh = theFinalH - anObstacleH;
		dv = theFinalV - anObstacleV;
		theFinalDistance = sqrt(dh*dh + dv*dv);
		if (theFinalDistance > aRadius - ENDPOINT_EPSILON)
			return true;
	}
	
	//	Report a potential collision.
	return false;
}


static void AddGaussianError(ModelData *md)
{
	//	Add a Gaussian error to itsGameOf.Pool.itsCueGrip,
	//	which we assume has already been set in preparation
	//	for the computer's shot in human vs. computer mode.
	
	Complex	theExactVector,
			theErrorVector,
			theFinalVector;
	double	theLength;

	static double	theStandardDeviation[5] = {0.08, 0.04, 0.03, 0.02, 0.00};
	
	//	Think of itsGameOf.Pool.itsCueGrip.itsHV as a complex number
	//	and multiply by a Gaussian random complex number
	//	with mean 1 and standard deviation depending
	//	on the difficulty level.

	CueVector(md, &theExactVector.re, &theExactVector.im, &theLength);

	theErrorVector.re = 1.0 + theStandardDeviation[md->itsDifficultyLevel] * GetRandomGaussian();
	theErrorVector.im =       theStandardDeviation[md->itsDifficultyLevel] * GetRandomGaussian();

	theFinalVector = ComplexMultiply(theExactVector, theErrorVector);
	
	//	Don't exceed length 0.5 - CUE_DEAD_ZONE.
	theLength = sqrt(theFinalVector.re * theFinalVector.re
				   + theFinalVector.im * theFinalVector.im);
	if (theLength > 0.5 - CUE_DEAD_ZONE)
	{
		theFinalVector.re *= (0.5 - CUE_DEAD_ZONE) / theLength;
		theFinalVector.im *= (0.5 - CUE_DEAD_ZONE) / theLength;
	}

	//	Set itsGameOf.Pool.itsCueGrip to the final value.
	SetCueGripForVector(md, theFinalVector.re, theFinalVector.im);
}


static double GetRandomGaussian(void)
{
	//	Return a Gaussian random number with mean 0 and standard deviation 1.

	//	Thanks to Robin Lock for telling me how to do this.

	double	u1,	//	uniform on (0,1]
			u2,	//	uniform on [0,1]
			x;	//	Gaussian with mean 0 and standard deviation 1

	do
	{
		u1 = (double) rand() / (double) RAND_MAX;
	} while (u1 == 0.0);

	u2 = (double) rand() / (double) RAND_MAX;

	x = sqrt(-2*log(u1)) * cos(2*PI*u2);

	return x;
}


static Complex ComplexMultiply(
	Complex	z,
	Complex	w)
{
	Complex zw;
	
	zw.re = z.re * w.re
		  - z.im * w.im;

	zw.im = z.re * w.im
		  + z.im * w.re;
	
	return zw;
}


static void PoolSimulationUpdate(ModelData *md)
{
	switch (md->itsSimulationStatus)
	{
		case SimulationPoolRollBalls:
		
			//	Advance the simulation by one frame.
			Shoot2(md, md->itsSimulationDeltaTime);

			//	Have the balls come to rest?
			if ( ! BallsAreRolling(md) )
			{
				SimulationEnd(md);
				Shoot3(md);
			}
			break;
		
		case SimulationPoolPlaceCueStick:

			//	PoolDraw() will check md->itsSimulationStatus
			//	and md->itsSimulationElapsedTime and scale
			//	the cue stick length accordingly.

			//	When the cue stick animation is finished,
			//	pause so that the user may appreciate it.
			if (md->itsSimulationElapsedTime >= STICK_PLACE_TIME)
			{
				SimulationEnd(md);
				SimulationBegin(md, SimulationPoolPauseBeforeShooting);
			}

			break;
		
		case SimulationPoolPauseBeforeShooting:

			//	After displaying the ready-to-go pool cue for one second...
			if (md->itsSimulationElapsedTime >= STICK_PAUSE_TIME)
			{
				SimulationEnd(md);	//	...terminate the waiting period and
				Shoot1(md);			//	...let the balls roll!
			}

			break;

		default:
			break;
	}
}


static void PoolDraw(ModelData *md)
{
	Placement		thePlacement;
	unsigned int	i;
	double			theCueVectorH,
					theCueVectorV,
					theCueVectorLength,
					theDirectionH,
					theDirectionV,
					theNearEnd,
					theFarEnd;

	//	Draw the background.
	DrawFundamentalDomain(md, md->itsGameOf.Pool.itsTextureNameTable, 8);

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(1.0, 1.0, 0.0, 1.0),
					PREMULTIPLY_RGBA(0.5, 1.0, 1.0, 1.0) );
	
	//	Draw the pocket.
	thePlacement.itsH		= 0.0;
	thePlacement.itsV		= 0.0;
	thePlacement.itsFlip	= false;
	thePlacement.itsAngle	= 0.0;
	thePlacement.itsSizeH	= 2 * HOLE_RADIUS;
	thePlacement.itsSizeV	= 2 * HOLE_RADIUS;
	DrawSprite(md, md->itsGameOf.Pool.itsTextureNamePocket, &thePlacement);
	
	//	Draw the balls.
	for (i = 0; i < NUM_POOL_BALLS; i++)
		if (md->itsGameOf.Pool.itsBallInPlay[i])
			DrawSprite(	md,
						md->itsGameOf.Pool.itsTextureNameBall[i],
						&md->itsGameOf.Pool.itsBallPlacement[i]);

	//	Draw the cue stick if necessary.
	if ( ! md->itsGameIsOver
	 && md->itsGameOf.Pool.itsCueStatus != PoolCueNone)
	{
		CueVector(md, &theCueVectorH, &theCueVectorV, &theCueVectorLength);

		if (theCueVectorLength > 0.0)
		{
			theDirectionH = theCueVectorH / theCueVectorLength;
			theDirectionV = theCueVectorV / theCueVectorLength;
			
			if (md->itsSimulationStatus == SimulationPoolPlaceCueStick
			 && md->itsSimulationElapsedTime <= STICK_PLACE_TIME)
			{
				theCueVectorLength *= md->itsSimulationElapsedTime / STICK_PLACE_TIME;
			}

			theNearEnd	= CUE_DEAD_ZONE;
			theFarEnd	= CUE_DEAD_ZONE  +  STICK_FACTOR * theCueVectorLength;

			DrawSegment(md,
						NULL,
						md->itsGameOf.Pool.itsTextureNameStick,
						md->itsGameOf.Pool.itsBallPlacement[0].itsH  -  ((theNearEnd + theFarEnd)/2) * theDirectionH,
						md->itsGameOf.Pool.itsBallPlacement[0].itsV  -  ((theNearEnd + theFarEnd)/2) * theDirectionV,
						theFarEnd - theNearEnd,
						0.02,
						atan2(theDirectionV, theDirectionH),
						1);
			DrawSegment(md,
						NULL,
						md->itsGameOf.Pool.itsTextureNameSight,
						md->itsGameOf.Pool.itsBallPlacement[0].itsH  +  ((theNearEnd + theFarEnd)/2) * theDirectionH,
						md->itsGameOf.Pool.itsBallPlacement[0].itsV  +  ((theNearEnd + theFarEnd)/2) * theDirectionV,
						theFarEnd - theNearEnd,
						0.01,
						atan2(theDirectionV, theDirectionH),
						1);
		}

#ifdef TORUS_GAMES_FOR_OPENGL_ES
		//	Draw the cue stick's grip point, if appropriate.
		if (md->itsGameOf.Pool.itsCueStatus == PoolCueActive)
		{
			md->itsGameOf.Pool.itsCueGrip.itsFlip	= false;
			md->itsGameOf.Pool.itsCueGrip.itsAngle	= 0.0;
			md->itsGameOf.Pool.itsCueGrip.itsSizeH	= GRIP_SIZE;
			md->itsGameOf.Pool.itsCueGrip.itsSizeV	= GRIP_SIZE;

			DrawColorWithAlphaMask(	md,
									md->itsTextureNamesCommon[TextureCommonHalo],
									&md->itsGameOf.Pool.itsCueGrip,
									PREMULTIPLY_RGBA(0.0, 0.5, 1.0, 1.0));
		}
#endif
	}
}
