//	TorusGamesRectangleWithEndcaps.c
//
//	Makes, binds and draws a Vertex Array Object for a rectangle with endcaps.
//
//	Employ the otherwise unused third coordinate to encode the endcap positions:
//
//		A non-endcap vertex will have coordinates (x, y, 0),
//		and the matrix multiplication will look something like
//
//			         (h 0)
//			(x y 0 1)(0 v)
//			         (v 0)	//	third row gets ignored
//			         (a b)
//
//		where h is the rectangle's width (excluding endcaps),
//		v is its height, and (a,b) is the location of its center.
//		The simple matrix shown may be postmultiplied by another matrix
//		to provide rotation and reflection.
//
//		An endcap vertex will have coordinates (x, y, 0.5),
//		and the matrix multiplication will look something like
//
//			           (h 0)
//			(x y 0.5 1)(0 v)
//			           (v 0)
//			           (a b)
//
//		Here the third row extends the width by exactly half the height.
//
//	The fourth coordinate is always 1 and is implicit 
//	(the vertex shader will know to append it).
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <stddef.h>	//	for offsetof()


//	The rectangle with endcaps Vertex Buffer Object (VBO) will contain
//	the following data for each of its vertices.
typedef struct
{
	float	pos[3],	//	position (x,y,c) where c is the endcap coordinate described above
			tex[2];	//	texture coordinates (u,v)
} RectangleWithEndcapsVBOData;


static void	BindRectangleWithEndcapsVBO(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag);


void MakeRectangleWithEndcapsVAO(
	unsigned int	aVertexArrayName,
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName,	//	ignored
	unsigned int	anOpenGLVersion)
{
	static const RectangleWithEndcapsVBOData	theVertices[8] =
	{
		{{-0.5, +0.5, -0.5}, {0.0, 1.0}},
		{{-0.5, -0.5, -0.5}, {0.0, 0.0}},
		{{-0.5, +0.5,  0.0}, {0.5, 1.0}},
		{{-0.5, -0.5,  0.0}, {0.5, 0.0}},
		{{+0.5, +0.5,  0.0}, {0.5, 1.0}},
		{{+0.5, -0.5,  0.0}, {0.5, 0.0}},
		{{+0.5, +0.5, +0.5}, {1.0, 1.0}},
		{{+0.5, -0.5, +0.5}, {1.0, 0.0}},
	};

	if (RUNNING_OPENGL_3)
		;	//	uncomment the following two lines:
//	if (anOpenGLVersion >= VERSION_NUMBER(3, 0))
//		glBindVertexArray(0);

	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);
	glBufferData(GL_ARRAY_BUFFER,
					8 * sizeof(RectangleWithEndcapsVBOData),
					theVertices,
					GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	if (RUNNING_OPENGL_3)
		;	//	uncomment the following several lines:
//	if (anOpenGLVersion >= VERSION_NUMBER(3, 0))
//	{
//		glBindVertexArray(aVertexArrayName);
//		BindRectangleWithEndcapsVBO(aVertexBufferName, anIndexBufferName);
//		glBindVertexArray(0);
//	}
UNUSED_PARAMETER(anIndexBufferName);	//	REMOVE THIS AFTER UNCOMMENTING ABOVE LINES.
}


void BindRectangleWithEndcapsVAO(
	unsigned int	aVertexArrayName,
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName,	//	ignored
	unsigned int	anOpenGLVersion,
	bool			aShadersFlag)	//	REMOVE THIS PARAMETER WHEN REMOVING OPENGL 1.X SUPPORT FROM DESKTOP VERSION.
									//	HMMM... OR MAYBE NOT -- IT SEEMS TO BE NEEDED FOR GLES 2 SUPPORT.
{
	//	When OpenGL 2.1 support eventually gets removed,
	//	this function may be eliminated and the caller may instead
	//	call glBindVertexArray(aVertexArrayName) directly.

	if (RUNNING_OPENGL_3)
		;	//	uncomment the following several lines:
//	if (anOpenGLVersion >= VERSION_NUMBER(3, 0))
//		glBindVertexArray(aVertexArrayName);
//	else
		BindRectangleWithEndcapsVBO(aVertexBufferName, anIndexBufferName, aShadersFlag);
}

static void BindRectangleWithEndcapsVBO(
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName,	//	ignored
	bool			aShadersFlag)	//	REMOVE THIS PARAMETER WHEN REMOVING OPENGL 1.X SUPPORT FROM DESKTOP VERSION.
									//	HMMM... OR MAYBE NOT -- IT SEEMS TO BE NEEDED FOR GLES 2 SUPPORT.
{
	//	Under OpenGL 2.1 or earlier, this function 
	//		gets called once per frame, from BindRectangleWithEndcapsVAO().
	//	Under OpenGL 3.0 or later, this function
	//		gets called only once, from MakeRectangleWithEndcapsVAO().
	//	When OpenGL 2.1 support eventually gets removed,
	//		this function may be eliminated and its contents
	//		may be moved directly into MakeRectangleWithEndcapsVAO().

	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);

	if (aShadersFlag)
	{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		glEnableVertexAttribArray(ATTRIBUTE_POSITION);
		glVertexAttribPointer(ATTRIBUTE_POSITION,  3, GL_FLOAT, GL_FALSE, sizeof(RectangleWithEndcapsVBOData), (void *)offsetof(RectangleWithEndcapsVBOData, pos));

		glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
		glVertexAttribPointer(ATTRIBUTE_TEX_COORD, 2, GL_FLOAT, GL_FALSE, sizeof(RectangleWithEndcapsVBOData), (void *)offsetof(RectangleWithEndcapsVBOData, tex));
#endif
	}
	else
	{
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		glVertexPointer  (3, GL_FLOAT, sizeof(RectangleWithEndcapsVBOData), (void *)offsetof(RectangleWithEndcapsVBOData, pos));
		glTexCoordPointer(2, GL_FLOAT, sizeof(RectangleWithEndcapsVBOData), (void *)offsetof(RectangleWithEndcapsVBOData, tex));
#endif
	}

	UNUSED_PARAMETER(anIndexBufferName);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

void DrawRectangleWithEndcapsVAO(void)
{
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 8);
}
