//	TorusGamesWordSearch.c
//
//	Code specific to the WordSearch game.
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


#ifdef TORUS_GAMES_FOR_OPENGL_ES
#define WORDSEARCH_CHARACTER_TEXTURE_SIZE	WORDSEARCH_CHARACTER_TEXTURE_SIZE_MOBILE
#else
#define WORDSEARCH_CHARACTER_TEXTURE_SIZE	WORDSEARCH_CHARACTER_TEXTURE_SIZE_DESKTOP
#endif


//	When testing various word search puzzle data sets,
//	let the computer solve the puzzles to save time.
//#define WORD_SEARCH_AUTO_SOLVE

//	When producing screenshots, use pre-determined colors.
//#define WORD_SEARCH_SCREENSHOT_COLORS

//	How close to the center of a cell (in cell units) must a click or tap be
//	in order to start a selection rather than a scroll.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
#define WORD_SEARCH_SELECTION_HIT_RADIUS				0.4
#else
#define WORD_SEARCH_SELECTION_HIT_RADIUS				0.3
#endif

//	In the touch-based interface, when continuing a previously started selection,
//	how close (in board units) must the new touch be to the previous selection's 
//	hot spot to count as a hit?
#define WORD_SEARCH_SELECTION_CONTINUATION_HIT_RADIUS	0.1

//	In the touch-based interface, the user taps to end a selection.
//	How close must the touch point stay to the hot point in order to count as a tap?
//	[OLD VERSION:  How much accidental movement should we allow in the definition of a tap?]
#define WORD_SEARCH_TAP_TOLERANCE						0.05


//	Public functions
static void	WordSearchShutDown(ModelData *md);
static void	WordSearchSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void	WordSearchShutDownTextures(ModelData *md);
static void	WordSearchReset(ModelData *md);
static void	WordSearchHandMoved(ModelData *md);
static bool	WordSearchDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void	WordSearchDragObject(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
static void	WordSearchDragEnd(ModelData *md);
static void	WordSearchSimulationUpdate(ModelData *md);
static void	WordSearchDraw(ModelData *md);

//	Private functions
static bool	ScratchWordFromList(ModelData *md, unsigned int aStartH, unsigned int aStartV, signed int aDeltaH, signed int aDeltaV);
static void	NormalizeCellCoordinates(ModelData *md, signed int *aCellH, signed int *aCellV);
static bool	SameWord(Char16 aWordA[MAX_WORDSEARCH_WORD_LENGTH + 1], Char16 aWordB[MAX_WORDSEARCH_WORD_LENGTH + 1]);
static void	RemoveSpacesAndHyphens(Char16 aWord[MAX_WORDSEARCH_WORD_LENGTH + 1]);
#ifdef TORUS_GAMES_FOR_OPENGL_ES
static void	GetHotPoint(ModelData *md, Placement *aHotPoint);
#endif


void WordSearchSetUp(ModelData *md)
{
	unsigned int	h,
					v;

	//	Initialize function pointers.
	md->itsGameShutDown			= &WordSearchShutDown;
	md->itsGameSetUpTextures	= &WordSearchSetUpTextures;
	md->itsGameShutDownTextures	= &WordSearchShutDownTextures;
	md->itsGameReset			= &WordSearchReset;
	md->itsGameHandMoved		= &WordSearchHandMoved;
	md->itsGameDragBegin		= &WordSearchDragBegin;
	md->itsGameDragObject		= &WordSearchDragObject;
	md->itsGameDragEnd			= &WordSearchDragEnd;
	md->itsGameCharacterInput	= NULL;
	md->itsGameSpecialCommand	= NULL;
	md->itsGameSimulationUpdate	= &WordSearchSimulationUpdate;
	md->itsGameDraw				= &WordSearchDraw;

	//	Initialize variables for robust error handling.
	md->itsGameOf.WordSearch.itsPuzzleIsValid	= false;
	md->itsGameOf.WordSearch.itsPuzzleSize		= 0;
	
	//	Note the initial language.
	md->itsGameOf.WordSearch.itsPreviousLanguageCode[0] = gLanguageCode[0];
	md->itsGameOf.WordSearch.itsPreviousLanguageCode[1] = gLanguageCode[1];
	md->itsGameOf.WordSearch.itsPreviousLanguageCode[2] = 0;

	//	Start at the beginning of each puzzle list.
	md->itsGameOf.WordSearch.itsIndexTorus		= 0;
	md->itsGameOf.WordSearch.itsIndexKlein		= 0;

	//	Set up an empty board to ensure safe texture initialization.
	for (h = 0; h < MAX_WORDSEARCH_SIZE; h++)
		for (v = 0; v < MAX_WORDSEARCH_SIZE; v++)
			md->itsGameOf.WordSearch.itsBoard[h][v] = ' ';

	//	No selection is pending.
	md->itsGameOf.WordSearch.itsWordSelectionIsPending = false;
	
	//	Initialize game-specific texture names.
	md->itsGameOf.WordSearch.itsTextureNameBackground	= 0;
	md->itsGameOf.WordSearch.itsTextureNameMark			= 0;
	for (h = 0; h < MAX_WORDSEARCH_SIZE; h++)
		for (v = 0; v < MAX_WORDSEARCH_SIZE; v++)
			md->itsGameOf.WordSearch.itsTextureNameCell[h][v] = 0;

	//	Reset the word search puzzle.
	WordSearchReset(md);
}


static void WordSearchShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	WordSearchShutDownTextures(md);
}


static void WordSearchSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	unsigned int	h,
					v;

	SetUpOneTexture(md, &md->itsGameOf.WordSearch.itsTextureNameBackground,	"WordSearch/Paper.rgba",	TextureWrap,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.WordSearch.itsTextureNameMark,		"WordSearch/Mark.rgba",		TextureClamp,	TextureRGBA,	aFirstError);

	for (h = 0; h < MAX_WORDSEARCH_SIZE; h++)
		for (v = 0; v < MAX_WORDSEARCH_SIZE; v++)
		{
			SetUpOneTexture(md, &md->itsGameOf.WordSearch.itsTextureNameCell[h][v],	NULL,				TextureClamp,	TextureAlpha,	aFirstError);
			SetAlphaTextureFromCharacter(
				md->itsGameOf.WordSearch.itsTextureNameCell[h][v],
				md->itsGameOf.WordSearch.itsBoard[h][v],
				md->itsFonts[FontWordSearchPuzzle],
				WORDSEARCH_CHARACTER_TEXTURE_SIZE,
				md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
				aFirstError);
		}
}


static void WordSearchShutDownTextures(ModelData *md)
{
	unsigned int	h,
					v;

	ShutDownOneTexture(&md->itsGameOf.WordSearch.itsTextureNameBackground	);
	ShutDownOneTexture(&md->itsGameOf.WordSearch.itsTextureNameMark			);

	for (h = 0; h < MAX_WORDSEARCH_SIZE; h++)
		for (v = 0; v < MAX_WORDSEARCH_SIZE; v++)
			ShutDownOneTexture(&md->itsGameOf.WordSearch.itsTextureNameCell[h][v]);
}


static void WordSearchReset(ModelData *md)
{
	Char16			theRawPuzzleData[1024],
					*r;	//	the read position
	unsigned int	theNumPuzzles,
					*theIndex,
					h,
					v,
					i,
					j;
#ifdef WORD_SEARCH_AUTO_SOLVE
	signed int		dh,
					dv;
	unsigned int	theLength;
#endif

	//	For safety, set the validity flag before doing anything else.
	//	On Win32, ErrorMessage() calls MessageBox() which apparently can send
	//	a WM_PAINT message to the main window *before* this call
	//	to WordSearchReset() even returns.  That is, we risk painting
	//	the window using invalid data.
	md->itsGameOf.WordSearch.itsPuzzleIsValid = false;
	
	//	Cancel any pending selection.
	md->itsGameOf.WordSearch.itsWordSelectionIsPending = false;
	
	//	If the user changed languages, reset the puzzle indices to 0 
	//	so the user will get the easiest puzzle in the new language.
	if ( ! SAME_LANGUAGE_CODE(gLanguageCode, md->itsGameOf.WordSearch.itsPreviousLanguageCode) )
	{
		md->itsGameOf.WordSearch.itsIndexTorus = 0;
		md->itsGameOf.WordSearch.itsIndexKlein = 0;

		md->itsGameOf.WordSearch.itsPreviousLanguageCode[0] = gLanguageCode[0];
		md->itsGameOf.WordSearch.itsPreviousLanguageCode[1] = gLanguageCode[1];
		md->itsGameOf.WordSearch.itsPreviousLanguageCode[2] = 0;
	}

	//	Read the raw puzzle data as a string resource.
	theNumPuzzles	= GetNumPuzzles(GameWordSearch, md->itsTopology);
	theIndex		= md->itsTopology == TopologyTorus ?
						&md->itsGameOf.WordSearch.itsIndexTorus :
						&md->itsGameOf.WordSearch.itsIndexKlein;
	*theIndex		%= theNumPuzzles;	//	unnecessary but safe
	GetRawPuzzleData(	GameWordSearch,
						md->itsTopology,
						*theIndex,
						theRawPuzzleData,
						BUFFER_LENGTH(theRawPuzzleData));

	//	Increment the index for next time.
	(*theIndex)++;
	*theIndex %= theNumPuzzles;

	//	Initialize a "read pointer".
	r = theRawPuzzleData;

	//	Confirm the leading 'T' or 'K'.
	if (*r++ != (md->itsTopology == TopologyTorus ? 'T' : 'K'))
		goto WordSearchResetError;

	//	Parse the puzzle size.
	if (*r >= '1' && *r <= '9')
	{
		md->itsGameOf.WordSearch.itsPuzzleSize = *r++ - '0';
		if (*r >= '1' && *r <= '9')
		{
			md->itsGameOf.WordSearch.itsPuzzleSize *= 10;
			md->itsGameOf.WordSearch.itsPuzzleSize += *r++ - '0';
		}

		if (*r++ != '/')
			goto WordSearchResetError;
	}
	else
		goto WordSearchResetError;
	if (md->itsGameOf.WordSearch.itsPuzzleSize <= 0
	 || md->itsGameOf.WordSearch.itsPuzzleSize > MAX_WORDSEARCH_SIZE)
		goto WordSearchResetError;

	//	We've finished reading the header information.
	if (*r++ != '/')
		goto WordSearchResetError;

	//	Read the puzzle board and set all flips to false.
	//
	//	Note #1:  The internal v-coordinate runs bottom-to-top,
	//	while raw puzzle data runs top-to-bottom.
	//
	//	Note #2:  The internal coordinates are ordered (h,v),
	//	so the outer loop iterates over v, even though v is
	//	the second coordinate.
	for (v = md->itsGameOf.WordSearch.itsPuzzleSize; v-- > 0; )
	{
		for (h = 0; h < md->itsGameOf.WordSearch.itsPuzzleSize; h++)
		{
			md->itsGameOf.WordSearch.itsBoard[h][v] = *r++;

			if (md->itsTopology == TopologyTorus)
				md->itsGameOf.WordSearch.itsBoardFlips[h][v] = false;
			else
				md->itsGameOf.WordSearch.itsBoardFlips[h][v] = rand() % 2;
		}

		if (*r++ != '/')
			goto WordSearchResetError;
	}
	if (*r++ != '/')
		goto WordSearchResetError;

	//	Read the words to be found.
	for (i = 0; i < MAX_WORDSEARCH_NUM_WORDS; i++)
	{
		if (*r != '/')
		{
			for (j = 0; j < MAX_WORDSEARCH_WORD_LENGTH + 1; j++)
			{
				if (*r == 0)
					goto WordSearchResetError;

				if (*r != '/')
					md->itsGameOf.WordSearch.itsWords[i][j] = *r++;
				else
					md->itsGameOf.WordSearch.itsWords[i][j] = 0;
			}
			if (md->itsGameOf.WordSearch.itsWords[i][MAX_WORDSEARCH_WORD_LENGTH] != 0)
				goto WordSearchResetError;
			r++;
		}
		else
		{
			for (j = 0; j < MAX_WORDSEARCH_WORD_LENGTH + 1; j++)
				md->itsGameOf.WordSearch.itsWords[i][j] = 0;
		}
	}
	if (*r != '/')
		goto WordSearchResetError;

	//	Display the current word List.
	WordSearchSetList(md);

	//	No lines have been found yet.
	md->itsGameOf.WordSearch.itsNumLinesFound = 0;

	//	The game is not over.
	md->itsGameIsOver = false;
	
	//	We'll need new textures.
	md->itsPreparedTextures = false;

	//	Abort any pending simulation.
	SimulationEnd(md);

	//	We're ready to go.
	md->itsGameOf.WordSearch.itsPuzzleIsValid = true;

#ifdef WORD_SEARCH_AUTO_SOLVE

	ErrorMessage(	"Warning:  Auto-solve is enabled.",
					"For testing purposes only",
					false);

	//	Note:  Use auto-solve to find only words of length at most itsPuzzleSize.
	//	Find longer words by hand to make sure they respect conditions
	//	about not extending more than half a board width beyond
	//	the fundamental square.

	for (h = 0; h < md->itsGameOf.WordSearch.itsPuzzleSize; h++)
		for (v = 0; v < md->itsGameOf.WordSearch.itsPuzzleSize; v++)
			for (dh = -1; dh <= +1; dh++)
				for (dv = -1; dv <= +1; dv++)
					if (dh != 0 || dv != 0)
						for (theLength = 1; theLength <= md->itsGameOf.WordSearch.itsPuzzleSize - 1; theLength++)	//	see note above
						{
							md->itsGameOf.WordSearch.itsStartH		= h;
							md->itsGameOf.WordSearch.itsStartV		= v;
							md->itsGameOf.WordSearch.itsDeltaH		= dh * (signed int) theLength;
							md->itsGameOf.WordSearch.itsDeltaV		= dv * (signed int) theLength;
							
							//	The R, G and B components should be premultiplied by the alpha component,
							//	but as long as the alpha component is 1.0 there's no need.
							md->itsGameOf.WordSearch.itsColor[0]	= (float)rand() / (float)RAND_MAX;
							md->itsGameOf.WordSearch.itsColor[1]	= (float)rand() / (float)RAND_MAX;
							md->itsGameOf.WordSearch.itsColor[2]	= (float)rand() / (float)RAND_MAX;
							md->itsGameOf.WordSearch.itsColor[3]	= 1.0;
							
							WordSearchDragEnd(md);
						}
#endif

	return;

WordSearchResetError:
	ErrorMessage(	"Invalid Word Search Data",
					"Internal Error",
					false);
	return;
}


static void WordSearchHandMoved(ModelData *md)
{
	double			theHandH,
					theHandV;
	unsigned int	h,
					v;

	//	Make sure our data is valid.
	if ( ! md->itsGameOf.WordSearch.itsPuzzleIsValid )
		return;

	//	Convert hand coordinates from [-0.5, +0.5] to [0.0, PuzzleSize].
	theHandH = md->itsGameOf.WordSearch.itsPuzzleSize*(md->itsHandPlacement.itsH + 0.5);
	theHandV = md->itsGameOf.WordSearch.itsPuzzleSize*(md->itsHandPlacement.itsV + 0.5);

	//	Find the integer coordinates (h,v) of the hit cell.
	h = (unsigned int)(signed int) floor(theHandH);	//	-1 would map to 0xFFFFFFFF and get caught below.
	v = (unsigned int)(signed int) floor(theHandV);

	//	Exclude values on the boundary and
	//	protect against unexpected errors.
	if (h >= md->itsGameOf.WordSearch.itsPuzzleSize
	 || v >= md->itsGameOf.WordSearch.itsPuzzleSize)
		return;

	//	The orientation of the symbol relative to the fundamental
	//	domain should match the current orientation of the hand
	//	relative to the fundamental domain.
	md->itsGameOf.WordSearch.itsBoardFlips[h][v] = md->itsHandPlacement.itsFlip;
}


static bool WordSearchDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	Placement	theHotPoint;
#endif
	double		theHandH,
				theHandV;

#ifdef WORD_SEARCH_SCREENSHOT_COLORS
	static unsigned int	theColorCount = 0;
#endif

	UNUSED_PARAMETER(aShiftOrControlKeyIsDown);

	//	If the game is already over, ignore all hits.
	if (md->itsGameIsOver)
		return false;

	//	Make sure our data are valid.
	if ( ! md->itsGameOf.WordSearch.itsPuzzleIsValid )
		return false;

	//	The touch-based and mouse-based versions of torus word search handle drags differently.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Touch-based interface
	
	//	Is a selection already in progress?
	if (md->itsGameOf.WordSearch.itsWordSelectionIsPending)
	{
		//	Locate the hot point.
		GetHotPoint(md, &theHotPoint);

		//	Typically, the drag is reversed if and only if
		//	theHotPoint.itsFlip disagrees with itsHandPlacement.itsFlip.
		//	The exceptional case -- when theHotPoint sits near the top of the fundamental domain
		//	while itsHandPlacement sits near the bottom, or vice versa -- gets corrected for
		//	within the call to ShortestDistance(), immediately below.
		md->itsDragIsReversed = (theHotPoint.itsFlip != md->itsHandPlacement.itsFlip);

		//	If the user has touched theHotPoint...
		if (ShortestDistance(
					md->itsHandPlacement.itsH,
					md->itsHandPlacement.itsV,
					theHotPoint.itsH,
					theHotPoint.itsV,
					md->itsTopology,
					&md->itsDragIsReversed)
				< WORD_SEARCH_SELECTION_CONTINUATION_HIT_RADIUS)
		{
			//	...then let the pending selection continue.
			//	That is, start a new drag which will be part of the same pending selection.
			md->itsGameOf.WordSearch.itsDragStartH	= md->itsHandPlacement.itsH;
			md->itsGameOf.WordSearch.itsDragStartV	= md->itsHandPlacement.itsV;
			md->itsGameOf.WordSearch.itsDragWasTap	= true;

			//	All done!
			return true;
		}
		else	//	But if the user touches elsewhere,
		{
			//	treat the touch as a scroll.
			//	The pre-existing pending selection remains valid but suspended,
			//	awaiting future touches for its final completion.
			return false;
		}
	}
	
	//	No selection is in progress, so let the code below start a new one.
#else
	//	Mouse-based interface
	//	Selections are never left in suspension, so no extra code is needed here.
#endif
	
	//	If we return false for any reason, then no selection will be pending.
	md->itsGameOf.WordSearch.itsWordSelectionIsPending = false;

	//	Convert hand coordinates from [-0.5, +0.5] to [0.0, PuzzleSize].
	theHandH = md->itsGameOf.WordSearch.itsPuzzleSize*(md->itsHandPlacement.itsH + 0.5);
	theHandV = md->itsGameOf.WordSearch.itsPuzzleSize*(md->itsHandPlacement.itsV + 0.5);

	//	Find the integer coordinates (h,v) of the hit cell.
	md->itsGameOf.WordSearch.itsStartH = (unsigned int)(signed int) floor(theHandH);	//	-1 would map to 0xFFFFFFFF and get caught below.
	md->itsGameOf.WordSearch.itsStartV = (unsigned int)(signed int) floor(theHandV);

	//	Exclude values on the boundary and protect against unexpected errors.
	if (md->itsGameOf.WordSearch.itsStartH >= md->itsGameOf.WordSearch.itsPuzzleSize
	 || md->itsGameOf.WordSearch.itsStartV >= md->itsGameOf.WordSearch.itsPuzzleSize)
		return false;

	//	Are we near the center of the hit cell?
	//	If not, treat this hit as a scroll.
	md->itsGameOf.WordSearch.itsDeltaH = theHandH - (md->itsGameOf.WordSearch.itsStartH + 0.5);
	md->itsGameOf.WordSearch.itsDeltaV = theHandV - (md->itsGameOf.WordSearch.itsStartV + 0.5);
	if (md->itsGameOf.WordSearch.itsDeltaH * md->itsGameOf.WordSearch.itsDeltaH
	  + md->itsGameOf.WordSearch.itsDeltaV * md->itsGameOf.WordSearch.itsDeltaV
	  >  WORD_SEARCH_SELECTION_HIT_RADIUS  *  WORD_SEARCH_SELECTION_HIT_RADIUS )
		return false;

	//	The orientation of the symbol relative to the fundamental
	//	domain should match the current orientation of the hand
	//	relative to the fundamental domain.
	md->itsGameOf.WordSearch.itsBoardFlips[md->itsGameOf.WordSearch.itsStartH][md->itsGameOf.WordSearch.itsStartV] = md->itsHandPlacement.itsFlip;

	//	Normally aDragIsReversed specifies the relative parity
	//	of the hand and object being dragged.  In the present case,
	//	however, we're not so much dragging an object as making
	//	a mark on the board itself.  So let itsDragIsReversed
	//	record the parity of the hand relative to the board, 
	//	at the time of the MouseDown.
	md->itsDragIsReversed = md->itsHandPlacement.itsFlip;	//	initially "false" in touch-based UI

	//	Choose a random opaque color.
	//
	//	The R, G and B components should be premultiplied by the alpha component,
	//	but as long as the alpha component is 1.0 there's no need.
	md->itsGameOf.WordSearch.itsColor[0] = (float)rand() / (float)RAND_MAX;
	md->itsGameOf.WordSearch.itsColor[1] = (float)rand() / (float)RAND_MAX;
	md->itsGameOf.WordSearch.itsColor[2] = (float)rand() / (float)RAND_MAX;
	md->itsGameOf.WordSearch.itsColor[3] = 1.0;
#ifdef WORD_SEARCH_SCREENSHOT_COLORS
	if ((theColorCount++) & 0x00000001)
	{
		//	magenta
		md->itsGameOf.WordSearch.itsColor[0] = 1.0;
		md->itsGameOf.WordSearch.itsColor[1] = 0.0;
		md->itsGameOf.WordSearch.itsColor[2] = 0.5;
		md->itsGameOf.WordSearch.itsColor[3] = 1.0;
	}
	else
	{
		//	cyan
		md->itsGameOf.WordSearch.itsColor[0] = 0.0;
		md->itsGameOf.WordSearch.itsColor[1] = 0.5;
		md->itsGameOf.WordSearch.itsColor[2] = 1.0;
		md->itsGameOf.WordSearch.itsColor[3] = 1.0;
	}
#endif

	//	A new word selection has begin.
	md->itsGameOf.WordSearch.itsWordSelectionIsPending = true;
	
	//	The first drag of that (possibly multi-drag) word selection has begun
	//	(relevant only in touch interface).
	md->itsGameOf.WordSearch.itsDragStartH	= ((md->itsGameOf.WordSearch.itsStartH + 0.5) / md->itsGameOf.WordSearch.itsPuzzleSize) - 0.5;
	md->itsGameOf.WordSearch.itsDragStartV	= ((md->itsGameOf.WordSearch.itsStartH + 0.5) / md->itsGameOf.WordSearch.itsPuzzleSize) - 0.5;
	md->itsGameOf.WordSearch.itsDragWasTap	= false;	//	A selection's first drag never counts as a terminating tap.

	return true;
}


static void WordSearchDragObject(
	ModelData	*md,
	double		aHandLocalDeltaH,
	double		aHandLocalDeltaV)
{
	double	theNewDeltaH,
			theNewDeltaV,
			theStartH,
			theStartV,
			theFinishH,
			theFinishV;

	//	Before moving anything, check how big the delta would be.
	//	Visualized in the board's universal cover, the delta stays
	//	in the coordinates of the original image of the fundamental domain.
	//	(Accumulating numerical error isn't a problem, but if it were we could
	//	adjust the delta to stay synchronized with the hand.)
	theNewDeltaH = md->itsGameOf.WordSearch.itsDeltaH
				+ md->itsGameOf.WordSearch.itsPuzzleSize *
				(md->itsDragIsReversed ? -aHandLocalDeltaH : aHandLocalDeltaH);
	theNewDeltaV = md->itsGameOf.WordSearch.itsDeltaV
				+ md->itsGameOf.WordSearch.itsPuzzleSize * aHandLocalDeltaV;

	//	The TorusGames rendering algorithm will draw the word marker
	//	correctly iff it does not extend more than 0.5 board widths
	//	beyond the image of fundamental domain containing the center
	//	of the marker line.  A little thought confirms that the following
	//	code is both correct and optimal.  Note that if we cannot
	//	validly extend the marker line, we simply return without
	//	updating anything (not even the hand cursor, which the user
	//	will perceive as blocked).
	theStartH	= ((md->itsGameOf.WordSearch.itsStartH + 0.5) / md->itsGameOf.WordSearch.itsPuzzleSize) - 0.5;
	theStartV	= ((md->itsGameOf.WordSearch.itsStartV + 0.5) / md->itsGameOf.WordSearch.itsPuzzleSize) - 0.5;
	theFinishH	= theStartH + (theNewDeltaH / md->itsGameOf.WordSearch.itsPuzzleSize);
	theFinishV	= theStartV + (theNewDeltaV / md->itsGameOf.WordSearch.itsPuzzleSize);
	if (theFinishH > (theStartH > 0.0 ? +2.0 : +1.0)
	 || theFinishH < (theStartH < 0.0 ? -2.0 : -1.0)
	 || theFinishV > (theStartV > 0.0 ? +2.0 : +1.0)
	 || theFinishV < (theStartV < 0.0 ? -2.0 : -1.0))
		return;

	//	Update the delta.
	md->itsGameOf.WordSearch.itsDeltaH = theNewDeltaH;
	md->itsGameOf.WordSearch.itsDeltaV = theNewDeltaV;

	//	Move the hand.
	md->itsHandPlacement.itsH += md->itsHandPlacement.itsFlip
								? -aHandLocalDeltaH : aHandLocalDeltaH;
	md->itsHandPlacement.itsV += aHandLocalDeltaV;
	NormalizePlacement(&md->itsHandPlacement, md->itsTopology);
	
	//	If the user's finger has travelled a non-trivial distance
	//	from the start of the present drag, then the drag is not a tap.
	//	This is relevant only in the touch-based interface, not the mouse-based one.
	//	In the touch-based interface, a selection may consist of several separate drags,
	//	terminated by a tap.
	if (ShortestDistance(
				md->itsGameOf.WordSearch.itsDragStartH,
				md->itsGameOf.WordSearch.itsDragStartV,
				md->itsHandPlacement.itsH,
				md->itsHandPlacement.itsV,
				md->itsTopology,
				NULL)
			> WORD_SEARCH_TAP_TOLERANCE)
	{
		md->itsGameOf.WordSearch.itsDragWasTap = false;
	}

	//	Flip the underlying letter if necessary.
	WordSearchHandMoved(md);
}


static void WordSearchDragEnd(ModelData *md)
{
	signed int		theDeltaH,
					theDeltaV;
	WordSearchLine	*theNewlyFoundLine;
	bool			theWordsRemainFlag;
	unsigned int	i;

	//	The touch-based and mouse-based versions of torus word search handle drags differently.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Touch-based interface
	
	//	If the selection was not a simple tap, then even though the drag is over,
	//	let the selection remain pending, awaiting subsequent drags.
	if ( ! md->itsGameOf.WordSearch.itsDragWasTap )
		return;

	//	A simple tap ends both the drag and the move that the drag is a part of.
	//	Continue with the code below.
#else
	//	Mouse-based interface
	//	Each selection consists of a single drag.
#endif

	//	The selection is complete.
	md->itsGameOf.WordSearch.itsWordSelectionIsPending = false;

	//	What integer distance did the mouse travel?
	theDeltaH = (signed int) floor(md->itsGameOf.WordSearch.itsDeltaH + 0.5);
	theDeltaV = (signed int) floor(md->itsGameOf.WordSearch.itsDeltaV + 0.5);

	//	Accept only horizontal, vertical and diagonal lines.
	if (theDeltaH != 0
	 && theDeltaV != 0
	 && ABS(theDeltaH) != ABS(theDeltaV))
		return;

	//	Reject invalid words.
	if
	(
		! ScratchWordFromList(	md,
								md->itsGameOf.WordSearch.itsStartH,
								md->itsGameOf.WordSearch.itsStartV,
								+theDeltaH,
								+theDeltaV)
	 &&
		! ScratchWordFromList(	md,
								md->itsGameOf.WordSearch.itsStartH + theDeltaH,
								md->itsGameOf.WordSearch.itsStartV + theDeltaV,
								-theDeltaH,
								-theDeltaV)
	)
		return;

	//	Just to be safe, double check that we're
	//	not accepting too many lines.
	if (md->itsGameOf.WordSearch.itsNumLinesFound >= MAX_WORDSEARCH_NUM_WORDS)
		return;

	//	Record the new line.
	theNewlyFoundLine = &md->itsGameOf.WordSearch.itsLines[md->itsGameOf.WordSearch.itsNumLinesFound];
	theNewlyFoundLine->itsStartH	= md->itsGameOf.WordSearch.itsStartH;
	theNewlyFoundLine->itsStartV	= md->itsGameOf.WordSearch.itsStartV;
	theNewlyFoundLine->itsDeltaH	= theDeltaH;
	theNewlyFoundLine->itsDeltaV	= theDeltaV;
	for (i = 0; i < 4; i++)
		theNewlyFoundLine->itsColor[i] = md->itsGameOf.WordSearch.itsColor[i];
	md->itsGameOf.WordSearch.itsNumLinesFound++;

	//	Still more words to be found?
	theWordsRemainFlag = false;
	for (i = 0; i < MAX_WORDSEARCH_NUM_WORDS; i++)
		if (md->itsGameOf.WordSearch.itsWords[i][0] != 0)
			theWordsRemainFlag = true;

	//	If no words remain, the user has won!
	if ( ! theWordsRemainFlag )
	{
		md->itsGameIsOver = true;
		PlayTheSound(SoundWordSearchComplete);
		SimulationBegin(md, SimulationWordSearchFlash);
	}
}


static bool ScratchWordFromList(
	ModelData		*md,
	unsigned int	aStartH,
	unsigned int	aStartV,
	signed int		aDeltaH,
	signed int		aDeltaV)
{
	unsigned int	theAbsDeltaH,
					theAbsDeltaV,
					theLength,
					i,
					j;
	signed int		theCurrentH,
					theCurrentV,
					theIncrementH,
					theIncrementV,
					theNormalizedH,
					theNormalizedV;
	Char16			theCandidate[MAX_WORDSEARCH_WORD_LENGTH + 1];

	//	Re-confirm that the word is horizontal, vertical or diagonal.
	theAbsDeltaH = ABS(aDeltaH);
	theAbsDeltaV = ABS(aDeltaV);
	if (theAbsDeltaH != 0
	 && theAbsDeltaV != 0
	 && theAbsDeltaH != theAbsDeltaV)
		return false;

	//	How long is the word?
	theLength = MAX(theAbsDeltaH, theAbsDeltaV) + 1;

	//	Reject words that are too long.
	if (theLength > MAX_WORDSEARCH_WORD_LENGTH)
		return false;

	//	Assemble the candidate word, filling unused spaces with zeros.
	//	We're guaranteed to have at least one zero.

	theCurrentH = (signed int) aStartH;
	theCurrentV = (signed int) aStartV;

	theIncrementH = SIGN(aDeltaH);
	theIncrementV = SIGN(aDeltaV);

	for (i = 0; i < MAX_WORDSEARCH_WORD_LENGTH + 1; i++)
	{
		if (i < theLength)
		{
			theNormalizedH = theCurrentH;
			theNormalizedV = theCurrentV;
			NormalizeCellCoordinates(md, &theNormalizedH, &theNormalizedV);

			theCandidate[i] = md->itsGameOf.WordSearch.itsBoard[theNormalizedH][theNormalizedV];

			theCurrentH += theIncrementH;
			theCurrentV += theIncrementV;
		}
		else
			theCandidate[i] = 0;
	}

	//	Does the candidate appear on the word list?
	//	If so, remove it.
	for (i = 0; i < MAX_WORDSEARCH_NUM_WORDS; i++)
	{
		if (SameWord(theCandidate, md->itsGameOf.WordSearch.itsWords[i]))
		{
			//	Clear the word from the list.
			for (j = 0; j < MAX_WORDSEARCH_WORD_LENGTH + 1; j++)
				md->itsGameOf.WordSearch.itsWords[i][j] = 0;

			//	Refresh the Word List window.
			WordSearchSetList(md);

			//	Report success.
			return true;
		}
	}

	//	The candidate wasn't on the list.
	return false;
}


static void NormalizeCellCoordinates(
	ModelData	*md,
	signed int	*aCellH,
	signed int	*aCellV)
{
	while (*aCellV >= (signed int) md->itsGameOf.WordSearch.itsPuzzleSize)
	{
		*aCellV -= md->itsGameOf.WordSearch.itsPuzzleSize;
		if (md->itsTopology == TopologyKlein)
			*aCellH = (md->itsGameOf.WordSearch.itsPuzzleSize - 1) - *aCellH;
	}

	while (*aCellV < 0)
	{
		*aCellV += md->itsGameOf.WordSearch.itsPuzzleSize;
		if (md->itsTopology == TopologyKlein)
			*aCellH = (md->itsGameOf.WordSearch.itsPuzzleSize - 1) - *aCellH;
	}

	while (*aCellH >= (signed int) md->itsGameOf.WordSearch.itsPuzzleSize)
		*aCellH -= md->itsGameOf.WordSearch.itsPuzzleSize;

	while (*aCellH < 0)
		*aCellH += md->itsGameOf.WordSearch.itsPuzzleSize;
}


static bool SameWord(
	Char16	aWordA[MAX_WORDSEARCH_WORD_LENGTH + 1],
	Char16	aWordB[MAX_WORDSEARCH_WORD_LENGTH + 1])
{
	Char16			theWordA[MAX_WORDSEARCH_WORD_LENGTH + 1],
					theWordB[MAX_WORDSEARCH_WORD_LENGTH + 1];
	unsigned int	i;

	//	Copy the two words.
	//	Convert to lower case as we go along (this makes a difference
	//	only in languages that make an uppercase/lowercase distinction --
	//	in other languages ToLowerCase() should have no effect),
	//	and promote small kana to large kana as we go along (this makes
	//	a difference only in Japanese -- in other languages PromoteSmallKana()
	//	has no effect).
	for (i = 0; i < MAX_WORDSEARCH_WORD_LENGTH + 1; i++)
	{
		theWordA[i] = PromoteSmallKana(ToLowerCase(aWordA[i]));
		theWordB[i] = PromoteSmallKana(ToLowerCase(aWordB[i]));
	}

	//	Remove spaces and hyphens from both words.
	RemoveSpacesAndHyphens(theWordA);
	RemoveSpacesAndHyphens(theWordB);

	//	Compare the lowercase, spaceless, hyphenless words.
	for (i = 0; i < MAX_WORDSEARCH_WORD_LENGTH + 1; i++)
	{
		if (theWordA[i] != theWordB[i])
			return false;

		if (theWordA[i] == 0)
			return true;
	}

	return false;	//	should never occur
}


static void RemoveSpacesAndHyphens(Char16 aWord[MAX_WORDSEARCH_WORD_LENGTH + 1])
{
	unsigned int	r,	//	read index
					w;	//	write index

	//	Copy the word in place, omitting all spaces and hyphens.
	w = 0;
	for (r = 0; r < MAX_WORDSEARCH_WORD_LENGTH + 1; r++)
	{
		if (aWord[r] != ' '
		 && aWord[r] != '-')
			aWord[w++] = aWord[r];
	}

	//	If one or more spaces were omitted,
	//	fill the remainder of the string with zeros.
	while (w < MAX_WORDSEARCH_WORD_LENGTH + 1)
		aWord[w++] = 0;
}


static void WordSearchSimulationUpdate(ModelData *md)
{
	switch (md->itsSimulationStatus)
	{
		case SimulationWordSearchFlash:
			md->itsFlashFlag = (((unsigned int) floor(5.0 * md->itsSimulationElapsedTime)) % 2) ? true : false;
			if (md->itsSimulationElapsedTime >= 1.4)
			{
				SimulationEnd(md);
				md->itsFlashFlag = false;
			}
			break;

		default:
			break;
	}
}


static void WordSearchDraw(ModelData *md)
{
	Placement		theCellPlacement;
	unsigned int	h,
					v,
					i;
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	Placement		theHotPoint;
#endif

	static float	theBlackColor[4] = PREMULTIPLY_RGBA(0.0, 0.0, 0.0, 1.0);

	//	Draw the background.
	DrawFundamentalDomain(md, md->itsGameOf.WordSearch.itsTextureNameBackground, 17);

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(1.0, 0.0, 0.0, 1.0),
					PREMULTIPLY_RGBA(0.0, 0.0, 1.0, 1.0) );

	//	Draw the puzzle.
	if (md->itsGameOf.WordSearch.itsPuzzleIsValid)
	{
		//	Mark the words, except when flashing at the end of the game.
		if (! md->itsGameIsOver
		 || ! md->itsFlashFlag )
		{
			//	Mark all previously found words.
			for (i = 0; i < md->itsGameOf.WordSearch.itsNumLinesFound; i++)
			{
				DrawSegmentWithEnds(
					md,
					md->itsGameOf.WordSearch.itsLines[i].itsColor,
					md->itsGameOf.WordSearch.itsTextureNameMark,
					((md->itsGameOf.WordSearch.itsLines[i].itsStartH + 0.5) + 0.5*md->itsGameOf.WordSearch.itsLines[i].itsDeltaH)/md->itsGameOf.WordSearch.itsPuzzleSize - 0.5,
					((md->itsGameOf.WordSearch.itsLines[i].itsStartV + 0.5) + 0.5*md->itsGameOf.WordSearch.itsLines[i].itsDeltaV)/md->itsGameOf.WordSearch.itsPuzzleSize - 0.5,
					sqrt(md->itsGameOf.WordSearch.itsLines[i].itsDeltaH*md->itsGameOf.WordSearch.itsLines[i].itsDeltaH
							+ md->itsGameOf.WordSearch.itsLines[i].itsDeltaV*md->itsGameOf.WordSearch.itsLines[i].itsDeltaV)
						/ md->itsGameOf.WordSearch.itsPuzzleSize,
					1.00 / md->itsGameOf.WordSearch.itsPuzzleSize,
					atan2(md->itsGameOf.WordSearch.itsLines[i].itsDeltaV, md->itsGameOf.WordSearch.itsLines[i].itsDeltaH));
			}

			//	If we're dragging, mark the tentatively selected word.
			if (md->itsGameOf.WordSearch.itsWordSelectionIsPending)
			{
				DrawSegmentWithEnds(
					md,
					md->itsGameOf.WordSearch.itsColor,
					md->itsGameOf.WordSearch.itsTextureNameMark,
					((md->itsGameOf.WordSearch.itsStartH + 0.5) + 0.5*md->itsGameOf.WordSearch.itsDeltaH)/md->itsGameOf.WordSearch.itsPuzzleSize - 0.5,
					((md->itsGameOf.WordSearch.itsStartV + 0.5) + 0.5*md->itsGameOf.WordSearch.itsDeltaV)/md->itsGameOf.WordSearch.itsPuzzleSize - 0.5,
					sqrt(md->itsGameOf.WordSearch.itsDeltaH*md->itsGameOf.WordSearch.itsDeltaH + md->itsGameOf.WordSearch.itsDeltaV*md->itsGameOf.WordSearch.itsDeltaV)
							/ md->itsGameOf.WordSearch.itsPuzzleSize,
					1.00 / md->itsGameOf.WordSearch.itsPuzzleSize,
					atan2(md->itsGameOf.WordSearch.itsDeltaV, md->itsGameOf.WordSearch.itsDeltaH));
			}
		}

		//	Initialize a general-purpose cell Placement.
		theCellPlacement.itsAngle	= 0.0;
		theCellPlacement.itsSizeH	= 1.0 / md->itsGameOf.WordSearch.itsPuzzleSize;
		theCellPlacement.itsSizeV	= 1.0 / md->itsGameOf.WordSearch.itsPuzzleSize;

		//	Draw the cells.
		for (h = 0; h < md->itsGameOf.WordSearch.itsPuzzleSize; h++)
		{
			theCellPlacement.itsH = -0.5 + (0.5 + h)*theCellPlacement.itsSizeH;

			for (v = 0; v < md->itsGameOf.WordSearch.itsPuzzleSize; v++)
			{
				theCellPlacement.itsV		= -0.5 + (0.5 + v)*theCellPlacement.itsSizeV;
				theCellPlacement.itsFlip	= md->itsGameOf.WordSearch.itsBoardFlips[h][v];

				DrawColorWithAlphaMask(	md,
										md->itsGameOf.WordSearch.itsTextureNameCell[h][v],
										&theCellPlacement,
										theBlackColor);
			}
		}
		
#ifdef TORUS_GAMES_FOR_OPENGL_ES
		//	Draw the selection's hot spot, if relevant.
		if (md->itsGameOf.WordSearch.itsWordSelectionIsPending)
		{
			GetHotPoint(md, &theHotPoint);
			DrawColorWithAlphaMask(	md,
									md->itsTextureNamesCommon[TextureCommonHalo],
									&theHotPoint,
									PREMULTIPLY_RGBA(0.0, 0.5, 1.0, 1.0));
		}
#endif
	}
}


#ifdef TORUS_GAMES_FOR_OPENGL_ES
static void GetHotPoint(
	ModelData	*md,		//	input
	Placement	*aHotPoint)	//	output
{
	double		theStartH,
				theStartV,
				theFinishH,
				theFinishV;

	if (md->itsGameOf.WordSearch.itsWordSelectionIsPending)
	{
		//	Locate the selection.
		theStartH	= ((md->itsGameOf.WordSearch.itsStartH + 0.5) / md->itsGameOf.WordSearch.itsPuzzleSize) - 0.5;
		theStartV	= ((md->itsGameOf.WordSearch.itsStartV + 0.5) / md->itsGameOf.WordSearch.itsPuzzleSize) - 0.5;
		theFinishH	= theStartH + (md->itsGameOf.WordSearch.itsDeltaH / md->itsGameOf.WordSearch.itsPuzzleSize);
		theFinishV	= theStartV + (md->itsGameOf.WordSearch.itsDeltaV / md->itsGameOf.WordSearch.itsPuzzleSize);

		//	Express the finishing point as a Placement.
		//	Conceptually, the finishing point inherits the starting point's unflipped chirality.
		aHotPoint->itsH		= theFinishH;
		aHotPoint->itsV		= theFinishV;
		aHotPoint->itsFlip	= false;
		aHotPoint->itsAngle	= 0.0;
		aHotPoint->itsSizeH	= 1.0 / md->itsGameOf.WordSearch.itsPuzzleSize;
		aHotPoint->itsSizeV	= 1.0 / md->itsGameOf.WordSearch.itsPuzzleSize;

		//	Normalize the finishing point to sit within the fundamental domain.
		//	The normalization will adjust the finishing point's chirality as required.
		NormalizePlacement(aHotPoint, md->itsTopology);
	}
	else
	{
		//	This case should never occur, but provide something to be safe.
		aHotPoint->itsH		= 0.0;
		aHotPoint->itsV		= 0.0;
		aHotPoint->itsFlip	= false;
		aHotPoint->itsAngle	= 0.0;
		aHotPoint->itsSizeH	= 1.0;
		aHotPoint->itsSizeV	= 1.0;
	}
}	
#endif
