//	TorusGamesMacOSUtilities.c
//
//	A few platform-dependent functions that the platform-independent code will need.
//	(Along with a couple global variables whose use I'll need to ponder further.)
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#import <Cocoa/Cocoa.h>
#include <AudioToolbox/AudioToolbox.h>	//	for MIDI player
#include <pthread.h>	//	for pthread_create()
#include <unistd.h>		//	for usleep()
#include "TorusGames-Common.h"
#include "TorusGamesMacOSUtilities.h"
	
	
//	Package up information for the new-thread wrapper function.
typedef struct
{
	ModelData	*md;
	void		(*itsStartFuction)(ModelData *);
	bool		itsArgumentsHaveBeenCopied;
} ThreadStartData;


//	Maintain a global MIDI player.
MusicPlayer		gMIDIPlayer		= NULL;
MusicSequence	gMIDISequence	= NULL;


//	Put a platform-dependent wrapper around the platform-independent thread starter function.
static void		*StartFunctionWrapper(void *aParam);


//	Define a SoundStopper object to release NSSound objects when they're done playing.
@interface SoundStopper : NSObject
{
}
- (id)initWithSound:(NSSound *)aSound;
- (void)sound:(NSSound *)sound didFinishPlaying:(BOOL)finishedPlaying;
@end
@implementation SoundStopper
- (id)initWithSound:(NSSound *)aSound
{
	self = [super init];
	if (self != nil)
	{
		[aSound retain];
		[self retain];
	}
	return self;
}
- (void)sound:(NSSound *)sound didFinishPlaying:(BOOL)finishedPlaying
{
	[sound setDelegate:nil];
	[sound release];
	[self release];
}
@end


NSString *GetLocalizedTextAsNSString(
	const char	*aKey)
{
	return [NSString stringWithUTF8String:GetLocalizedText(aKey)];
}


bool GetBasePath(
	char			*aPath,
	unsigned int	aPathSize)
{
	return [[[[NSBundle mainBundle] resourcePath] stringByAppendingString:@"/"]
		getCString:aPath maxLength:aPathSize encoding:NSUTF8StringEncoding];
}


ErrorText GetFileContents(
	const char		*aPathName,		//	input, zero-terminated UTF-8 string containing absolute path name
	unsigned int	*aNumRawBytes,	//	output, the file size in bytes
	Byte			**someRawBytes)	//	output, the file's contents as raw bytes
{
	//	Technical Note:  I had hoped that setlocale() + fopen() would
	//	let me read the file within the platform-independent part
	//	of the code, but alas on WinXP it works when aPathName 
	//	contains only ASCII characters but fails when aPathName 
	//	contains non-trivial Unicode characters, for example when
	//	I temporarily change "Textures" to "Textures日本".
	//	So I reluctantly resorted to platform-dependent code
	//	to read the file.

	NSData	*theFileContents;
	
	//	Read the file.
	theFileContents = [NSData dataWithContentsOfFile:[NSString stringWithUTF8String:aPathName]];
	if (theFileContents == nil)
		return "Couldn't get file contents as NSData in GetFileContents()";
	
	//	Note the number of bytes.
	*aNumRawBytes = [theFileContents length];
	if (*aNumRawBytes == 0)
		return "File contains 0 bytes in GetFileContents()";

	//	Copy the data.  The caller will own this copy, and must
	//	eventually free it with a call to FreeRawTextureBytes().
	*someRawBytes = GET_MEMORY(*aNumRawBytes);
	if (*someRawBytes == NULL)
	{
		*aNumRawBytes = 0;
		return "Couldn't allocate buffer in GetFileContents()";
	}
	[theFileContents getBytes:(void *)(*someRawBytes) length:(*aNumRawBytes)];
	
	//	Success!
	return NULL;
}

void FreeFileContents(
	unsigned int	*aNumRawBytes,
	Byte			**someRawBytes)
{
	*aNumRawBytes = 0;
	FREE_MEMORY_SAFELY(*someRawBytes);
}


void SetAlphaTextureFromCharacter(
	unsigned int	aTextureName,
	Char16			aCharacter,
	OpaqueFont		anOpaqueFont,	//	unused
	unsigned int	aSize,
	bool			aShadersFlag,
	ErrorText		*aFirstError)
{
	ErrorText				theErrorMessage			= NULL;
	Byte					*theAlphaBytes			= NULL;
	CGContextRef			theBitmapContext		= NULL;
	UniCharCount			theRunLengthArray[1]	= {kATSUToTextEnd};
	ATSUStyle				theStyleArray[1]		= {NULL};
	unsigned int			theCellHeight			= 0;
	Fixed					theFontSize				= 0,	//	Overwritten below.
							theBaseline				= 0;
	ATSUAttributeTag		theStyleTags[1]			= {kATSUSizeTag};
	UInt32					theStyleValueSizes[1]	= {sizeof(Fixed)};
	ATSUAttributeValuePtr	theStyleValues[1]		= {&theFontSize};
	ATSUTextLayout			theATSUTextLayout		= NULL;
	ATSUAttributeTag		theLayoutTags[1]		= {kATSUCGContextTag};
	UInt32					theLayoutValueSizes[1]	= {sizeof(CGContextRef)};
	ATSUAttributeValuePtr	theLayoutValues[1]		= {&theBitmapContext};
	Rect					theImageRect			= {0, 0, 0, 0};

	//	Assumes desired OpenGL context is already active.
	
	//	Note:  We could in theory set up the ATSUI structures once and for all, in advance,
	//	rather than creating them fresh each time SetAlphaTextureFromCharacter() gets called.
	//	However,
	//		1.	the present code is fast enough (~3ms on my iMac) and, more importantly,
	//		2.	most of the time (~2ms on my iMac) is spent in the call to SetTextureImageFancy(),
	//	so there's little to gain by pre-initializing the ATSUI structures.
	//	Instead, let's keep the code simple, with everything right here in one place.

	//	OpenGL (and the underlying graphics card) likes textures with power-of-two dimensions.
	if ( ! IsPowerOfTwo(aSize) )
	{
		theErrorMessage = "Character texture size must be a power of two.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Allocate temporary memory for the bitmapped image.
	//	The image will be an alpha-only image with 1 byte per pixel (no RGB!).
	theAlphaBytes = (Byte *) GET_MEMORY(aSize*aSize*1);
	if (theAlphaBytes == NULL)
	{
		theErrorMessage = "Couldn't allocate theAlphaBytes.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}
	
	//	Create a Core Graphics context using theAlphaBytes.
	theBitmapContext = CGBitmapContextCreate(
							theAlphaBytes,
							aSize,
							aSize,
							8,
							aSize*1,
							NULL,
							kCGImageAlphaOnly);
	if (theBitmapContext == NULL)
	{
		theErrorMessage = "Couldn't create theBitmapContext.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}
	
	//	Even though Core Graphics and ATSUI both use a consistent coordinate system
	//	with the y-coordinate running bottom-to-top, it seems that CGBitmapContextCreate()
	//	runs the rows of the buffer (theAlphaBytes) top-to-bottom, apparently
	//	for compatibility with the old QuickDraw coordinate system.  (For details, go to
	//		http://developer.apple.com/samplecode/TypeServicesForUnicode/listing1.html
	//	and search for "upside down".)
	//	To simulate a bottom-to-top image, let's flip the coordinates.
	//	(The other alternative, if this strategy ever becomes problematic,
	//	would be to manually swap the rows of the image after all drawing is complete.)
	CGContextScaleCTM(theBitmapContext, 1, -1);
	CGContextTranslateCTM(theBitmapContext, 0, -((float)aSize));

	//	Clear the image.
	CGContextClearRect(theBitmapContext, (CGRect) { {0, 0}, {aSize, aSize} } );

	//	ATSUI wants theFontSize in points, where 1pt = 1/72in.
	//	We are given aSize in pixels, with no notion of inches anywhere in sight.
	//	Traditionally one assumed 72 pixels/inch (so 1pt = 1px)
	//	but that assumption seems shaky.  Hope for the best.
	theCellHeight = aSize;	

	//	ATSUI wants the font height excluding the descent,
	//	so we must manually fudge things a bit to allow room for descenders.
	theFontSize	= FloatToFixed( 0.80 * theCellHeight );
	theBaseline	= FloatToFixed( 0.20 * theCellHeight );

	//	Create a style that specifies the font size.
	//	We've already initialized theStyleValues to contain a pointer to theFontSize.
	if (ATSUCreateStyle(&theStyleArray[0]) != noErr)
	{
		theErrorMessage = "Couldn't create the ATSU Style object.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}
	if (ATSUSetAttributes(
			theStyleArray[0],
			1,
			theStyleTags,
			theStyleValueSizes,
			theStyleValues)
		!= noErr)
	{
		theErrorMessage = "Couldn't set ATSU Style attributes.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Create the main text layout structure.
	//	The text arraying contains aCharacter alone.
	if (ATSUCreateTextLayoutWithTextPtr(
			&aCharacter,
			kATSUFromTextBeginning,
			kATSUToTextEnd,
			1,	//	1 unichar in buffer
			1,	//	1 style run
			theRunLengthArray,
			theStyleArray,
			&theATSUTextLayout)
		!= noErr)
	{
		theErrorMessage = "Couldn't create theATSUTextLayout.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}
	
	//	If the default font can't handle aCharacter,
	//	let ATSUI find a font that can.
	(void) ATSUSetTransientFontMatching(theATSUTextLayout, true);
	
	//	Assign theBitmapContext to theATSUTextLayout.
	//	We've already initialized theLayoutValues to contain a pointer to theBitmapContext.
	if (ATSUSetLayoutControls(
			theATSUTextLayout,
			1,
			theLayoutTags,
			theLayoutValueSizes,
			theLayoutValues)
		!= noErr)
	{
		theErrorMessage = "ATSUSetLayoutControls() failed.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}
	
	//	Measure the width of the character as it will be rendered,
	//	so we can center it correctly.
	if (ATSUMeasureTextImage(
			theATSUTextLayout,
			0,
			1,
			FloatToFixed(0),
			FloatToFixed(0),
			&theImageRect)
		!= noErr)
	{
		theErrorMessage = "ATSUMeasureTextImage() failed.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Draw the character into theBitmapContext.
	if (ATSUDrawText(
			theATSUTextLayout,
			kATSUFromTextBeginning,
			kATSUToTextEnd,
			FloatToFixed( (aSize - theImageRect.right) /2.0 ),
			theBaseline)
		!= noErr)
	{
		theErrorMessage = "ATSUDrawText() failed.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Convert the raw bitmap image to an OpenGL alpha-only texture.
	SetTextureImageFancy(aTextureName, aSize, aSize, 1, theAlphaBytes, aShadersFlag);

CleanUpSetAlphaTextureFromCharacter:

	if (theATSUTextLayout != NULL)
		ATSUDisposeTextLayout(theATSUTextLayout);
	
	if (theStyleArray[0] != NULL)
		ATSUDisposeStyle(theStyleArray[0]);

	if (theBitmapContext != NULL)
		CGContextRelease(theBitmapContext);

	FREE_MEMORY_SAFELY(theAlphaBytes);

	if (theErrorMessage != NULL)
	{
		//	Substitute a transparent texture for the missing one.
		SetTextureImageFancy(aTextureName, 1, 1, 1, (Byte [1]){0x00}, aShadersFlag);
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}


void SetUpMIDIPlayer(void)
{
	if (NewMusicPlayer(&gMIDIPlayer) != noErr)
		gMIDIPlayer = NULL;

	gMIDISequence = NULL;
}

void ShutDownMIDIPlayer(void)
{
	if (gMIDIPlayer != NULL)
	{
		MusicPlayerStop(gMIDIPlayer);
		DisposeMusicPlayer(gMIDIPlayer);
		gMIDIPlayer = NULL;
	}
	
	if (gMIDISequence != NULL)
	{
		DisposeMusicSequence(gMIDISequence);
		gMIDISequence = NULL;
	}
}

void PlayMIDI(
	const char	*aRelativePath,	//	zero-terminated UTF-8 string with path separator '/'
	SoundType	aSound)			//	sound index permits caching in iPhone Torus Games
{
//	SInt32			theMajorVersionNumber,
//					theMinorVersionNumber;
	MusicSequence	theNewSequence	= NULL;
	
	UNUSED_PARAMETER(aSound);
	
	//	On the one hand, Mac OS 10.4 does not officially support MusicSequenceFileLoadData().
	//	(I think it included something similar, but with a different syntax.)
//	Gestalt(gestaltSystemVersionMajor, &theMajorVersionNumber);
//	Gestalt(gestaltSystemVersionMinor, &theMinorVersionNumber);
//	if (theMajorVersionNumber == 10
//	 && theMinorVersionNumber <= 4)
//		return;

	//	On the other hand, the following code was running fine
	//	on my own Mac OS 10.4.x systems until I upgraded to 10.6 a few months ago.
	//	The only difference is that Xcode 3.2.1 now complains about it.
	//	So probably MusicSequenceFileLoadData is OK,
	//	but let's check the function pointer to be safe.
	if (MusicSequenceFileLoadData == NULL)
		return;

	if (gMIDIPlayer != NULL)
	{
		//	Stop any music that might already be playing.
		MusicPlayerStop(gMIDIPlayer);

		//	Create a new sequence, start it playing,
		//	and remember it so we can free it later.
		if (NewMusicSequence(&theNewSequence) == noErr)
		{
			if
			(
				MusicSequenceFileLoadData(
					theNewSequence,
					(CFDataRef) [NSData dataWithContentsOfFile:
						[[[[NSBundle mainBundle] resourcePath]
							stringByAppendingString:@"/Sounds - Desktop/"]
							stringByAppendingString:[NSString stringWithUTF8String:aRelativePath]]],
					kMusicSequenceFile_MIDIType,
					0
					) == noErr
			 &&
				MusicPlayerSetSequence(gMIDIPlayer, theNewSequence) == noErr
			)
			{
				//	Start the new music playing.
				MusicPlayerStart(gMIDIPlayer);
				
				//	Destroy the old sequence and remember the new one.
				DisposeMusicSequence(gMIDISequence);
				gMIDISequence	= theNewSequence;
				theNewSequence	= NULL;
			}
			else
			{
				DisposeMusicSequence(theNewSequence);
				theNewSequence = NULL;
			}
		}
	}
}

void PlayWAV(
	const char	*aRelativePath,	//	zero-terminated UTF-8 string with path separator '/'
	SoundType	aSound)			//	sound index permits caching in iPhone Torus Games
{
	NSSound	*theSound	= nil;
	
	UNUSED_PARAMETER(aSound);

	//	Create theSound.
	theSound = [[[NSSound alloc]
					initWithContentsOfFile:	[[[[NSBundle mainBundle] resourcePath]
							stringByAppendingString:@"/Sounds - Desktop/"]
							stringByAppendingString:[NSString stringWithUTF8String:aRelativePath]]
					byReference:			YES] autorelease];
	
	//	We'll need to release theSound after it has finished playing.
	//	To know when it's finished, create a SoundStopper object
	//	to receive the sound:didFinishPlaying: message.
	//	Cocoa objects do not retain their delegates,
	//	so the SoundStopper must retain both theSound and itself,
	//	and must release both theSound and itself when
	//	theSound has finished playing.
	[theSound setDelegate:[[[SoundStopper alloc] initWithSound:theSound] autorelease]];
	
	//	Start theSound playing.
	[theSound play];
}

void ClearSoundCache(void)
{
	//	The Mac OS Torus Games do not maintain a sound cache.
}


void StartNewThread(
	ModelData	*md,
	void		(*aStartFuction)(ModelData *))
{
	//	Put a wrapper around aStartFuction to isolate details
	//	like _cdecl vs. _stdcall from the rest of the program.
	
	ThreadStartData	theThreadStartData;
	pthread_t		theNewThreadID;	//	We'll ignore the new thread's ID.
	
	theThreadStartData.md							= md;
	theThreadStartData.itsStartFuction				= aStartFuction;
	theThreadStartData.itsArgumentsHaveBeenCopied	= false;

	pthread_create(&theNewThreadID, NULL, StartFunctionWrapper, (void *)&theThreadStartData);
	
	while ( ! theThreadStartData.itsArgumentsHaveBeenCopied )
		SleepBriefly();
}

static void *StartFunctionWrapper(void *aParam)
{
	ThreadStartData	theThreadStartData;
	
	//	Copy the input parameters to our own local memory.
	theThreadStartData = *(ThreadStartData *)aParam;
	
	//	Let the caller know we've made a copy,
	//	so it can return safely.
	((ThreadStartData *)aParam)->itsArgumentsHaveBeenCopied = true;
	
	//	Call itsStartFunction.
	//	Whatever calling convention the compiler prefers is fine by us.
	(*theThreadStartData.itsStartFuction)(theThreadStartData.md);
	
	return NULL;
}


void SleepBriefly(void)
{
	//	Normally the platform-independent code doesn't sleep;
	//	it returns controls to the platform-dependent code
	//	which manages idle time as it sees fit.
	//	However, when Chess waits for its secondary thread
	//	to terminate, it needs to do something to avoid hogging
	//	CPU cycles while it waits.
	
	usleep(10000);	//	10000 µs = 10 ms
}


void ErrorMessage(
	ErrorText	aMessage,	//	UTF8
	ErrorText	aTitle,		//	UTF8
	bool		aFatalFlag)	//	Exit the program after user dismisses the error alert?
{
	//	NSRunAlertPanel works better than NSBeginAlertSheet
	//	because the former blocks while the latter does not.
	//	This is especially important when displaying a message
	//	during the shutdown sequence (for example the memory-leak message),
	//	when the application would quit out from under any pending
	//	NSBeginAlertSheet() calls.

	NSRunAlertPanel(	[NSString stringWithUTF8String:aTitle  ],
						[NSString stringWithUTF8String:aMessage],
						@"OK",
						NULL,
						NULL);
	
	if (aFatalFlag)
		exit(1);
}
