//	TorusGames-Win32-Utilities.c
//
//	Provides platform-dependent supporting functions
//	for the Win32 version of Torus Games.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Win32.h"
#include <string.h>		//	for wcslen() and wcscpy()
#include <mmsystem.h>	//	comment out:  typedef UINT FAR   *LPUINT;
#include <process.h>


typedef struct
{
	ModelData	*md;
	void		(*itsStartFuction)(ModelData *);
	bool		itsArgumentsHaveBeenCopied;
} ThreadStartData;


static bool			ConvertEndOfLineMarkers(WCHAR *aTextBuffer, unsigned int aTextBufferLength);
static void __cdecl	StartFunctionWrapper(void *aParam);


#define CONVERSION_ERROR_STRING		L"<error converting text>"

WCHAR *GetLocalizedTextAsWCHAR(
	const char		*aKey,				//	zero-terminated UTF-8 string
	WCHAR			*aValueBuffer,		//	zero-terminated WCHAR string
	unsigned int	aValueBufferLength)	//	length of aValueBuffer in WCHARs (not bytes),
										//		including space for the terminating zero
{
	if (aValueBufferLength == 0)
		ErrorMessage("GetLocalizedTextAsWCHAR() received zero-length output buffer.", "Internal Error", true);

	if (MultiByteToWideChar(
			CP_UTF8, MB_ERR_INVALID_CHARS,
			GetLocalizedText(aKey), -1,
			aValueBuffer, aValueBufferLength) == 0
	 || ! ConvertEndOfLineMarkers(aValueBuffer, aValueBufferLength) )
	{
		if (aValueBufferLength >= wcslen(CONVERSION_ERROR_STRING) + 1)
			wcscpy(aValueBuffer, CONVERSION_ERROR_STRING);
		else
		if (aValueBufferLength >= 2)
			wcscpy(aValueBuffer, L"?");
		else
			aValueBuffer[0] = 0;
	}

	return aValueBuffer;
}

static bool ConvertEndOfLineMarkers(
	WCHAR			*aTextBuffer,		//	zero-terminated WCHAR string
	unsigned int	aTextBufferLength)	//	length of aTextBuffer in WCHARs (not bytes),
										//		including space for the terminating zero;
										//		typically greater than the space occupied
										//		by the string itself
{
	//	Convert "\n" to "\r\n" if aValueBuffer is big enough to hold it.

	WCHAR	*r,	//	read  position
			*w;	//	write position

	//	Set r to the terminating zero.
	//	Set w a little bit further on, 
	//	to allow room to insert '\r' characters.
	r = w = aTextBuffer;
	while (*r != 0)
	{
		if (*r == L'\n')
			w++;
		r++;
		w++;
	}
	
	//	If we've got room to insert the '\r' characters...
	if (w < aTextBuffer + aTextBufferLength)
	{
		//	then read the string backwards...
		do
		{
			//	copying all characters...
			*w = *r;

			//	and inserting '\r' characters as necessary.
			if (*r == L'\n'
			 && --w >= aTextBuffer)	//	error check is logically unnecessary
			{
				*w = L'\r';
			}

		} while (--r >= aTextBuffer && --w >= aTextBuffer);
			
		return true;
	}
	else
		return false;
}


bool GetBasePath(
	char			*aPath,			//	output buffer, zero-terminated UTF-8 string
	unsigned int	aPathSize)		//	output buffer size, including space for terminating zero
{
	bool			theReturnValue	= false;
	WCHAR			*thePath		= NULL;	//	zero-terminated UTF-16 string
	unsigned int	theTerminator,
					i;

	//	Allocate a buffer for the UTF-16 version of the path.
	//	One UTF-16 character is guaranteed to be enough
	//	to handle any string that will ultimately fit
	//	into a UTF-8 path of aPathSize bytes.
	thePath = (WCHAR *) GET_MEMORY(aPathSize * sizeof(WCHAR));
	if (thePath == NULL)
		goto CleanUpGetBasePath;

	//	Get an absolute path for our executable file.
	//	GetModuleFileName() will return the length of the string,
	//	which we interpret as the position of the terminating zero.
	//	If the true path is too long for the buffer, presumably
	//	GetModuleFileName() allows room for the terminating zero,
	//	so assume a path that just barely fits has been truncated.
	//	If GetModuleFileName() fails for any other reason,
	//	it should return 0.
	theTerminator = GetModuleFileName(NULL, thePath, aPathSize);
	if (theTerminator == 0
	 || theTerminator >= aPathSize - 1)
	{
		goto CleanUpGetBasePath;
	}

	//	Remove the name of executable file itself,
	//	leaving only the path to the directory,
	//	including the trailing backslash.
	while (theTerminator > 0 && thePath[theTerminator - 1] != '\\')
		thePath[--theTerminator] = 0;
	
	//	Convert the path separator from '\' to '/'.
	for (i = 0; i < theTerminator; i++)
		if (thePath[i] == '\\')
			thePath[i] = '/';

	//	Convert to UTF-8.
	if (WideCharToMultiByte(CP_UTF8, 0, thePath, -1, aPath, aPathSize, NULL, NULL) == 0)
		goto CleanUpGetBasePath;

	//	Success!
	theReturnValue = true;

CleanUpGetBasePath:
	
	FREE_MEMORY_SAFELY(thePath);
	
	if ( ! theReturnValue )
		aPath[0] = 0;
	
	return theReturnValue;
}


ErrorText GetFileContents(
	const char		*aPathName,		//	absolute path as zero-terminated UTF-8 string
	unsigned int	*aNumRawBytes,	//	output, size of file's contents
	Byte			**someRawBytes)	//	output, copy of file's contents
{
	//	Technical Note:  I had hoped that setlocale() + fopen() would
	//	let me read the file within the platform-independent part
	//	of the code, but alas on WinXP it works when aPathName 
	//	contains only ASCII characters but fails when aPathName 
	//	contains non-trivial Unicode characters, for example when
	//	I temporarily change "Textures" to "Textures日本".
	//	So I reluctantly resorted to platform-dependent code
	//	to read the file.

	ErrorText		theErrorMessage	= NULL;
	unsigned int	thePathLength	= 0;
	WCHAR			*thePathName	= NULL;
	HANDLE			theFile			= INVALID_HANDLE_VALUE;
	unsigned long	theNumBytesRead	= 0;

	//	Convert aPathName from UTF-8 to UTF-16.
	//	One UTF-16 character per UTF-8 character
	//	is guaranteed to suffice (even if the UTF-16 text
	//	includes surrogate pairs).
	thePathLength = 0;
	while (aPathName[thePathLength++] != 0)	//	counts the terminating zero as well
		;
	thePathName = (WCHAR *) GET_MEMORY(thePathLength * sizeof(WCHAR));
	if (thePathName == NULL)
	{
		theErrorMessage = "Couldn't get memory for UTF-16 path name in GetFileContents().";
		goto CleanUpGetFileContents;
	}
	if (MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aPathName, -1, thePathName, thePathLength) == 0)
	{
		theErrorMessage = "Couldn't convert path name from UTF-8 to UTF-16 in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Open the file.
	theFile = CreateFile(thePathName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (theFile == INVALID_HANDLE_VALUE)
	{
		theErrorMessage = "Couldn't open file in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Get the file size.
	*aNumRawBytes = GetFileSize(theFile, NULL);
	if (*aNumRawBytes == INVALID_FILE_SIZE)
	{
		theErrorMessage = "Couldn't get file size in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Allocate a buffer to read the file contents into.
	*someRawBytes = (Byte *) GET_MEMORY(*aNumRawBytes);
	if (*someRawBytes == NULL)
	{
		theErrorMessage = "Couldn't allocate memory for file's contents in GetFileContents().";
		goto CleanUpGetFileContents;
	}

	//	Read the file contents into the buffer.
	if (ReadFile(theFile, *someRawBytes, *aNumRawBytes, &theNumBytesRead, NULL) == 0
	 || theNumBytesRead != *aNumRawBytes)
	{
		theErrorMessage = "Couldn't read file's contents in GetFileContents().";
		goto CleanUpGetFileContents;
	}

CleanUpGetFileContents:
	
	if (theFile != INVALID_HANDLE_VALUE)
		CloseHandle(theFile);

	FREE_MEMORY_SAFELY(thePathName);
	
	if (theErrorMessage != NULL)
	{
		*aNumRawBytes = 0;
		FREE_MEMORY_SAFELY(*someRawBytes);
	}

	return theErrorMessage;
}

void FreeFileContents(
	unsigned int	*aNumRawBytes,
	Byte			**someRawBytes)
{
	*aNumRawBytes = 0;
	FREE_MEMORY_SAFELY(*someRawBytes);
}


void PlayMIDI(
	const char	*aRelativePath,	//	zero-terminated UTF-8 string with path separator '/'
	SoundType	aSound)			//	sound index permits caching in iPhone Torus Games
{
	//	Very short MIDI files (sound effects) open acceptably fast.
	//	Longer MIDI files (songs) load fast on Win95/98 but take tens of seconds
	//	to load on WinXP (!) and must therefore be pre-loaded in a separate thread
	//	(see "MIDI thread - old/TorusGamesMusic.c" for working code).

	char	theFullPath[2048]			= "";
	WCHAR	theFullPathUTF16[2048]		= L"",
			theOpenCommand[2048 + 128]	= L"";	//	fixed length string is dangerous!
	
	UNUSED_PARAMETER(aSound);

	if (GetAbsolutePath("Sounds - Desktop", aRelativePath, theFullPath, BUFFER_LENGTH(theFullPath)) != NULL)
		return;

	if (MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, theFullPath, -1, theFullPathUTF16, BUFFER_LENGTH(theFullPathUTF16)) == 0)
		return;

	//	Format theOpenCommand.
	//	(Dangerous code:  Watch out for buffer overruns!)
	wsprintf(theOpenCommand, L"open \"%s\" type sequencer alias TheSound", theFullPathUTF16);

	mciSendString(L"stop all",				0, 0, 0);
	mciSendString(L"close all",				0, 0, 0);
	mciSendString(theOpenCommand,			0, 0, 0);
	mciSendString(L"play TheSound from 0",	0, 0, 0);
}

void PlayWAV(
	const char	*aRelativePath,	//	zero-terminated UTF-8 string with path separator '/'
	SoundType	aSound)			//	sound index permits caching in iPhone Torus Games
{
	char	theFullPath[2048]		= "";
	WCHAR	theFullPathUTF16[2048]	= L"";
	
	UNUSED_PARAMETER(aSound);

	if (GetAbsolutePath("Sounds - Desktop", aRelativePath, theFullPath, BUFFER_LENGTH(theFullPath)) != NULL)
		return;

	if (MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, theFullPath, -1, theFullPathUTF16, BUFFER_LENGTH(theFullPathUTF16)) == 0)
		return;

	PlaySound(theFullPathUTF16, NULL, SND_FILENAME | SND_ASYNC | SND_NOWAIT);
}

void ClearSoundCache(void)
{
	//	The Win32 Torus Games do not maintain a sound cache.
}


void SetAlphaTextureFromCharacter(
	unsigned int	aTextureName,
	Char16			aCharacter,
	OpaqueFont		anOpaqueFont,	//	pointer to a Font (pixel size should be a little less than aSize for good appearance)
	unsigned int	aSize,			//	must be a power of two
	bool			aShaderFlag,
	ErrorText		*aFirstError)
{
	ErrorText		theErrorMessage	= NULL;
	HDC				theScreenDC		= NULL,
					theOffscreenDC	= NULL;
	HBITMAP			theBitmap		= NULL;
	TEXTMETRIC		theTextMetrics;
	Byte			*theRGBABytes	= NULL;
	BITMAPINFO		theBitmapInfo	=	{
											{sizeof(BITMAPINFOHEADER), 0, 0, 1, 32, BI_RGB, 0, 0, 0, 0, 0},
											{{0,0,0,0}}
										};
	Byte			*theAlphaBytes	= NULL;
	unsigned int	i;

	//	Assumes desired OpenGL context is already active.

	if ( ! IsPowerOfTwo(aSize) )
	{
		theErrorMessage = "Character texture size must be a power of two.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	theScreenDC = CreateDC(L"DISPLAY", NULL, NULL, NULL);
	//	Don't panic if theScreenDC == NULL.
	//	Even without it we can still get a default monochrome bitmap.

	theOffscreenDC = CreateCompatibleDC(theScreenDC);	//	OK even if theScreenDC == NULL
	if (theOffscreenDC == NULL)
	{
		theErrorMessage = "Couldn't create device context.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Unfortunately SelectObject() will insist on either a monochrome bitmap
	//	or a bitmap with the same color organization as the device
	//	(see Petzold page 666), so we can't use an obvious greyscale bitmap.
	//	Normally we'll use a color bitmap (for good antialiasing), but
	//	if for any reason theScreenDC == NULL we'll settle for monochrome
	//	(no antialiasing) instead.
	if (theScreenDC != NULL)
		theBitmap = CreateCompatibleBitmap(theScreenDC, aSize, aSize);		//	same pixel format as screen
	else
		theBitmap = CreateCompatibleBitmap(theOffscreenDC, aSize, aSize);	//	monochrome
	if (theBitmap == NULL)
	{
		theErrorMessage = "Couldn't create bitmap.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}
	SelectObject(theOffscreenDC, theBitmap);

	SelectObject(theOffscreenDC, anOpaqueFont->itsHFONT);
	GetTextMetrics(theOffscreenDC, &theTextMetrics);
	SetTextColor(theOffscreenDC, 0x00FFFFFF);
	SetBkMode(theOffscreenDC, TRANSPARENT);
	SetTextAlign(theOffscreenDC, TA_CENTER | TA_BOTTOM);
	TextOutW(theOffscreenDC, aSize/2, (aSize + theTextMetrics.tmHeight)/2, &aCharacter, 1);	//	works even on Win 95/98/ME !

	theRGBABytes = (Byte *) GET_MEMORY(aSize*aSize*4);
	if (theRGBABytes == NULL)
	{
		theErrorMessage = "Couldn't allocate theRGBABytes.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	theBitmapInfo.bmiHeader.biWidth  = aSize;
	theBitmapInfo.bmiHeader.biHeight = aSize;
	if (aSize != (unsigned int) GetDIBits(theOffscreenDC, theBitmap, 0, aSize, theRGBABytes, &theBitmapInfo, DIB_RGB_COLORS))
	{
		theErrorMessage = "Couldn't read bitmap bits.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	theAlphaBytes = (Byte *) GET_MEMORY(aSize*aSize*1);
	if (theAlphaBytes == NULL)
	{
		theErrorMessage = "Couldn't allocate theAlphaBytes.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Use the first component (red? blue?) of each pixel in theBitmap
	//	as the alpha value for the texture.
	for (i = 0; i < aSize*aSize; i++)
		theAlphaBytes[i] = theRGBABytes[4*i + 0];

	SetTextureImageFancy(aTextureName, aSize, aSize, 1, theAlphaBytes, aShaderFlag);

CleanUpSetAlphaTextureFromCharacter:

	FREE_MEMORY_SAFELY(theAlphaBytes);

	FREE_MEMORY_SAFELY(theRGBABytes);

	if (theScreenDC != NULL)
		DeleteDC(theScreenDC);

	if (theOffscreenDC != NULL)		//	Delete theOffscreenDC before deleting objects selected into it.
		DeleteDC(theOffscreenDC);

	if (theBitmap != NULL)
		DeleteObject(theBitmap);

	if (theErrorMessage != NULL)
	{
		//	Substitute a transparent texture for the missing one.
		SetTextureImageFancy(aTextureName, 1, 1, 1, (Byte [1]){0x00}, aShaderFlag);
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}


void StartNewThread(
	ModelData	*md,
	void		(*aStartFuction)(ModelData *))
{
	//	Put a wrapper around aStartFuction to isolate details
	//	like _cdecl vs. _stdcall from the rest of the program.
	
	ThreadStartData	theThreadStartData;
	
	theThreadStartData.md							= md;
	theThreadStartData.itsStartFuction				= aStartFuction;
	theThreadStartData.itsArgumentsHaveBeenCopied	= false;

	_beginthread(StartFunctionWrapper, 0, (void *)&theThreadStartData);
	
	while ( ! theThreadStartData.itsArgumentsHaveBeenCopied )
		SleepBriefly();
}


static void __cdecl StartFunctionWrapper(void *aParam)
{
	ThreadStartData	theThreadStartData;
	
	//	Copy the input parameters to our own local memory.
	theThreadStartData = *(ThreadStartData *)aParam;
	
	//	Let the caller know we've made a copy,
	//	so it can return safely.
	((ThreadStartData *)aParam)->itsArgumentsHaveBeenCopied = true;
	
	//	Call itsStartFunction.
	//	Whatever calling convention the compiler prefers is fine by us.
	(*theThreadStartData.itsStartFuction)(theThreadStartData.md);
}


void SleepBriefly(void)
{
	//	Normally the platform-independent code doesn't sleep;
	//	it returns control to the platform-dependent code
	//	which manages idle time as it sees fit.
	//	However, when Chess waits for its secondary thread
	//	to terminate, it needs to do something to avoid hogging
	//	CPU cycles while it waits.
	
	Sleep(10);	//	10 milliseconds
}


void ErrorMessage(
	ErrorText	aMessage,	//	UTF-8
	ErrorText	aTitle,		//	UTF8
	bool		aFatalFlag)	//	Exit the program after user dismisses the error alert?
{
	//	Translate the incoming messages from UTF-8 to UTF-16
	//	and display them with MessageBoxW().

	bool			theMessageWasGiven	= false;
	unsigned int	theMessageLength	= 0,	//	number of unicode characters including terminating zero
					theTitleLength		= 0;
	WCHAR			*theMessageBuffer	= NULL,	//	buffer for UTF-16 translation including terminating zero
					*theTitleBuffer		= NULL;
	DWORD			theTextDirection;

	theMessageLength	= MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aMessage, -1, NULL, 0);
	if (theMessageLength == 0)		//	MultiByteToWideChar() failed, possibly because of bad input
		goto CleanUpErrorMessage;
	theMessageBuffer	= GET_MEMORY(theMessageLength * sizeof(WCHAR));
	if (theMessageBuffer == NULL)	//	deep trouble
		goto CleanUpErrorMessage;
	MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aMessage, -1, theMessageBuffer, theMessageLength);

	theTitleLength		= MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aTitle, -1, NULL, 0);
	if (theTitleLength == 0)		//	MultiByteToWideChar() failed, possibly because of bad input
		goto CleanUpErrorMessage;
	theTitleBuffer		= GET_MEMORY(theTitleLength * sizeof(WCHAR));
	if (theTitleBuffer == NULL)		//	deep trouble
		goto CleanUpErrorMessage;
	MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, aTitle, -1, theTitleBuffer, theTitleLength);

	theTextDirection = LanguageReadsRightToLeft() ? (MB_RTLREADING | MB_RIGHT) : 0;

	MessageBoxW(NULL, theMessageBuffer, theTitleBuffer, MB_OK | MB_TASKMODAL | theTextDirection);

	theMessageWasGiven = true;
	
	if (aFatalFlag)
		ExitProcess(1);

CleanUpErrorMessage:

	FREE_MEMORY_SAFELY(theMessageBuffer);
	FREE_MEMORY_SAFELY(theTitleBuffer  );

	//	If something went wrong...
	if ( ! theMessageWasGiven )
	{
		//	report the present failure as well as...
		MessageBoxA(NULL,
					"Original error message to follow...",
					"Error in ErrorMessage()",
					MB_OK | MB_TASKMODAL | MB_ICONERROR);

		//	the original error message.  The UTF-8 will be interpreted
		//	relative to the local code page, but at least any 7-bit characters
		//	should display OK.
		MessageBoxA(NULL,
					aMessage,
					aTitle,
					MB_OK | MB_TASKMODAL | MB_ICONERROR);
	
		if (aFatalFlag)
			ExitProcess(2);
	}
}


bool LanguageReadsRightToLeft(void)
{
	if (SAME_LANGUAGE_CODE(gLanguageCode, "ar")		//	Arabic
	 || SAME_LANGUAGE_CODE(gLanguageCode, "fa"))	//	Persian
		return true;
	else
		return false;
}
