// 	TorusGamesMainViewController.m
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#import <QuartzCore/QuartzCore.h>	//	for Core Animation
//#import <QuartzCore/CADisplayLink.h>
#import "TorusGamesMainViewController.h"
#import "TorusGamesModelData.h"
#import "TorusGamesEAGLView.h"
#import "TorusGamesKeyboardView.h"
#import "TorusGamesPodOSUtilities.h"


//	Run animations at the display's full refresh rate, typically 60 fps.
//	Most of the time the Torus Games aren't animating anything at all,
//	so for those few seconds here and there when the user is moving something,
//	we can afford to animate it as smoothly as the graphics hardware allows,
//	without fear of undue drain on the battery.
#define FRAME_INTERVAL	1

//	Parameters for spinning the board while the game resets.
#define RESET_ANIMATION_DURATION	1.0		//	seconds
#define RF1							0.25	//	reset shrinkage factor at 90° and 270°
#define RF2							0.0625	//	reset shrinkage factor at 180°

//	For interpreting flicks.
#define FLICK_RESOLUTION_MAX_TIME			0.05	//	seconds
#define	FLICK_RESOLUTION_MIN_DISTANCE		0.05	//	intrinsic units in [-0.5, +0.5] square

//	For hiding the toolbar when a Crossword puzzle is active.
#define TOOLBAR_CENTER_VISIBLE	((CGPoint){ 320.0/2.0, 480.0 - 44.0/2.0})
#define TOOLBAR_CENTER_HIDDEN	((CGPoint){-320.0/2.0, 480.0 - 44.0/2.0})


static Byte gBackgroundColor[3] = {0x80, 0xA0, 0xC0};


static size_t	GetBevelPixels(void *info, void *buffer, off_t position, size_t count);


@implementation TorusGamesMainViewController


- (id)initWithCoder:(NSCoder *)decoder
{
	if ((self = [super initWithCoder:decoder]) != nil)
	{
		//	Just to be safe...
		itsAnimationTimer = nil;

		//	No touch is being resolved.
		itsTouchIsResolved			= true;	//	really itsTouchIsResolved is undefined
		itsUnresolvedTouchPoint		= CGPointZero;
		itsUnresolvedTouchMotion	= CGPointZero;
		itsUnresolvedTouchDuration	= 0.0;
		itsUnresolvedTouchTimer		= nil;

		//	What time are we starting at?
		itsLastRenderTime = [[NSDate date] timeIntervalSince1970];
		
		//	No keyboard yet.
		itsKeyboard = nil;
		
		//	Let the Game Choice tab be the default.
		itsDefaultTab = TabPageGameChoice;
	}

	return self;
}

- (void)awakeFromNib
{
	//	Set the background color here, rather than in Interface Builder,
	//	so that the background and the bevel and use the same gBackgroundColor.
	[itsEnclosingView setBackgroundColor:[UIColor
		colorWithRed:	gBackgroundColor[0]/255.0
		green:			gBackgroundColor[1]/255.0
		blue:			gBackgroundColor[2]/255.0
		alpha:			1.0]];
	
	//	Create an image for the shaded bevel,
	//	basing the colors on gBackgroundColor.
	[self makeBevel];

	[super awakeFromNib];
}


- (void)applicationDidFinishLaunching:(UIApplication *)application
{
	NSArray		*thePreferredLanguages;
	NSString	*thePreferredLanguage;

	//	Initialize the language code and dictionary to the user's default language.
	//
	//	Note #1.  On Mac OS this code works great when called from 
	//	applicationDidFinishLaunching:, but not when called from awakeFromNib
	//	(in which case translations of top-level menu items would fail to appear,
	//	even though translations of submenu items appeared just fine).
	//
	//	Note #2.  preferredLocalizations determines language availability
	//	based on what localizations it finds in our program's application bundle.
	//	Thus, as a practical matter, localizing InfoPList.strings to a new language
	//	will bring that language to the system's attention.
	//	
	//	Note #3.  On Mac OS, one would expect preferredLocalizations to return 
	//	an NSArray containing all localizations supported by our mainBundle and 
	//	acceptable to the user, listed in order of the user's preferences.  
	//	In practice, however, preferredLocalizations always returns a 1-element array,
	//	even when several acceptable languages are available.  Apple's documentation 
	//	is evasive on this question, saying only that preferredLocalizations
	//
	//		Returns one or more localizations contained in the receiver’s bundle 
	//		that the receiver uses to locate resources based on the user’s preferences.
	//
	//	Fortunately preferredLocalizations reliably returns the user's most preferred
	//	localization, from among those supported.
	thePreferredLanguages	= [[NSBundle mainBundle] preferredLocalizations];
	thePreferredLanguage	= ([thePreferredLanguages count] > 0 ?			//	If a preferred language exists...
								[thePreferredLanguages objectAtIndex:0] :	//	...use it.
								@"en");										//	Otherwise fall back to English.
	RefreshLocalization([thePreferredLanguage UTF8String]);
	
	//	If itsGLView failed to load a context,
	//	tell the user what happened and quit now.
	//	This is unlikely to occur as long as we're supporting
	//	OpenGL ES 1.1 as well as OpenGL ES 2.0,
	//	but if I someday remove GLES 1.1 support and
	//	the user runs the app on a (by then old) GLES 1.1 device,
	//	we'll want it to fail gracefully.
	//	When that day comes, I should provide a localized message
	//	explaining what happened, and perhaps directing the user
	//	to a (by then old) GLES 1.1 version of the app.
	if ([itsGLView context] == nil)
	{
		ErrorMessage("Couldn't load OpenGL ES context.  App cannot continue.", "Graphics Unavailable", true);
		return;
	}

	[self changeGamePodOS:GameNone];
	[self refreshCrosswordTextEntry];
	[self startAnimation];
	
	SetUpMIDIPlayer();	//	In the iPhone Torus Games, this
						//	sets up an M4A player, not a MIDI player.

	//	Show the settings panel so the user may choose a game.
	[self showTabBarController];
}


- (void)applicationDidReceiveMemoryWarning:(UIApplication *)application
{
	//	At any given moment, Torus Games holds only those resources
	//	(textures, sounds, etc.) required by the active game,
	//	so there's really not much we can do to free memory.
	
	//	As a token gesture we can free the active game's cached sounds,
	//	but I doubt that will fully resolve a memory shortage.
	ClearSoundCache();
}


- (void)applicationWillTerminate:(UIApplication *)application
{
	//	Stop the animation timer.
	[self stopAnimation];
	
	//	Release the settings panel, if it's open.
	//	The UIViewController documentation guarantees that
	//	dismissModalViewControllerAnimated will dismiss all
	//	view controllers above it on the stack.
	if ([self modalViewController] != nil)
		[self dismissModalViewControllerAnimated:NO];

	//	Free the ModelData.
	[itsModelData clearModelDataWithContext:[itsGLView context]];

	//	Release the localization dictionary.
	RefreshLocalization("--");
	
	//	We should be clean.
	if (gMemCount != 0)
		NSLog(@"Memory Leak:  gMemCount == %d.  Please report this to Jeff Weeks at www.geometrygames.org/contact.html", gMemCount);
	else
		NSLog(@"Torus Games memory usage is clean:  memory allocated equals memory freed.  :-)");
}


- (void)dealloc
{
	//	The rules regarding memory management of NIB objects are complicated,
	//	and differ between iPhone OS and Mac OS (!).  For details, see
	//
	//		http://developer.apple.com/documentation/Cocoa/Conceptual/MemoryMgmt/Articles/mmNibObjects.html
	//	
	//	Let's hope the following is correct.
	//
	//	Hmmm... the point is moot, because this method never gets called!  
	//	This is a known oddity, well documented on the web.
	//	It's no doubt related to the fact that in main.m, UIApplicationMain() never returns.
	//	Perhaps some overly clever programmer at Apple decided that at a certain point
	//	the program should simply exit(), so that it shuts down faster without
	//	"wasting time" on cleanup?
	[super dealloc];
}


- (void)makeBevel
{
	CGColorSpaceRef		theColorSpace	= NULL;
	CGDataProviderRef	theDataProvider	= NULL;
	CGImageRef			theCGImage		= NULL;
	UIImage				*theUIImage		= nil;

	static CGDataProviderDirectCallbacks	theCallbackFunctions =
											{
												0,
												NULL,
												NULL,
												GetBevelPixels,
												NULL
											};
	
	theColorSpace = CGColorSpaceCreateDeviceRGB();
	if (theColorSpace == NULL)
		goto CleanUpMakeBevel;
	
	theDataProvider = CGDataProviderCreateDirect(NULL, 300 * 300 * 4, &theCallbackFunctions);
	if (theDataProvider == NULL)
		goto CleanUpMakeBevel;

	theCGImage = CGImageCreate(
		300,
		300,
		8,
		32,
		300*4,
		theColorSpace,
		kCGImageAlphaNoneSkipLast,
		theDataProvider,
		NULL,
		false,
		kCGRenderingIntentDefault);
	if (theCGImage == NULL)
		goto CleanUpMakeBevel;
	
	theUIImage = [[UIImage alloc] initWithCGImage:theCGImage];
	if (theUIImage == nil)
		goto CleanUpMakeBevel;
	
	[itsBevel setImage:theUIImage];

CleanUpMakeBevel:

	if (theUIImage != nil)
	{
		[theUIImage release];
		theUIImage = nil;
	}

	if (theCGImage != NULL)
	{
		CGImageRelease(theCGImage);
		theCGImage = NULL;
	}
	
	if (theDataProvider != NULL)
	{
		CGDataProviderRelease(theDataProvider);
		theDataProvider = NULL;
	}
	
	if (theColorSpace != NULL)
	{
		CGColorSpaceRelease(theColorSpace);
		theColorSpace = NULL;
	}
}

static size_t GetBevelPixels(
	void	*info,
	void	*buffer,
	off_t	position,
	size_t	count)
{
	unsigned int	theNumBytesRemaining,
					thePixel,
					theChannel,
					theRow,
					theCol,
					t,	//	integer weighting factor, as parts in 32
					theBlend;

	theNumBytesRemaining = count;

	thePixel	= position / 4;
	theChannel	= position % 4;
	
	theRow		= thePixel / 300;
	theCol		= thePixel % 300;
	
	while (true)
	{
		if (theRow >= 300)
			return 0;	//	should never occur

		if (theRow < 6 && (theCol >= theRow && theCol <= (300 - 1) - theRow))
		{
			//	topmost 6 rows, including pixels on the diagonal
			
			if (theCol == theRow || theCol == (300 - 1) - theRow)
			{	//	diagonal
				t			= 0;
				theBlend	= 0x00;
			}
			else
			{	//	non-diagonal
				switch (theRow)	
				{
					case 0:		t = 2;	break;
					case 1:		t = 4;	break;
					default:	t = 8;	break;
				}
				theBlend = 0x00;
			}
		}
		else
		if (theRow >= 300 - 6 && (theCol >= (300 - 1) - theRow && theCol <= theRow))
		{
			//	bottommost 6 rows, including pixels on the diagonal

			if (theCol == (300 - 1) - theRow || theCol == theRow)
			{	//	diagonal
				switch (theRow)
				{
					case (300 - 1):	t =  3;	break;
					case (300 - 2):	t =  6;	break;
					default:		t = 12;	break;
				}
			}
			else
			{	//	non-diagonal
				switch (theRow)
				{
					case (300 - 1):	t =  4;	break;
					case (300 - 2):	t =  8;	break;
					default:		t = 16;	break;
				}
			}
			theBlend = 0xFF;
		}
		else
		if (theCol < 6)
		{
			//	leftmost 6 columns, excluding pixels on the diagonal
			switch (theCol)
			{
				case 0:		t = 2;	break;
				case 1:		t = 4;	break;
				default:	t = 8;	break;
			}
			theBlend = 0xFF;
		}
		else
		if (theCol >= 300 - 6)
		{
			//	rightmost 6 columns, excluding pixels on the diagonal
			switch (theCol)
			{
				case (300 - 1):	t = 2;	break;
				case (300 - 2):	t = 4;	break;
				default:		t = 8;	break;
			}
			theBlend = 0xFF;
		}
		else
		{
			//	generic interior point
			t			= 0;
			theBlend	= 0xFF;
		}

		//	On the first pass through the main loop
		//	we could be starting in any row, column or channel.
		//	On subsequent passes we'll be at channel 0.
		//	Either way, set values for whatever channels remain
		//	in the current pixel...
		while (theChannel < 3)		//	R, G, B channels
		{
			*((Byte *)buffer++) = ((32 - t) * gBackgroundColor[theChannel++]
								  +    t    * theBlend)
								>> 5;	//	same as /32, but maybe quicker
			if (--theNumBytesRemaining == 0)	//	CGDataProviderCreateDirect() is unlikely to ask us
				break;							//		for fractional pixels, but in theory it could.
		}
		if (theNumBytesRemaining == 0)
			break;
		*((Byte *)buffer++) = 0xFF;	//	alpha channel -- current implementation ignores the value
		if (--theNumBytesRemaining == 0)
			break;
		
		//	...and move on to the next pixel.
		theChannel = 0;
		theCol++;
		if (theCol == 300)
		{
			theCol = 0;
			theRow++;
		}
	}
	
	//	The above loop terminates when it either encounters a bad
	//	row number (theRow >= 300) in which case it returns 0 immediately,
	//	or when it finishes writing the requested bytes
	//	(theNumBytesRemaining == 0) in which case we arrive here and report success.
	return count;	//	total number of bytes written
}


- (void)startAnimation
{
	if (itsAnimationTimer == nil)
	{

		itsAnimationTimer = [NSClassFromString(@"CADisplayLink")
								displayLinkWithTarget:	self
								selector:				@selector(pleaseRedraw:)];
		[itsAnimationTimer setFrameInterval:FRAME_INTERVAL];
		[itsAnimationTimer addToRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
	}
}

- (void)stopAnimation
{
	if (itsAnimationTimer != nil)
	{
		[itsAnimationTimer invalidate];
		itsAnimationTimer = nil;
	}
}

- (void)pleaseRedraw:(CADisplayLink *)aDisplayLink
{
	double	theCurrentTime,
			theElapsedTime;

	//	Compute the time since the last redraw.
	//
	//	Note:  timeIntervalSince1970 reports a usable value on the iPhone/iPod Touch itself,
	//	even though on the simulator it reports only an integer number of seconds.
	theCurrentTime = [[NSDate date] timeIntervalSince1970];
	theElapsedTime = theCurrentTime - itsLastRenderTime;
	itsLastRenderTime = theCurrentTime;
	
	//	If the device is managing to keep up with the requested frame rate,
	//	then duration*frameInterval gives the elapsed time between frames.
	//	If the device can't keep up with the frame rate, then the animation
	//	will visibly slow down.
	//
	//		Advantage:  If we skip a frame, we get back on track immediately.
	//
	//		Disadvantage:  If the user runs the app on slower software
	//		than I'm testing it on, the animation could run visibly slowly.
	//
	//	Maybe better to stick with the more pedestrian approach given above,
	//	just to be safe?
//	if (itsAnimationTimer != nil)
//		theElapsedTime = [itsAnimationTimer duration] * [itsAnimationTimer frameInterval];
//	else
//		theElapsedTime = 0.0;

	//	Is an animation active?
	if (SimulationWantsUpdates([itsModelData modelData]))
	{
		//	Update the simulation.
		SimulationUpdate([itsModelData modelData], theElapsedTime);

		//	Redraw.
		[itsGLView drawViewUsing:itsModelData];

//NSLog(@"frame rate = %lf", 1.0 / theElapsedTime);
	}
}


- (void)touchBeganAt:(CGPoint)aPoint	//	coordinates in range [-0.5, +0.5]
{
#ifdef INTERPRET_FLICKS
	itsTouchIsResolved			= false;
	itsUnresolvedTouchPoint		= aPoint;
	itsUnresolvedTouchMotion	= CGPointZero;
	itsUnresolvedTouchDuration	= 0.0;
	itsUnresolvedTouchTimer		= [NSTimer	scheduledTimerWithTimeInterval:FLICK_RESOLUTION_MAX_TIME
											target:self
											selector:@selector(touchResolvedByTimeout:)
											userInfo:nil
											repeats:NO];
#else
	//	Pass the touch-began into the platform-independent code.
	TouchBegan([itsModelData modelData], aPoint.x, aPoint.y, false);
#endif
}

- (void)touchMovedAcross:(CGPoint)aMotion	//	coordinate changes in range [-0.5, +0.5]
				  during:(double)aDuration	//	elapsed time in seconds
{
#ifdef INTERPRET_FLICKS
	double	theDistanceSquared;

	if ( ! itsTouchIsResolved )
	{
		itsUnresolvedTouchMotion.x += aMotion.x;
		itsUnresolvedTouchMotion.y += aMotion.y;
		itsUnresolvedTouchDuration += aDuration;
		
		theDistanceSquared	= itsUnresolvedTouchMotion.x * itsUnresolvedTouchMotion.x
							+ itsUnresolvedTouchMotion.y * itsUnresolvedTouchMotion.y;
		
		if (theDistanceSquared > FLICK_RESOLUTION_MIN_DISTANCE*FLICK_RESOLUTION_MIN_DISTANCE)
		{
			[self touchResolvesToFlick:true];
		}
	}
	else
	{
		//	The touch has already been resolved.
		//	Pass the touch-motion into the platform-independent code.
		TouchMoved([itsModelData modelData], aMotion.x, aMotion.y, aDuration);
	}
#else
	//	Pass the touch-motion into the platform-independent code.
	TouchMoved([itsModelData modelData], aMotion.x, aMotion.y, aDuration);
#endif
}

- (void)touchEnded
{
#ifdef INTERPRET_FLICKS
	if ( ! itsTouchIsResolved )				//	If the touch remains unresolved, then
		[self touchResolvesToFlick:false];	//	...it was just a brief tap, not a flick.
#endif

	//	Pass the touch-end into the platform-independent code.
	TouchEnded([itsModelData modelData]);
}

- (void)touchCancelled
{
#ifdef INTERPRET_FLICKS
	if ( ! itsTouchIsResolved )
	{
		//	Forget the unresolved touch.
		itsTouchIsResolved = true;
		[itsUnresolvedTouchTimer invalidate];
		itsUnresolvedTouchTimer = nil;
	}
#endif

	//	Pass the touch-cancelled into the platform-independent code.
	TouchCancelled([itsModelData modelData]);
}

#ifdef INTERPRET_FLICKS

- (void)touchResolvedByTimeout:(NSTimer*)theTimer
{
	//	FLICK_RESOLUTION_MAX_TIME has elapsed
	//	and the touch has failed to move FLICK_RESOLUTION_MIN_DISTANCE,
	//	so consider the touch to be a regular touch, not a flick.
	[self touchResolvesToFlick:false];
}

- (void)touchResolvesToFlick:(bool)aFlickFlag
{
	TouchBegan([itsModelData modelData], itsUnresolvedTouchPoint.x, itsUnresolvedTouchPoint.y, aFlickFlag);
	TouchMoved([itsModelData modelData], itsUnresolvedTouchMotion.x, itsUnresolvedTouchMotion.y, itsUnresolvedTouchDuration);

	itsTouchIsResolved = true;
	[itsUnresolvedTouchTimer invalidate];
	itsUnresolvedTouchTimer = nil;
}

#endif


- (void)setMessageText:(NSString *)aMessage
{
	//	Adjust itsMessageField's font size to accommodate aMessage.
	[self setMessageFontSizeForMessage:aMessage];

	//	Write aMessage into itsMessageField.
	[itsMessageField setText:aMessage];
}


- (void)setMessageFontSizeForMessage:(NSString *)aMessage
{
	bool	theFontSizeHasChanged;
	UIFont	*theFont;
	float	theFontSize;
	CGSize	theActualSize,
			thePotentialSize;

	//	Alas setAdjustsFontSizeToFitWidth works 
	//	for single-line text only, so we must manually adjust
	//	the font size to fit aMessage into itsMessageField.
	
	//	If no text is present, do nothing.
	if (aMessage == nil)
		return;

	//	In almost all cases our default 21-point font will work fine
	//	and no changes will be required.  Take care not to change
	//	the font except in those rare cases where it's really necessary.
	theFontSizeHasChanged = false;

	//	If necessary, restore our default 21-point font.
	theFont = [itsMessageField font];
	if ([theFont pointSize] != 21.0)
	{
		theFont = [theFont fontWithSize:21.0];
		theFontSizeHasChanged = true;
	}
	theFontSize = 21.0;
	
	//	thePotentialSize is the same as theActualSize,
	//	but with an effectively unlimited height.
	theActualSize		= [itsMessageField frame].size;
	thePotentialSize	= (CGSize) {theActualSize.width, 1024.0};
	
	//	As long as the required size exceeds theActualSize,
	//	keep shrinking the font's point size.
	while ([aMessage sizeWithFont:theFont constrainedToSize:thePotentialSize].height
				> theActualSize.height
	    && theFontSize > 8.0 )
	{
		theFontSize				-= 1.0;
		theFont					= [theFont fontWithSize:theFontSize];
		theFontSizeHasChanged	= true;
	}

	//	If theFontSizeHasChanged, install the new font.
	if (theFontSizeHasChanged)
		[itsMessageField setFont:theFont];
}


- (IBAction)resetGameWithAnimation
{
	CAKeyframeAnimation	*theAnimation;

	//	Stop the usual animation timer, so that the user won't see
	//	the newly reset game until both
	//
	//		(1)	the reset animation has reached its halfway point
	//	and
	//		(2)	the new game is ready.
	[self stopAnimation];
	
	//	Start the Reset Game animation running in a separate thread.
	theAnimation = [CAKeyframeAnimation animationWithKeyPath:@"transform"];
	[theAnimation setDuration:RESET_ANIMATION_DURATION];
	[theAnimation setValues:[NSArray arrayWithObjects:
			[NSValue valueWithCATransform3D:(CATransform3D)
				{	+1.0,  0.0,  0.0,  0.0,
					 0.0, +1.0,  0.0,  0.0,
					 0.0,  0.0, +1.0,  0.0,
					 0.0,  0.0,  0.0, +1.0	}],
			[NSValue valueWithCATransform3D:(CATransform3D)
				{	 0.0, +RF1,  0.0,  0.0,
					-RF1,  0.0,  0.0,  0.0,
					 0.0,  0.0, +1.0,  0.0,
					 0.0,  0.0,  0.0, +1.0	}],
			[NSValue valueWithCATransform3D:(CATransform3D)
				{	-RF2,  0.0,  0.0,  0.0,
					 0.0, -RF2,  0.0,  0.0,
					 0.0,  0.0, +1.0,  0.0,
					 0.0,  0.0,  0.0, +1.0	}],
			[NSValue valueWithCATransform3D:(CATransform3D)
				{	 0.0, -RF1,  0.0,  0.0,
					+RF1,  0.0,  0.0,  0.0,
					 0.0,  0.0, +1.0,  0.0,
					 0.0,  0.0,  0.0, +1.0	}],
			[NSValue valueWithCATransform3D:(CATransform3D)
				{	+1.0,  0.0,  0.0,  0.0,
					 0.0, +1.0,  0.0,  0.0,
					 0.0,  0.0, +1.0,  0.0,
					 0.0,  0.0,  0.0, +1.0	}],
			nil
		]];
	[theAnimation setDelegate:nil];
	[[itsGLViewContainer layer] addAnimation:theAnimation forKey:@"reset game"];
	
	//	Schedule a timer to fire half way through the reset animation.
	//	Note that if the resetGame call (immediately below) takes longer
	//	than this, we won't get the halfwayPoint message until resetGame
	//	has returned, because those two calls take place in the same thread.
	[NSTimer	scheduledTimerWithTimeInterval:(0.5 * RESET_ANIMATION_DURATION)
				target:self
				selector:@selector(halfwayPoint)
				userInfo:nil
				repeats:NO];
	
	//	The animation won't start until we pass control back to the run loop,
	//	so set a timer to reset the game more-or-less immediately after
	//	we return control to the run loop.
	[NSTimer	scheduledTimerWithTimeInterval:(0.000001)
				target:self
				selector:@selector(resetGame)
				userInfo:nil
				repeats:NO];
}

- (void)resetGame
{
	//	While the animation runs and the halfway timer waits,
	//	reset the game and load new textures etc.
	//	The former is slow for the jigsaw puzzle, and
	//	the latter can be slow for games like the Crossword and Word Search,
	//	which need to prepare a lot of data-dependent textures,
	//	which is why we let the game prepare them now,
	//	while the animation runs, rather than waiting until we want 
	//	to draw the new game.
	ResetGame([itsModelData modelData]);
	[EAGLContext setCurrentContext:[itsGLView context]];
	(void) SetUpGraphicsAsNeeded([itsModelData modelData]);
	
	//	OK, the game is now reset.  Either the halfway timer will have already fired
	//	and the halfwayPoint message will be waiting for us, or the halfway timer
	//	won't yet have fired and we'll wait for it.
}

- (void)halfwayPoint
{
	//	The game has been reset and the animation has reached its halfway point
	//	(or maybe further), so resume normal updates.
	[self startAnimation];
	
	//	When the user solves a puzzle, the Crossword shows the toolbar
	//	and leaves it there.  Wait a couple seconds, then hide the toolbar again,
	//	so the user can access the keyboard.
	if ([itsModelData modelData]->itsGame == GameCrossword)
		[self performSelector:@selector(hideToolbarWithAnimation) withObject:nil afterDelay:1.0];
}


- (void)changeGamePodOS:(GameType)aNewGame
{
	[itsModelData changeGame:aNewGame context:[itsGLView context]];

	[self refreshToolbar];
}


- (void)refreshGameLanguageWithReset:(bool)aResetFlag
{
	//	The method tableView:willSelectRowAtIndexPath: should have
	//	already called RefreshLocalization(), which sets gLanguageCode
	//	and rebuilds gDictionary.
	
	//	Refresh toolbar button labels.
	[self refreshToolbar];
		
	//	Refresh any game-specific text.
	switch ([itsModelData modelData]->itsGame)
	{
		case GameCrossword:
			if (aResetFlag)
				ResetGame([itsModelData modelData]);
			[self refreshCrosswordTextEntry];	//	install new keyboard
			break;

		case GameWordSearch:
			if (aResetFlag)
				ResetGame([itsModelData modelData]);
			break;
		
		case GamePool:
			//	No need to disturb the Pool game in progress.
			//	Just refresh the current status message in the new language.
			PoolSetStatusMessage([itsModelData modelData]);
			break;
		
		default:
			//	Do nothing.
			break;
	}
}


- (void)refreshToolbar
{
	char	*theResetKey;

	//	Refresh the Reset button.
	switch ([itsModelData modelData]->itsGame)
	{
		case GameNone:			theResetKey = "S: Reset";					break;
		case GameIntro:			theResetKey = "S: Reset Practice Board";	break;
		case GameTicTacToe:		theResetKey = "S: Erase Tic-Tac-Toe Board";	break;
		case GameMaze:			theResetKey = "S: New Maze";				break;
		case GameCrossword:		theResetKey = "S: New Crossword Puzzle";	break;
		case GameWordSearch:	theResetKey = "S: New Word Search";			break;
		case GameJigsaw:		theResetKey = "S: New Jigsaw Puzzle";		break;
		case GameChess:			theResetKey = "S: Reset Chess Board";		break;
		case GamePool:			theResetKey = "S: Reset Pool Balls";		break;
		case GameApples:		theResetKey = "S: New Apples";				break;
		default:				theResetKey = NULL;							break;
	}
	[itsResetButton setTitle:GetLocalizedTextAsNSString(theResetKey)];
	
	//	Refresh the Settings button.
	[itsSettingsButton setTitle:GetLocalizedTextAsNSString("Settings")];
}


- (void)refreshCrosswordTextEntry
{
	bool	theGameIsCrossword;

	//	Update the window layout, the keyboard and the toolbar.
	
	//	Which layout to use?
	theGameIsCrossword = ([itsModelData modelData]->itsGame == GameCrossword);

	//	Remove the old keyboard, if any.
	[itsKeyboard removeFromSuperview];
	[itsKeyboard release];
	itsKeyboard = nil;
	
	//	Adjust the layout.
	if (theGameIsCrossword)
	{
		//	Leave extra room for the keyboard.
		//	Of the 480 - 320 = 160 available vertical pixels,
		//	put 48 at the top and 112 at the bottom.
		//
		//	Note:  The 320×320 square dedicated to the game board
		//	allows for a 10-pixel margin on all four sides
		//	of a 300×300 container view.  The container view
		//	includes the bevel as well as the game board itself.
		[itsMessageField setFrame:(CGRect){
			{TOP_ZONE_MARGIN, 0.0},
			{SCREEN_WIDTH - 2*TOP_ZONE_MARGIN, TOP_ZONE_CROSSWORD + MIDDLE_ZONE_MARGIN}}];
		[itsGLViewContainer setFrame:(CGRect){
			{MIDDLE_ZONE_MARGIN, TOP_ZONE_CROSSWORD + MIDDLE_ZONE_MARGIN},
			{SCREEN_WIDTH - 2*MIDDLE_ZONE_MARGIN, SCREEN_WIDTH - 2*MIDDLE_ZONE_MARGIN}}];
	}
	else
	{
		//	Center the game board without leaving room for a keyboard.
		[itsMessageField setFrame:(CGRect){
			{TOP_ZONE_MARGIN, 0.0},
			{SCREEN_WIDTH - 2*TOP_ZONE_MARGIN, TOP_ZONE_NORMAL + MIDDLE_ZONE_MARGIN}}];
		[itsGLViewContainer setFrame:(CGRect){
			{MIDDLE_ZONE_MARGIN, TOP_ZONE_NORMAL + MIDDLE_ZONE_MARGIN},
			{SCREEN_WIDTH - 2*MIDDLE_ZONE_MARGIN, SCREEN_WIDTH - 2*MIDDLE_ZONE_MARGIN}}];
	}

	//	Adjust itsMessageField's font size 
	//	to accommodate whatever text it may already contain.
	//
	//	Note:  tabBarControllerDidFinishWithSettings calls changeGamePodOS 
	//	before calling refreshCrosswordTextEntry, so in practice itsMessageField 
	//	will often contain new text that we need to adapt the font size to.
	[self setMessageFontSizeForMessage:[itsMessageField text]];

	//	Create a new keyboard, if the Crossword is active.
	if (theGameIsCrossword)
	{
		itsKeyboard = [[TorusGamesKeyboardView alloc] initWithController:self modelData:itsModelData];
		[[self view] addSubview:itsKeyboard];
		[[self view] sendSubviewToBack:itsKeyboard];
	}
	
	//	Hide itsToolbar for a Crossword puzzle,
	//	so it doesn't cover the keyboard.
	//	Show itsToolbar for all other games.
	//	Delay the command so that the user can see
	//	the toolbar go away, and will thus have a clue
	//	where it went and how s/he might get it back when needed.
	if (theGameIsCrossword)
		[self performSelector:@selector(hideToolbarWithAnimation) withObject:nil afterDelay:2.0];
	else
		[self performSelector:@selector(showToolbarWithAnimation) withObject:nil afterDelay:1.0];
}

- (IBAction)showToolbarBriefly:(id)sender
{
	//	The user has touched the "toolbar stub" 
	//	at the lower left corner of the screen.
	
	//	Show the toolbar, but then...
	[self showToolbarWithAnimation];

	//	...after a couple seconds, hide it again 
	//	so it doesn't cover the bottom row of keys.
	[self performSelector:@selector(hideToolbarWithAnimation) withObject:nil afterDelay:2.0];
}

- (void)showToolbarWithAnimation
{
	[UIView beginAnimations:@"show toolbar" context:nil];
	[itsToolbar setCenter:TOOLBAR_CENTER_VISIBLE];
	[UIView commitAnimations];
}

- (void)hideToolbarWithAnimation
{
	[UIView beginAnimations:@"hide toolbar" context:nil];
	[itsToolbar setCenter:TOOLBAR_CENTER_HIDDEN];
	[UIView commitAnimations];
}


- (IBAction)showTabBarController
{
	ModelData					*md;
	TorusGamesTabBarController	*theTabBarController;
	
	md = [itsModelData modelData];

	itsPreviousSettings.itsGame				= md->itsGame;
	itsPreviousSettings.itsTopology			= md->itsTopology;
	itsPreviousSettings.itsHumanVsComputer	= md->itsHumanVsComputer;
	itsPreviousSettings.itsLanguageCode[0]	= gLanguageCode[0];
	itsPreviousSettings.itsLanguageCode[1]	= gLanguageCode[1];
	itsPreviousSettings.itsLanguageCode[2]	= 0;

	[self stopAnimation];
	
	theTabBarController = [[TorusGamesTabBarController alloc] initWithParent:self settings:itsPreviousSettings tab:itsDefaultTab];
	[theTabBarController setModalTransitionStyle:UIModalTransitionStyleFlipHorizontal];
	[self presentModalViewController:theTabBarController animated:YES];
	[theTabBarController release];
}

- (void)tabBarControllerDidFinishWithSettings:(TorusGamesSettings)someNewSettings tab:(TabPageTag)aLastTab
{
	ModelData	*md;
	
	//	Remember the last tab viewed, so it can be the default for next time.
	itsDefaultTab = aLastTab;
	
	//	Fetch a pointer to the ModelData.
	md = [itsModelData modelData];

	//	Install someNewSettings with minimal disruption to the game in progress.
	if (someNewSettings.itsGame != itsPreviousSettings.itsGame)
	{
		//	We need to reset the game.
		//	But if the user has, say, also changed the topology,
		//	then we want to avoid resetting the game twice.
		//	To be totally safe, switch to the null game, set the new 
		//	play mode, topology and language (very fast for the null game!),
		//	and then switch to the new game.
		[self changeGamePodOS:GameNone];
		[self setMessageText:@""];
		ChangeTopology(md, someNewSettings.itsTopology);
		ChangeHumanVsComputer(md, someNewSettings.itsHumanVsComputer);
		[self refreshGameLanguageWithReset:NO];
		[self changeGamePodOS:someNewSettings.itsGame];
		[self refreshCrosswordTextEntry];
	}
	else
	{
		//	ChangeHumanVsComputer() is always fast.  It never resets the game.
		if (someNewSettings.itsHumanVsComputer != itsPreviousSettings.itsHumanVsComputer)
			ChangeHumanVsComputer(md, someNewSettings.itsHumanVsComputer);
		
		//	Normally refreshing the game language will reset the game
		//	iff the game requires it.
		//	But in the case that ChangeTopology() is about to reset
		//	the game anyhow, pass refreshGameLanguageWithReset:NO 
		//	to ask that refreshGameLanguage: suppress its own reset.
		if ( ! SAME_LANGUAGE_CODE(someNewSettings.itsLanguageCode, itsPreviousSettings.itsLanguageCode) )
			[self refreshGameLanguageWithReset:
				(someNewSettings.itsTopology == itsPreviousSettings.itsTopology)];
		
		//	ChangeTopology() always resets the game.
		if (someNewSettings.itsTopology != itsPreviousSettings.itsTopology)
			ChangeTopology(md, someNewSettings.itsTopology);
	}

	[self pleaseRedraw:nil];

	[self dismissModalViewControllerAnimated:YES];

	[self startAnimation];
}

@end
