//	TorusGamesPodOSCallbacks.m
//
//	A few platform-dependent functions that the platform-independent code
//	will call to access resources.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#import <UIKit/UIKit.h>
#import "TorusGames-Common.h"
#import "TorusGamesPodOSCallbacks.h"
#import "TorusGamesPodOSUtilities.h"
#import "TorusGamesMainViewController.h"


static NSString		*GetGameString(GameType aGame);
static NSString		*GetTopologyString(TopologyType aTopology);
static unsigned int	LengthOfUnicodeString(Char16 *aString);


unsigned int GetNumPuzzles(
	GameType		aGame,
	TopologyType	aTopology)
{
	NSString		*theNumPuzzlesAsString	= nil;
	unsigned int	theNumPuzzlesAsInt		= 1;

	theNumPuzzlesAsString = GetLocalizedTextAsNSString(
								[[[[(aGame == GameWordSearch ? @"NumMobile" : @"Num")
									stringByAppendingString:GetTopologyString(aTopology)]
									stringByAppendingString:GetGameString(aGame)]
									stringByAppendingString:@"Puzzles"]
									UTF8String]);

	theNumPuzzlesAsInt = [theNumPuzzlesAsString intValue];

	return theNumPuzzlesAsInt;
}


void GetRawPuzzleData(
	GameType		aGame,
	TopologyType	aTopology,
	unsigned int	aPuzzleIndex,
	Char16			*aBuffer,
	unsigned int	aBufferLength)	//	number of Char16's (not bytes!) available in aBuffer, including terminating zero
{
	ErrorText		theError			= NULL;
	NSString		*theIndexString		= nil,
					*theRawPuzzleData	= nil;
	unsigned int	theNumDataBytes,
					i;

	//	Fill the output buffer with zeros for clean output upon success
	//	or robust error handling upon failure.
	for (i = 0; i < aBufferLength; i++)
		aBuffer[i] = 0;
	
	theIndexString = [NSString stringWithFormat:@"%d",aPuzzleIndex];

	theRawPuzzleData = GetLocalizedTextAsNSString(
						[[[	GetTopologyString(aTopology)
							stringByAppendingString: GetGameString(aGame)]
							stringByAppendingString: theIndexString]
							UTF8String]);
	
	theNumDataBytes = [theRawPuzzleData length];	//	length with no terminating zero
	
	if (theNumDataBytes == 0)
	{
		theError = "Couldn't read theRawPuzzleData.";
		goto CleanUpGetRawPuzzleData;
	}

	if (aBufferLength >= theNumDataBytes + 1)		//	allow for terminating zero
	{
		[theRawPuzzleData getCharacters:aBuffer];	//	the string itself
		aBuffer[theNumDataBytes] = 0;				//	the terminating zero (redundant but safe)
	}
	else
	{
		theError = "Output buffer too small for theRawPuzzleData.";
		goto CleanUpGetRawPuzzleData;
	}

CleanUpGetRawPuzzleData:

	if (theError != NULL)
		ErrorMessage(theError, "Internal Error in GetRawPuzzleData()", false);
}


static NSString *GetGameString(GameType aGame)
{
	switch (aGame)
	{
		case GameCrossword:		return @"Crossword";
		case GameWordSearch:	return @"WordSearch";
		default:
			ErrorMessage("Invalid GameType", "Internal Error in GetGameString()", false);
			return @"";
	}
}


static NSString *GetTopologyString(TopologyType aTopology)
{
	switch (aTopology)
	{
		case TopologyTorus:		return @"Torus";
		case TopologyKlein:		return @"Klein";
		default:
			ErrorMessage("Invalid TopologyType", "Internal Error in GetTopologyString()", false);
			return @"";
	}
}


void CrosswordSetClues(
	ModelData	*md,
	Char16		*aClueAcross,
	Char16		*aClueDown,
	Char16		aHotDirection)
{
	Char16		*theClue;
	NSString	*theClueString;

	theClue			= (aHotDirection == L'↓' ? aClueDown : aClueAcross);
	theClueString	= [NSString stringWithCharacters:theClue length:LengthOfUnicodeString(theClue)];
	[((TorusGamesMainViewController *)[[UIApplication sharedApplication] delegate]) setMessageText:theClueString];
}


void WordSearchSetList(ModelData *md)
{
	NSMutableString	*theWordList	= nil;
	bool			theFirstWord	= true;
	NSString		*theWord		= nil;
	unsigned int	i;

	theWordList = [NSMutableString stringWithCapacity:1024];

	for (i = 0; i < MAX_WORDSEARCH_NUM_WORDS; i++)
	{
		if (md->itsGameOf.WordSearch.itsWords[i][0] != 0)
		{
			if ( ! theFirstWord )
				[theWordList appendString:@", "];
			else
				theFirstWord = false;

			theWord = [NSString
						stringWithCharacters:	md->itsGameOf.WordSearch.itsWords[i]
						length:					LengthOfUnicodeString(md->itsGameOf.WordSearch.itsWords[i])];
			[theWordList appendString:theWord];
		}
	}

	[((TorusGamesMainViewController *)[[UIApplication sharedApplication] delegate]) setMessageText:theWordList];
}


void PoolSetStatusMessage(ModelData *md)
{
	NSString	*theSubject,
				*thePredicate,
				*theKey,
				*theMessage;

	//	Determine the active subject.
	if (md->itsHumanVsComputer)
	{
		if (md->itsGameOf.Pool.itsWhoseTurn == md->itsGameOf.Pool.itsComputerPlayer)
			theSubject = @"Computer";
		else
			theSubject = @"Human";
	}
	else
	{
		if (md->itsGameOf.Pool.itsWhoseTurn == PoolPlayerA)
			theSubject = @"PlayerA";
		else
			theSubject = @"PlayerB";
	}
	
	//	Determine the appropriate predicate.
	if (md->itsGameOf.Pool.itsWinFlag)
		thePredicate = @"HasWon";
	else
	if (md->itsGameOf.Pool.itsLoseFlag)
		thePredicate = @"HasLost";
	else
		switch (md->itsGameOf.Pool.itsWhoseTurn == PoolPlayerA ?
					md->itsGameOf.Pool.itsPlayerAGoal :
					md->itsGameOf.Pool.itsPlayerBGoal)
		{
			case PoolUndecided:	thePredicate = @"Shoots";				break;
			case PoolSolids:	thePredicate = @"ShootsForSolids";		break;
			case PoolStripes:	thePredicate = @"ShootsForStripes";		break;
			case PoolBlack:		thePredicate = @"ShootsForFourBall";	break;
			default:			thePredicate = @"";						break;	//	unused
		}

	//	Assemble theKey.
	theKey = [theSubject stringByAppendingString:thePredicate];
	
	//	Fetch the message.
	theMessage = GetLocalizedTextAsNSString([theKey UTF8String]);

	//	Set the message.
	[((TorusGamesMainViewController *)[[UIApplication sharedApplication] delegate]) setMessageText:theMessage];
}


static unsigned int LengthOfUnicodeString(Char16 *aString)
{
	unsigned int	theLength;
	
	theLength = 0;

	if (aString != NULL)
	{
		while (*aString++ != 0)
			theLength++;
	}

	return theLength;
}
