//	TorusGamesPodOSUtilities.m

#import <UIKit/UIKit.h>
#import <AudioToolbox/AudioToolbox.h>
#import <AVFoundation/AVAudioPlayer.h>
#include <pthread.h>	//	for pthread_create()
#include <unistd.h>		//	for usleep()
#import "CMGlyphDrawing.h"
#include "TorusGames-Common.h"
#include "TorusGamesPodOSUtilities.h"

//	CGFontGetGlyphsForUnichars() is undocumented.
//	Use it with caution, or better yet try not to use it at all.
//extern void CGFontGetGlyphsForUnichars(CGFontRef, UniChar[], CGGlyph[], size_t);


//	Cache SystemSoundIDs so the sounds can be reused,
//	and so they can be released when the user changes games.
static SystemSoundID	gCachedSoundsWAV[NumSounds];	//	values auto-initialize to 0
static AVAudioPlayer	*gCachedSoundsM4A[NumSounds];


//	Package up information for the new-thread wrapper function.
typedef struct
{
	ModelData	*md;
	void		(*itsStartFuction)(ModelData *);
	bool		itsArgumentsHaveBeenCopied;
} ThreadStartData;


static bool		CharacterIsKana(Char16 aCharacter);
static bool		CharacterIsHangul(Char16 aCharacter);
static CGGlyph	UnicharToGlyph(CGFontRef aFont, Char16 aCharacter);
//	Put a platform-dependent wrapper around the platform-independent thread starter function.
static void		*StartFunctionWrapper(void *aParam);


NSString *GetLocalizedTextAsNSString(
	const char	*aKey)
{
	return [NSString stringWithUTF8String:GetLocalizedText(aKey)];
}


bool GetBasePath(
	char			*aPath,
	unsigned int	aPathSize)
{
	return [[[[NSBundle mainBundle] resourcePath] stringByAppendingString:@"/"]
		getCString:aPath maxLength:aPathSize encoding:NSUTF8StringEncoding];
}


ErrorText GetFileContents(
	const char		*aPathName,		//	input, zero-terminated UTF-8 string containing absolute path name
	unsigned int	*aNumRawBytes,	//	output, the file size in bytes
	Byte			**someRawBytes)	//	output, the file's contents as raw bytes
{
	//	Technical Note:  I had hoped that setlocale() + fopen() would
	//	let me read the file within the platform-independent part
	//	of the code, but alas on WinXP it works when aPathName 
	//	contains only ASCII characters but fails when aPathName 
	//	contains non-trivial Unicode characters, for example when
	//	I temporarily change "Textures" to "Textures日本".
	//	So I reluctantly resorted to platform-dependent code
	//	to read the file.

	NSData	*theFileContents;
	
	//	Read the file.
	theFileContents = [NSData dataWithContentsOfFile:[NSString stringWithUTF8String:aPathName]];
	if (theFileContents == nil)
		return "Couldn't get file contents as NSData in GetFileContents()";
	
	//	Note the number of bytes.
	*aNumRawBytes = [theFileContents length];
	if (*aNumRawBytes == 0)
		return "File contains 0 bytes in GetFileContents()";

	//	Copy the data.  The caller will own this copy, and must
	//	eventually free it with a call to FreeRawTextureBytes().
	*someRawBytes = GET_MEMORY(*aNumRawBytes);
	if (*someRawBytes == NULL)
	{
		*aNumRawBytes = 0;
		return "Couldn't allocate buffer in GetFileContents()";
	}
	[theFileContents getBytes:(void *)(*someRawBytes) length:(*aNumRawBytes)];
	
	//	Success!
	return NULL;
}

void FreeFileContents(
	unsigned int	*aNumRawBytes,
	Byte			**someRawBytes)
{
	*aNumRawBytes = 0;
	FREE_MEMORY_SAFELY(*someRawBytes);
}


void SetAlphaTextureFromCharacter(
	unsigned int	aTextureIndex,
	Char16			aCharacter,
	OpaqueFont		anOpaqueFont,	//	unused
	unsigned int	aSize,
	bool			aShadersFlag,
	ErrorText		*aFirstError)
{
	ErrorText		theErrorMessage		= NULL;
	Byte			*theAlphaBytes		= NULL;
	CGContextRef	theBitmapContext	= NULL;
	NSString		*theFontName		= NULL;
	CGFontRef		theFont				= NULL;
	CGGlyph			theGlyph;
	float			theCellSize,
					theFontSize,
					theBaseline;
	int				theGlyphAdvance;
	float			theGlyphWidth;

	//	Assumes desired OpenGL context is already active.

	//	OpenGL (and the underlying graphics card) likes textures with power-of-two dimensions.
	if (IsPowerOfTwo(aSize) == false)
	{
		theErrorMessage = "Character texture size must be a power of two.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Allocate temporary memory for the bitmapped image.
	//	The image will be an alpha-only image with 1 byte per pixel (no RGB!).
	theAlphaBytes = (Byte *) GET_MEMORY(aSize*aSize*1);
	if (theAlphaBytes == NULL)
	{
		theErrorMessage = "Couldn't allocate theAlphaBytes.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}
	
	//	Create a Core Graphics context using theAlphaBytes.
	theBitmapContext = CGBitmapContextCreate(
							theAlphaBytes,
							aSize,
							aSize,
							8,
							aSize*1,
							NULL,
							kCGImageAlphaOnly);
	if (theBitmapContext == NULL)
	{
		theErrorMessage = "Couldn't create theBitmapContext.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}
	
	//	Even though Core Graphics uses a coordinate system with the y-coordinate 
	//	running bottom-to-top, it seems that CGBitmapContextCreate()
	//	runs the rows of the buffer (theAlphaBytes) top-to-bottom, apparently
	//	for compatibility with the old QuickDraw coordinate system.  (For details, go to
	//		http://developer.apple.com/samplecode/TypeServicesForUnicode/listing1.html
	//	and search for "upside down".)
	//	To simulate a bottom-to-top image, let's flip the coordinates.
	//	(The other alternative, if this strategy ever becomes problematic,
	//	would be to manually swap the rows of the image after all drawing is complete.)
	CGContextScaleCTM(theBitmapContext, 1, -1);
	CGContextTranslateCTM(theBitmapContext, 0, -((float)aSize));

	//	Clear the image.
	CGContextClearRect(theBitmapContext, (CGRect) { {0, 0}, {aSize, aSize} } );

	//	Set the font.
	//	Helvetica doesn't support Japanese, so use 
	//	HiraKakuProN-W3 or HiraKakuProN-W6 for kana.
	if (CharacterIsKana(aCharacter))
		theFontName = @"HiraKakuProN-W6";
	else
	if (CharacterIsHangul(aCharacter))
		theFontName = @"AppleGothic";
	else
		theFontName = @"Helvetica";
	theFont = CGFontCreateWithFontName((CFStringRef)theFontName);
	if (theFont == NULL)
	{
		theErrorMessage = "Couldn't load requested font.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Converting aCharacter to a CGGlyph is harder than you'd think.
	//	Do the conversion in a separate function.
	theGlyph = UnicharToGlyph(theFont, aCharacter);
	if (theGlyph == 0 && aCharacter == L'・')		//	HiraKakuProN-W6 seems to lack the katakana 「・」,
		theGlyph = UnicharToGlyph(theFont, L'•');	//	so substitute a bullet '•' instead.
	if (theGlyph == 0)
	{
		theErrorMessage = "Font does not provide requested character.";
		goto CleanUpSetAlphaTextureFromCharacter;
	}

	//	Set the font size.
	theCellSize = aSize;
	theFontSize = 0.8 * theCellSize;
	theBaseline = 0.2 * theCellSize;
	CGContextSetFont(theBitmapContext, theFont);
	CGContextSetFontSize(theBitmapContext, theFontSize);	//	"text space units" = pixels?
	
	//	How wide is the given glyph?
	CGFontGetGlyphAdvances(theFont, &theGlyph, 1, &theGlyphAdvance);
	theGlyphWidth = theFontSize * (float)theGlyphAdvance / (float)CGFontGetUnitsPerEm(theFont); 

	//	Draw the glyph.
	CGContextShowGlyphsAtPoint(	theBitmapContext,
								0.5*(aSize - theGlyphWidth),
								theBaseline,
								&theGlyph,
								1);

	//	Convert the raw bitmap image to an OpenGL alpha-only texture.
	SetTextureImageFancy(aTextureIndex, aSize, aSize, 1, theAlphaBytes, aShadersFlag);

CleanUpSetAlphaTextureFromCharacter:

	if (theFont != NULL)
		CGFontRelease(theFont);

	if (theBitmapContext != NULL)
		CGContextRelease(theBitmapContext);

	FREE_MEMORY_SAFELY(theAlphaBytes);

	if (theErrorMessage != NULL)
	{
		//	Substitute a transparent texture for the missing one.
		SetTextureImageFancy(aTextureIndex, 1, 1, 1, (Byte [1]){0x00}, aShadersFlag);
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}

static bool CharacterIsKana(
	Char16	aCharacter)
{
	if (aCharacter >= L'ぁ' && aCharacter <= L'ん')	//	hiragana
		return true;
	
	if (aCharacter >= L'ァ' && aCharacter <= L'ン')	//	katakana
		return true;
	
	return false;
}

static bool CharacterIsHangul(
	Char16	aCharacter)
{
	//	Test for a precomposed Hangul syllable
	//	(unicode range 0xAC00 - 0xD7A3).
	if (aCharacter >= L'가' && aCharacter <= L'힣')
		return true;
	
	//	Test for a Hangul compatibility jamo
	//	(unicode range 0x3131 - 0x3163).
	if (aCharacter >= L'ㄱ' && aCharacter <= L'ㅣ')
		return true;
	
	return false;
}

static CGGlyph UnicharToGlyph(
	CGFontRef	aFont,
	Char16		aCharacter)
{
//	CGGlyph	theGlyphArray[8];	//	Allow lots of extra space,
								//	because the unichar to glyph mapping
								//	is not guaranteed to be one-to-one.

	//	Apparently there is no officially supported way 
	//	to get the glyph code corresponding to a given unichar (!).
	//	Apple has a function that does exactly what we need, but it's undocumented.  :-(
//	CGFontGetGlyphsForUnichars(aFont, &aCharacter, theGlyphArray, 1);
//	return theGlyphArray[0];

	CGGlyph	theGlyph;

	//	James J. Hu has provided a workaround for CGFontGetGlyphsForUnichars().
	CMFontGetGlyphsForUnichars(aFont, &aCharacter, &theGlyph, 1);
	return theGlyph;
}


void SetUpMIDIPlayer(void)	//	In the iPhone Torus Games, this applies to M4A, not MIDI.
{
	NSString		*theRelativePath,
					*theAbsolutePath;
	NSURL			*theFileURL;
	AVAudioPlayer	*theDummyPlayer;

	//	The first time the program plays an M4A sound,
	//	AVAudioPlayer takes so long to get set up that
	//	the first sound may get missed entirely.
	//	To speed things along, get AVAudioPlayer warmed up
	//	ahead of time.  Calling prepareToPlay and then
	//	releasing theDummyPlayer seems to do the trick.

	theRelativePath	= @"Dummy.m4a";
	theAbsolutePath	= [[NSBundle mainBundle]
						pathForResource:	theRelativePath
						ofType:				nil
						inDirectory:		@"Sounds - Mobile"];
	theFileURL		= [NSURL fileURLWithPath:theAbsolutePath isDirectory:NO];
	theDummyPlayer	= [[AVAudioPlayer alloc] initWithContentsOfURL:theFileURL error:nil];
	
	[theDummyPlayer prepareToPlay];

	[theDummyPlayer release];
}

void ShutDownMIDIPlayer(void)
{
}

void PlayMIDI(
	const char	*aRelativePath,	//	zero-terminated UTF-8 string with path separator '/'
	SoundType	aSound)			//	sound index permits caching in iPhone Torus Games
{
	NSString	*theRelativePath,
				*theAbsolutePath;
	NSURL		*theFileURL;

	//	The iPhone OS does not support MIDI files.
	//	Play an equivalent M4A file, prepared in advance.

	//	Just to be safe...
	if ((unsigned int)aSound >= NumSounds)
		return;	//	should never occur
	
	//	If the requested sound isn't already loaded, load it now.
	if (gCachedSoundsM4A[aSound] == nil)
	{
		//	Technical note:  Be careful here -- if the requested file is not present,
		//	fileURLWithPath:isDirectory: will throw an exception and crash!
		theRelativePath				= [[[NSString stringWithUTF8String:aRelativePath]
										stringByDeletingPathExtension]		//	remove .mid
										stringByAppendingString:@".m4a"];	//	append .m4a
		theAbsolutePath				= [[NSBundle mainBundle]
										pathForResource:	theRelativePath
										ofType:				nil
										inDirectory:		@"Sounds - Mobile"];
		theFileURL					= [NSURL fileURLWithPath:theAbsolutePath isDirectory:NO];
		gCachedSoundsM4A[aSound]	= [[AVAudioPlayer alloc] initWithContentsOfURL:theFileURL error:nil];
	}
	
	//	If the requested sound is already playing, stop it.
	//	Otherwise the "play" command (see below) will likely have no effect.
	if ([gCachedSoundsM4A[aSound] isPlaying])
		[gCachedSoundsM4A[aSound] stop];
	
	//	Restart the sound from the beginning each time.
	//	(Let's hope the AVAudioPlayer caches the decoded sound,
	//	so it doesn't have to decode the M4A file over and over.)
	[gCachedSoundsM4A[aSound] setCurrentTime:0.0];

	//	There's no need to call [gCachedSoundsM4A[aSound] prepareToPlay],
	//	given that we'll be calling [gCachedSoundsM4A[aSound] play] immediately anyhow.
	
	//	Start the sound playing.
	[gCachedSoundsM4A[aSound] play];
}

void PlayWAV(	//	can also play AIF
	const char	*aRelativePath,	//	zero-terminated UTF-8 string with path separator '/'
	SoundType	aSound)			//	sound index permits caching in iPhone Torus Games
{
	//	Technical Note:  I tried funnelling PlayWAV() and PlayMIDI() into
	//	a single unified PlaySoundFile() function, which played all sounds
	//	with an AVAudioPlayer, exactly as in PlayMIDI() above.
	//	This basically worked, except that each call to PlaySoundFile()
	//	would block the main thread for a brief but perceptible amount of time.
	//	The web page 
	//
	//		http://www.blumtnwerx.com/blog/2009/05/adventures-in-sound-land/
	//
	//	observes that each call to [anAVAudioPlayer play] blocks the main thread
	//	for 0.04-0.05 sec, even though the sound itself plays asynchronously
	//	in a separate thread.  The problem was most noticeable in the Pool game,
	//	where the motion of the balls would hesitate slightly each time two balls
	//	collided and made a sound.  The initial break (of the pool balls) was
	//	disasterous, because it generated a great many sounds and thus a long delay.
	//	(The user doesn't hear all those sounds, because each succeeding one cancels
	//	the previous one almost immediately.)  Moral of the story:  use AudioServices,
	//	not AVAudioPlayer, for time-critical sounds.
	//
	//	Afterthought, for future reference:  If absolutely necessary, I could
	//	call [anAVAudioPlayer play] from a separate thread, but that hardly
	//	advances the present goal of simplifying the code.  So for now I'm happy
	//	to leave well enough alone, and let AudioServices play the pool sounds

	//	Just to be safe...
	if ((unsigned int)aSound >= NumSounds)
		return;	//	should never occur
	
	//	If the requested sound isn't already loaded, load it now.
	if (gCachedSoundsWAV[aSound] == 0)
	{
		AudioServicesCreateSystemSoundID(
			CFBundleCopyResourceURL(CFBundleGetMainBundle(),
									(CFStringRef) [NSString stringWithUTF8String:aRelativePath],
									(CFStringRef) nil,
									(CFStringRef) @"Sounds - Mobile"),
			&gCachedSoundsWAV[aSound]);
	}
	
	//	Start the sound playing.
	AudioServicesPlaySystemSound(gCachedSoundsWAV[aSound]);
}

void ClearSoundCache(void)
{
	unsigned int	i;
	
	//	Warning:  I'm assuming that a valid SystemSoundID is never zero.
	//	I haven't seen that assumption documented anywhere,
	//	but the fact that AudioServicesCreateSystemSoundID() sets
	//	the SystemSoundID value to zero when it can't find the requested file
	//	is a good sign.
	
	for (i = 0; i < NumSounds; i++)
	{
		if (gCachedSoundsWAV[i] != 0)
		{
			AudioServicesDisposeSystemSoundID(gCachedSoundsWAV[i]);
			gCachedSoundsWAV[i] = 0;
		}

		if (gCachedSoundsM4A[i] != nil)
		{
			[gCachedSoundsM4A[i] stop];
			[gCachedSoundsM4A[i] release];
			gCachedSoundsM4A[i] = nil;
		}
	}
}


void StartNewThread(
	ModelData	*md,
	void		(*aStartFuction)(ModelData *))
{
	//	Put a wrapper around aStartFuction to isolate details
	//	like _cdecl vs. _stdcall from the rest of the program.
	
	ThreadStartData	theThreadStartData;
	pthread_t		theNewThreadID;	//	We'll ignore the new thread's ID.
	
	theThreadStartData.md							= md;
	theThreadStartData.itsStartFuction				= aStartFuction;
	theThreadStartData.itsArgumentsHaveBeenCopied	= false;

	pthread_create(&theNewThreadID, NULL, StartFunctionWrapper, (void *)&theThreadStartData);
	
	while (theThreadStartData.itsArgumentsHaveBeenCopied == false)
		SleepBriefly();
}

static void *StartFunctionWrapper(void *aParam)
{
	ThreadStartData	theThreadStartData;
	
	//	Copy the input parameters to our own local memory.
	theThreadStartData = *(ThreadStartData *)aParam;
	
	//	Let the caller know we've made a copy,
	//	so it can return safely.
	((ThreadStartData *)aParam)->itsArgumentsHaveBeenCopied = true;
	
	//	Call itsStartFunction.
	//	Whatever calling convention the compiler prefers is fine by us.
	(*theThreadStartData.itsStartFuction)(theThreadStartData.md);
	
	return NULL;
}


void SleepBriefly(void)
{
	//	Normally the platform-independent code doesn't sleep;
	//	it returns controls to the platform-dependent code
	//	which manages idle time as it sees fit.
	//	However, when Chess waits for its secondary thread
	//	to terminate, it needs to do something to avoid hogging
	//	CPU cycles while it waits.
	
	usleep(10000);	//	10000 µs = 10 ms
}


//	The following code is messier than I would like.
//	The problem is that if a function calls
//
//		ErrorMessage(...);
//		exit(...);
//
//	the program will quit before the message gets displayed.
//	The following code lets the ErrorMessage() call create
//	a UIAlertView and return immediately.  When the UIAlertView
//	gets dismissed, it calls its delegate, which quits the program.

@interface ErrorMessageDelegate : NSObject <UIAlertViewDelegate>
{
	bool	itsFatalFlag;
}
- (id)initWithFatalFlag:(bool)aFatalFlag;
- (void)alertView:(UIAlertView *)alertView clickedButtonAtIndex:(NSInteger)buttonIndex;
@end

static UIAlertView			*gAlert		= nil;
static ErrorMessageDelegate	*gDelegate	= nil;

void ErrorMessage(
	ErrorText	aMessage,	//	UTF8
	ErrorText	aTitle,		//	UTF8
	bool		aFatalFlag)	//	Exit the program after user dismisses the error alert?
{
	if (gAlert != nil || gDelegate != nil)
		return;

	gDelegate	= [[ErrorMessageDelegate alloc] initWithFatalFlag:aFatalFlag];
	gAlert		= [[UIAlertView alloc]
					initWithTitle:		[NSString stringWithUTF8String:aTitle]
					message:			[NSString stringWithUTF8String:aMessage]
					delegate:			gDelegate
					cancelButtonTitle:	@"OK"
					otherButtonTitles:	nil];
	[gAlert show];
}

@implementation ErrorMessageDelegate

- (id)initWithFatalFlag:(bool)aFatalFlag
{
	if ((self = [super init]) != nil)
	{
		itsFatalFlag = aFatalFlag;
	}
	return self;
}

- (void)alertView:(UIAlertView *)alertView clickedButtonAtIndex:(NSInteger)buttonIndex
{
	[gAlert release];
	gAlert = nil;

	[gDelegate release];
	gDelegate = nil;
	
	if (itsFatalFlag)
		exit(1);
}

@end

